/////////////////////////////////////////////////////////////////////////////////////
// This file is part of the PngOptimizer application
// Copyright (C) 2002/2008 Hadrien Nilsson - psydk.org
// For conditions of distribution and use, see copyright notice in PngOptimizer.h
/////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "POEngine.h"

///////////////////////////////////////////////////////////////////////////////
const wchar k_szCannotDumpTry[] = L"Cannot dump internal try, check available memory";
const wchar k_szCannotWriteOnDevice[] = L"Cannot write on device";
const wchar k_szCannotWriteUncomplete[] = L"Write uncomplete, device may be full";
const wchar k_szNotEnoughMemoryForInternalConversion[] = L"Not enough memory for internal conversion";
const wchar k_szUnsupportedPixelFormat[] = L"Unsupported pixel format";

const wchar k_szCannotLoadFile[] = L"Cannot load file";
const wchar k_szFileTooBig[] = L"File is too big";
const wchar k_szUnsupportedFileFormat[] = L"Unsupported file format";
const wchar k_szCannotLoadRawSourceContent[] = L"Cannot load raw source content";
const wchar k_szNotEnoughMemoryToFlattenPaletteImage[] = L"Not enough memory to flatten palette image";
const wchar k_szNotEnoughMemoryToDeflatePaletteImage[] = L"Not enough memory to deflate palette image";
const wchar k_szNotEnoughMemoryToConvertTo24Bits[] = L"Not enough memory to convert to 24 bits";

const wchar k_szCannotGetFileAttributes[] = L"Cannot get file attributes";
const wchar k_szFileIsReadOnly[] = L"File is read-only";
const wchar k_szCannotPerformBackup[] = L"Cannot perform backup, previous backup deletion failed";
const wchar k_szCannotPerformBackupRenameFailed[] = L"Cannot perform backup, rename failed";
const wchar k_szNotEnoughMemoryToKeepOriginalFile[] = L"Not enough memory to keep original file";
const wchar k_szCorruptedChunkStructure[] = L"Corrupted chunk structure";
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
DynamicMemoryFile& POEngine::ResultManager::GetCandidate()
{
	// Get the largest, except if 0
	int64 nSize0 = m_dmf0.GetPosition();
	int64 nSize1 = m_dmf1.GetPosition();

	if( nSize0 == 0 )
	{
		return m_dmf0;
	}

	if( nSize1 == 0 )
	{
		return m_dmf1;
	}

	DynamicMemoryFile& dmf = ( nSize1 > nSize0 ) ? m_dmf1 : m_dmf0;

	// Open the dynamic memory file with allocation performed
	
	// The -64 is to be gentle with the memory allocator that needs room for its headers
	const int32 nFirstAlloc = 512 * 1024 - 64;
	dmf.Open(nFirstAlloc);
	dmf.SetPosition(0);

	return dmf;
}

DynamicMemoryFile& POEngine::ResultManager::GetSmallest()
{
	// Get the smallest, except if 0
	int64 nSize0 = m_dmf0.GetPosition();
	int64 nSize1 = m_dmf1.GetPosition();

	if( nSize0 == 0 )
	{
		return m_dmf1;
	}

	if( nSize1 == 0 )
	{
		return m_dmf0;
	}
	DynamicMemoryFile& dmf = ( nSize0 < nSize1 ) ? m_dmf0 : m_dmf1;
	return dmf;
}

void POEngine::ResultManager::Reset()
{
	m_dmf0.SetPosition(0);
	m_dmf1.SetPosition(0);
}

///////////////////////////////////////////////////////////////////////////////

POEngine::POEngine()
{
	m_bBackupOldPngFiles = true;
	m_bKeepInterlacing = false;
	m_bKeepBackgroundColor = false;
	m_bAvoidGreyWithSimpleTransparency = true;
}

POEngine::~POEngine()
{
}

bool POEngine::IsBlackAndWhite(const Palette& pal, bool& bShouldSwap)
{
	Color32 col0 = pal.m_aColors[0];
	Color32 col1 = pal.m_aColors[1];
	
	bShouldSwap = false;

	if( col0.IsEqualRgb(Color32::black) && col1.IsEqualRgb(Color32::white))
		return true;

	if( col0.IsEqualRgb(Color32::white) && col1.IsEqualRgb(Color32::black))
	{
		bShouldSwap = true;
		return true;
	}

	if( col0.GetAlpha() == 0 && col1.IsEqualRgb(Color32::white))
		return true;

	if( col0.GetAlpha() == 0 && col1.IsEqualRgb(Color32::black))
	{
		bShouldSwap = true;
		return true;
	}

	return false;
}
/////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////
bool POEngine::IsGreyPalette(const Palette& pal)
{
	for(int32 i = 0; i < pal.m_nCount; ++i)
	{
		uint8 r, g, b;
		pal.m_aColors[i].ToRgb(r, g, b);

		if( r != g || r != b || g != b)
		{
			return false;
		}
	}
	return true;
}

void POEngine::SortByLuminance(Palette& pal, uint8* pBuffer, int32 nPixelCount)
{
	const int32 nColCount = pal.m_nCount;

	//////////////////////////////////////////////////////////////////
	// Sort opaque alphas (255) according to their intensity
	int32 nLuminanceSortStart = 0;
	for(int32 i = 0; i < nColCount; ++i)
	{
		if( pal.m_aColors[i].GetAlpha() == 255)
		{
			// We can start here
			nLuminanceSortStart = i;
			break;
		}
	}
	
	uint32 anLuminances[256];
	uint32 anLumSortedIndexes[256];
		
	for(int32 i = 0; i < 256; ++i)
	{
		anLumSortedIndexes[i] = i;
		anLuminances[i] = 0;
	}

	for(int32 i = nLuminanceSortStart; i < nColCount; ++i)
	{
		uint32 r, g, b;
		pal.m_aColors[i].ToRgb(r, g, b);
		
		// Y = R * 0.299 + G * 0.587 + B * 0.114;
		uint32 y = r * 299 + g * 587 + b * 114;
		anLuminances[i] = y;
	}

	const int32 nLumColCount = nColCount - nLuminanceSortStart;
	uint32* paLumElements = anLumSortedIndexes + nLuminanceSortStart;
	uint32* paLumSortAttributs = anLuminances + nLuminanceSortStart;
	
	chustd::Sort::ByteSort(paLumElements, paLumSortAttributs, nLumColCount);

	// Swap colors from the palette in order to sort them
	Palette palLumSorted;
	palLumSorted.m_nCount = nColCount;

	for(int32 i = 0; i < nColCount; ++i)
	{
		palLumSorted.m_aColors[i] = pal.m_aColors[ anLumSortedIndexes[i] ];
	}
	
	// conv2
	uint8 aConvTable2[256];
	for(int32 i = 0; i < 256; ++i)
	{
		int32 n = anLumSortedIndexes[i];
		aConvTable2[n] = uint8(i);
	}

	pal = palLumSorted;
	
	//////////////////////////////////////////////////////////////////
	// Change the pixel indexes so they point on the new palette
	for(int32 i = 0; i < nPixelCount; ++i)
	{
		uint8 nOldIndex = pBuffer[i];
		uint8 nNewIndex = aConvTable2[ nOldIndex ];
		
		pBuffer[i] = nNewIndex;
	}
}

void POEngine::SortByPopulation(Palette& pal, uint8* pBuffer, int32 nPixelCount)
{
	const int32 nColCount = pal.m_nCount;

	//////////////////////////////////////////////////////////////////
	// Sort opaque alphas (255) according to their population
	int32 nPopulationSortStart = 0;
	for(int32 i = 0; i < nColCount; ++i)
	{
		if( pal.m_aColors[i].GetAlpha() == 255)
		{
			// We can start here
			nPopulationSortStart = i;
			break;
		}
	}
	
	uint32 anPopulations[256];
	uint32 anPopSortedIndexes[256];
		
	for(int32 i = 0; i < 256; ++i)
	{
		anPopSortedIndexes[i] = i;
		anPopulations[i] = 0;
	}

	for(int32 i = 0; i < nPixelCount; ++i)
	{
		uint8 nIndex = pBuffer[i];
		anPopulations[nIndex]++;
	}

	const int32 nPopColCount = nColCount - nPopulationSortStart;
	uint32* paPopElements = anPopSortedIndexes + nPopulationSortStart;
	uint32* paPopSortAttributs = anPopulations + nPopulationSortStart;
	
	chustd::Sort::ByteSort(paPopElements, paPopSortAttributs, nPopColCount);

	// Reverse the sort
	for(int32 i = 0; i < nPopColCount / 2; ++i)
	{
		const int32 nOffsetFar = nPopColCount - i - 1;
		uint32 nPos = paPopElements[i];
		paPopElements[i] = paPopElements[nOffsetFar];
		paPopElements[nOffsetFar] = nPos;
	}

	// Create the new palette in reverse order of their population
	// (the most popular have the lowest index)
	Palette palPopSorted;
	palPopSorted.m_nCount = nColCount;

	for(int32 i = 0; i < nColCount; ++i)
	{
		palPopSorted.m_aColors[i] = pal.m_aColors[ anPopSortedIndexes[i] ];
	}
	
	// Create the converstion table old indexes --> new indexes
	uint8 aConvTable2[256];
	for(int32 i = 0; i < 256; ++i)
	{
		int32 n = anPopSortedIndexes[i];
		aConvTable2[n] = uint8(i);
	}

	pal = palPopSorted;

	//////////////////////////////////////////////////////////////////
	// Change the pixel indexes so they point on the new palette
	for(int32 i = 0; i < nPixelCount; ++i)
	{
		uint8 nOldIndex = pBuffer[i];
		uint8 nNewIndex = aConvTable2[ nOldIndex ];
		
		pBuffer[i] = nNewIndex;
	}
}

bool POEngine::FlattenBufferAndDump(Palette& palNewNew, PngDumpSettings& ds, int32 nPixelCount, uint8* pNewBuffer)
{
	const int32 nColCount = palNewNew.m_nCount;
	const int32 nWidth = ds.m_nWidth;
	const int32 nHeight = ds.m_nHeight;

	if( nColCount <= 2)
	{
		// 1 bit per pixel
		ds.m_epf = pf1bppIndexed;
		
		// Maybe it's a black&white picture, in that case there is no need for a palette
		bool bShouldSwap;
		if( IsBlackAndWhite(palNewNew, bShouldSwap))
		{
			uint8 nAlpha0 = palNewNew.m_aColors[0].GetAlpha();
			uint8 nAlpha1 = palNewNew.m_aColors[1].GetAlpha(); 
			
			if( (nAlpha0 != 255 || nAlpha1 != 255) && m_bAvoidGreyWithSimpleTransparency)
			{
				// A black&white picture with transparency, we stay in palette mode
			}
			else
			{
				if( bShouldSwap)
				{
					Color32 colTmp = palNewNew.m_aColors[0];
					palNewNew.m_aColors[0] = palNewNew.m_aColors[1];
					palNewNew.m_aColors[1] = colTmp;
					
					// Inversion : 0 => 1   et 1 => 0
					for(int32 i = 0; i < nPixelCount; ++i)
					{
						uint8 nIndex = pNewBuffer[i];
						pNewBuffer[i] = uint8(nIndex ^ 1);
					}
				}
				
				// Check if there was transparency in this paletted image, and if it is
				// compatible with a pure black&white image with no palette

				// In grayscale, only one index can define an alpha of 0
				uint8 a0 = palNewNew.m_aColors[0].GetAlpha();
				uint8 a1 = palNewNew.m_aColors[1].GetAlpha();
				if( a0 == 255 && a1 == 255)
				{
					// No transparency at all
					ds.m_bUseTransparentColor = false;
					ds.m_epf = pf1bppGrayScale;
				}
				else if( a0 == 0 && a1 == 255)
				{
					// Transparency for index 0
					ds.m_bUseTransparentColor = true;
					ds.m_nTransGrey = 0;
					ds.m_epf = pf1bppGrayScale;
				}
				else if( a0 == 255 && a1 == 0)
				{
					// Transparency for index 1
					ds.m_bUseTransparentColor = true;
					ds.m_nTransGrey = 1;
					ds.m_epf = pf1bppGrayScale;
				}
				else
				{
					// Another configuration of transparency, we cannot keep the image in black&white
					// We stay in palette mode...
				}
			}
		}
	}
	else if( nColCount <= 4)
	{
		// 2 bits per pixel
		ds.m_epf = pf2bppIndexed;
	}
	else if( nColCount <= 16)
	{
		// 4 bits per pixel
		ds.m_epf = pf4bppIndexed;
	}
	else
	{
		// 8 bits per pixel
		ds.m_epf = pf8bppIndexed;
	}
	
	//	Build the new buffer
	int32 nBitPerPix = ImageType::SizeofPixelInBits(ds.m_epf);
	
	/////////////////////////////////////////////////////////
	// Convert the pixel buffer so one byte => one or several pixels
	
	// Note : there is padding at the end of each row	
	uint8 nCurrentByte = 0; // Current byte value
	int32 nUsedBitCount = 0; // Number of bits used inside the byte
	
	const int32 nImgRowSize = ImageType::ComputeByteWidth(ds.m_epf, nWidth);
	
	ByteArray rbResult;
	if( !rbResult.SetSize(nImgRowSize * nHeight))
	{
		AddError(k_szNotEnoughMemoryToFlattenPaletteImage);
		return false;
	}

	uint8* const pResultBuffer = rbResult.GetBuffer();
	
	uint8* pResultBufferTmp = pResultBuffer;
	
	for(int32 iRow = 0; iRow < nHeight; iRow++)
	{
		nCurrentByte = 0;
		nUsedBitCount = 0;
		
		for(int32 iCol = 0; iCol < nWidth; iCol++)
		{
			uint8 nVal = pNewBuffer[iRow * nWidth + iCol];
			
			nUsedBitCount += nBitPerPix;
			nCurrentByte |= (nVal << (8 - nUsedBitCount));
			
			if( nUsedBitCount == 8)
			{
				pResultBufferTmp[0] = nCurrentByte;
				++pResultBufferTmp;
				
				nCurrentByte = 0;
				nUsedBitCount = 0;
			}
		}
		
		if( nUsedBitCount != 0)
		{
			pResultBufferTmp[0] = nCurrentByte;
			++pResultBufferTmp;
		}
	}
	
	ds.m_pBuffer = pResultBuffer;
	ds.m_pPalette = &palNewNew;
	
	// Test 1
	// - No filtering
	// - ZLib default strategy
	// - ZLib normal memory load
	ds.m_nFiltering = 0;
	ds.m_nZLibStrategy = PngDumpSettings::zlibStrategyDefault;
	ds.m_nZLibWindowBitsAndMem = 0;
	if( !Png::Dump(m_resultmgr.GetCandidate(), ds))
	{
		AddError(k_szCannotDumpTry);
		return false;
	}

	// Test 2
	// - No filtering
	// - ZLib default strategy
	// - ZLib low memory load (yes, sometimes it can increase compression ! I don't know why... ZLib master, anyone ?
	ds.m_nFiltering = 0;
	ds.m_nZLibStrategy = PngDumpSettings::zlibStrategyDefault;
	ds.m_nZLibWindowBitsAndMem = PngDumpSettings::zlibWindowBitsAndMemLow;
	if( !Png::Dump(m_resultmgr.GetCandidate(), ds))
	{
		AddError(k_szCannotDumpTry);
		return false;
	}

	// Test 3
	// - Filtering
	// - ZLib filter strategy
	// - ZLib normal memory load
	ds.m_nFiltering = 1;
	ds.m_nZLibStrategy = PngDumpSettings::zlibStrategyFilter;
	ds.m_nZLibWindowBitsAndMem = 0;
	if( !Png::Dump(m_resultmgr.GetCandidate(), ds))
	{
		AddError(k_szCannotDumpTry);
		return false;
	}

	// Test 4
	// - Filtering
	// - ZLib default strategy
	// - ZLib normal memory load
	ds.m_nFiltering = 1;
	ds.m_nZLibStrategy = PngDumpSettings::zlibStrategyDefault;
	ds.m_nZLibWindowBitsAndMem = 0;
	if( !Png::Dump(m_resultmgr.GetCandidate(), ds))
	{
		AddError(k_szCannotDumpTry);
		return false;
	}
	
	return true;
}

bool POEngine::DumpBestResultToFile(const String& strNewFileName)
{
	DynamicMemoryFile& dmf = m_resultmgr.GetSmallest();
	int32 nSizeToDump = (int32) dmf.GetPosition();

	File file;
	if( !file.Open(strNewFileName, File::modeWrite))
	{
		AddError(k_szCannotWriteOnDevice);
		return false;
	}

	int32 nWritten = file.Write(dmf.GetContent().GetBuffer(), nSizeToDump);
	if( nWritten != nSizeToDump)
	{
		AddError(k_szCannotWriteUncomplete);
		return false;
	}
	file.Close();
	return true;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////
// ds : input image information
bool POEngine::OptimizePaletteMode(PngDumpSettings& ds)
{
	PixelFormat pf = ds.m_epf;
	const int32 nWidth = ds.m_nWidth;
	const int32 nHeight = ds.m_nHeight;
	uint8* const pBuffer = (uint8*) ds.m_pBuffer;
	const Palette* pPalette = ds.m_pPalette;

	ASSERT( pf1bppIndexed <= pf && pf <= pf8bppIndexed);

	// Mode image with palette
	
	int32 nBitPerPix = ImageType::SizeofPixelInBits(pf);
	
	/////////////////////////////////////////////////////////
	// Convert the buffer so one byte => one pixel
	// This is to speed-up incoming algorithm. Before dumping, the buffer will be flatten back.
	
	const int32 nPixelCount = nWidth * nHeight;
	
	ByteArray rbNew;
	if( !rbNew.SetSize(nPixelCount))
	{
		AddError(k_szNotEnoughMemoryToDeflatePaletteImage);
		return false;
	}

	uint8* pNewBuffer = rbNew.GetBuffer();
	
	int32 nMask = ~((~0L) << nBitPerPix);
	int32 nShift = 8 - nBitPerPix;
	
	int32 nByteOld = 0; // Current byte index in old buffer
	int32 nByteNew = 0; // Current byte index in new buffer
	for(int32 y = 0; y < nHeight; ++y)
	{
		for(int32 x = 0; x < nWidth; ++x)
		{
			const uint8 nCurrentByte = pBuffer[nByteOld];
			
			uint8 nIndex = uint8( (nCurrentByte >> nShift) & nMask);
			
			nShift -= nBitPerPix;
			if( nShift < 0)
			{
				nShift = 8 - nBitPerPix;
				nByteOld += 1;
			}
			
			pNewBuffer[nByteNew] = nIndex;
			nByteNew += 1;
		}
		
		// Some padding bits remain
		if( nShift != 8 - nBitPerPix)
		{
			nShift = 8 - nBitPerPix;
			nByteOld += 1;
		}
	}
	
	/////////////////////////////////////////////////////////
	// Count the number of times a color appears
	int32 aColCounts[256];
	Memory::Zero32(aColCounts, 256);
	
	for(int32 i = 0; i < nPixelCount; ++i)
	{
		aColCounts[ pNewBuffer[i] ]++;
	}
	
	/////////////////////////////////////////////////////////
	// Remove holes in the palette and create the conversion table old index -> new index
	int32 nColCount = 0; // Number of colors in the palette
	int32 nHoleAt = -1;
	uint8 aNoHolesTable[256]; // Conversion table
	int32 iNoHoles = 0;
	
	Palette palNew; // New palette without holes
	
	for(int32 i = 0; i < 256; ++i)
	{
		if( aColCounts[i] != 0)
		{
			nColCount++;
			
			if( nHoleAt >= 0)
			{
				// One of the previous color was not used
				//bHolesInPaletteFound = true; // Commented out, this is for debugging only
			}
			
			aNoHolesTable[i] = (uint8) iNoHoles;
			palNew.m_aColors[iNoHoles] = pPalette->m_aColors[i];
			
			iNoHoles++;
		}
		else
		{
			if( nHoleAt < 0)
			{
				nHoleAt = i;
			}
		}
	}
	
	palNew.m_nCount = nColCount;
	
	//////////////////////////////////////////////////////////////////
	// Sort the palette according to an increasing alpha
	
	// Buyild the elements to sort (indexes in the palette)
	// Get back the sort attributes (the alpha of each color)
	uint32 aAlphaSortedIndexes[256];
	uint8 aAlphas[256];
	for(int32 i = 0; i < 256 ;++i)
	{
		aAlphaSortedIndexes[i] = i;
		aAlphas[i] = palNew.m_aColors[i].GetAlpha();
	}
	
	uint32* paElements = aAlphaSortedIndexes;
	uint8* paSortAttributs = aAlphas;
	
	chustd::Sort::ByteSortAtt(paElements, paSortAttributs, nColCount);
	
	Palette palNewNew; // New sorted palette
	palNewNew.m_nCount = nColCount;
	
	for(int32 i = 0; i < nColCount; ++i)
	{
		palNewNew.m_aColors[i] = palNew.m_aColors[ aAlphaSortedIndexes[i]];
	}
	
	// ori -> noholes -> alpha
	uint8 aConvTable[256];
	for(int32 i = 0; i < 256; ++i)
	{
		int32 n = aAlphaSortedIndexes[i];
		aConvTable[n] = uint8(i);
	}
	
	//////////////////////////////////////////////////////////////////
	// Change the pixel buffer so every index points on the new palette
	for(int32 i = 0; i < nPixelCount; ++i)
	{
		uint8 nOldIndex = pNewBuffer[i];
		uint8 nNewIndex = aNoHolesTable[nOldIndex];
		uint8 nNewIndex2 = aConvTable[ nNewIndex ];
		
		pNewBuffer[i] = nNewIndex2;
	}

	SortByLuminance(palNewNew, pNewBuffer, nPixelCount);
	if( !FlattenBufferAndDump(palNewNew, ds, nPixelCount, pNewBuffer))
	{
		return false;
	}

	SortByPopulation(palNewNew, pNewBuffer, nPixelCount);
	if( !FlattenBufferAndDump(palNewNew, ds, nPixelCount, pNewBuffer))
	{
		return false;
	}

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Find an unused color among the pixels
// Needs memory allocation, thus can fail
bool POEngine::FindUnusedColorHardcoreMethod(const uint8* pRgba, int32 nPixelCount, 
                                             uint8& nRed, uint8& nGreen, uint8& nBlue)
{
	ByteArray aSorted;
	if( !aSorted.SetSize(nPixelCount * 4))
	{
		return false;
	}

	ByteArray aSorted2;
	if( !aSorted2.SetSize(nPixelCount * 4))
	{
		return false;
	}

	uint32* pRgba32 = (uint32*) pRgba;

	// ByteSort function that works the same on both endian mode
	chustd::Sort::ByteSortLittleEndian( pRgba32, (uint32*)aSorted.GetPtr(), (uint32*)aSorted2.GetPtr(), nPixelCount);

	// This first entry is pure black with alpha set to 0,
	// If we are in this function it's because it is not suitable so we start at 1

	if( k_ePlatformByteOrder == boBigEndian )
	{
		// Convert uint32 to little endian
		IFile::Swap32( (uint32*) aSorted.GetPtr(), nPixelCount);
	}

	uint32* pSorted32 = (uint32*) aSorted.GetPtr();

	uint32 nA = 0x00000000; // ABGR in register
	for(int32 i = 1; i < nPixelCount; ++i)
	{
		uint32 nB = (0x00ffffff & pSorted32[i]);
		uint32 nDif = nB - nA;
		if( nDif > 1)
		{
			// Found a useable range of colors to be used as transparent
			break;
		}
		nA = nB; // Next entry
	}

	if( nA == 0x00ffffff)
	{
		// This can only occurs if all 16777216 colors are present in the picture
		// Like a 4096*4096 picture
		return false;
	}
	
	uint32 nColor = nA + 1;

	nRed = uint8(nColor & 0x000000ff);
	nGreen = uint8((nColor & 0x0000ff00) >> 8);
	nBlue = uint8((nColor & 0x00ff0000) >> 16);
	
	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////
// Find an unused color among the pixels
// rbRgb [in] : pixel buffer
// nRed, nGreen, nBlue [out] : color not present in the pixel buffer
// Returns true upon success
bool POEngine::FindUnusedColor(const ByteArray& aRgb, uint8& nRed, uint8& nGreen, uint8& nBlue)
{
	// Check with some candidates
	FixArray<Color32, 7> aCandidates;
	aCandidates[0].SetRgb(1, 0, 0);
	aCandidates[1].SetRgb(1, 1, 0);
	aCandidates[2].SetRgb(1, 1, 1);
	aCandidates[3].SetRgb(0, 1, 1);
	aCandidates[4].SetRgb(0, 0, 1);
	aCandidates[5].SetRgb(1, 0, 1);
	aCandidates[6].SetRgb(0, 1, 0);
	
	const int32 nPixelCount = aRgb.GetSize() / 3;
	
	foreach(aCandidates, iCandidate)
	{
		bool bGoodCandidate = true;

		uint8 cr, cg, cb;
		aCandidates[iCandidate].ToRgb(cr, cg, cb);

		uint8* pPixelsRgb = aRgb.GetPtr();
		for(int32 i = 0; i < nPixelCount; ++i)
		{
			uint8 r = pPixelsRgb[0];
			uint8 g = pPixelsRgb[1];
			uint8 b = pPixelsRgb[2];

			if( r == cr && g == cg && b == cb )
			{
				// Found, not a good candidate
				bGoodCandidate = false;
				break;
			}

			pPixelsRgb += 3;
		}

		if( bGoodCandidate )
		{
			nRed = cr;
			nGreen = cg;
			nBlue = cb;
			return true;
		}
	}
	return false;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
bool POEngine::Optimize32BitsMode(PngDumpSettings& ds)
{
	ASSERT(ds.m_epf == pf32bppRgba);

	const int32 nWidth = ds.m_nWidth;
	const int32 nHeight = ds.m_nHeight;
	const int32 nPixelCount = nWidth * nHeight;

	///////////////////////////////////////////////////////////////////////////////////////////////
	// First step : set to 0 every color which alpha is 0
	
	// It will allow a potential optimisation 32 bits --> palette mode
	// We set the fully transparent color to 0 as 4 bytes to 0 are nicely compressed

	uint8*  pPixels8  = (uint8*) ds.m_pBuffer;

	for(int32 iPixel = 0; iPixel < nPixelCount; ++iPixel)
	{
		// In memory : RGBA, RGBA, RGBA
		//             0123  0123  0123
		uint8 nAlpha = pPixels8[3];

		if( nAlpha == 0 )
		{
			// Alpha set to 0, set everything to 0
			uint32* pPixels32 = (uint32*) pPixels8;
			pPixels32[0] = 0x00000000;
		}

		pPixels8 += 4;
	}

	// Test 0
	// We save the result as-is, sometimes a small 32 bits image can be smaller than the same one in 24 or 8 bits
	if( nPixelCount <= (64 * 64))
	{
		// Our definition of "small" is 64*64
		if( !Png::Dump(m_resultmgr.GetCandidate(), ds))
		{
			AddError(k_szCannotDumpTry);
			return false;
		}
	}

	///////////////////////////////////////////////////////////////////////////////////////////////
	// Verify the alpha channel is really necessary
	uint8* pBuffer = (uint8*) ds.m_pBuffer;
	const uint8* pSrcBuffer = pBuffer;

	ByteArray rbNewRgb;
	if( !rbNewRgb.SetSize(nPixelCount * 3))
	{
		AddError(k_szNotEnoughMemoryToConvertTo24Bits);
		return false;
	}

	uint8* pNewBuffer = rbNewRgb.GetBuffer();

	// To count how many times the same alpha appears in the image
	int32 anAlphas[256];
	Memory::Zero32(anAlphas, 256);

	for(int32 i = 0; i < nPixelCount; ++i)
	{
		uint8 r = pSrcBuffer[0];
		uint8 g = pSrcBuffer[1];
		uint8 b = pSrcBuffer[2];
		uint8 nAlpha = pSrcBuffer[3];

		pNewBuffer[0] = r;
		pNewBuffer[1] = g;
		pNewBuffer[2] = b;

		anAlphas[nAlpha] += 1;
		
		pSrcBuffer += 4;
		pNewBuffer += 3;
	}

	// All pixels are opaque, no need for an alpha channel, use the 24 bits optimizer
	if( anAlphas[255] == nPixelCount)
	{
		ds.m_pBuffer = rbNewRgb.GetBuffer();
		ds.m_epf = pf24bppRgb;
	
		return Optimize24BitsMode(ds);
	}

	// Ok, maybe we can keep the 24 bits buffer if every alpha is set to 255 except for one color

	if( (anAlphas[0] + anAlphas[255]) == nPixelCount)
	{
		// Yes we can, but the color to be used to mean a pixel is totaly transparent must not appear
		// elsewhere the picture as being an opaque pixel.

		// It means we must choose a color not used in the picture, that's a bit tricky

		////////////////////////////////////////////////////
		// First, check if black is a used color
		
		bool bBlackAsTransparentColor = true;
		uint8* pPixelsRgba = (uint8*) ds.m_pBuffer;

		for(int32 i = 0; i < nPixelCount; ++i)
		{
			uint8 r = pPixelsRgba[0];
			uint8 g = pPixelsRgba[1];
			uint8 b = pPixelsRgba[2];
			uint8 nAlpha = pPixelsRgba[3];
			if( nAlpha == 255)
			{
				if( r == 0 && g == 0 && b == 0 )
				{
					// Yes, black is a used color, we cannot use it as a the transparent color
					bBlackAsTransparentColor = false;
					break;
				}
			}
			
			pPixelsRgba += 4;
		}
		////////////////////////////////////////////////////
		
		uint8 nTransRed = 0;
		uint8 nTransGreen = 0;
		uint8 nTransBlue = 0;

		bool bContinueIn24Bits = bBlackAsTransparentColor;

		if( !bBlackAsTransparentColor)
		{
			// Find an alternative transparent color
			if( FindUnusedColor(rbNewRgb, nTransRed, nTransGreen, nTransBlue)
			  || FindUnusedColorHardcoreMethod((uint8*)ds.m_pBuffer, nPixelCount, nTransRed, nTransGreen, nTransBlue))
			{
				bContinueIn24Bits = true;

				// Ok, we have our color to be use for transparency :)
				// Convert every pixel to that color

				uint8* pPixelsRgb = rbNewRgb.GetPtr();
				uint8* pPixelsRgba = (uint8*) ds.m_pBuffer;

				for(int32 i = 0; i < nPixelCount; ++i)
				{
					uint8 nAlpha = pPixelsRgba[3];
					if( nAlpha == 0)
					{
						// Put the transparent color
						pPixelsRgb[0] = nTransRed;
						pPixelsRgb[1] = nTransGreen;
						pPixelsRgb[2] = nTransBlue;
					}
					
					pPixelsRgba += 4;
					pPixelsRgb += 3;
				}
			}
		}
		
		if( bContinueIn24Bits )
		{
			// Now we have our 24 bits image + one color for transparency, continue with 24 bits optimization...
			ds.m_pBuffer = rbNewRgb.GetBuffer();
			ds.m_epf = pf24bppRgb;
			ds.m_bUseTransparentColor = true;
			ds.m_nTransRed = nTransRed;
			ds.m_nTransGreen = nTransGreen;
			ds.m_nTransBlue = nTransBlue;

			return Optimize24BitsMode(ds);
		}
	}

	///////////////////////////////////////////////////////////////////
	// Dump to memory

	ds.m_pBuffer = pBuffer;
	ds.m_epf = pf32bppRgba;

	// Test 1
	//  - No filtering
	//  - ZLib default strategy
	ds.m_nFiltering = 0;
	ds.m_nZLibStrategy = PngDumpSettings::zlibStrategyDefault;
	if( !Png::Dump(m_resultmgr.GetCandidate(), ds))
	{
		return false;
	}

	// Test 2
	//  - Filtering
	//  - ZLib filter strategy
	ds.m_nFiltering = 1;
	ds.m_nZLibStrategy = PngDumpSettings::zlibStrategyFilter;
	if( !Png::Dump(m_resultmgr.GetCandidate(), ds))
	{
		return false;
	}

	// Test 3
	//  - Filtering
	//  - ZLib default strategy
	ds.m_nFiltering = 1;
	ds.m_nZLibStrategy = PngDumpSettings::zlibStrategyDefault;
	if( !Png::Dump(m_resultmgr.GetCandidate(), ds))
	{
		return false;
	}

	return true;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////
bool POEngine::Optimize24BitsMode(PngDumpSettings& ds)
{
	ASSERT(ds.m_epf == pf24bppRgb);

	const uint8* pBuffer = (const uint8*) ds.m_pBuffer;
	const int32 nWidth = ds.m_nWidth;
	const int32 nHeight = ds.m_nHeight;

	////////////////////////////////////////////////////////////////
	// Some test to check if we can switch to palette mode
	Palette palTest;
	palTest.m_nCount = 0;

	const uint8* pSrcBuffer = pBuffer;
	const int32 nPixelCount = nWidth * nHeight;

	ByteArray rbNew;
	if( !rbNew.SetSize(nPixelCount))
	{
		return false;
	}
	uint8* pNewBuffer = rbNew.GetBuffer();

	bool bTooMuchColors = false;
	for(int32 i = 0; i < nPixelCount; ++i)
	{
		uint8 r = pSrcBuffer[0];
		uint8 g = pSrcBuffer[1];
		uint8 b = pSrcBuffer[2];
		
		// Find the color in the palette
		int32 iCol = 0;
		for(; iCol < palTest.m_nCount; ++iCol)
		{
			const Color32& col = palTest.m_aColors[iCol];
			uint8 rp, gp, bp;
			col.ToRgb(rp, gp, bp);

			if( r == rp && g == gp && b == bp)
			{
				// Found !
				break;
			}
		}

		pNewBuffer[0] = uint8(iCol); // Fill the new buffer

		if( iCol == palTest.m_nCount)
		{
			// The 24 bit color was not in the palette

			// Add a new entry in the palette
			if( palTest.m_nCount == 256)
			{
				// The palette is full
				bTooMuchColors = true;
				break;
			}

			// New entry in the palette
			palTest.m_aColors[ palTest.m_nCount].SetRgb(r, g, b);
			palTest.m_nCount++;
		}

		pSrcBuffer += 3;
		pNewBuffer += 1;
	}

	// If the picture is very small, we may enlarge it if we switch to palette mode
	const int32 nMaxSizeOrigin = nPixelCount * 3;
	const int32 nMaxSizeNew = nPixelCount + palTest.m_nCount * 3 + 12; // 12 = min size chunk
	if( nMaxSizeOrigin < nMaxSizeNew)
	{
		bTooMuchColors = true;
	}
	
	///////////////////////////////////////////////////////////////////
	ds.m_pBuffer = pBuffer;
	ds.m_epf = pf24bppRgb;

	// Test 1
	//  - Filtering
	//  - ZLib default strategy
	ds.m_nFiltering = 1;
	ds.m_nZLibStrategy = PngDumpSettings::zlibStrategyDefault;
	if( !Png::Dump(m_resultmgr.GetCandidate(), ds))
	{
		AddError(k_szCannotDumpTry);
		return false;
	}

	// Test 2
	//  - Filtering
	//  - ZLib filter strategy
	ds.m_nFiltering = 1;
	ds.m_nZLibStrategy = PngDumpSettings::zlibStrategyFilter;
	if( !Png::Dump(m_resultmgr.GetCandidate(), ds))
	{
		AddError(k_szCannotDumpTry);
		return false;
	}

	// Test 3
	//  - No filtering
	//  - ZLib default strategy
	ds.m_nFiltering = 0;
	ds.m_nZLibStrategy = PngDumpSettings::zlibStrategyDefault;
	if( !Png::Dump(m_resultmgr.GetCandidate(), ds))
	{
		AddError(k_szCannotDumpTry);
		return false;
	}

	///////////////////////////////////////////////////////////////////

	if( bTooMuchColors)
	{
		// Cannot convert to palette mode, we stop here
		return true;
	}

	// Add transparency to the palette
	if( ds.m_bUseTransparentColor)
	{
		// Find the transparent color
		int32 iCol = 0;
		for(; iCol < palTest.m_nCount; ++iCol)
		{
			const Color32& col = palTest.m_aColors[iCol];
			uint8 rp, gp, bp;
			col.ToRgb(rp, gp, bp);

			if( ds.m_nTransRed == rp && ds.m_nTransGreen == gp && ds.m_nTransBlue == bp)
			{
				// Aha ! :)
				palTest.m_aColors[iCol].SetAlpha(0);
				break;
			}
		}
	}

	// Now we have a 8 bits indexed image, we can continue with the palette mode optimizer
	ds.m_pBuffer = rbNew.GetBuffer();
	ds.m_pPalette = &palTest;
	ds.m_epf = pf8bppIndexed;

	return OptimizePaletteMode(ds);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////
bool POEngine::OptimizeGrayScale(PngDumpSettings& ds)
{
	ASSERT( pf1bppGrayScale <= ds.m_epf && ds.m_epf <= pf16bppGrayScale );

	// Test 1
	ds.m_nFiltering = 0;
	ds.m_nZLibStrategy = PngDumpSettings::zlibStrategyDefault;
	if( !Png::Dump(m_resultmgr.GetCandidate(), ds))
	{
		return false;
	}
	
	// Test 2
	ds.m_nFiltering = 1;
	ds.m_nZLibStrategy = PngDumpSettings::zlibStrategyDefault;
	if( !Png::Dump(m_resultmgr.GetCandidate(), ds))
	{
		return false;
	}

	// Test 3
	ds.m_nFiltering = 1;
	ds.m_nZLibStrategy = PngDumpSettings::zlibStrategyFilter;
	if( !Png::Dump(m_resultmgr.GetCandidate(), ds))
	{
		return false;
	}

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
void POEngine::BgraToRgba(PngDumpSettings& ds)
{
	uint8* pBuffer = (uint8*) ds.m_pBuffer;
	int32 nPixelCount = ds.m_nWidth * ds.m_nHeight;

	for(int32 i = 0; i < nPixelCount; ++i)
	{
		uint8 b = pBuffer[0];
		uint8 r = pBuffer[2];
		pBuffer[0] = r;
		pBuffer[2] = b;

		pBuffer += 4;
	}

	ds.m_epf = pf32bppRgba;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
void POEngine::BgrToRgb(PngDumpSettings& ds)
{
	uint8* pBuffer = (uint8*) ds.m_pBuffer;
	int32 nPixelCount = ds.m_nWidth * ds.m_nHeight;

	for(int32 i = 0; i < nPixelCount; ++i)
	{
		uint8 b = pBuffer[0];
		uint8 r = pBuffer[2];
		pBuffer[0] = r;
		pBuffer[2] = b;

		pBuffer += 3;
	}

	ds.m_epf = pf24bppRgb;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
bool POEngine::Rgb16ToRgb24(PngDumpSettings& ds, ByteArray& aNewBuffer)
{
	uint8* pSrc = (uint8*) ds.m_pBuffer;
	int32 nPixelCount = ds.m_nWidth * ds.m_nHeight;

	if( !aNewBuffer.SetSize(nPixelCount * 3))
	{
		return false;
	}

	uint8* pDst = aNewBuffer.GetPtr();

	if( ds.m_epf == pf16bppRgb555)
	{
		for(int32 i = 0; i < nPixelCount; i++)
		{
			uint16 nPixel = uint16( (pSrc[0] << 8) | pSrc[1]);
			pSrc += 2;

			uint8 r = uint8((nPixel & 0x7c00) >> 10);
			uint8 g = uint8((nPixel & 0x03e0) >> 5);
			uint8 b = uint8((nPixel & 0x001f));

			r = uint8( (r << 3) + ((r >> 2) & 0x07));
			g = uint8( (g << 3) + ((g >> 2) & 0x07));
			b = uint8( (b << 3) + ((b >> 2) & 0x07));

			pDst[0] = uint8(r);
			pDst[1] = uint8(g);
			pDst[2] = uint8(b);
			pDst += 3;
		}
	}
	else if( ds.m_epf == pf16bppRgb565)
	{
		for(int32 i = 0; i < nPixelCount; i++)
		{
			uint16 nPixel = uint16( (pSrc[0] << 8) | pSrc[1]);
			pSrc += 2;

			uint8 r = uint8( (nPixel & 0xf800) >> 11 );
			uint8 g = uint8( (nPixel & 0x07e0) >> 5 );
			uint8 b = uint8( (nPixel & 0x001f) );

			r = uint8( (r << 3) + ((r >> 2) & 0x07));
			g = uint8( (g << 2) + ((g >> 1) & 0x03));
			b = uint8( (b << 3) + ((b >> 2) & 0x07));

			pDst[0] = uint8(r);
			pDst[1] = uint8(g);
			pDst[2] = uint8(b);
			pDst += 3;
		}
	}
	else
	{
		return false;
	}

	ds.m_epf = pf24bppRgb;
	ds.m_pBuffer = aNewBuffer.GetPtr();
	return true;
}

/////////////////////////////////////////////////////////////////////////////////////////////////
// Public version of the optimizing function
bool POEngine::OptimizeExternalBuffer(PngDumpSettings& ds, const String& strNewFileName)
{
	m_astrErrors.SetSize(0);
	m_resultmgr.Reset();

	return Optimize(ds, strNewFileName);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
bool POEngine::Optimize(PngDumpSettings& ds, const String& strNewFileName)
{
	ByteArray aTmpPixels;

	if( ds.m_epf == pf32bppBgra)
	{
		// Yuk ! Change this
		BgraToRgba(ds);
	}
	else if( ds.m_epf == pf24bppBgr)
	{
		// Yuk ! Change this aswell
		BgrToRgb(ds);
	}
	else if( ds.m_epf == pf16bppRgb555 || ds.m_epf == pf16bppRgb565)
	{
		// Change this, but harder to do this time because we need to reallocate
		if( !Rgb16ToRgb24(ds, aTmpPixels))
		{
			AddError(k_szNotEnoughMemoryForInternalConversion);
			return false;
		}
	}

	ds.m_nZLibCompressionLevel = 9; // Best compression

	PixelFormat pf = ds.m_epf;

	bool bOptimizeOk = false;

	if( pf1bppIndexed <= pf && pf <= pf8bppIndexed)
	{
		bOptimizeOk = OptimizePaletteMode(ds);
	}
	else if( pf == pf32bppRgba)
	{
		bOptimizeOk = Optimize32BitsMode(ds);
	}
	else if( pf == pf24bppRgb)
	{
		bOptimizeOk = Optimize24BitsMode(ds);
	}
	else if( pf1bppGrayScale <= pf && pf <= pf16bppGrayScale)
	{
		bOptimizeOk = OptimizeGrayScale(ds);
	}
	else if( pf == pf48bppRgb || pf == pf64bppRgba
	         || pf == pf16bppGrayScaleAlpha || pf == pf32bppGrayScaleAlpha )
	{
		// Other uncommon pixel formats
		
		// Test 1 : do a regular dump
		bOptimizeOk = Png::Dump(m_resultmgr.GetCandidate(), ds);

		if( bOptimizeOk )
		{
			// Test 2
			//  - ZLib strategy : filter
			ds.m_nZLibStrategy = PngDumpSettings::zlibStrategyFilter;

			bOptimizeOk = Png::Dump(m_resultmgr.GetCandidate(), ds);
		}
	}
	else
	{
		AddError(k_szUnsupportedPixelFormat);
		return false;
	}

	if( !bOptimizeOk)
	{
		return false;
	}

	return DumpBestResultToFile(strNewFileName);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
void POEngine::PrintText(const String& strText, enum TextColor eColor)
{
	m_eventProgressing.Fire(strText, eColor);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Build a new PNG from another one (given by file) excluding unwanted chunks
bool POEngine::InsertCleanOriginalPngAsResult(IFile& file)
{
	if( !Png::IsPng(file))
	{
		// Not a PNG
		ASSERT(0); // Programming error
		return false;
	}

	const int64 nFileSize = file.GetSize();
	const int32 nFileSize32 = int32(nFileSize);

	IFile& msIn = file;

	DynamicMemoryFile& dmf = m_resultmgr.GetCandidate();
	dmf.SetByteOrder(boBigEndian);
	
	// Ensure capacity
	if( !dmf.SetPosition(nFileSize32))
	{
		AddError(k_szNotEnoughMemoryToKeepOriginalFile);
		return false;
	}
	dmf.SetPosition(0);

	Png::WriteSignature(dmf);
		
	ChunkedFile chfIn(msIn);

	bool bOkHandled = true;
	bool bIENDFound = false;
	
	while( !bIENDFound)
	{
		bool bChunkPresent = chfIn.BeginChunkRead();
		if( !bChunkPresent)
		{
			break;
		}

		bool bKeepChunk = false;
		if( chfIn.m_nChunk == Png::chunk_bKGD && m_bKeepBackgroundColor)
		{
			// We keep that chunk
			bKeepChunk = true;
		}
		else
		{
			switch( chfIn.m_nChunk)
			{
			case Png::chunk_IEND:
				bIENDFound = true;
			case Png::chunk_IHDR:
			case Png::chunk_PLTE:
			case Png::chunk_IDAT:
			case Png::chunk_tRNS:
				bKeepChunk = true;
				break;

			default:
				break;
			}
		}

		if( bKeepChunk)
		{
			// We keep that chunk
			dmf.Write32( chfIn.m_nSizeof);
			dmf.Write32( chfIn.m_nChunk);
			dmf.WriteFromFile( chfIn, chfIn.m_nSizeof);

			if( !chfIn.EndChunkRead())
			{
				bOkHandled = false;	
			}
			else
			{
				// We bypassed the CRC computation, so we just write the one we read
				dmf.Write32(chfIn.m_nCrc);
			}
		}
		else
		{
			if( !chfIn.EndChunkRead())
			{
				// Unexpected end of stream, CRC not present or chunk mishandled (read too much)
				bOkHandled = false;
				break;
			}
		}
	}

	if( !(bOkHandled && bIENDFound))
	{
		AddError(k_szCorruptedChunkStructure);
		return false;
	}
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
bool POEngine::OptimizeFileNoBackup(const String& strFileName, const String& strNewFileName)
{
	m_astrErrors.SetSize(0);
	m_resultmgr.Reset();

	File fileImage;
	if( !fileImage.Open(strFileName, File::modeRead))
	{
		AddError(k_szCannotLoadFile);
		return false;
	}

	const int64 nFileSize = fileImage.GetSize();
	if( nFileSize > MAX_INT32)
	{
		AddError(k_szFileTooBig);
		return false;
	}

	///////////////////////////////////////////////////////////////////////////
	ImageLoader imgloader;
	if( !imgloader.InstanciateLosslessFormat(fileImage))
	{
		AddError(k_szUnsupportedFileFormat);
		return false;
	}

	ImageType& img = *imgloader.m_pImageType;
	bool bLoadOk = false;

	if( imgloader.m_nType == ImageLoader::typePng)
	{
		// Insert a clean version of the source PNG
		// "clean" means the same PNG expect some unwanted chunks (like the gamma chunk)
		
		// As we insert a copy of the source file, the source file becomes a candidate for the best result,
		// thus if we cannot achieve a better compression than the original file, we just dump the original file
		if( !InsertCleanOriginalPngAsResult(fileImage))
		{
			return false;
		}

		fileImage.Close(); // No need for the file anymore

		//////////////////////////////////////////////////////////
		// By the way, let's load the PNG from memory
		
		// GetSmallest returns the clean PNG copy we just inserted
		DynamicMemoryFile& dmf = m_resultmgr.GetSmallest();
		
		dmf.SetPosition(0);
		bLoadOk = img.LoadFromFile(dmf);
		dmf.SetPosition(nFileSize);
	}
	else
	{
		bLoadOk = img.LoadFromFile(fileImage);
		fileImage.Close();
	}

	if( !bLoadOk)
	{
		String strImgErr = img.GetLastErrorString();
		
		String strErr = L"Cannot load image : ";
		strErr = strErr + strImgErr;
		
		AddError(strErr);
		return false;
	}

	const int32 nWidth = img.GetWidth();
	const int32 nHeight = img.GetHeight();
	uint8* const pBuffer = (uint8*) img.GetBuffer();
	const Palette* pPalette = img.GetPalette();
	PixelFormat pf = img.GetPixelFormat();
	
	//////////////////////////////////////////////////////////////
	PngDumpSettings ds; // Final dump structure
	ds.m_nWidth = nWidth;
	ds.m_nHeight = nHeight;
	ds.m_nZLibCompressionLevel = 9; // Best compression
	ds.m_pBuffer = pBuffer;
	ds.m_pPalette = pPalette;
	ds.m_epf = pf;

	if( imgloader.m_nType == ImageLoader::typePng)
	{
		const Png& png = (const Png&) img;
		ds.m_bUseTransparentColor = png.IsSimpleTransparencyUsed();
		ds.m_nTransGrey = png.GetGreyTransIndex();
		png.GetTransIndexes(ds.m_nTransRed, ds.m_nTransGreen, ds.m_nTransBlue);

		if( m_bKeepInterlacing)
		{
			if( png.IsInterlaced())
			{
				ds.m_bInterlace = true;
			}
		}

		if( m_bKeepBackgroundColor)
		{
			// We only keep the background color for non-palette images
			// Why ?
			// 1. Tracking the correct background index through a palette reorganization is very tricky
			// 2. Keeping background color is used so 32 bits images are not blended with gray
			// in Internet Explorer 6
			//
			// So let's wait before a real need comes

			if( pf1bppIndexed <= pf && pf <= pf8bppIndexed)
			{
				// Give up
			}
			else
			{
				ds.m_bUseBackgroundColor = png.HasBackgroundColor();
				ds.m_nBkGrey = png.GetBkGrey();
				ds.m_nBkRed = png.GetBkRed();
				ds.m_nBkGreen = png.GetBkGreen();
				ds.m_nBkBlue = png.GetBkBlue();
			}
		}
	}
	else if(  imgloader.m_nType == ImageLoader::typeGif)
	{
		const Gif& gif = (const Gif&) img;
		if( gif.IsAnimated())
		{
			PrintText(L"[Animated GIF : converting first frame] ", tcGifFirstFrame);
		}
	}
	
	return Optimize(ds, strNewFileName);
}


///////////////////////////////////////////////////////////////////////////////////////////////////
void POEngine::OptimizeFileOrDir(const String& strFilePath, const String& strDisplayDir)
{
	m_astrErrors.SetSize(0);

	bool bSrcIsDir = false;
	bool bSrcIsReadOnly = false;
	if( !File::GetFileAttributes(strFilePath, bSrcIsDir, bSrcIsReadOnly))
	{
		// We cannot even read file attributes, so we just stop there for this file
		return;
	}

	// We need the name alone
	String strDirOnly, strNameOnly;
	File::SplitPath(strFilePath, strDirOnly, strNameOnly);

	if( bSrcIsDir)
	{
		// A directory, we optimize it before we continue
		StringArray astrSubFileNames = Directory::GetFileNames(strFilePath, L"*.*");

		// Set the directory to display
		String strNewDisplayDir;
		if( !strDisplayDir.IsEmpty())
		{
			strNewDisplayDir = strDisplayDir + L"/";
		}
		strNewDisplayDir = strNewDisplayDir + strNameOnly;

		OptimizeFiles(strFilePath, astrSubFileNames, strNewDisplayDir);
		return;
	}

	// Filter by extension
	String strExt = File::GetExtension(strFilePath);
	strExt = strExt.ToLowerCase();

	bool bExtOk = (strExt == L"png" || strExt == L"gif" || strExt == L"bmp" || strExt == L"tga");
	if( !bExtOk)
	{
		return;
	}
			
	///////////////////////////////////////////////////////////////////////
	// Write the waiting message during the optimization
	if( strExt == L"png")
	{
		PrintText(L"Optimizing ", tcActionVerb);
	}
	else
	{
		PrintText(L"Converting ", tcActionVerb);
	}
	///////////////////////////////////////////////////////////////////////
	
	if( !strDisplayDir.IsEmpty())
	{
		// Display the sub-directory
		PrintText(strDisplayDir + L"/", tcFilePath);
	}
	PrintText(strNameOnly, tcFilePath);
	
	PrintText(L"... ", tcRegularInfo);

	// Get the source file size
	uint32 nOldSize = (uint32) File::GetSize(strFilePath);

	strDirOnly = File::AddSeparator(strDirOnly);

	String strOldFilePath;
	String strNewFilePath;
	
	// If the source file is a PNG, we rename that source file with a "_" at the beginning of its name
	if( strExt == L"png")
	{
		if( bSrcIsReadOnly)
		{
			// File is read only, we won't be able to modify it
			AddError(k_szFileIsReadOnly);
			return;
		}

		if( m_bBackupOldPngFiles)
		{
			String strBackupFilePath = strDirOnly + L"_" + strNameOnly;

			// Delete a possible previous backup file
			if( File::Exists(strBackupFilePath))
			{
				if( !File::Delete(strBackupFilePath))
				{
					// We cannot delete the old backup file, we stop here
					AddError(k_szCannotPerformBackup);
					return;
				}
			}

			if( File::Rename(strFilePath, strBackupFilePath))
			{
				// Rename possible
				strOldFilePath = strBackupFilePath;
				strNewFilePath = strFilePath;
			}
			else
			{
				// We cannot rename for unknown reason
				AddError(k_szCannotPerformBackupRenameFailed);
				return;
			}
		}
		else
		{
			strOldFilePath = strFilePath;
			strNewFilePath = strFilePath;
		}
	}
	else
	{
		// Not a PNG file
		strOldFilePath = strFilePath;
		strNewFilePath = strDirOnly + File::RemoveExtension(strNameOnly) + L".png";
	}

	// TMP DEBUG : to comparer size before and after, uncomment the line above
	//strNewFilePath = File::AddFileNamePrefix(strNewFilePath, L"optimized-");

	if( OptimizeFileNoBackup(strOldFilePath, strNewFilePath))
	{
		PrintText(L" (OK) ", tcActionOk);

		uint32 nNewSize = (uint32) File::GetSize(strNewFilePath);

		if( nOldSize > 0 && nNewSize > 0)
		{
			uint32 nRatio = (nNewSize * 100) / nOldSize;
			
			String strOldSize = String::FromInt(nOldSize);
			String strNewSize = String::FromInt(nNewSize);
			String strRatio = String::FromInt(nRatio);

			StringBuilder sb;
			sb += strOldSize;
			sb += L" bytes ";
			sb += L"-> "; // Arrow to the right
			sb += strNewSize;
			sb += L" bytes ";

			PrintText(sb.ToString(), tcSizeInfo);

			sb.Empty();

			sb += '(';
			sb += strRatio;
			sb += L"% of the original size)\n"; // End of line here
			
			if( nNewSize > nOldSize )
			{
				PrintText(sb.ToString(), tcFileEnlarged);
			}
			else if( nNewSize == nOldSize )
			{
				PrintText(sb.ToString(), tcFileSizeSame);
			}
			else
			{
				PrintText(sb.ToString(), tcFileReduced);
			}
		}
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////
void POEngine::OptimizeFiles(const StringArray& astrFilePaths)
{
	OptimizeFiles(L"", astrFilePaths, L"");
}

///////////////////////////////////////////////////////////////////////////////////////////////////
void POEngine::OptimizeFiles(const String& strDir, const StringArray& astrFileNames)
{
	OptimizeFiles(strDir, astrFileNames, L"");
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// strBaseDir :    directory name containing the files
// astrFileNames : file path relative to base dir
// strDisplayDir : directory name to display
void POEngine::OptimizeFiles(const String& strBaseDir, const StringArray& astrFileNames, const String& strDisplayDir)
{
	const int32 nFileCount = astrFileNames.GetSize();
	for(int32 iFile = 0; iFile < nFileCount; ++iFile)
	{
		String strFilePath = File::AddSeparator(strBaseDir, L'\\') + astrFileNames[iFile];

		OptimizeFileOrDir(strFilePath, strDisplayDir);
			
		String strLastError = GetLastErrorString();
		if( !strLastError.IsEmpty())
		{
			PrintText(L" (KO) ", tcActionFail);
			PrintText( strLastError + L"\n", tcErrorMsg);

			m_astrErrors.RemoveAll();
		}
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////
void POEngine::AddError(const String& str)
{
	m_astrErrors.Add(str);
}

String POEngine::GetLastErrorString() const
{
	const int32 nErrCount = m_astrErrors.GetSize();
	if( nErrCount == 0 )
	{
		return String();
	}

	String str = m_astrErrors[0];
	for(int32 i = 1; i < nErrCount; ++i)
	{
		str = str + L"\n" + m_astrErrors[i];
	}
	return str;
}
