///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#ifndef CHUSTD_ARRAY_H
#define CHUSTD_ARRAY_H

#include "Master.h"
#include "Memory.h"
#include "String.h"

// The "new" operator has been redefined in CppExtension.h with CHUDEBUG_NEW
// We undef the "new" preprocessor macro so we can use the real new to
// perform "placement new"s. The macro is defined again at the end of this file
#ifdef CHUDEBUG_NEW
#define ARRAY_OLDNEW 1
#undef new
#endif

///////////////////////////////////////////////////////////////
// Already defined placement new and delete in the file <new>
inline void* __cdecl operator new(size_t, void* p)
{
	return p;
}

inline void __cdecl operator delete(void *, void *)
{
	return;
}
///////////////////////////////////////////////////////////////

namespace chustd {

// Array: manages a linear dynamic array of elements:
// classes, structures, atomic types
// Warning:
// - stored classes must have a default constructor
// - do not use pointer or references on element of the array if you plan
// to resize the array
// If you want to keep pointers valid, you may use CPtrArray instead
template <class T> class Array
{
public:
	// Adds an element at the end of the array
	int32 Add();
	int32 Add(const T& t);
	int32 Add(const T* paTs, int32 nCount);
	int32 Add(const Array<T>& arr);

	// Adds an unexisting element to the list
	// Returns true if the element could be added
	bool AddUnique(const T& t);

	// Initializes the array with an external array
	// The array is resized and contains copies of the given values
	void Set(const T* paTs, int32 nCount);

	// Initializes all element with a value
	void Set(const T& t);

	// Gets the number of elements stored
	inline int32 GetSize() const { return m_nCount; }

	// Returns true if the array contains no element
	inline bool IsEmpty() const { return ( m_nCount == 0 ); }

	// Sets the number of elements (does not free memory, see RemoveAll and FreeExtra)
	// Returns true upon success
	bool SetSize(int32 nSize);
	
	// Gets an element at a specific position
	inline T& operator [] (int32 nIndex)
	{
		ASSERTMSG(nIndex >= 0 && nIndex < m_nCount, "Array::[] - Bad index");
		return m_paTs[nIndex];
	}

	// Gets an element at a specific position
	inline const T& operator [] (int32 nIndex) const
	{
		ASSERTMSG(nIndex >= 0 && nIndex < m_nCount, "Array::[] - Bad index");
		return m_paTs[nIndex];
	}

	// Gets an element at a specific position
	inline T& GetAt(int32 nIndex)
	{
		ASSERTMSG(nIndex >= 0 && nIndex < m_nCount, "Array::GetAt - Bad index");
		return m_paTs[nIndex];
	}

	// Gets an element at a specific position
	inline const T& GetAt(int32 nIndex) const
	{
		ASSERTMSG(nIndex >= 0 && nIndex < m_nCount, "Array::GetAt - Bad index");
		return m_paTs[nIndex];
	}


	const T& GetLast() const
	{
		ASSERTMSG(m_nCount > 0, "Array::GetLast - Count is 0");
		return m_paTs[m_nCount - 1];
	}

	T& GetLast()
	{
		ASSERTMSG(m_nCount > 0, "Array::GetLast - Count is 0");
		return m_paTs[m_nCount - 1];
	}

	// Removes an element at a specific position
	// Extra elements are shifted in order to cap the hole
	void RemoveAt(int32 nIndex);
	
	// Removes a given element of the list
	// Returns true if the element could be found
	bool Remove(const T& t);

	/// Removes the last element of the list
	void RemoveLast();

	// Removes all the elements and frees all memory allocated
	void RemoveAll();

	// Gets the position of an element
	// Returns -1 if no element found
	int32 Find(const T& t) const;
	int32 Find(const T* paTs, int32 nCount) const;

	// Inserts an element at a specific position
	void InsertAt(int32 nIndex);
	void InsertAt(int32 nIndex, const T& t);

	void InsertUniqueAt(int32 nIndex, const T& t);

	// Moves all elements from the current array to another array
	void MoveTo(Array<T>& aReceiver);

	// Swap the content of two arrays
	void Swap(Array<T>& aFriend);

	// Swap two values
	void Swap(int32 nIndex1, int32 nIndex2);
	
	// Gets a pointer on the first element of the array
	inline T* GetBuffer() const { return m_paTs; }
	inline T* GetPtr() const { return m_paTs; }

	inline int32 SizeofElement() { return sizeof(T); }

	///////////////////////////////////////////
	// Sets the minimum wanted allocation
	// Returns true if the allocation could be done
	bool EnsureCapacity(int32 nCapacity)
	{
		return EnsureCapacityEx(nCapacity, m_nCount, true);
	}
	
	// Gets the number of allocated elements
	int32 GetCapacity() const { return m_nRbvAndAllocated & 0x7fffffff; }

	// Remove all extra allocated memory
	void FreeExtra();

	void SetAsReturnValue() { m_nRbvAndAllocated |= 0x80000000; }
	bool IsReturnedByValue() const { return (m_nRbvAndAllocated & 0x80000000) != 0; }

	///////////////////////////////////////////
	// Makes a copy of another array
	Array<T>& operator = (const Array<T>& arr);

	Array();
	Array(int32 nCount);
	Array(const Array<T>& arr);
	Array(const T* paTs, int32 nCount);
	virtual ~Array();

protected:
	T* m_paTs;          // 'T' buffer
	int32 m_nCount;     // Number of elements in the array
	
	// Packed value
	// m_nRbvAndAllocated & 0x80000000 : boolean set to 1 to avoid a full copy when passing by value (Rbv)
	// m_nRbvAndAllocated & 0x7fffffff : number of allocated elements (but not necessarely constructed)
	uint32 m_nRbvAndAllocated;
	
protected:
	inline void CallConstructors(int32 nStartIndex, int32 nCount);
	inline void CallDestructors(int32 nStartIndex, int32 nCount);

	virtual void DeletePointedData(int32 nStartIndex, int32 nCount);

private:
	// When one element is added to the array, this compute how much memory should be allocated
	static int32 GetMinCapacity();
	bool EnsureCapacityEx(int32 nCapacity, int32 nElementToCopy, bool bFreeMemory = true);
	void InsertAtNoConstruct(int32 nIndex);
};

template <class T>
Array<T>::Array()
{
	m_paTs = null;
	m_nCount = 0;
	m_nRbvAndAllocated = 0;
}

template <class T>
Array<T>::Array(int32 nCount)
{
	m_paTs = null;
	m_nCount = 0;
	m_nRbvAndAllocated = 0;

	if( !SetSize(nCount))
	{
		ASSERT(0); // Better than nothing
	}
}

template <class T>
Array<T>::Array(const Array<T>& arr)
{
	if( arr.IsReturnedByValue())
	{
		m_paTs = arr.m_paTs;
		m_nCount = arr.m_nCount;
		m_nRbvAndAllocated = arr.GetCapacity(); // Set the rbv bit to 0

		// Bad, bad, but well, advanced C++
		Array<T>& arr2 = (Array<T>&) arr;

		arr2.m_paTs = null;
		arr2.m_nCount = 0;
		arr2.m_nRbvAndAllocated = 0;
	}
	else
	{
		m_paTs = null;
		m_nCount = 0;
		m_nRbvAndAllocated = 0;

		Add(arr.m_paTs, arr.m_nCount);
	}
}

template <class T>
Array<T>::Array(const T* paTs, int32 nCount)
{
	m_paTs = null;
	m_nCount = 0;
	m_nRbvAndAllocated = 0;

	Add(paTs, nCount);
}

template <class T>
Array<T>::~Array()
{
	RemoveAll();
}

template <class T>
void Array<T>::RemoveAll()
{
	if(m_paTs == null)
		return;

	CallDestructors(0, m_nCount);
	Memory::Free64(m_paTs);

	m_paTs = null;
	m_nCount = 0;
	m_nRbvAndAllocated = 0;
}

template <class T>
void Array<T>::MoveTo(Array<T>& aReceiver)
{
	aReceiver.RemoveAll();
	aReceiver.m_nCount = m_nCount;
	aReceiver.m_nRbvAndAllocated = GetCapacity(); // The rbv bit is set to 0
	aReceiver.m_paTs = m_paTs;

	m_paTs = null;
	m_nCount = 0;
	m_nRbvAndAllocated = 0;
}

template <class T>
void Array<T>::Swap(Array<T>& aFriend)
{
	// Swap nCount
	int32 nTmp = m_nCount;
	m_nCount = aFriend.m_nCount;
	aFriend.m_nCount = nTmp;

	// Swap m_nRbvAndAllocated, keep the original rbv bit
	nTmp = m_nRbvAndAllocated;
	m_nRbvAndAllocated &= 0x80000000;
	m_nRbvAndAllocated |= (aFriend.m_nRbvAndAllocated & 0x7fffffff);
	
	aFriend.m_nRbvAndAllocated &= 0x80000000;
	aFriend.m_nRbvAndAllocated |= (nTmp & 0x7fffffff);

	// Swap buffers
	T* pTmp = m_paTs;
	m_paTs = aFriend.m_paTs;
	aFriend.m_paTs = pTmp;
}

template <class T>
void Array<T>::Swap(int32 nIndex1, int32 nIndex2)
{
	ASSERTMSG( 0 <= nIndex1 && nIndex1 <= m_nCount, "Array::Swap - Bad index 1");
	ASSERTMSG( 0 <= nIndex2 && nIndex2 <= m_nCount, "Array::Swap - Bad index 2");

	Memory::Swap(m_paTs + nIndex1, m_paTs + nIndex2, sizeof(T));
}

// Assignment operator
template <class T>
Array<T>& Array<T>::operator = (const Array<T>& arr)
{
	if( this == &arr )
		return *this;

	if( m_nCount < arr.m_nCount)
	{
		// For each element call the assignment operator
		for(int32 i = 0; i < m_nCount; i++)
		{
			m_paTs[i] = arr.m_paTs[i];
		}
		
		// Construct new elements by copy
		Add(arr.m_paTs + m_nCount, arr.m_nCount - m_nCount);
	}
	else
	{
		// THIS array is bigger than the OTHER array

		// Destroy extra elements
		int32 nDestroyStartIndex = arr.m_nCount;
		int32 nDestroyCount = m_nCount - arr.m_nCount;
		CallDestructors(nDestroyStartIndex, nDestroyCount);

		// For each element call the assignment operator
		for(int32 i = 0; i < arr.m_nCount; i++)
		{
			m_paTs[i] = arr.m_paTs[i];
		}
		m_nCount = arr.m_nCount;
	}
	
	return *this;
}

// Adds an empty element at the end of the array
// Returns the index of the added element, -1 upon error
template <class T>
int32 Array<T>::Add()
{
	if( GetCapacity() <= m_nCount)
	{
		// Not enough memory, must reallocate
		int32 nNewCapacity = (m_nCount == 0) ? 1 : m_nCount * 2;
		if( !EnsureCapacity(nNewCapacity))
		{
			return -1;
		}
	}

	CallConstructors(m_nCount, 1);

	const int32 nCurrentPos = m_nCount;
	m_nCount++;

	return nCurrentPos;
}

// Adds an element to the end of the array
// Returns the index of the added element, -1 upon error
template <class T>
int32 Array<T>::Add(const T& t)
{
	if( GetCapacity() <= m_nCount)
	{
		// Not enough memory, must reallocate
		int32 nNewCapacity = (m_nCount == 0) ? 1 : m_nCount * 2;
		if( !EnsureCapacity(nNewCapacity))
		{
			return -1;
		}
	}

	// Call the copy constructor for the new array element
	::new( (void*)(m_paTs + m_nCount)) T(t);

	const int32 nCurrentPos = m_nCount;
	m_nCount++;

	return nCurrentPos;
}

// Adds several elements to the end of the array
// Returns the index of the last added element, -1 upon error
template <class T>
int32 Array<T>::Add(const T* paTs, int32 nCount)
{
	if( nCount == 0)
		return m_nCount - 1;

	if( !EnsureCapacity(m_nCount + nCount))
	{
		return -1;
	}
	T* pNew = m_paTs + m_nCount;

	for(int32 i = 0; i < nCount; i++)
	{
		// Call the copy constructor for the new array element
		::new( (void*)(pNew + i)) T( paTs[i] );
	}
	m_nCount += nCount;
	return m_nCount - 1;
}

// Adds several elements to the end of the array
// Returns the index of the last added element
template <class T>
int32 Array<T>::Add(const Array<T>& arr)
{
	return Add(arr.m_paTs, arr.m_nCount);
}

template <class T>
void Array<T>::Set(const T* paTs, int32 nCount)
{
	if( !SetSize(nCount))
	{
		ASSERT(0); // Better than nothing
		return;
	}

	// Use the assignment operator
	for(int32 i = 0; i < m_nCount; i++)
	{
		m_paTs[i] = paTs[i];
	}
}

template <class T>
void Array<T>::Set(const T& t)
{
	for(int32 i = 0; i < m_nCount; i++)
	{
		m_paTs[i] = t;
	}
}

// Remove the element at a specific index
template <class T>
void Array<T>::RemoveAt(int32 nIndex)
{
	ASSERTMSG(nIndex >= 0 && nIndex < m_nCount, "Array::RemoveAt - Bad index");
	
	// Number of element on the right of the element to remove, including the element to remove
	const int32 nRightCount = m_nCount - nIndex;
	
	// Destroy the element to be removed
	CallDestructors(nIndex, 1);
	if( nRightCount > 1)
	{
		// Translate the last elements to the left
		// memove is needed instead of memcpy in our case
		Memory::Move(m_paTs + nIndex, m_paTs + nIndex + 1, sizeof(T) * (nRightCount - 1) );
	}
	m_nCount--;
}

template <class T>
void Array<T>::InsertAtNoConstruct(int32 nIndex)
{
	const int32 nShiftCount = m_nCount - nIndex; // Number of element to translate

	// Is there any free space so we can translate the current content of the array ?
	if( GetCapacity() <= m_nCount)
	{
		// Not enough memory
		T* pOld = m_paTs;

		int32 nNewCapacity = (m_nCount == 0) ? 1 : m_nCount * 2;
		if( !EnsureCapacityEx(nNewCapacity, nIndex, false))
		{
			ASSERT(0); // Better than nothing
			return;
		}

		if( nShiftCount > 0)
		{
			Memory::Copy(m_paTs + nIndex + 1, pOld + nIndex, sizeof(T) * nShiftCount);
		}
		Memory::Free64(pOld);
	}
	else
	{
		if( nShiftCount > 0)
		{
			// Element translation to the right
			Memory::Move(m_paTs + nIndex + 1, m_paTs + nIndex, sizeof(T) * nShiftCount);
		}
	}
}

template <class T>
void Array<T>::InsertAt(int32 nIndex)
{
	ASSERTMSG(nIndex >= 0 && nIndex <= m_nCount, "Array::InsertAt - Bad index");
	if( nIndex == m_nCount)
	{
		Add();
		return;
	}

	InsertAtNoConstruct(nIndex);

	CallConstructors(nIndex, 1);

	m_nCount++;
}

template <class T>
void Array<T>::InsertAt(int32 nIndex, const T& t)
{
	ASSERTMSG(nIndex >= 0 && nIndex <= m_nCount, "Array::InsertAt - Bad index");
	if( nIndex == m_nCount)
	{
		Add(t);
		return;
	}

	InsertAtNoConstruct(nIndex);

	// Call the copy constructor for the new array element
	::new( (void*)(m_paTs + nIndex)) T(t);

	m_nCount++;
}

template <class T>
void Array<T>::InsertUniqueAt(int32 nIndex, const T& t)
{
	ASSERTMSG(nIndex >= 0 && nIndex <= m_nCount, "Array::InsertAt - Bad index");
	if( nIndex == m_nCount)
	{
		AddUnique(t);
		return;
	}

	if( Find(t) >= 0)
	{
		// Already exists
		return;
	}

	InsertAtNoConstruct(nIndex);

	// Call the copy constructor for the new array element
	::new( (void*)(m_paTs + nIndex)) T(t);

	m_nCount++;
}

template <class T>
void Array<T>::FreeExtra()
{
	if( GetCapacity() == m_nCount)
		return;

	// If GetCapacity() == 0, it means m_nCount is 0. So the previous test check this case.
	// If m_nCount is not 0, the object may by corrupted.
	ASSERT( GetCapacity() > 0);

	if( m_nCount > 0)
	{
		// Some data to copy
		const int32 nByteCount = sizeof(T) * m_nCount;
		void* paNewTs = Memory::Alloc64(nByteCount);
		if( paNewTs == null)
		{
			ASSERT(0); // Better than nothing
			return;
		}

		// Copy of the old element into the new array
		int32 nInt64Count = Memory::ByteCountToInt64Count(nByteCount);
		Memory::Copy64(paNewTs, m_paTs, nInt64Count);
	
		// Free the memory used by the old array
		Memory::Free64(m_paTs);

		m_paTs = (T*) paNewTs;
	}
	else
	{
		// The array is empty
		Memory::Free64(m_paTs);
		m_paTs = null;
	}
	
	// Keep the rbv bit
	m_nRbvAndAllocated = m_nCount | (m_nRbvAndAllocated & 0x80000000);
}

template <class T>
int32 Array<T>::GetMinCapacity()
{
	// 480 bytes or 2 elements is the first size
	// 480 bytes -> 240 bytes for each element
	// With 480 bytes this allows an overhead of 32 bytes to reach 512 bytes,
	// thus making the memory allocator happy

	int32 nWantsByteCount = 480;
	int32 nAllocate = nWantsByteCount / sizeof(T);
	if( nAllocate < 2)
	{
		nAllocate = 2;
	}
	return nAllocate;
}

// Allocate as much memory as wanted if necessary
// If allocation is done, previous elements are copied to the new buffer
// nCapacity : allocation wanted in number of elements
// nElementToCopy : number of elements to copy from the old buffer (used for insertion)
// bFreeMemory : true to free the previous buffer (used for insertion)
// bAllocateOnZero : true to allocate when Capacity is zero (used for add and insertion)
// Returns true upon success (allocation succeed)
template <class T>
bool Array<T>::EnsureCapacityEx(int32 nCapacity, int32 nElementToCopy, bool bFreeMemory /* = true */)
{
	ASSERT(nCapacity >= 0);

	if( GetCapacity() < nCapacity)
	{
		// Needs reallocation

		//////////////////////////////////////////////////////////
		// Check that we do not allocate a too small buffer
		int32 nMinCount = GetMinCapacity();
		if( nCapacity < nMinCount)
		{
			nCapacity = nMinCount;
		}
		//////////////////////////////////////////////////////////

		int32 nNewByteCount = nCapacity * sizeof(T);
		void* paNewTs = Memory::Alloc64(nNewByteCount);
		if( paNewTs == null)
		{
			return false;
		}

		if( nElementToCopy > 0)
		{
			// Copy of the old element into the new array
			const int32 nOldByteCount = sizeof(T) * nElementToCopy;
			int32 nInt64Count = Memory::ByteCountToInt64Count(nOldByteCount);
			Memory::Copy64(paNewTs, m_paTs, nInt64Count);
		}

		// bFreeMemory is used for insertion functions :
		// one should not free memory when inserting elements, because the second part of the old
		// array must be copied
		if( GetCapacity() > 0 && bFreeMemory)
		{
			// Free the memory used by the old array
			Memory::Free64(m_paTs);
		}

		m_paTs = (T*) paNewTs;
		
		// Keep the rbv bit
		m_nRbvAndAllocated = nCapacity | (m_nRbvAndAllocated & 0x80000000);
	}
	return true;
}


// Shrinks or narrows the array
template <class T>
bool Array<T>::SetSize(int32 nSize)
{
	ASSERT(nSize >= 0);

	if( nSize == m_nCount)
		return true;
	
	if( nSize > 0)
	{
		if( !EnsureCapacity(nSize))
		{
			return false;
		}
	}

	if( nSize < m_nCount)
	{
		// Shrinking, must destroy some elements
		CallDestructors(nSize, m_nCount - nSize);
	}
	else
	{
		// Growing, must construct some elements
		CallConstructors(m_nCount, nSize - m_nCount);
	}
	m_nCount = nSize;
	return true;
}

template <class T>
inline void Array<T>::CallConstructors(int32 nStartIndex, int32 nCount)
{
	T* paEls = m_paTs + nStartIndex;

	for(int32 i = 0; i < nCount; ++i)
	{
		T* pElToConstruct = paEls + i;
		::new( (void*)(pElToConstruct)) T;
	}
}

template <class T>
inline void Array<T>::CallDestructors(int32 nStartIndex, int32 nCount)
{
	DeletePointedData(nStartIndex, nCount);

	for(int32 i = 0; i < nCount; ++i)
	{
		(m_paTs + nStartIndex + i)->~T(); // Do not forget the parenthesis !
	}
}

template <class T>
void Array<T>::DeletePointedData(int32 /*nStartIndex*/, int32 /*nCount*/)
{

}

// Returns true if the element could be inserted
template <class T>
bool Array<T>::AddUnique(const T& t)
{
	for(int32 i = 0; i < m_nCount; ++i)
	{
		if( m_paTs[i] == t)
			return false; // Element already there
	}
	
	Add(t);
	return true;
}

// Returns true if the element could be removed
template <class T>
bool Array<T>::Remove(const T& t)
{
	for(int32 i = 0; i < m_nCount; ++i)
	{
		if( m_paTs[i] == t)
		{
			// Found !
			RemoveAt(i);
			return true;
		}
	}
	return false;
}

template <class T>
void Array<T>::RemoveLast()
{
	if( m_nCount == 0 )
	{
		return;
	}

	const int nLastIndex = m_nCount - 1;
	RemoveAt(nLastIndex);
}



template <class T>
int32 Array<T>::Find(const T& t) const
{
	for(int32 i = 0; i < m_nCount; ++i)
	{
		if( m_paTs[i] == t)
		{
			// Found !
			return i;
		}
	}
	return -1;
}

template <class T>
int32 Array<T>::Find(const T* paTs, int32 nCount) const
{
	const int32 nMaxMain = m_nCount - nCount;
	if( nMaxMain < 0)
		return -1;

	for(int32 i = 0; i <= nMaxMain; ++i)
	{
		int32 j = 0;
		for(; j < nCount; ++j)
		{
			if( m_paTs[i + j] != paTs[j])
			{
				break;
			}
		}

		if( j == nCount)
			return i;
	}
	return -1;
}

/////////////////////////////////////////////////////////////////////////////////////
// Predefined arrays
class ByteArray : public Array<uint8>
{
public:
	ByteArray() {}
	~ByteArray() {}
};

class Int32Array : public Array<int32>
{
public:
	Int32Array() {}
	~Int32Array() {}
};

class UInt32Array : public Array<uint32>
{
public:
	UInt32Array() {}
	~UInt32Array() {}
};

class StringArray : public Array<chustd::String>
{
public:
	StringArray() {}
	~StringArray() {}
};


// This class is used to create a ByteArray from an external buffer
// TODO : maybe we need a better architecture to convert external buffers to ByteArrays
class ByteArrayBox
{
public:
	ByteArrayBox(uint8* pExternalBuffer, int32 nSize);
	virtual ~ByteArrayBox(); // TODO : remove the virtual when ByteArray will be a self-contained class with no virtual

	const ByteArray& ToByteArray() const;

	void Set(uint8* pExternalBuffer, int32 nSize);

private:
	uint8* m_paTs;      // byte buffer
	int32 m_nCount;     // Number of elements in the array
	
	// Packed value
	// m_nRbvAndAllocated & 0x80000000 : boolean set to 1 to avoid a full copy when passing by value (Rbv)
	// m_nRbvAndAllocated & 0x7fffffff : number of allocated elements (but not necessarely constructed)
	uint32 m_nRbvAndAllocated;
};
/////////////////////////////////////////////////////////////////////////////////////

} // namespace chustd

/////////////////////////////////////////////////////////////////////////////////////

// Redefined the CHUDEBUG_NEW macro
#ifdef ARRAY_OLDNEW
#define new CHUDEBUG_NEW
#endif

#endif // ndef CHUSTD_ARRAY_H

