///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "Master.h"
#include "Bmp.h"
#include "Array.h"
#include "Memory.h"
#include "Math.h"

//////////////////////////////////////////////////////////////////////
using namespace chustd;

static const uint8 k_BmpSignature[2] = {'B', 'M'};
//////////////////////////////////////////////////////////////////////

void Bmp::BmpHeader::SwapBytes()
{
	nFileSize =     IFile::Swap32(nFileSize);
	nReserved =     IFile::Swap32(nReserved);
	nOffsetToData = IFile::Swap32(nOffsetToData);
	nBitmapHeaderSize = IFile::Swap32(nBitmapHeaderSize);
	nWidth =  IFile::Swap32(nWidth);
	nHeight = IFile::Swap32(nHeight);
	nPlanes = IFile::Swap16(nPlanes);
	nDepth =  IFile::Swap16(nDepth);
	nCompression =     IFile::Swap32(nCompression);
	nBitmapDataSize =  IFile::Swap32(nBitmapDataSize);
	nXPelsPerMeter =   IFile::Swap32(nXPelsPerMeter);
	nYPelsPerMeter =   IFile::Swap32(nYPelsPerMeter);
	nColorCount =      IFile::Swap32(nColorCount);
	nImportantColors = IFile::Swap32(nImportantColors);
}

//////////////////////////////////////////////////////////////////////
Bmp::Bmp()
{
	Initialize();	
}

Bmp::~Bmp()
{

}

void Bmp::Initialize()
{
	m_nWidth = 0;
	m_nHeight = 0;
	m_nLastError = 0;

	m_epf = pfUnknown;
}

bool Bmp::IsBmp(IFile& file)
{
	const int32 nSignSize = sizeof(k_BmpSignature);
	uint8 aRead[nSignSize];
		
	if( file.Read(aRead, nSignSize) != nSignSize )
	{
		return false;
	}

	if( !chustd::Memory::IsEqual(k_BmpSignature, aRead, nSignSize) )
	{
		return false;
	}
	return true;
}


bool Bmp::LoadFromFile(IFile& file)
{
	/////////////////////////////////////////////////////
	// Reset the object
	FreeBuffer();
	Initialize();

	// The BMP format is little endian
	TmpEndianMode tem(file, boLittleEndian);

	/////////////////////////////////////////////////////
	// Read the header first
	if( !IsBmp(file) )
	{
		m_nLastError = errNotABmpFile;
		return false;
	}

	BmpHeader bh;
	const int32 nHeaderSize = sizeof(BmpHeader);
	if( file.Read(&bh, nHeaderSize) != nHeaderSize )
	{
		m_nLastError = uncompleteFile;
		return false;
	}

	if( file.ShouldSwapBytes() )
	{
		bh.SwapBytes();
	}

	////////////////////////////////////////////
	// Determine the version
	int32 nBmpVersion = 0;
	if( bh.nBitmapHeaderSize < 40 )
	{
		m_nLastError = errBadHeaderSize;
		return false;
	}

	if( bh.nBitmapHeaderSize == 40 )
	{
		nBmpVersion = 3;
	}
	else
	{
		nBmpVersion = 4;
	}
	////////////////////////////////////////////


	m_nWidth = bh.nWidth;
	m_nHeight = bh.nHeight;

	bool bRowOrderTopToBottom = false;
	if( m_nHeight < 0 )
	{
		// The row order is top to bottom
		bRowOrderTopToBottom = true;
		m_nHeight = -m_nHeight;
	}

	if( !( 0 <= bh.nCompression && bh.nCompression <= compBitfields) )
	{
		m_nLastError = errUnsupportedCompressionFormat;
		return false;
	}

	if( bh.nCompression == compRle8bits || bh.nCompression == compRle4bits )
	{
		if( bh.nDepth != 4 && bh.nDepth != 8 )
		{
			m_nLastError = errUnconsistentDepthAndCompression;
			return false;
		}
	}

	if( bh.nCompression == compBitfields )
	{
		if( bh.nDepth != 16 && bh.nDepth != 32 )
		{
			m_nLastError = errUnconsistentDepthAndCompression;
			return false;
		}
	}

	if( bh.nDepth <= 8 )
	{
		if( bh.nColorCount == 0 )
		{
			// Valid value, means use the maximum possible number of colors
			// for the current depth
			bh.nColorCount = 1 << bh.nDepth;
		}

		if( bh.nColorCount > 256 )
		{
			// Too many colors
			m_nLastError = errUnconsistentDepthAndColorCount;
			return false;
		}
	}
	else
	{
		// True colors picture, reset the nColorCount value which is not always 0
		// Some picture editors put 16777216 inside
		bh.nColorCount = 0;
	}

	////////////////////////////////////////////////
	// Read the palette
	uint8 aColors[256 * 4];
	const int32 nPaletteSize = bh.nColorCount * 4;
	if( file.Read(aColors, nPaletteSize) != nPaletteSize )
	{
		m_nLastError = uncompleteFile;
		return false;
	}
	
	m_ColorTable.m_nCount = bh.nColorCount;

	for(int32 i = 0; i < bh.nColorCount; ++i)
	{
		uint8 b = aColors[4 * i + 0];
		uint8 g = aColors[4 * i + 1];
		uint8 r = aColors[4 * i + 2];
		
		m_ColorTable.m_aColors[i].SetRgb(r, g, b);
	}
	
	////////////////////////////////////////////////
	// Get the pixel format
	PixelFormat pf = GetPfFromDepth(bh.nDepth);
	
	////////////////////////////////////////////////
	// Refine the pixel format
	uint32 nMaskRed = 0;
	uint32 nMaskGreen = 0;
	uint32 nMaskBlue = 0;
	uint32 nMaskAlpha = 0;

	if( bh.nCompression == compBitfields )
	{
		// Read the masks
		if( !(file.Read32(nMaskRed) && file.Read32(nMaskGreen) && file.Read32(nMaskBlue)) )
		{
			m_nLastError = uncompleteFile;
			return false;
		}

		if( nBmpVersion >= 4 )
		{
			file.Read32(nMaskAlpha);
		}

		if( bh.nDepth == 16 )
		{
			if( nMaskRed == 0x7C00 && nMaskGreen == 0x03E0 && nMaskBlue == 0x001F )
			{
				pf = pf16bppRgb555;
			}
			else if( nMaskRed == 0xF800 && nMaskGreen == 0x07E0 && nMaskBlue == 0x001F )
			{
				pf = pf16bppRgb565;
			}
			else
			{
				m_nLastError = errUnsupportedBitfieldsFormat;
				return false;
			}
		}
		else if( bh.nDepth == 32 )
		{
			if( nMaskRed == 0x00FF0000 && nMaskGreen == 0x0000FF00 && nMaskBlue == 0x000000FF )
			{
				pf = pf32bppBgra;
			}
			else
			{
				m_nLastError = errUnsupportedBitfieldsFormat;
				return false;
			}
		}
		else
		{
			m_nLastError = errUnsupportedBitfieldsFormat;
			return false;
		}
	}
	
	m_epf = pf;
	//////////////////////////////////////////////////

	///////////////////////////////////////////////////
	// Read the pixels
	
	int64 nCurPos = file.GetPosition();
	file.SetPosition(bh.nOffsetToData);

	bool bReadOk = ReadPixelData(file, bh);
	if( bReadOk )
	{
		if( pf == pf32bppBgra && nMaskAlpha != 0xFF000000 )
		{
			// The byte used for alpha is not used, so we set every alpha bytes to 255
			SetAlphaFullOpaque();
		}

		if( !bRowOrderTopToBottom )
		{
			// Flip is needed
			FlipVertical();
		}
	}

	file.SetPosition(bh.nFileSize);	
	return bReadOk;
}


bool Bmp::ReadPixelData(IFile& file, const BmpHeader& bh)
{
	uint16 nDepth = bh.nDepth;

	if( nDepth == 32 )
	{
		if( !ReadPixelsFromFile(file) )
		{
			return false;
		}

		return true;
	}
	
	if( nDepth == 24 )
	{
		if( !ReadPixelsFromFile(file) )
		{
			return false;
		}

		RemovePadding();
		return true;
	}
	
	if( nDepth == 16 )
	{
		if( !ReadPixelsFromFile(file) )
		{
			return false;
		}
		RemovePadding();

		// Arrange values from GGGBBBBB_ARRRRRGG to ARRRRRGG_GGGBBBBB in memory
		const int32 nPixelCount = m_nWidth * m_nHeight;
		IFile::Swap16( (uint16*) m_aPixelBuffer.GetPtr(), nPixelCount);

		return true;
	}
	
	if( nDepth == 8 || nDepth == 4 || nDepth == 1 )
	{
		// There is padding at the end of each row in chustd image buffers for depth < 8
		const int32 nNiceBytePerRow = ImageType::ComputeByteWidth(m_epf, m_nWidth);

		///////////////////////////////////////////////////////////////
		// Preparing Src values

		int32 nSrcBytesPerRow = ComputeBmpBytesPerRow(nNiceBytePerRow);
		int32 nSrcRowPadding = nSrcBytesPerRow - nNiceBytePerRow;

		int32 nDataSize = nSrcBytesPerRow * m_nHeight;
		
		ByteArray rbSrc;
		rbSrc.SetSize(nDataSize);

		if( bh.nCompression == compRle8bits || bh.nCompression == compRle4bits )
		{
			// Should uncompress first
			int32 nCompressedSize = bh.nFileSize - (int32) file.GetPosition();
			
			ByteArray rbCompressedData;
			rbCompressedData.SetSize(nCompressedSize);
			uint8* const pCompressedData = rbCompressedData.GetPtr();

			const int32 nActuallyRead = file.Read(pCompressedData, nCompressedSize);
			if( nActuallyRead != nCompressedSize )
			{
				m_nLastError = errUnexpectedEndOfFile;
				return false;
			}

			if( bh.nCompression == compRle4bits )
			{
				// RleUncompress works with a destination buffer of 8 bits pixels
				rbSrc.SetSize( rbSrc.GetSize() * 2);
			}

			if( !RleUncompress(rbCompressedData, rbSrc, m_nWidth, int32(bh.nDepth)) )
			{
				m_nLastError = errBadCompressedData;
				return false;
			}
			
			// RleUncompress removes the uint32 alignment of bmp
			nSrcRowPadding = 0;
		}
		else if( bh.nCompression == compNone )
		{
			file.Read(rbSrc.GetPtr(), rbSrc.GetSize());
		}
		else
		{
			ASSERT(0);
			return false;
		}

		int32 nSrcOffset = 0;
		uint8* const pSrc = rbSrc.GetPtr();
		///////////////////////////////////////////////////////////////
		
		///////////////////////////////////////////////////////////////
		// Preparing Dst values
		const int32 nDstBytesPerRow = nNiceBytePerRow;

		int32 nDstOffset = 0;
		const int32 nNextDstRowOffset = 0;

		if( !m_aPixelBuffer.SetSize(nDstBytesPerRow * m_nHeight) )
		{
			m_nLastError = notEnoughMemory;
			return false;
		}
		uint8* const pDst = m_aPixelBuffer.GetPtr();

		for(int32 iVert = 0; iVert < m_nHeight; iVert++)
		{
			if( bh.nCompression == compRle4bits )
			{
				// Repack with two pixels per byte
				for(int32 iHrz = 0; iHrz < m_nWidth; iHrz++)
				{
					if( (iHrz & 1) == 0 )
					{
						pDst[nDstOffset + iHrz / 2] = uint8(pSrc[nSrcOffset + iHrz] << 4);
					}
					else
					{
						pDst[nDstOffset + iHrz / 2] |= pSrc[nSrcOffset + iHrz];
					}
				}
				nSrcOffset += m_nWidth;
				nDstOffset += nDstBytesPerRow;
			}
			else
			{
				for(int32 iHrz = 0; iHrz < nDstBytesPerRow; iHrz++)
				{
					pDst[nDstOffset] = pSrc[nSrcOffset];
					
					nSrcOffset += 1;
					nDstOffset += 1;
				}
			}
			nSrcOffset += nSrcRowPadding;
			nDstOffset -= nNextDstRowOffset;
		}
		return true;
	}

	m_nLastError = errDepthNotSupported;
	return false;
}

////////////////////////////////////////////////////////////////
// Uncompress this weirdo RLE BMP compressed pixel data for depths of 4 and 8 bits
bool Bmp::RleUncompress(const ByteArray& rbSrc, ByteArray& rbDst, int32 nNiceBytePerRow, int32 nDepth)
{
	const uint8* pSrc = rbSrc.GetPtr();
	const int32 nSrcSize = rbSrc.GetSize();

	uint8* pDst = rbDst.GetPtr();
	const int32 nDstSize = rbDst.GetSize();

	enum { stateCalm, stateGetColorIndex, stateGetEscape, 
		stateGetHrzDelta, stateGetVrtDelta, stateGetSinglePixels };

	int nState = stateCalm;

	// Values stored during the parsing
	uint8 nRepeatCount = 0;
	uint8 nHrzDelta = 0;
	uint8 nVrtDelta = 0;
	uint8 nAbsRepeatCount = 0;

	int32 iSrc = 0;
	int32 iDst = 0;

	int32 iDstStartRow = 0;

	while( iSrc < nSrcSize)
	{
		if( nState == stateCalm )
		{
			uint8 nCode = pSrc[iSrc];
			if( nCode > 0 )
			{
				nState = stateGetColorIndex;
				nRepeatCount = nCode;
			}
			else
			{
				nState = stateGetEscape;
			}
			++iSrc;
		}
		else if( nState == stateGetColorIndex )
		{
			uint8 nColorIndex = pSrc[iSrc];

			if( iDst + nRepeatCount > nDstSize )
			{
				return false;
			}
			
			if( nDepth == 8 )
			{
				for(int32 i = 0; i < nRepeatCount; ++i)
				{
					pDst[iDst + i] = nColorIndex;
				}
			}
			else
			{
				for(int32 i = 0; i < nRepeatCount; ++i)
				{
					pDst[iDst + i] = uint8(((i & 1) == 0) ? (nColorIndex >> 4) : (nColorIndex & 0x0f));
				}
			}
			iDst += nRepeatCount;
			nState = stateCalm;
			++iSrc;
		}
		else if( nState == stateGetEscape )
		{
			uint8 nEscape = pSrc[iSrc];

			if( nEscape == 0 )
			{
				// 0 = End of line
				const int32 nDstNextRow = iDstStartRow + nNiceBytePerRow;

				// Fill the remaining _ROW_ with 0
				int32 nFillCount = nDstNextRow - iDst;
				if( nFillCount != 0 )
				{
					Memory::Zero(pDst + iDst, nFillCount);
				}

				// A new row is starting
				iDst = nDstNextRow;
				iDstStartRow = nDstNextRow;

				nState = stateCalm;
			}
			else if( nEscape == 1 )
			{
				// End of bitmap

				// Fill the remaining _DST-DATA_ with 0
				int32 nFillCount = nDstSize - iDst;
				if( nFillCount != 0 )
				{
					Memory::Zero(pDst + iDst, nFillCount);
				}
				return true;
			}
			else if( nEscape == 2 )
			{
				// nDelta
				nState = stateGetHrzDelta;
			}
			else
			{
				// Absolute mode
				
				nAbsRepeatCount = nEscape;
				nState = stateGetSinglePixels;
			}

			++iSrc;
		}
		else if( nState == stateGetSinglePixels )
		{
			int32 nSrcByteCount;
			if( nDepth == 4 )
			{
				nSrcByteCount = Math::DivCeil(nAbsRepeatCount, 2);
			}
			else
			{
				nSrcByteCount = nAbsRepeatCount;
			}

			if( iSrc + nSrcByteCount > nSrcSize )
			{
				return false;
			}

			if( iDst + nAbsRepeatCount > nDstSize )
			{
				return false;
			}
		
			if( nDepth == 8 )
			{
				for(int32 i = 0; i < nAbsRepeatCount; ++i)
				{
					pDst[iDst + i] = pSrc[iSrc + i];
				}
			}
			else
			{
				for(int32 i = 0; i < nAbsRepeatCount; ++i)
				{
					uint8 nColorIndex = pSrc[iSrc + i / 2];
					pDst[iDst + i] = uint8(((i & 1) == 0) ? (nColorIndex >> 4) : (nColorIndex & 0x0f));
				}
			}

			iSrc += nSrcByteCount;
			iDst += nAbsRepeatCount;

			// In absolute mode, each run must be aligned on a word boundary
			if( (nSrcByteCount & 1) != 0 )
			{
				++iSrc;
			}

			nState = stateCalm;
		}
		else if( nState == stateGetHrzDelta )
		{
			nHrzDelta = pSrc[iSrc];
			nState = stateGetVrtDelta;

			++iSrc;
		}
		else if( nState == stateGetVrtDelta )
		{
			nVrtDelta = pSrc[iSrc];
			
			// Move the current position
			const int32 nNewDst = iDst + nVrtDelta * nNiceBytePerRow + nHrzDelta;
			if( nNewDst > nDstSize )
				return false;

			int32 nFillCount = nNewDst - iDst;

			Memory::Zero(pDst + iDst, nFillCount);

			iDst = nNewDst;
			iDstStartRow += nVrtDelta * nNiceBytePerRow;

			nState = stateCalm;

			iSrc++;
		}
		else
		{
			ASSERT(0);
			return false;
		}
	}
	return false; // Should end with an EndOfBitmap escape
}

const Palette* Bmp::GetPalette() const
{
	return &m_ColorTable;
}

PixelFormat Bmp::GetPixelFormat() const
{
	return m_epf;
}

int32 Bmp::ComputeBmpBytesPerRow(int32 nExpectedBytesPerRow)
{
	const int32 nRemainBit1 = (nExpectedBytesPerRow >> 1) & 1;
	const int32 nRemainBit0 = (nExpectedBytesPerRow >> 0) & 1;
	const int32 nNewByteCount = ( (nExpectedBytesPerRow >> 2) + (nRemainBit1 | nRemainBit0)) * 4;
	return nNewByteCount;
}

PixelFormat Bmp::GetPfFromDepth(uint16 nDepth) const
{
	if( nDepth == 32 )
	{
		return pf32bppBgra;
	}
	else if( nDepth == 24 )
	{
		return pf24bppBgr;
	}
	else if( nDepth == 16 )
	{
		return pf16bppRgb555;
	}
	else if( nDepth == 8 )
	{
		return pf8bppIndexed;
	}
	else if( nDepth == 4 )
	{
		return pf4bppIndexed;
	}
	else if( nDepth == 1 )
	{
		return pf1bppIndexed;
	}
	
	return pfUnknown;
}

bool Bmp::Dump(const String& strFileName, const BmpDumpSettings& ds)
{
	File file;
	if( !file.Open(strFileName, File::modeWrite) )
		return false;

	return Dump(file, ds);
}

bool Bmp::Dump(IFile& file, const BmpDumpSettings& ds)
{
	const uint8* pSrc = (const uint8*) ds.m_pBuffer;
	const int32 nWidth = ds.m_nWidth;
	const int32 nHeight = ds.m_nHeight;
	PixelFormat epf = ds.m_epf;

	int16 nDepth = (int16) ImageType::SizeofPixelInBits(epf);
	if( nDepth != 24 )
	{
		// Not supported
		return false;
	}

	TmpEndianMode tem(file, boLittleEndian);

	file.Write(k_BmpSignature, sizeof(k_BmpSignature));

	// Complete file size in bytes
	int64 nFieldPosition_FileSize = file.GetPosition();
	
	// Offset from beginning of the file to the beginning of the image data
	int64 nFieldPosition_OffsetToData = 4 + 4 + file.GetPosition();
	
	// Number of colors
	int32 nColorCount = 0;
	if( ds.m_pPalette != null )
	{
		nColorCount = ds.m_pPalette->m_nCount;
	}

	BmpHeader bh;
	bh.nFileSize = 0;
	bh.nReserved = 0;
	bh.nOffsetToData = 0;
	bh.nBitmapHeaderSize = 40;
	bh.nWidth = nWidth;
	bh.nHeight = nHeight;
	bh.nPlanes = 1;
	bh.nDepth = nDepth;  // Possible values : 1, 4, 8, 16, 24, 32
	bh.nCompression = 0; // 0=None (BI_RGB), 1=RLE 8bits, 2=RLE 4bits, 3=BI_BITFIELDS
	bh.nBitmapDataSize = 0; // Rounded to the next 4 byte boundary
	bh.nXPelsPerMeter = 2834;
	bh.nYPelsPerMeter = 2834;
	bh.nColorCount = nColorCount;
	bh.nImportantColors = 0;

	if( file.ShouldSwapBytes() )
	{
		bh.SwapBytes();
	}
	
	int32 nWantedWritten = sizeof(BmpHeader);
	if( file.Write(&bh, nWantedWritten) != nWantedWritten )
	{
		return false;
	}

	////////////////////////////////////////////////
	// Write the palette
	if( ds.m_pPalette != null )
	{
		uint8 aPalBuf[256 * 4];
		uint8* pPal = aPalBuf;

		for(int32 i = 0; i < ds.m_pPalette->m_nCount; ++i)
		{
			ds.m_pPalette->m_aColors[i].ToRgba( pPal[2], pPal[1], pPal[0], pPal[3]);

			pPal += 4;
		}

		const int32 nPalByteSize = ds.m_pPalette->m_nCount * 4;
		if( file.Write(aPalBuf, nPalByteSize) != nPalByteSize )
		{
			return false;
		}
	}
	////////////////////////////////////////////////

	//////////////////////////////////////////////////
	uint32 nOffsetToData = (uint32) file.GetPosition();

	if( nDepth == 24 )
	{
		// Each row in a bmp file must be uint32 aligned
		int32 nPad = (int32(nWidth * 3) & 0x0003);
		if( nPad != 0 )
		{
			nPad = 4 - nPad;
		}

		ByteArray aRow;
		aRow.SetSize(nWidth * 3);

		for(int32 iRow = 0; iRow < nHeight; ++iRow)
		{
			const uint8* pWriteSrc = pSrc + ((nHeight - 1) - iRow) * nWidth * 3;
			
			if( epf == pf24bppBgr )
			{
				file.Write(pWriteSrc, nWidth * 3);
			}
			else if( epf == pf24bppRgb )
			{
				// Swap red and blue
				for(int32 i = 0; i < nWidth; ++i)
				{
					uint8 r = pWriteSrc[3 * i];
					uint8 g = pWriteSrc[3 * i + 1];
					uint8 b = pWriteSrc[3 * i + 2];
					aRow[3 * i] = b;
					aRow[3 * i + 1] = g;
					aRow[3 * i + 2] = r;
				}
				file.Write(aRow.GetPtr(), aRow.GetSize());
			}
			
			if( nPad != 0 )
			{
				int32 nDummy = 0;
				file.Write(&nDummy, nPad);
			}
		}
	}
	//////////////////////////////////////////////////

	uint32 nFileSize = (uint32) file.GetPosition();
	file.SetPosition(nFieldPosition_FileSize);
	if( !file.Write32(nFileSize) )
	{
		return false;
	}

	file.SetPosition(nFieldPosition_OffsetToData);
	if( !file.Write32(nOffsetToData) )
	{
		return false;
	}

	/////////////////////////////////
	return true;
}


String Bmp::GetLastErrorString() const
{
	switch(m_nLastError)
	{
	case Bmp::errNotABmpFile:
		return L"Not a BMP file";

	case Bmp::errBadHeaderSize:
		return L"Bad header size";

	case Bmp::errUnconsistentDepthAndCompression:
		return L"Unconsistent depth and compression";

	case Bmp::errUnconsistentDepthAndColorCount:
		return L"Unconsistent depth and color count";

	case Bmp::errUnsupportedCompressionFormat:
		return L"Unsupported compression format";

	case Bmp::errUnsupportedBitfieldsFormat:
		return L"Unsupported bitfields format";

	case Bmp::errDepthNotSupported:
		return L"Depth not supported (accepted : 32, 24, 16, 8, 4, 1) for BMP files";

	case Bmp::errUnexpectedEndOfFile:
		return L"Unexpected end of file";

	case Bmp::errBadCompressedData:
		return L"Bad compressed data";

	case Bmp::errTooManyColors:
		return L"Too many colors";
	}

	return ImageType::GetLastErrorString();
}

bool Bmp::ReadPixelsFromFile(IFile& file)
{
	int32 nSrcBytesPerRow = ImageType::ComputeByteWidth(m_epf, m_nWidth);
	int32 nBmpBytesPerRow = ComputeBmpBytesPerRow(nSrcBytesPerRow);
	const int32 nDataSize = nBmpBytesPerRow * m_nHeight;
	
	if( !m_aPixelBuffer.SetSize(nDataSize) )
	{
		m_nLastError = notEnoughMemory;
		return false;
	}

	uint8* pPixels = m_aPixelBuffer.GetPtr();

	int32 nRead = file.Read(pPixels, nDataSize);
	if( nRead != nDataSize )
	{
		m_nLastError = uncompleteFile;
		return false;
	}

	return true;
}

// Remove the last unused bytes used by the Bmp pixels so the row are 32 bits aligned
void Bmp::RemovePadding()
{
	int32 nWantedBytesPerRow = ImageType::ComputeByteWidth(m_epf, m_nWidth);
	int32 nBmpBytesPerRow = ComputeBmpBytesPerRow(nWantedBytesPerRow);

	if( nWantedBytesPerRow == nBmpBytesPerRow )
	{
		// Not needed ! :)
		return;
	}

	uint8* pSrc = m_aPixelBuffer.GetPtr();
	uint8* pDst = pSrc;

	const int32 nHeight = m_nHeight;
	for(int iY = 0; iY < nHeight; ++iY)
	{
		Memory::Move(pDst, pSrc, nWantedBytesPerRow);

		pSrc += nBmpBytesPerRow;
		pDst += nWantedBytesPerRow;
	}
}
