///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "Master.h"
#include "CppExtension.h"
#include "FormatType.h"
#include "Array.h"
#include "Memory.h"
#include "File.h"

void CheckPlatformCompatibility()
{
	// Check type sizes
	if( sizeof(int16) != 2)
	{
		ASSERTMSG(0, "Compatibility error: int16 size is not 16 bits!!!\n");
	}

	if( sizeof(int32) != 4)
	{
		ASSERTMSG(0, "Compatibility error: int32 size is not 32 bits!!!\n");
	}

	// int64 is a bit weird so we do not test it here

	if( sizeof(float32) != 4)
	{
		ASSERTMSG(0, "Compatibility error: float32 size is not 32 bits!!!\n");
	}

	if( sizeof(float64) != 8)
	{
		ASSERTMSG(0, "Compatibility error: float64 size is not 64 bits!!!\n");
	}

	if( sizeof(wchar) != 2)
	{
		ASSERTMSG(0, "Compatibility error: wchar size is not 16 bits!!!\n");
	}

	// Check endian mode
	bool bLittleEndian = false;
	const uint8 aVals[4] = {0x12, 0x34, 0x56, 0x78};
	uint32 nVal = ((const uint32*) aVals) [0];
	if( nVal != 0x12345678)
	{
		// The CPU is little endian (yuck ! :-6)
		bLittleEndian = true;
	}
	if( chustd::k_ePlatformByteOrder != chustd::boLittleEndian && bLittleEndian)
	{
		ASSERTMSG(0, "Compatibility error: bad endian mode!\n"
			"comp_bLittleEndian value in Master.h is not correct!!!\n");
	}
}

#undef double
#undef float

#ifdef WIN32
#include <windows.h>
#pragma comment(lib, "user32.lib") // Used for MessageBoxA
#endif // WIN32

void ChustdSzCat(char* pszDst, const char* pszSrc, int nSizeofBuffer)
{
	if( pszSrc == null)
	{
		return;
	}

	int i = 0;
	for(; i < nSizeofBuffer - 1; ++i)
	{
		if( pszDst[i] == 0)
			break;
	}

	for(; i < nSizeofBuffer - 1; ++i)
	{
		const char& c = pszSrc[0];
		pszDst[i] = c;
		if( c == 0)
			break;
		pszSrc++;
	}
	pszDst[nSizeofBuffer - 1] = 0;
}

void ChustdToAscii(char* sz, wchar* wsz)
{
	while(true)
	{
		const wchar& c = wsz[0];
		sz[0] = char(c);
		if( c == 0)
			break;
		wsz++;
		sz++;
	}
}

void ChustdAssertMessage(const char* pszFileName, int32 nLine, const char* pszMsg)
{
	static const char szTitle[] = "ASSERT failure";
	
	char szBuffer[300];
	szBuffer[0] = 0;
	ChustdSzCat(szBuffer, pszFileName, 300);
	ChustdSzCat(szBuffer, " (", 300);
	
	wchar szLine[16];
	char szLine2[16];
	chustd::FormatInt32(nLine, szLine);
	ChustdToAscii(szLine2, szLine);
	ChustdSzCat(szBuffer, szLine2, 300);
	ChustdSzCat(szBuffer, ")", 300);

	ChustdSzCat(szBuffer, "\n\n", 300);
	ChustdSzCat(szBuffer, pszMsg, 300);

#ifdef WIN32
	typedef int (WINAPI *PFN_MESSAGEBOXA)(HWND, LPCSTR, LPCSTR, UINT);

	// Manually loads user32.dll as this is the only dependency in chustd
	HMODULE hMod = LoadLibraryW(L"user32.dll");
	if( hMod )
	{
		PFN_MESSAGEBOXA pfnMessageBoxA = (PFN_MESSAGEBOXA) GetProcAddress(hMod, "MessageBoxA");
		pfnMessageBoxA(NULL, szBuffer, szTitle, MB_TASKMODAL | MB_OK | MB_TOPMOST);
		FreeLibrary(hMod);
	}
#elif CHUOS
	//chuos_SystemConsole_WriteLine(szBuffer);
#else
	fprintf(stderr, "ASSERT failure - %s(%d) : %s", pszFileName, nLine, pszMsg);
#endif // WIN32
}

void ChustdDebugMessage(const char* pszMsg)
{
#ifdef WIN32
	OutputDebugStringA(pszMsg);
	OutputDebugStringA("\n");
#elif CHUOS
	//chuos_SystemConsole_WriteLine(pszMsg);
#else
	fprintf(stderr, "%s\n", pszMsg);
#endif // WIN32
}

//////////////////////////////////////////////////////////////////////////////////////////////
void ChustdTraceLine(const chustd::String& str)
{
	static bool bCreated = false;

	uint32 nOpenMode = 0;
	if( !bCreated)
	{
		nOpenMode = chustd::IFile::modeWrite;
		bCreated = true;
	}
	else
	{
		nOpenMode = chustd::IFile::modeAppend;
	}

	chustd::File file;
	if( file.Open(L"chustd-trace.txt", nOpenMode) )
	{
		// Avoid code units conversion by using the platform byte order
		file.SetByteOrder(chustd::k_ePlatformByteOrder);

		if( file.GetSize() == 0)
		{
			// Put the BOM
			file.Write16(uint16(0xFEFF));
		}
		file.WriteString(str);

#ifdef _WIN32 // Let's be kind with Windows
		file.Write16(uint16(0x000D));
#endif

		file.Write16(uint16(0x000A));
	}
}

//////////////////////////////////////////
void ChustdDebugPrintLeak(void* p, int32 nSize, const char* pszFileName, int32 nLine)
{
	wchar szPointer[20];
	chustd::FormatPtr(p, szPointer);
	
	wchar szSize[16];
	chustd::FormatInt32(nSize, szSize, 6, L' ');

	char szBuffer[512];
	szBuffer[0] = 0;

	char szTmp[16];
	
	ChustdToAscii(szTmp, szPointer);
	ChustdSzCat(szBuffer, szTmp, sizeof(szBuffer));
	ChustdSzCat(szBuffer, " / ", sizeof(szBuffer));
	
	ChustdToAscii(szTmp, szSize);
	ChustdSzCat(szBuffer, szTmp, sizeof(szBuffer));
	ChustdSzCat(szBuffer, " : ", sizeof(szBuffer));

	// Display the content of the buffer
	uint8* pData = (uint8*) p;
	
	int32 nShowSize = 16;
	
	// Values show
	for(int32 i = 0; i < nShowSize; ++i)
	{
		if( i < nSize)
		{
			wchar szData[16];
			chustd::FormatUInt8Hex( pData[i], szData);

			ChustdToAscii(szTmp, szData);
			ChustdSzCat(szBuffer, szTmp, sizeof(szBuffer));
			ChustdSzCat(szBuffer, " ", sizeof(szBuffer));
		}
		else
		{
			ChustdSzCat(szBuffer, "   ", sizeof(szBuffer));
		}
	}

	ChustdSzCat(szBuffer, "  ", sizeof(szBuffer));

	// Ascii show
	for(int32 i = 0; i < nShowSize; ++i)
	{
		if( i < nSize)
		{
			uint8 nChar = pData[i];
			if( nChar <= 31)
			{
				nChar = '.';
			}
			szTmp[0] = nChar;
			szTmp[1] = 0;
			ChustdSzCat(szBuffer, szTmp, sizeof(szBuffer));
		}
		else
		{
			ChustdSzCat(szBuffer, " ", sizeof(szBuffer));
		}
	}

	if( pszFileName != null && nLine > 0)
	{
		ChustdSzCat(szBuffer, "   ", sizeof(szBuffer));
		ChustdSzCat(szBuffer, pszFileName, sizeof(szBuffer));

		wchar szData[16];
		chustd::FormatInt32(nLine, szData);
		ChustdToAscii(szTmp, szData);

		ChustdSzCat(szBuffer, " (", sizeof(szBuffer));
		ChustdSzCat(szBuffer, szTmp, sizeof(szBuffer));
		ChustdSzCat(szBuffer, ")", sizeof(szBuffer));
	}

	ChustdDebugMessage(szBuffer);
}

void CheckMemory()
{
	uint32 nAllocated = chustd::Memory::GetAllocated();
	if( nAllocated != 0)
	{
		// Memory leaks
		
		char szBuffer[300];
		szBuffer[0] = 0;
		ChustdSzCat(szBuffer, "!!! ---> Memory leaks detected : ", sizeof(szBuffer));
		
		wchar szLine[16];
		chustd::FormatUInt32(nAllocated, szLine);
		
		char szLine2[16];
		ChustdToAscii(szLine2, szLine);
		ChustdSzCat(szBuffer, szLine2, sizeof(szBuffer));
		ChustdSzCat(szBuffer, " bytes", sizeof(szBuffer));

		ChustdDebugMessage(szBuffer);

		chustd::Memory::DebugPrintBlocks();
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////
typedef void (__cdecl *PFN_ExeInit)();

// The .CRT$XCU section is an array of pointer to pre-initialisation functions
// (like calls to global objects constructors)
// We do not support $XIA $XIZ (C initializers), $XPA $XPZ (C pre-terminators), $XTA $XTZ (C terminators)

// Create a reference to the first entry (with the help of the  A  letter) of the .CRT$XCU section
#pragma data_seg(".CRT$XCAFirst")
PFN_ExeInit g_apfnXcuBegin[] = { 0 };

// Create a reference to the last entry (with the help of the  Z  letter) of the .CRT$XCU section
#pragma data_seg(".CRT$XCZLast")
PFN_ExeInit g_apfnXcuEnd[] = { 0 };

#pragma data_seg()  // Reset

#if _MSC_VER >= 1400
#pragma comment(linker, "/merge:.CRT=.rdata") // In VC8, .CRT is read-only, so merge with .rdata
#else
#pragma comment(linker, "/merge:.CRT=.data") // In VC6, it won't work to merge with .rdata, use .data
#endif

// Call pre-init functions which pointers are stored as an array in the .CRT$XCU section
void CallExeInitFunctions()
{
	const PFN_ExeInit* papfnExeInits = g_apfnXcuBegin;
	const int nCount = int(g_apfnXcuEnd - g_apfnXcuBegin);
	
	for(int i = 0; i < nCount; ++i)
	{
		PFN_ExeInit pfnExeInit = papfnExeInits[i];
		if( pfnExeInit != 0)
		{
			pfnExeInit();
		}
	}
}


//////////////////////////////////////////////////////////////////
extern int32 __cdecl ChuEntryPoint();

extern "C"
{

/////////////////////////////////////////////////////////////////////////////
unsigned int _fltused = 1;

/////////////////////////////////////////////////////////////////////////////
#if defined(_M_IX86)

__declspec( naked )
void __cdecl _chkesp()
{
	__asm
	{
		jz	EverythingFine
		int     3
EverythingFine:
		ret
    }
}

/////////////////////////////////////////////////////////////////////////////
// _chkstk function is called when the local variables exceed 4096 bytes
// eax : in : number of bytes needed
const uint32 PAGE_SIZE = 4096;

__declspec( naked )
void __cdecl _chkstk()
{
	__asm
	{
        push ecx
        cmp  eax, PAGE_SIZE  // Needs more than one page ?
        lea  ecx, [esp] + 8  // ecx : new  stack pointer (+8 : for return address and pushed ecs)
        jb   short OnePage  // Less than one page

DoLoop:
        sub  ecx, PAGE_SIZE
        sub  eax, PAGE_SIZE

        test dword ptr[ecx], eax  // Touch the page (exception --> page growth)

        cmp  eax, PAGE_SIZE
        jae  DoLoop

OnePage:
        sub  ecx, eax
        mov  eax, esp

        test dword ptr[ecx], eax // Touch

        mov  esp, ecx // New stack pointer

        mov  ecx, dword ptr[eax]     // Restore ecx
        mov  eax, dword ptr[eax + 4] // Restore return address

        push eax
        ret
	}
}

__declspec(naked) void __cdecl _ftol()
{
	__asm
	{
		sub	esp,4
		fst	dword ptr [esp]
		mov	eax,dword ptr [esp]
		and	eax,0x80000000
		xor	eax,0xbeffffff
		mov	dword ptr [esp],eax
		fadd	dword ptr [esp]
		fistp	dword ptr [esp]
		pop	eax
		ret
	}
}

#endif // defined(_M_IX86)

/////////////////////////////////////////////////////////////////////////////
int _purecall()
{
	INTERRUPT_EXECUTION();
	return 0;
}

/////////////////////////////////////////////////////////////////////////////
// atexit
typedef void (__cdecl*PFN_AtExit)();

chustd::Array<PFN_AtExit>& GetAtExitArray()
{
	// Don't make it global, so we are sure its constructor is called before any usage
	static chustd::Array<PFN_AtExit> apfnAtExit;

	return apfnAtExit;
}


int atexit(PFN_AtExit pfn)
{
	GetAtExitArray().Add(pfn);
	return 0;
}

static void CallAtExit()
{
	chustd::Array<PFN_AtExit>& apfnAtExit = GetAtExitArray();

	const int32 nSize = apfnAtExit.GetSize();

	// Exit functions are called in reverse order
	for( int32 i = nSize - 1; i >= 0; --i )
	{
		PFN_AtExit pfn = apfnAtExit[i];
		pfn();
	}
}

void WinMainCRTStartup()
{
	CheckPlatformCompatibility();
	chustd::Memory::InitAllocator();
	CallExeInitFunctions();
	int32 nExitCode = ChuEntryPoint();
	CallAtExit();
	CheckMemory();
	::ExitProcess(nExitCode);
}

void mainCRTStartup()
{
	CheckPlatformCompatibility();
	chustd::Memory::InitAllocator();
	CallExeInitFunctions();
	int32 nExitCode = ChuEntryPoint();
	CallAtExit();
	CheckMemory();
	::ExitProcess(nExitCode);
}

int __stdcall _DllMainCRTStartup(void* /*hModule*/, uint32 nReasonForCall, void* /*pReserved*/)
{
	switch(nReasonForCall)
	{
	case 1:
		// DLL_PROCESS_ATTACH
		CheckPlatformCompatibility();
		chustd::Memory::InitAllocator();
		CallExeInitFunctions();
		break;

	case 2:
		// DLL_THREAD_ATTACH
		break;

	case 3:
		// DLL_THREAD_DETACH
		break;
	
	case 0:
		// DLL_PROCESS_DETACH
		CallAtExit();
		CheckMemory();
		break;
	}
	return 1;
}

////////////////////////////////////////////////////////////////
// This is a workaround for VC8

// The compiler creates calls to memcpy and memset for small loops. memcpy and memset
// do not exist in chustd.

// Solution 1: remove that behaviour. Is there a flag for that ? Could not find any.

// Solution 2 : provide here a definition of memcpy and memset. A bit tricky, but it works.

// In Master.h, memcpy is defined as memcpy_plop  so the compiler won't complain 
// that memcpy has another linkage when it parses the file declaring memcpy.
// It will just think a function called memcpy_plop exists.

// Then, we need to provide our own memcpy function, but as memcpy is an intrinsic
// function, it means a call to memcpy can be replaced by... a small loop ! So
// we disable intrinsic replacement for memcpy.

// The same problem applies for memset.

#undef memcpy
#undef memset

void* memcpy(void* pDst, const void* pSrc, int nSize);

#pragma function(memcpy)

void* memcpy(void* pDst, const void* pSrc, int nSize)
{
	chustd::Memory::Copy(pDst, pSrc, nSize);
	return pDst;
}

void* memset(void* pDst, int nValue, int nSize);

#pragma function(memset)

void* memset(void* pDst, int nValue, int nSize)
{
	chustd::Memory::Set(pDst, uint8(nValue), nSize);
	return pDst;
}

} // extern "C"

