///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "Master.h"
#include "Date.h"
#include "String.h"
#include "StringBuilder.h"
#include "FormatType.h"
#include "FixArray.h"
#include "ScanType.h"
#include "TimeStamp.h"

using namespace chustd;
//////////////////////////////////////////////////////////////////////
Duration::Duration()
{
	m_nStamp = 0;
}

Duration::Duration(int64 nChuTimeStamp) : m_nStamp(nChuTimeStamp)
{

}

Duration::Duration(int32 nDays, int8 nHours, int8 nMinutes, int8 nSeconds, int16 nMilliseconds)
{
	if( !(-23 <= nHours && nHours <= 23))
	{
		nHours = 0;
	}

	if( !(-59 <= nMinutes && nMinutes <= 59))
	{
		nMinutes = 0;
	}

	if( !(-59 <= nSeconds && nSeconds <= 59))
	{
		nSeconds = 0;
	}

	int64 nStamp = nDays;

	nStamp *= 24;
	nStamp += nHours;

	nStamp *= 60;
	nStamp += nMinutes;

	nStamp *= 60;
	nStamp += nSeconds;

	nStamp *= 1000;
	nStamp += nMilliseconds;

	m_nStamp = nStamp;
}

// Returns the time stamp in seconds
int64 Duration::GetStampInSec() const
{
	return m_nStamp / 1000;
}

Duration Duration::FromSeconds(int64 nSeconds)
{
	Duration dur;
	dur.m_nStamp = 	1000;
	dur.m_nStamp *= nSeconds;
	return dur;
}

void Duration::Split(int32& nDays, int8& nHours, int8& nMinutes, int8& nSeconds, int16& nMilliseconds) const
{
	int64 nNext = m_nStamp;

	nMilliseconds = int16(nNext % 1000);
	nNext /= 1000;

	nSeconds = int8(nNext % 60);
	nNext /= 60;

	nMinutes = int8(nNext % 60);
	nNext /= 60;

	nHours = int8(nNext % 24);
	nNext /= 24;

	nDays = int32(nNext);
}

void Duration::FormatFrench(String& str) const
{
	int32 nDays;
	int8 nHours;
	int8 nMinutes;
	int8 nSeconds;
	int16 nMilliseconds;
	Split(nDays, nHours, nMinutes, nSeconds, nMilliseconds);

	String strDays, strHours, strMinutes, strSeconds;
	wchar szBuffer[12];

	if( nDays > 0)
	{
		if( nDays == 1)
		{
			strDays = L"1 jour ";
		}
		else
		{
			FormatInt32(nDays, szBuffer);
			strDays = String(szBuffer) + L" jours";
		}
	}

	if( nHours > 0)
	{
		if( nHours == 1)
		{
			strHours = L"1 heure ";
		}
		else
		{
			FormatInt32(nHours, szBuffer);
			strHours = String(szBuffer) + L" heures";
		}
	}

	if( nMinutes > 0)
	{
		if( nMinutes == 1)
		{
			strMinutes = L"1 minute ";
		}
		else
		{
			FormatInt32(nMinutes, szBuffer);
			strMinutes = String(szBuffer) + L" minutes";
		}
	}
	
	if( nSeconds > 0)
	{
		if( nSeconds == 1)
		{
			strSeconds = L"1 seconde";
		}
		else
		{
			FormatInt32(nSeconds, szBuffer);
			strSeconds = String(szBuffer) + L" secondes";
		}
	}

	str = strDays + strHours + strMinutes + strSeconds;
}

void Duration::FormatSmall(class String& str) const
{
	int32 nDays;
	int8 nHours;
	int8 nMinutes;
	int8 nSeconds;
	int16 nMilliseconds;
	Split(nDays, nHours, nMinutes, nSeconds, nMilliseconds);

	String strDays, strHours, strMinutes, strSeconds;
	wchar szBuffer[12];

	if( nDays > 0)
	{
		FormatInt32(nDays, szBuffer);
		strDays = String(szBuffer) + L" j";
	}

	if( nHours > 0)
	{
		FormatInt32(nHours, szBuffer);
		strHours = String(szBuffer) + L" h";
	}

	if( nMinutes > 0)
	{
		FormatInt32(nMinutes, szBuffer);
		strMinutes = String(szBuffer) + L" m";
	}
	
	if( nSeconds > 0)
	{
		FormatInt32(nSeconds, szBuffer);
		strSeconds = String(szBuffer) + L" s";
	}

	str = strDays + strHours + strMinutes + strSeconds;
}
//////////////////////////////////////////////////////////////////////
Date::Date()
{
	m_nYear = 0;
	m_nMonth = 0;
	m_nDay = 0;
	m_nHour = 0;
	m_nMinute = 0;
	m_nSecond = 0;

	m_nStamp = 0;
}

// static
Date Date::GetNow()
{
	Date date;

#ifdef WIN32
	SYSTEMTIME st;
	::GetLocalTime(&st); // OS Dependent
	
	// The FILETIME structure is a 64-bit value representing the number 
	// of 100-nanosecond intervals since January 1, 1601.
	FILETIME ft;
	::SystemTimeToFileTime(&st, &ft);

	int64 nStamp = ft.dwHighDateTime;
	nStamp <<= 32;
	nStamp |= ft.dwLowDateTime;

	date.m_nStamp = nStamp / 10000;
	
#elif CHUOS
	date.m_nStamp = chuos_GetLocalTimeStamp();
#endif
	
	date.UpdateFieldsFromStamp();
	return date;
}

void Date::UpdateFieldsFromStamp()
{
	uint16 nMillisecond = 0;
	chustd::SplitChuTimeStamp(m_nStamp, m_nYear, m_nMonth, m_nDay, m_nHour, m_nMinute, m_nSecond, nMillisecond);
}

// This function computes a timestamp from the splitted fields
void Date::UpdateStampFromFields()
{
	uint16 nMillisecond = 0;
	m_nStamp = chustd::MakeChuTimeStamp(m_nYear, m_nMonth, m_nDay, m_nHour, m_nMinute, m_nSecond, nMillisecond);
}

void Date::Set(int16 nYear, uint8 nMonth, uint8 nDay, uint8 nHour, uint8 nMinute, uint8 nSecond)
{
	// TODO: think if we should check correctness here
	m_nYear = nYear;
	m_nMonth = nMonth;
	m_nDay = nDay;
	m_nHour = nHour;
	m_nMinute = nMinute;
	m_nSecond = nSecond;

	UpdateStampFromFields();
}

// Formats date only as YYYY-MM-DD (ISO 8601)
String Date::FormatDate(DateFormat eFormat) const
{
	StringBuilder sbFinal;
	wchar szBuffer[12];

	FormatInt32(m_nYear, szBuffer, 4, L'0');
	sbFinal += szBuffer;

	if( eFormat != dfCompact )
	{
		sbFinal += L"-";
	}
	FormatInt32(m_nMonth, szBuffer, 2, L'0');
	sbFinal += szBuffer;

	if( eFormat != dfCompact )
	{
		sbFinal += L"-";
	}
	FormatInt32(m_nDay, szBuffer, 2, L'0');
	sbFinal += szBuffer;

	return sbFinal.ToString();
}

String Date::FormatTime(DateFormat eFormat) const
{
	StringBuilder sbFinal;
	wchar szBuffer[12];

	FormatInt32(m_nHour, szBuffer, 2, L'0');
	sbFinal += szBuffer;

	if( eFormat != dfCompact )
	{
		sbFinal += L":";
	}
	FormatInt32(m_nMinute, szBuffer, 2, L'0');
	sbFinal += szBuffer;


	if( eFormat != dfCompact )
	{
		sbFinal += L":";
	}
	FormatInt32(m_nSecond, szBuffer, 2, L'0');
	sbFinal += szBuffer;

	return sbFinal.ToString();
}

String Date::Format(DateFormat eFormat) const
{
	String strDate = FormatDate(eFormat);
	String strTime = FormatTime(eFormat);

	if( eFormat == dfExpandedT || eFormat == dfCompact)
	{
		return strDate + L"T" + strTime;
	}

	return strDate + L" " + strTime;
}

// Gets date as YYYY-MM-DD hh:mm:ss
// Works if last parts are missing, for instance : YYYY-MM
// Returns the number of read members
int32 Date::Scan(const String& strDate)
{
	//int32 nCount = strDate.Scan("%?-%?-%? %?:%?:%?", m_nYear, m_nMonth, m_nDay, m_nHour, m_nMinute, m_nSecond);
	
	const wchar* psz = strDate.GetBuffer();
	int32 nAdvance = 0;
	
	FixArray<bool, 12> abParseStatus;

	abParseStatus[0] = ScanType(psz, nAdvance, L'', &m_nYear);
	abParseStatus[1] = psz[nAdvance] == L'-';
	nAdvance++;
	
	abParseStatus[2] = ScanType(psz, nAdvance, L'', &m_nMonth);
	abParseStatus[3] = psz[nAdvance] == L'-';
	nAdvance++;

	abParseStatus[4] = ScanType(psz, nAdvance, L'', &m_nDay);
	abParseStatus[5] = psz[nAdvance] == L' ';
	nAdvance++;

	abParseStatus[6] = ScanType(psz, nAdvance, L'', &m_nHour);
	abParseStatus[7] = psz[nAdvance] == L':';
	nAdvance++;

	abParseStatus[8] = ScanType(psz, nAdvance, L'', &m_nMinute);
	abParseStatus[9] = psz[nAdvance] == L':';
	nAdvance++;

	abParseStatus[10] = ScanType(psz, nAdvance, L'', &m_nSecond);
	abParseStatus[11] = true;
	
	int32 nCount = 0;
	for(int32 i = 0; i < abParseStatus.GetSize(); i += 2)
	{
		if( !abParseStatus[i])
			break;

		nCount++;
	}

	/////////////////////////////////////////////////////////////////
	// Check fields integrity
	if( nCount >= 6 && !( m_nSecond <= 59) )
	{
		nCount = 5;
	}
	if( nCount >= 5 && !( m_nMinute <= 59) )
	{
		nCount = 4;
	}
	if( nCount >= 4 && !( m_nHour <= 23) )
	{
		// Just take the year, the month and the day as the hour is invalid
		nCount = 3;
	}
	if( nCount >= 3 && !( 1 <= m_nDay && m_nDay <= 31) )
	{
		// Just take the year and the month as the day is invalid
		nCount = 2;
	}
	if( nCount >= 2 && !( 1 <= m_nMonth && m_nMonth <= 12) )
	{
		// Just take the year as the month is invalid
		nCount = 1;
	}
	/////////////////////////////////////////////////////////////////

	uint8* const apn[] = {null, &m_nMonth, &m_nDay, &m_nHour, &m_nMinute, &m_nSecond };
	int32 iClear = nCount;

	for(; iClear < 3; ++iClear)
	{
		apn[iClear][0] = 1;
	}

	for(; iClear < 6; ++iClear)
	{
		apn[iClear][0] = 0;
	}

	UpdateStampFromFields();

	return nCount;
}

void Date::AddDays(int32 nCount)
{
	int64 nOneDay = 24 * 60 * 60;
	nOneDay *= 1000;
	m_nStamp += nCount * nOneDay;

	UpdateFieldsFromStamp();
}

void Date::SetTime(uint8 nHour, uint8 nMinute, uint8 nSecond)
{
	if( !(0 <= nHour && nHour <= 23))
	{
		return;
	}

	if( !(0 <= nMinute && nMinute <= 59))
	{
		return;
	}

	if( !(0 <= nSecond && nSecond <= 59))
	{
		return;
	}

	m_nHour = nHour;
	m_nMinute = nMinute;
	m_nSecond = nSecond;

	UpdateStampFromFields();
}

const Duration Date::operator - (const Date& t2) const
{
	Duration dur;
	dur.m_nStamp = m_nStamp - t2.m_nStamp;

	return dur;
}

bool Date::operator > (const Date& t2) const
{
	return m_nStamp > t2.m_nStamp;
}

bool Date::operator < (const Date& t2) const
{
	return m_nStamp < t2.m_nStamp;
}

bool Date::operator >= (const Date& t2) const
{
	return m_nStamp >= t2.m_nStamp;
}

bool Date::operator <= (const Date& t2) const
{
	return m_nStamp <= t2.m_nStamp;
}

// Returns the time stamp in seconds
int64 Date::GetStampInSec() const
{
	return m_nStamp / 1000;
}

const Date& Date::operator += (Duration dur)
{
	m_nStamp += dur.m_nStamp;
	UpdateFieldsFromStamp();

	return *this;
}

Date Date::FromUnixTimeStamp(int64 nTime)
{
	// 0 == 1970-01-01
	nTime *= 1000;

	Date date;
	date.Set(1970);
	date.m_nStamp += nTime;

	date.UpdateFieldsFromStamp();
	return date;
}

int64 Date::ToUnixTimeStamp() const
{
	Date dateBase;
	dateBase.Set(1970);

	int64 nDif = m_nStamp - dateBase.m_nStamp;
	nDif /= 1000;

	if( nDif < 0)
		nDif = 0;

	return nDif;
}

int64 Date::GetChuTimeStamp() const
{
	return m_nStamp;
}

Date Date::FromChuTimeStamp(int64 nStamp)
{
	Date date;
	date.m_nStamp = nStamp;
	date.UpdateFieldsFromStamp();
	return date;
}

