///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#ifndef CHUSTD_DATE_H
#define CHUSTD_DATE_H

// In winbase.h #define GetCurrentTime == GetTickCount()
#ifdef GetCurrentTime
#undef GetCurrentTime
#endif

namespace chustd {

// Manages a duration between two dates
class Duration
{
	friend class Date;
public:
	
	void Split(int32& nDays, int8& nHours, int8& nMinutes, int8& nSeconds, int16& nMilliseconds) const;
	int64 GetStampInSec() const;

	Duration();
	Duration(int64 nChuTimeStamp);
	Duration(int32 nDays, int8 nHours, int8 nMinutes, int8 nSeconds, int16 nMilliseconds);

	void FormatFrench(class String& str) const;
	void FormatSmall(class String& str) const;

	static Duration FromSeconds(int64 nSeconds);

protected:
	int64 m_nStamp;
};

// Used by the Date::Format method
enum DateFormat
{
	dfExpanded,  // YYYY-MM-DD hh:mm:ss
	dfExpandedT, // YYYY-MM-DDThh:mm:ss
	dfCompact    // YYYYMMDDThhmmss
};

// Manages a date+time
class Date
{
public:
	uint16 GetYear() const { return m_nYear; }
	uint8 GetMonth() const { return m_nMonth; }
	uint8 GetDay() const { return m_nDay; }
	uint8 GetHour() const { return m_nHour; }
	uint8 GetMinute() const { return m_nMinute; }
	uint8 GetSecond() const { return m_nSecond; }
	
	int64 GetStampInSec() const;
	int64 GetChuTimeStamp() const;

	void AddDays(int32 nCount);
	void SetTime(uint8 nHour, uint8 nMinute, uint8 nSecond);

	static Date FromChuTimeStamp(int64 nStamp);

public:
	const Duration operator - (const Date& t2) const;
	bool operator > (const Date& t2) const;
	bool operator < (const Date& t2) const;
	bool operator >= (const Date& t2) const;
	bool operator <= (const Date& t2) const;
	
	const Date& operator += (Duration dur);

	static Date GetNow();
	
	static Date FromUnixTimeStamp(int64 nTime);
	int64 ToUnixTimeStamp() const;
	
	// Formats date and hour as YYYY-MM-DD hh:mm:ss (ISO 8601)
	String Format(DateFormat eFormat = dfExpanded) const;

	// Formats date only as YYYY-MM-DD
	String FormatDate(DateFormat eFormat = dfExpanded) const;

	// Formats time only as hh:mm:ss
	String FormatTime(DateFormat eFormat = dfExpanded) const;

	// Gets date as YYYY-MM-DD hh:mm:ss
	// Works if last pars are missing, for instance : YYYY-MM 
	// Returns the number of read members
	int32 Scan(const String& strDate);

	void Set(int16 nYear, uint8 nMonth = 1, uint8 nDay = 1, uint8 nHour = 0, uint8 nMinute = 0, uint8 nSecond = 0);

	Date();

protected:
	int64 m_nStamp;

	uint16 m_nYear;
	uint8 m_nMonth;
	uint8 m_nDay;

	uint8 m_nHour;
	uint8 m_nMinute;
	uint8 m_nSecond;

	uint8 m_nDummy;

protected:
	// Updates m_nYear, m_nMonth, m_nDay, m_nHour, m_nMinute, m_nSecond
	void UpdateFieldsFromStamp();

	// Updates m_nStamp
	void UpdateStampFromFields();
};

} // namespace chustd

#endif // ndef CHUSTD_DATE_H
