///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "Master.h"
#include "File.h"
#include "StringBuilder.h"

namespace chustd {
	class FileImpl
	{
	public:
#ifdef WIN32
		HANDLE m_handle; // HANDLE : type Win32
#elif CHUOS
		IFileSystem::Handle m_handle;
#endif
		uint32 m_nOpenMode;

	public:
		FileImpl()
		{
			m_handle = 0;
		}
	};
}

//////////////////////////////////////////////////////////////////////
using namespace chustd;
//////////////////////////////////////////////////////////////////////

File::File()
{
	m_pImpl = new FileImpl;
}

File::~File()
{
	Close();
	delete m_pImpl;
}

bool File::Open(const String& strFilePath, uint32 nMode/*=modeRead*/)
{
	if( strFilePath.IsEmpty())
		return false;

#ifdef WIN32
	int nAccess = 0;
	int nDisposition = 0;
	int nShareMode = FILE_SHARE_READ;

	bool bOpenRead = (nMode & modeRead) != 0;
	bool bOpenWrite = (nMode & modeWrite) != 0;
	bool bAppend = (nMode & modeAppend) != 0;

	if( !bAppend)
	{
		if( bOpenRead && !bOpenWrite)
		{
			// Read only on an existing file
			nAccess = GENERIC_READ;
			nDisposition = OPEN_EXISTING;
		}
		else if( !bOpenRead && bOpenWrite)
		{
			// Write only on a unexisting or existing file
			nAccess = GENERIC_WRITE;
			nDisposition = CREATE_ALWAYS;
		}
		else if( bOpenRead && bOpenWrite)
		{
			// If the file exists, open it and set at the start of it
			// If the file does not exist, create it
			nAccess = GENERIC_READ | GENERIC_WRITE;
			nDisposition = OPEN_ALWAYS;
		}
		else
		{
			// Bad open mode
			return false;
		}
	}
	else
	{
		// Note : the append mode in chustd has a different behaviour in comparison with the Libc:
		// A call to SetPosition followed by a call to Write, writes data at the position specified by SetPosition.
		// The Libc seeks to the end of the file before writing.

		// In append mode, the file is implicitely open in write mode
		nMode |= modeWrite;
		nAccess = GENERIC_WRITE;

		if( bOpenRead )
		{
			nAccess |= GENERIC_READ;
		}
		nDisposition = OPEN_ALWAYS;
	}

	HANDLE handle = CreateFileW(strFilePath.GetBuffer(), 
		nAccess, nShareMode, NULL, nDisposition, FILE_ATTRIBUTE_NORMAL, NULL);
	if( handle == INVALID_HANDLE_VALUE )
	{
		return false;
	}

	if( bAppend )
	{
		DWORD nRet = SetFilePointer(handle, 0, 0, FILE_END);
	}
#elif CHUOS
	/*
	IDrive* pDrive = Kernel::GetDriveFromPath(strFilePath);
	String strAbsoluteFilePath = pDrive->GetAbsolutePath(strFilePath);
	*/
	IFileSystem* pFileSystem = null;
	IFileSystem::Handle handle = pFileSystem->OpenFile(strFilePath, nMode);
#endif

	m_pImpl->m_handle = handle;
	m_pImpl->m_nOpenMode = nMode;
	return true;
}

void File::Close()
{
	if( !m_pImpl)
		return;

	if( m_pImpl->m_handle== null)
		return;

#ifdef WIN32
	CloseHandle(m_pImpl->m_handle);
#elif CHUOS
	IFileSystem* pFileSystem = null;
	pFileSystem->CloseFile(m_pImpl->m_handle);
#endif

	m_pImpl->m_handle = 0;
}

ByteOrder File::GetByteOrder() const
{
	if( !m_pImpl)
	{
		return boBigEndian;
	}

	if( (m_pImpl->m_nOpenMode & modeLittleEndian) != 0)
	{
		return boLittleEndian;
	}
	return boBigEndian;
}

void File::SetByteOrder(ByteOrder eByteOrder)
{
	if( !m_pImpl)
		return;

	if( eByteOrder == boLittleEndian)
	{
		m_pImpl->m_nOpenMode |= modeLittleEndian;
	}
	else if( eByteOrder == boBigEndian)
	{
		m_pImpl->m_nOpenMode &= !modeLittleEndian;
	}
	else
	{
		ASSERT(0);	
	}
}


int32 File::Read(void* pBuffer, int32 nSize)
{
	if( !m_pImpl)
		return 0;

	if( m_pImpl->m_handle == null)
		return 0;

	if( nSize <= 0)
		return 0;

#ifdef WIN32
	DWORD nRead = 0;
	BOOL bRet = ReadFile(m_pImpl->m_handle, pBuffer, nSize, &nRead, NULL);
#elif CHUOS
	IFileSystem* pFileSystem = null;
	int32 nRead = pFileSystem->ReadFile(m_pImpl->m_handle, pBuffer, nSize);
#endif
	return int32(nRead);
}

int32 File::Write(const void* pBuffer, int32 nSize)
{
	if( !m_pImpl)
		return 0;

	if( m_pImpl->m_handle == null)
		return 0;

	if( nSize <= 0)
		return 0;

#ifdef WIN32
	DWORD nWritten = 0;
	BOOL bRet = WriteFile(m_pImpl->m_handle, pBuffer, nSize, &nWritten, NULL);
#elif CHUOS
	IFileSystem* pFileSystem = null;
	int32 nWritten = pFileSystem->WriteFile(m_pImpl->m_handle, pBuffer, nSize);
#endif

	return int32(nWritten);
}

bool File::SetPosition(int64 nOffset, Whence eWhence)
{
	if( !m_pImpl)
		return false;

	if( m_pImpl->m_handle == null)
		return false;

#ifdef WIN32
	int32 nCFrom = 0;
	switch(eWhence)
	{
	case posBegin:
		nCFrom = FILE_BEGIN;
		break;
	case posCurrent:
		nCFrom = FILE_CURRENT;
		break;
	case posEnd:
		nCFrom = FILE_END;
		break;
	default:
		ASSERT(0); // Bad value for seek
		break;
	}
	
	LONG nHigh = LONG(nOffset >> 32);
	DWORD nLow = DWORD(nOffset & 0x00000000ffffffff);
	DWORD nRet = SetFilePointer(m_pImpl->m_handle, nLow, &nHigh, nCFrom);

	DWORD nError = 0;
	if( nRet == INVALID_SET_FILE_POINTER && (nError = GetLastError()) != NO_ERROR )
	{
		return false;
	}
	return true;

#elif CHUOS
	IFileSystem* pFileSystem = null;
	pFileSystem->SetFilePosition(m_pImpl->m_handle, nOffset, eWhence);
	return true;
#endif
}

int64 File::GetPosition() const
{
	if( !m_pImpl)
		return 0;

	if( m_pImpl->m_handle == null)
		return 0;

#ifdef WIN32
	LONG nHigh = 0;
	DWORD nRet = SetFilePointer(m_pImpl->m_handle, 0, &nHigh, FILE_CURRENT);

	DWORD nError = 0;
	if( nRet == INVALID_SET_FILE_POINTER && (nError = GetLastError()) != NO_ERROR )
	{
		return 0;
	}

	int64 nPosition = nHigh;
	nPosition <<= 32;
	nPosition |= nRet;

	return nPosition;

#elif CHUOS
	IFileSystem* pFileSystem = null;
	return pFileSystem->GetFilePosition(m_pImpl->m_handle);
#endif
}

int64 File::GetSize()
{
	if( !m_pImpl)
		return 0;

	if( !m_pImpl->m_handle)
		return 0;

#ifdef WIN32
	DWORD nSizeHigh = 0;
	DWORD nSizeLow = GetFileSize(m_pImpl->m_handle, &nSizeHigh);

	int64 nLength = nSizeHigh;
	nLength <<= 32;
	nLength |= nSizeLow;

	return nLength;

#elif CHUOS
	IFileSystem* pFileSystem = null;
	return pFileSystem->GetFileSize(m_pImpl->m_handle);
#endif
}

//////////////////////////////////////////////////////////////////////////
bool File::Rename(const String& strOldName, const String& strNewName)
{
#ifdef WIN32
	BOOL bRet = MoveFileW(strOldName.GetBuffer(), strNewName.GetBuffer());
	return bRet != FALSE;
#elif CHUOS
	// Get the file names alone
	String strDir, strDir2;
	String strOldName2, strNewName2;

	File::SplitPath(strOldName, strDir, strOldName2);
	File::SplitPath(strNewName, strDir2, strNewName2);

	IFileSystem* pFileSystem = null;
	return pFileSystem->RenameFile(strDir, strOldName2, strNewName2);
#endif
}

void File::SplitPath(const String& strPath, String& strDir, String& strFileName)
{
	const int32 nLength = strPath.GetLength();
	if( nLength == 0)
	{
		strDir.Empty();
		strFileName.Empty();
		return;
	}
	
	int32 iSeparator = nLength - 1;
	for(; iSeparator >= 0; --iSeparator)
	{
		uint16 c = strPath.GetAt(iSeparator);
		if( c == '\\' || c == '/' || c == ':')
		{
			break;
		}
	}

	if( iSeparator < 0)
	{
		// No separator found
		strDir.Empty();
		strFileName = strPath;
		return;
	}

	strDir = strPath.Left(iSeparator + 1); // +1 in order to include the separator
	strFileName = strPath.Right( (nLength - iSeparator) - 1);
}

String File::GetDrive(const String& strPath)
{
	const int32 nLength = strPath.GetLength();
	if( nLength == 0)
	{
		return String();
	}
	
	uint16 c = 0;

	int32 iSeparator = 0;
	for(; iSeparator < nLength; ++iSeparator)
	{
		c = strPath.GetAt(iSeparator);
		if( c == '\\' || c == '/' || c == ':')
		{
			break;
		}
	}

	if( iSeparator < 0)
	{
		// No separator found
		return String();
	}

	if( c == ':')
	{
		// Drive found
		return strPath.Left(iSeparator);
	}
	
	return String();
}

String File::AddSeparator(const String& strDir, wchar cModel /*='/'*/)
{
	const int32 nLength = strDir.GetLength();
	if( nLength > 0)
	{
		const wchar c = strDir.GetAt(nLength - 1); // Last character
		if( (c != '\\') && (c != '/') )
		{
			return strDir + String(&cModel, 1);
		}
	}
	return strDir;
}

String File::RemoveExtension(const String& strFilePath)
{
	const int32 nLength = strFilePath.GetLength();
	if( nLength > 0)
	{
		for(int32 i = nLength - 1; i >= 0; i--)
		{
			if( strFilePath.GetAt(i) == L'.')
			{
				return strFilePath.Left(i);
			}
		}
	}
	return strFilePath;
}


// Cuts string to parent's directory
// Does not leave the final separator
String File::CutToParent(const String& strPath)
{
	int32 nLength = strPath.GetLength();
	
	const wchar* pszPath = strPath.GetBuffer();

	// Is last character a separator ?
	wchar c = pszPath[nLength - 1];
	if( c == L':')
		return strPath; // A partition, do not modify the string

	if( c == L'\\' || c == L'/' )
		nLength--; // Yep, do not count the separator

	int32 i = nLength - 1;
	for(;i >= 0; i--)
	{
		if( pszPath[i] == L':' || pszPath[i] == L'\\' || pszPath[i] == L'/')
			break;
	}
	
	return strPath.Left(i + 1);
}

// Returns true if a path name is absolute

// plop/onk.txt returns false
// ../plop/onk.txt returns false

// file://plop/onk.txt returns true
// d:/plop/onk.txt returns true
// d:onk.txt returns true
// /plop/onk.txt returns true
bool File::IsAbsolutePath(const String& strPath)
{
	const int32 nLength = strPath.GetLength();
	if( nLength == 0)
		return false;

	const wchar* pszPath = strPath.GetBuffer();

	bool bSeparatorFound = false;
	int32 i = 0;
	for(; i < nLength; i++)
	{
		if( pszPath[i] == '/'
			|| pszPath[i] == '\\'
			|| pszPath[i] == ':')
		{
			bSeparatorFound = true;
			break;
		}
	}
	
	if( !bSeparatorFound)
		return false;

	if( pszPath[i] == ':')
		return true;

	if( pszPath[0] == '/' || pszPath[0] == '\\')
		return true;

	return false;
}

String File::GetExtension(const String& strFilePath)
{
	const int32 nLength = strFilePath.GetLength();
	String strExt;

	for(int32 i = nLength - 1; i >= 0; i--)
	{
		wchar c = strFilePath.GetAt(i);
		if( c == L'.')
		{
			const int32 nCount = nLength - 1 - i;
			strExt = strFilePath.Right(nCount);
			break;
		}
	}
	return strExt;
}

// Non static
// 2 Go Max
ByteArray File::GetContent()
{
	const int64 nCurrentPosition = GetPosition();
	int64 nFileLength = GetSize();
	
	if( nFileLength > 0x000000007fffffff)
	{
		nFileLength = 0x000000007fffffff;
	}
	
	ByteArray aContent;
	aContent.SetSize( int32(nFileLength));

	Read(aContent.GetPtr(), int32(nFileLength));
	
	SetPosition(nCurrentPosition);
	
	aContent.SetAsReturnValue();
	return aContent;
}

// Static
ByteArray File::GetContent(const String& strFilePath)
{
	File file;
	if( !file.Open(strFilePath, File::modeRead))
	{
		return ByteArray();
	}

	ByteArray aContent = file.GetContent();

	aContent.SetAsReturnValue();
	return aContent;
}

// Non static
String File::GetTextContent(const TextEncoding& te)
{
	ByteArray aContent = GetContent();
	return String::FromBytes(aContent, te);
}

// Static
String File::GetTextContent(const String& strFilePath, const TextEncoding& te)
{
	ByteArray aContent = GetContent(strFilePath);
	return String::FromBytes(aContent, te);
}

// Non static
StringArray File::GetLines(const TextEncoding& te)
{
	String strContent = GetTextContent(te);
	StringArray astrLines = strContent.SplitByEndlines();
	
	astrLines.SetAsReturnValue();
	return astrLines;
}

// Static
StringArray File::GetLines(const String& strFilePath, const TextEncoding& te)
{
	File file;
	if( !file.Open(strFilePath, File::modeRead))
	{
		return StringArray();
	}

	StringArray astrLines = file.GetLines(te);

	astrLines.SetAsReturnValue();
	return astrLines;
}


bool File::Exists(const String& strPath)
{
	bool bDirectory, bReadOnly;
	return GetFileAttributes(strPath, bDirectory, bReadOnly) && !bDirectory;
}

int64 File::GetSize(const String& strPath)
{
	File file;
	if( !file.Open(strPath))
		return 0;

	return file.GetSize();
}

bool File::GetFileTime(const String& strPath, uint64& nCreationTime, uint64& nLastAccessTime, uint64& nLastWriteTime)
{
	File file;
	if( !file.Open(strPath, File::modeRead) )
	{
		return false;
	}

#ifdef WIN32
	FILETIME timeCreation = {0};
	FILETIME timeLastAccess = {0};
	FILETIME timeLastWrite = {0};
	BOOL bGetTime = ::GetFileTime(file.m_pImpl->m_handle, &timeCreation, &timeLastAccess, &timeLastWrite);

	if( !bGetTime )
	{
		return false;
	}

	nCreationTime = *((uint64*)&timeCreation);
	nLastAccessTime = *((uint64*)&timeLastAccess);	
	nLastWriteTime = *((uint64*)&timeLastWrite);

#elif CHUOS
	Date dateCreation, dateModification;
	IFileSystem* pFileSystem = null;
	pFileSystem->GetFileTime(file.m_pImpl->m_handle, dateCreation, dateModification);
	
	nCreationTime = dateCreation.ToUnixTimeStamp();
	nLastAccessTime = 0;
	nLastWriteTime = dateModification.ToUnixTimeStamp();
#endif
	return true;
}

String File::GetThisProgramName()
{
#ifdef WIN32
	HINSTANCE hInstance = GetModuleHandle(NULL);

	wchar szBuffer[MAX_PATH];
	::GetModuleFileName(hInstance, szBuffer, MAX_PATH);

	String strProgram(szBuffer);
	return strProgram;

#elif CHUOS
	return String();
#endif
}


String File::GetAbsolutePath(const String& strCurrentDir, const String& strRelativePath, wchar cModel)
{
	wchar szModel[2] = {cModel, 0};

	ASSERT(strCurrentDir.StartsWith(szModel));

	String strPath;

	if( strRelativePath.StartsWith(szModel))
	{
		// Starts with the root dir, no need to cat with strCurrentDir
		strPath = strRelativePath;
	}
	else
	{
		strPath = File::AddSeparator(strCurrentDir);
		strPath = strPath + strRelativePath;
	}

	StringArray astrFragments = strPath.Split(cModel);
	StringArray astrStack;

	foreach(astrFragments, i)
	{
		String& strFrag = astrFragments[i];
		if( strFrag == L"..")
		{
			// Before : plop/abc/..
			// After:   plop/
			astrStack.RemoveLast();
		}
		else if( strFrag == L".")
		{
			// Before : plop/abc/.
			// After:   plop/abc/
		}
		else
		{
			if( !strFrag.IsEmpty())
			{
				astrStack.Add(strFrag);
			}
		}
	}

	if( astrStack.GetSize() == 0)
		return String(szModel, 1);

	StringBuilder sbResult;
	foreach(astrStack, i)
	{
		sbResult += cModel;
		sbResult += astrStack[i];
	}
	return sbResult.ToString();
}


String File::GetAbsolutePath(const String& strFilePath)
{
#ifdef WIN32
	const int nLength = ::GetFullPathName(strFilePath.GetBuffer(), 0, NULL, NULL);

	wchar* pBuffer = new wchar[nLength];
	wchar* pFilePart;
	::GetFullPathName(strFilePath.GetBuffer(), nLength, pBuffer, &pFilePart);

	String strFullPathName(pBuffer);

	delete[] pBuffer;

	return strFullPathName;

#elif CHUOS
	return strFilePath;
#endif
}

bool File::Copy(const String& strPathSrc, const String& strPathDst)
{
#ifdef WIN32
	BOOL bRet = ::CopyFileW(strPathSrc.GetBuffer(), strPathDst.GetBuffer(), FALSE);
	return bRet != 0;
#elif CHUOS
	IFileSystem* pFileSystem = null;
	return pFileSystem->CopyFile(strPathSrc, strPathDst);
#endif
}


bool File::GetFileAttributes(const String& strPath, bool& bDirectory, bool& bReadOnly)
{
#ifdef WIN32
	DWORD dwAttributs = ::GetFileAttributesW(strPath.GetBuffer());
	if( dwAttributs == INVALID_FILE_ATTRIBUTES)
		return false;

	bDirectory = (dwAttributs & FILE_ATTRIBUTE_DIRECTORY) != 0;
	bReadOnly = (dwAttributs & FILE_ATTRIBUTE_READONLY) != 0;

#elif CHUOS
	/*
	IFileSystem* pFileSystem = null;
	if( !pFileSystem->GetFileAttributes(strPath, bDirectory, bReadOnly))
		return false;
	*/
#endif
	return true;
}


bool File::Delete(const String& strPath)
{
#ifdef WIN32
	return ::DeleteFileW(strPath.GetBuffer()) != FALSE;

#elif CHUOS

	IFileSystem* pFileSystem = null;
	return pFileSystem->DeleteFile(strPath);
#endif
}

String File::AddFileNamePrefix(const String& strPath, const String& strPrefix)
{
	String strDir, strFileName;
	SplitPath(strPath, strDir, strFileName);

	String strNewPath = strDir + strPrefix + strFileName;
	return strNewPath;
}
