///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "Master.h"
#include "Font.h"
#include "String.h"
#include "Array.h"
#include "TextEncoding.h"
#include "File.h"
#include "System.h"

// text
#include "freetype/freetype.h"
#include "freetype/ftglyph.h"

namespace chustd {

////////////////////////////////////////
class FontLibrary
{
public:
	FT_Library m_ftLib;
	bool m_bInitOk;
	
public:
	FontLibrary()
	{
		if( FT_Init_FreeType( &m_ftLib) != 0)
		{
			m_bInitOk = false;
		}
		m_bInitOk = true;
	}
	~FontLibrary()
	{
		if( m_bInitOk)
		{
			FT_Done_FreeType( m_ftLib);
		}
	}
};

////////////////////////////////////////
class FontImpl
{
public:
	FT_Face m_ftFace;
	
	class SurfaceForGlyph : public Surface8
	{
	public:
		void SetBufferFromFTBitmap(FT_Bitmap& ftb);
	};
	mutable SurfaceForGlyph m_surfGlyph;

	int32 m_nAscent;
	int32 m_nDescent;
	int32 m_nExtraLeading; // Haut supplmentaire  ajouter entre deux lignes de texte (dfaut 0)

	FT_Glyph m_hGlyphA;
	
	FontImpl()
	{
		m_ftFace = 0;

		m_nAscent = 0;
		m_nDescent = 0;
		m_nExtraLeading = 0;

		m_hGlyphA = null;
	}

	~FontImpl()
	{
		// Test de cache
		//FT_Done_Glyph(m_hGlyphA);
	}
};

}

using namespace chustd;

//////////////////////////////////////////////////////////////////////
static FontLibrary g_ftlib;
//////////////////////////////////////////////////////////////////////

Font::Font()
{
	m_pImpl = new FontImpl;
}

Font::~Font()
{
	delete m_pImpl;
}

//////////////////////////////////////////////////////////////////////
void FontImpl::SurfaceForGlyph::SetBufferFromFTBitmap(FT_Bitmap& ftb)
{
	if( ftb.pixel_mode == ft_pixel_mode_grays)
	{
		// 8 bits par pixel, on peut utiliser directement le pointeur fourni par le slot
		SetBuffer((uint8*)ftb.buffer, ftb.width, ftb.rows, ftb.width);
	}
	else if( ftb.pixel_mode == ft_pixel_mode_mono)
	{
		// 1 bit par pixel, va falloir ruser
		Create(ftb.width, ftb.rows);

		uint8* pSrc = (uint8*) ftb.buffer;
		int32 nSrcBytePitch = ftb.pitch;

		uint8* pDst = (uint8*) GetBuffer();
		
		const int32 nHeight = GetHeight();
		const int32 nWidth = GetWidth();

		int32 iDst = 0;
		for(int32 iRow = 0; iRow < nHeight; ++iRow)
		{
			for(int32 iX = 0; iX < nWidth; ++iX)
			{
				int32 nByteOffset = (iRow * nSrcBytePitch) + iX / 8;
				int32 nLocalBitOffset = iX % 8;
					
				uint8 nMask = uint8(0x0080 >> nLocalBitOffset);
				uint8 nByte = pSrc[nByteOffset];
				nByte &= nMask;

				if( nByte == 0)
				{
					pDst[iDst] = 0x00;
				}
				else
				{
					pDst[iDst] = 0xff;
				}
				++iDst;
			}
		}
	}
}

bool Font::Create(const String& strFileName)
{
	if( !g_ftlib.m_bInitOk)
		return false;

	if( !m_pImpl)
		return false;

	if( m_pImpl->m_ftFace != null)
	{
		FT_Done_Face(m_pImpl->m_ftFace);
		m_pImpl->m_ftFace = null;
	}

	String strFilePath = strFileName;
	if( !File::Exists(strFilePath))
	{
		// Not found in local directory, try in OS fonts directory
		String strFontsDir = System::GetFontsDirectory();
		strFontsDir = File::AddSeparator(strFontsDir);
		strFilePath = strFontsDir + strFileName;
	}

	if( FT_New_Face(g_ftlib.m_ftLib, (char*) &strFilePath, 0, &m_pImpl->m_ftFace) != 0)
		return false;

	return true;
}

bool Font::SetSize(int32 nSize)
{
	if( m_pImpl == null)
		return false;

	FT_Face ftFace = m_pImpl->m_ftFace;

	int nRet = FT_Set_Pixel_Sizes(ftFace, 0, nSize);
	if( nRet != 0)
		return false;

	///////////////////////////////////////////////////////////////////
	// Compute ascent
	int32 nAscent = (ftFace->size->metrics.ascender) >> 6;
	int32 nAscent2 = 0;
	uint16 nUnicodeChar = 0x00c0; // A GRAVE ACCENT

	// Ne suffit pas toujours, on charge donc un glyph qui couvre bien tout l'ascent
	FT_GlyphSlot slot = ftFace->glyph;
	int nLoadRes = FT_Load_Char( ftFace, nUnicodeChar, FT_LOAD_RENDER);
	if( nLoadRes != 0)
	{
		return false;
	}
	nAscent2 = slot->bitmap_top;

	if( nAscent2 > nAscent)
	{
		// Ah bah oui. Exemple : Arial en taille 11 :
		// m_ftFace nous informe d'un ascent de 10 mais en ralit il vaut 11 ; youpi :p
		nAscent = nAscent2;
	}

	m_pImpl->m_nAscent = nAscent;
	
	///////////////////////////////////////////////////////////////////
	// Calcule le descent
	
	int32 nDescent = -(ftFace->size->metrics.descender) >> 6;
	
	m_pImpl->m_nDescent = nDescent;
	///////////////////////////////////////////////////////////////////

	// Test de cache
	//int error = FT_Load_Char( m_pImpl->m_ftFace, 'A', FT_LOAD_DEFAULT );
	//error = FT_Get_Glyph( m_pImpl->m_ftFace->glyph, &m_pImpl->m_hGlyphA );

	return true;
}

// Retourne la hauteur des caractres de la police courante
int32 Font::GetSize() const
{
	int32 nSize = m_pImpl->m_ftFace->size->metrics.y_ppem;
	return nSize;
}

int32 Font::GetAscent() const
{
	return m_pImpl->m_nAscent;
}

int32 Font::GetDescent() const
{
	return m_pImpl->m_nDescent;
}

void Font::SetExtraLeading(int32 nExtraLeading)
{
	m_pImpl->m_nExtraLeading = nExtraLeading;
}

int32 Font::GetLineHeight() const
{
	int32 nLineHeight = GetAscent() + GetDescent() + m_pImpl->m_nExtraLeading;
	return nLineHeight;
}

void Font::GetTextRect(const String& strText, int32& nRectWidth, int32& nRectHeight)
{
	StringArray aStrings = strText.SplitByEndlines();

	const int32 nPieceCount = aStrings.GetSize();
	const int32 nHeightPx = nPieceCount * GetLineHeight();

	int32 nMaxWidth = 0;
	for(int iLine = 0; iLine < nPieceCount; ++iLine)
	{
		const String& strLine = aStrings[iLine];
		int32 nWidth = GetTextWidth(strLine);
		if( nWidth > nMaxWidth)
		{
			nMaxWidth = nWidth;
		}
	}

	nRectWidth = nMaxWidth;
	nRectHeight = nHeightPx;
}

int32 Font::GetTextHeight(const chustd::String& strText, int32 nRectWidth)
{
	StringArray aStrings = SplitString(strText, nRectWidth);

	const int32 nPieceCount = aStrings.GetSize();
	const int32 nTotalHeight = nPieceCount * GetLineHeight();

	return nTotalHeight;
}

int32 Font::GetTextWidth(const chustd::String& strText) const
{
	const int32 nCharCount = strText.GetLength();

	FT_Face ftFace = m_pImpl->m_ftFace;

	if( ftFace == 0)
		return 0;

	FT_GlyphSlot slot = ftFace->glyph;

	int32 nPenX = 0;
	//int32 nPenY = 0;

	for(int32 i = 0; i < nCharCount; i++ )
	{
		// Load glyph image into the slot (erase previous one)
		uint16 nUnicodeChar = strText.GetAt(i);
		FT_Load_Char( ftFace, nUnicodeChar, FT_LOAD_DEFAULT );
		
		//int32 x = nPenX + slot->bitmap_left;
		//int32 y = nPenY - slot->bitmap_top;

		// Increment pen position 
		nPenX += slot->advance.x >> 6;
	}

	return nPenX;
}

const Surface8& Font::RenderChar(uint32 nCodePoint, int32& nOffsetX, int32& nOffsetY, int32& nAdvance) const
{
	FT_Face ftFace = m_pImpl->m_ftFace;

	if( ftFace == 0)
	{
		m_pImpl->m_surfGlyph.Fill(255);
		return m_pImpl->m_surfGlyph;
	}


	if(0)// nCodePoint == 'A' ) // // Test de cache
	{
		FT_Vector origin;
		origin.x = 32; /* 1/2 pixel in 26.6 format */
		origin.y = 0; 
		int error = FT_Glyph_To_Bitmap( &m_pImpl->m_hGlyphA, FT_RENDER_MODE_NORMAL, 0, 0); 
	
		FT_BitmapGlyph gb = (FT_BitmapGlyph) m_pImpl->m_hGlyphA;
		
		m_pImpl->m_surfGlyph.SetBufferFromFTBitmap(gb->bitmap);

		nOffsetX = gb->left;
		nOffsetY = -gb->top;

		nAdvance = gb->root.advance.x >> 16; // Advance stored in 16:16 in FT_Glyph
	}
	else
	{
		FT_GlyphSlot slot = ftFace->glyph;

		// Load glyph image into the slot (erase previous one)
		
		// 0 = pas d'erreur
		int32 nErrCode = FT_Load_Char( ftFace, nCodePoint, FT_LOAD_RENDER);
		if( nErrCode != 0)
		{
			m_pImpl->m_surfGlyph.Fill(255);
			return m_pImpl->m_surfGlyph;
		}

		m_pImpl->m_surfGlyph.SetBufferFromFTBitmap(slot->bitmap);

		nOffsetX = slot->bitmap_left;
		nOffsetY = -slot->bitmap_top;
	
		nAdvance = slot->advance.x >> 6; // Advance stored in 26:6 in FT_GlyphSlot
	}

	return m_pImpl->m_surfGlyph;
}

void Font::GetBlankPositions(const chustd::String& str, Int32Array& anPositions) const
{
	anPositions.SetSize(0);

	const int32 nLength = str.GetLength();
	if( nLength < 0)
		return;

	const uint16* psz = (uint16*) str.GetBuffer();
	for(int32 i = 0; i < nLength; i++)
	{
		if( String::IsBlankChar(psz[i]))
		{
			anPositions.Add(i);
		}
	}
}

chustd::StringArray Font::SplitString(const chustd::String& strOriginalText, int32 nRectWidth) const
{
	StringArray astrPieces;
	astrPieces.SetAsReturnValue();

	if( nRectWidth <= 0)
	{
		// Just split according to endlines
		return strOriginalText.SplitByEndlines();
	}

	// First split by endlines
	chustd::StringArray astrFirstPass = strOriginalText.SplitByEndlines();

	////////////////////////////////////////////////////////
	chustd::Int32Array anPositions;

	const int32 nFirstPassCount = astrFirstPass.GetSize();
	for(int32 iFPStr = 0; iFPStr < nFirstPassCount; iFPStr++)
	{
		const String& strText = astrFirstPass[iFPStr];
		
		int32 nTextWidth = GetTextWidth(strText);
		if( nTextWidth > nRectWidth)
		{
			// Text too large, try to split it
			GetBlankPositions(strText, anPositions);
			anPositions.Add( strText.GetLength());
			
			chustd::String strPreviousPiece;
			chustd::String strPiece;
			
			int32 nPreviousPieceWidth = 0;
			int32 nPieceWidth = 0;
			
			int32 nPreviousPieceLength = 0;
			int32 nPieceLength = 0;
			
			int32 nStartChar = 0;
			const int32 nPosCount = anPositions.GetSize();
			for(int32 i = 0; i < nPosCount; i++)
			{
				nPreviousPieceLength = nPieceLength;
				nPieceLength = anPositions[i] - nStartChar;
				
				strPreviousPiece = strPiece;
				strPiece = strText.Mid(nStartChar, nPieceLength);
				
				nPreviousPieceWidth = nPieceWidth;
				nPieceWidth = GetTextWidth(strPiece);
				
				if( nPieceWidth >= nRectWidth)
				{
					// We've gone too far, we keep the previous piece
					if( nPreviousPieceLength > 0)
					{
						astrPieces.Add(strPreviousPiece);
						
						// Get rid of the last piece that made the string too long
						i--;
						nPieceWidth = 0;
						nPieceLength = 0;
						strPiece.Empty();
					}
					else
					{
						// Cannot be cut
						astrPieces.Add(strPiece);
						
						nPieceWidth = 0;
						nPieceLength = 0;
						strPiece.Empty();
					}
					nStartChar = anPositions[i] + 1;
				}
			}
			
			// Add the last piece
			if( nPieceLength > 0)
			{
				astrPieces.Add(strPiece);
			}
		}
		else
		{
			astrPieces.Add(strText);	
		}
	}

	return astrPieces;
}

