///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "Master.h"
#include "Gif.h"

//////////////////////////////////////////////////////////////////////
using namespace chustd;

static const uint8 k_GifSignature[3] = {'G', 'I', 'F'};
//////////////////////////////////////////////////////////////////////

Gif::Gif()
{
	Initialize();
}

Gif::~Gif()
{

}

void Gif::Initialize()
{
	m_nWidth = 0;
	m_nHeight = 0;
	m_nLastError = 0;
	m_nCurrentFrame = 0;

	m_nBackgroundColorIndex = 0;
	m_bTransparencyUsed = false;
	m_nTransparentColorIndex = -1;
	m_nDisposalMethod = -1;
	m_nDelayTime = -1;
	m_eVersion = gifverNotSet;
	
	m_nLoopCount = 0;

	m_GlobalColorTable.m_nCount = 0;
	
	m_aFrames.SetSize(0);
}

PixelFormat Gif::GetPixelFormat() const
{
	// Even for depths < 8, each pixel takes one byte in memory
	return pf8bppIndexed;
}

const Palette* Gif::GetPalette() const
{
	const int32 nFrameCount = m_aFrames.GetSize();
	
	if( nFrameCount > 0 )
	{
		const GifAnimFrame& frame = m_aFrames[m_nCurrentFrame];
		if( frame.m_pPalette != null )
		{
			return frame.m_pPalette;
		}
	}

	return &m_GlobalColorTable;
}

bool Gif::IsInterlaced() const
{ 
	const int32 nFrameCount = m_aFrames.GetSize();
	
	if( nFrameCount > 0 )
	{
		const GifAnimFrame& frame = m_aFrames[m_nCurrentFrame];
		return frame.m_bInterlaced;
	}

	// Warn the user
	ASSERT(0);
	return false;
}

int32 Gif::GetFrameCount() const
{
	return m_aFrames.GetSize();
}

int32 Gif::GetFrameIndex() const
{
	return m_nCurrentFrame;
}

void Gif::SetFrameIndex(int32 nFrameIndex)
{
	const int32 nFrameCount = m_aFrames.GetSize();
	if( !(0 <= nFrameIndex && nFrameIndex < nFrameCount) )
	{
		ASSERT(0); // Bad index
		return;
	}

	m_nCurrentFrame = nFrameIndex;

	GifAnimFrame& frame = m_aFrames[nFrameIndex];
	AdjustPalette(frame);

	m_nWidth = frame.m_nWidth;
	m_nHeight = frame.m_nHeight;
	m_aPixelBuffer = frame.m_aPixelBuffer;
}

const FrameInfo* Gif::GetFrameInfo()
{
	int32 nFrameIndex = m_nCurrentFrame;

	const int32 nFrameCount = m_aFrames.GetSize();
	if( !(0 <= nFrameIndex && nFrameIndex < nFrameCount) )
	{
		ASSERT(0); // Bad index
		return null;
	}

	GifAnimFrame& frame = m_aFrames[nFrameIndex];
	frame.m_info.m_nClearIndex = GetClearIndex();
	return &frame.m_info;
}

void Gif::AdjustPalette(GifAnimFrame& frame)
{
	if( frame.m_pPalette == null )
	{
		// Sometimes the palette is global but the transparent index is 
		// not constant from frame to frame
		if( frame.m_nTransparentIndex >= 0 )
		{
			m_GlobalColorTable.SetAlphaFullOpaque();
			m_GlobalColorTable.m_aColors[ frame.m_nTransparentIndex].SetAlpha(0);
		}
	}
}

FrameInfo::Disposal Gif::GenDisposalFromGifDisposal(uint8 nGifDisposal)
{
	switch( nGifDisposal )
	{
	case gdmNotSpecified:
		return FrameInfo::dispNone;
	case gdmDoNotDispose:
		return FrameInfo::dispNone;
	case gdmRestoreToBkColor:
		return FrameInfo::dispClearToTransBlack;
	case gdmRestoreToPrevious:
		return FrameInfo::dispRestoreToPrevious;
	default:
		break;
	}
	return FrameInfo::dispNone;
}

int32 Gif::GetLoopCount() const
{
	return m_nLoopCount;
}

bool Gif::ReadTerminator(IFile& file)
{
	uint8 nBlockTerminator = 0xff;
	if( !file.Read8(nBlockTerminator))
	{
		m_nLastError = uncompleteFile;
		return false;
	}
	if( nBlockTerminator != 0)
	{
		m_nLastError = errBlockTerminatorNotFound;
		return false;
	}
	return true;
}

bool Gif::IsGif(IFile& file)
{
	const int32 nSignSize = sizeof(k_GifSignature);
	uint8 aRead[nSignSize];

	if( file.Read(aRead, nSignSize) != nSignSize)
	{
		return false;
	}

	if( !chustd::Memory::IsEqual(k_GifSignature, aRead, nSignSize))
	{
		return false;
	}
	return true;
}

bool Gif::LoadFromFile(IFile& file)
{
	/////////////////////////////////////////////////////
	// Reset the object
	FreeBuffer();
	Initialize();

	/////////////////////////////////////////////////////
	// Read the header first
	if( !IsGif(file))
	{
		m_nLastError = errNotAGifFile;
		return false;
	}

	/////////////////////////////////////////////////////
	// Read version (ex: "87a")
	uint8 aVersion[3];
	file.Read(aVersion, 3);

	if( Memory::IsEqual(aVersion, "87a", 3))
	{
		m_eVersion = gifver87a;
	}
	else if( Memory::IsEqual(aVersion, "89a", 3))
	{
		m_eVersion = gifver89a;
	}
	else
	{
		m_nLastError = errBadGifVersion;
		return false;
	}
	/////////////////////////////////////////////////////

	/////////////////////////////////////////////////////
	// Gif works in little endian mode
	TmpEndianMode tem(file, boLittleEndian);

	// Read common header to all gif files
	if( !ReadLogicalScreenDescriptor(file))
	{
		return false;
	}
	
	////////////////////////////////////////////////
	// The remaining data is splitted into blocks
	
	// Number of frames (> 1 means animated gif)
	bool bReadBlocks = true;
	bool bBlockOk = false;
	do
	{
		uint8 nBlockType;
		if( !file.Read8(nBlockType))
		{
			m_nLastError = uncompleteFile;
			return false;
		}
		
		if( nBlockType == 0x2c)
		{
			// 0x2c = ',' = ImageDescriptor
			
			int32 nAddOk = m_aFrames.Add();
			if( nAddOk < 0 )
			{
				// Allocation failure
				m_nLastError = notEnoughMemory;
				return false;
			}

			GifAnimFrame& frame = m_aFrames.GetLast();

			if( !ReadImageDescriptor(file, frame))
			{
				return false;
			}
			
			// Read pixels
			bBlockOk = ReadImageData(file, frame);
			if( bBlockOk)
			{
				bBlockOk = ReadTerminator(file);
			}

			// Modify palette according to a retrieved transparent index
			if( m_nTransparentColorIndex >= 0 )
			{
				// Set global flag, used when clearing the screen
				m_bTransparencyUsed = true;

				frame.m_nTransparentIndex = m_nTransparentColorIndex;

				if( frame.m_pPalette != null )
				{
					frame.m_pPalette->m_aColors[ m_nTransparentColorIndex].SetAlpha(0);
				}
				else
				{
					// The alpha will be set to 0 in the global palette during
					// a call to AdjustPalette, which occurs after loading and after
					// each call to SetCurrentFrame. This is because the transparent
					// index inside the global palette may vary from frame to frame.
				}
				m_nTransparentColorIndex = -1; // We are done with this value
			}

			frame.m_info.m_disposal = GenDisposalFromGifDisposal(m_nDisposalMethod);
			m_nDisposalMethod = 0;

			// Set delay time
			if( m_nDelayTime >= 0 )
			{
				frame.m_info.m_nDelayMs = m_nDelayTime * 10; // Gif is in 100th of second
				m_nDelayTime = -1;
			}
		}
		else if( nBlockType == 0x21)
		{
			// 0x21 = '!'
			bBlockOk = ReadExtensionBlock(file);
		}
		else if( nBlockType == 0x3b)
		{
			// End of image 0x3b = ';'
			bReadBlocks = false;
		}
		else
		{
			// Error, unknown block
			bReadBlocks = false;
		}
	}while( bReadBlocks && bBlockOk);

	if( !bBlockOk )
	{
		return false;
	}

	////////////////////////////////////////////////
	// Uninterlace frames pixel buffers
	foreach(m_aFrames, i)
	{
		if( m_aFrames[i].m_bInterlaced )
		{
			if( !UninterlaceBuffer(m_aFrames[i]))
			{
				return false;
			}
		}
	}

	////////////////////////////////////////////////
	// Prepare the main pixel buffer : background color and first frame pixels
	
	// Optimization for single frame GIFs
	const int nFrameCount = m_aFrames.GetSize();

	// Prepare the palette for the first frame in case it uses the Global palette
	if( nFrameCount > 0 )
	{
		AdjustPalette(m_aFrames[0]);
	}

	if( nFrameCount == 1 )
	{
		GifAnimFrame& frame = m_aFrames[0];
		if( frame.m_info.m_nOffsetX == 0 && frame.m_info.m_nOffsetY == 0 
			&& frame.m_nWidth == m_nWidth && frame.m_nHeight == m_nHeight )
		{
			// Zero copy optimization
			frame.m_aPixelBuffer.MoveTo(m_aPixelBuffer);
			return true;
		}
	}

	// Prepare the main pixel buffer and copy the first frame pixels into it
	const int32 nPixelCount = m_nWidth * m_nHeight;
	if( !m_aPixelBuffer.SetSize(nPixelCount))
	{
		m_nLastError = notEnoughMemory;
		return false;
	}
	
	ClearBackground();

	if( nFrameCount == 0 )
	{
		// No image found
		return true;
	}

	// Prepare the display of the first frame
	CopyFramePixels(m_aFrames[0], false);
	return true;
}

bool Gif::ReadLogicalScreenDescriptor(IFile& file)
{
	// Screen size when the gif was created. Usually ignored.
	uint16 nLogicalScreenWidth;
	uint16 nLogicalScreenHeight;
	uint8  nFields;
	uint8  nBackgroundColorIndex;
	uint8  nPixelAspectRatio;
	
	bool bReadOk = ( 
		file.Read16(nLogicalScreenWidth)
	 && file.Read16(nLogicalScreenHeight)
	 && file.Read8(nFields)
	 && file.Read8(nBackgroundColorIndex)
	 && file.Read8(nPixelAspectRatio));

	if( !bReadOk)
	{
		m_nLastError = uncompleteFile;
		return false;
	}

	bool bGlobalColorTableFlag = false;
	if( nFields & 0x80)
	{
		bGlobalColorTableFlag = true;
	}

	if( bGlobalColorTableFlag)
	{
		//uint8 nSortFlag = ((nFields & 0x08) >> 3);
		const uint8 nLsl = uint8( (nFields & 0x07) + 1);
		m_GlobalColorTable.m_nCount = (1 << nLsl);
	}
	else
	{
		m_GlobalColorTable.m_nCount = 0;
	}

	bool bOk = true;
	if( m_GlobalColorTable.m_nCount != 0 )
	{
		bOk = ReadColorTable(file, m_GlobalColorTable);
	}

	m_nBackgroundColorIndex = nBackgroundColorIndex;
	m_nWidth = nLogicalScreenWidth;
	m_nHeight = nLogicalScreenHeight;

	return bOk;
}

bool Gif::ReadColorTable(IFile& file, Palette& pal)
{
	const int32 nColorCount = pal.m_nCount;
	
	uint8 aBuf[256 * 3];
	const int32 nWantedRead = nColorCount * 3;
	if( file.Read(aBuf, nWantedRead) != nWantedRead)
	{
		m_nLastError = uncompleteFile;
		return false;
	}

	uint8* pSrc = aBuf;
	for(int32 i = 0; i < nColorCount; ++i )
	{
		uint8 r = pSrc[0];
		uint8 g = pSrc[1];
		uint8 b = pSrc[2];
				
		pal.m_aColors[i].SetRgb(r, g, b);

		pSrc += 3;
	}
	return true;
}

bool Gif::ReadExtensionBlock(IFile& file)
{
	uint8 nFunctionCode;
	if( !file.Read8(nFunctionCode))
	{
		m_nLastError = uncompleteFile;
		return false;
	}
	
	if( nFunctionCode == 0xf9 )
	{
		// Graphic Control Extension
		uint8  nByteCount;
		uint8  nPackedField;
		uint16 nDelayTime;
		uint8  nTransparentColorIndex;
		
		if( file.Read8(nByteCount)
		 && file.Read8(nPackedField)
		 && file.Read16(nDelayTime)
		 && file.Read8(nTransparentColorIndex) )
		{
			if( nPackedField & 0x01)
			{
				// The transparent color index is valid
				m_nTransparentColorIndex = nTransparentColorIndex;
			}

			m_nDisposalMethod = (nPackedField >> 2) & 0x07;
			m_nDelayTime = nDelayTime;

			return ReadTerminator(file);
		}
		else
		{
			m_nLastError = uncompleteFile;
			return false;
		}
	}
	else if( nFunctionCode == 0xff )
	{
		// Netscape Extension
		uint8  nByteCount;
		uint8  aNetscape[8]; // "NETSCAPE"
		uint8  aVersion[4];  // "2.0"
		uint8  nSubBlockByteCount;
		uint8  nOne;
		uint16 nIterationCount;

		if( file.Read8(nByteCount)
		 && file.Read(aNetscape, 8)
		 && file.Read(aVersion, 3)
		 && file.Read8(nSubBlockByteCount)
		 && file.Read8(nOne)
		 && file.Read16(nIterationCount) )
		{
			if( Memory::IsEqual(aNetscape, "NETSCAPE", 8) )
			{
				if( nSubBlockByteCount == 3 && nOne == 1 )
				{
					m_nLoopCount = nIterationCount;	
				}
			}

			return ReadTerminator(file);
		}
		else
		{
			m_nLastError = uncompleteFile;
			return false;
		}
	}

	while(true)
	{
		uint8 nByteCount = 0;
		if( !file.Read8(nByteCount))
		{
			m_nLastError = uncompleteFile;
			return false;
		}
		if( nByteCount == 0 )
			break;
			
		file.SetPosition(nByteCount, IFile::posCurrent);
	}
	return true;
}

// The byte type of this block (0x2c = ',') has already been read
bool Gif::ReadImageDescriptor(IFile& file, GifAnimFrame& frame)
{
	uint16 nImageLeft, nImageTop, nImageWidth, nImageHeight;
	
	bool bReadOk = file.Read16(nImageLeft) && file.Read16(nImageTop)
		&& file.Read16(nImageWidth) && file.Read16(nImageHeight);
	
	if( !bReadOk)
	{
		m_nLastError = uncompleteFile;
		return false;
	}

	// Check that the frame rect is inside the main rect
	const bool bRectOk = (0 <= nImageLeft && (nImageLeft + nImageWidth) <= m_nWidth)
		&& (0 <= nImageTop && (nImageTop + nImageHeight) <= m_nHeight);
	if( !bRectOk )
	{
		m_nLastError = errBadImageRect;
		return false;
	}

	frame.m_info.m_nOffsetX = nImageLeft;
	frame.m_info.m_nOffsetY = nImageTop;
	frame.m_nWidth = nImageWidth;
	frame.m_nHeight = nImageHeight;

	uint8 nFields;
	if( !file.Read8(nFields))
	{
		m_nLastError = uncompleteFile;
		return false;
	}
	
	bool bUseLocalColorMap = false;
	if( nFields & 0x80)
	{
		bUseLocalColorMap = true;
	}

	frame.m_bInterlaced = ((nFields & 0x40) != 0);

	if( !bUseLocalColorMap )
	{
		// This frame uses the global palette
		ASSERT(frame.m_pPalette == null);
		frame.m_pPalette = null;
		return true;
	}

	// A palette is following
	frame.m_pPalette = new Palette();
	if( frame.m_pPalette == null )
	{
		m_nLastError = notEnoughMemory;
		return false;
	}
	const int32 nLsl = (nFields & 0x07) + 1;
	frame.m_pPalette->m_nCount = (1 << nLsl);

	bool bOk = true;
	if( frame.m_pPalette->m_nCount != 0 )
	{
		bOk = ReadColorTable(file, *frame.m_pPalette);
	}

	return bOk;
}

void Gif::InitTable()
{
	// Next empty cell index in the string table.
	// The first cells ( 2^m_nCodeSize) are used by the roots.
	// The two nexts are used by specific codes (that's why we do + 2).
	const int32 nRootCount = (1 << m_nCodeSize);
	m_nNextEntry = nRootCount + 2;

	// Initialize the string table
	for(int32 i = 0; i < nRootCount; i++)
	{
		m_aStringTable[i].nValue = uint8(i);
		m_aStringTable[i].nLength = 1;
		m_aStringTable[i].nPrevious = -1;
	}

	// Compute the initial code sizes
	m_nCompressionSize = m_nCodeSize + 1;
}

// Returns the first character of the string
uint8 Gif::OutputString(int32 nIndex, uint8* pPixelBuffer)
{
	// As we store the previous values, we scan the string in reverse order
	uint8 nFirstChar = 0;
	int32 nLength = m_aStringTable[nIndex].nLength;
	for(int32 i = nLength - 1; i >= 0; i--)
	{
		ASSERT(nIndex >= 0 && nIndex < 4096);
		nFirstChar = m_aStringTable[nIndex].nValue;
		
		const int32 nOffset = m_nOutputIndex + i;
		ASSERT(nOffset >= 0);

		pPixelBuffer[nOffset] = nFirstChar;

		nIndex = m_aStringTable[nIndex].nPrevious;
	}
	m_nOutputIndex += nLength;

	return nFirstChar;
}

int32 Gif::ReadBits(uint8* pBuffer, int32 nLength, int32& nBitPosition)
{
	// <abcde> <fghij> <klmno>
	// is coded:
	// hijabcde xklmnofg

	int32 nBytePos1 = nBitPosition / 8;
	int32 nBytePos2 = (nBitPosition + nLength) / 8;
	if( ((nBitPosition + nLength) % 8) == 0)
	{
		// Do not count this byte
		nBytePos2--;
	}

	int32 nValue = 0;
	for(int32 i = nBytePos2; i >= nBytePos1; i--)
	{
		nValue <<= 8;
		nValue |= pBuffer[i];
	}

	const uint32 nMask = ~(0xffffffff << nLength);
	const int32 nBitOffset = nBitPosition % 8;

	nValue >>= nBitOffset;
	nValue &= nMask;

	nBitPosition += nLength;

	if( !(nValue <= m_nNextEntry))
	{
		// Corrupted GIF
		// Note: it is safe to notify an error in the return value
		// as normal codes are limited to 12 bits, here -1 is 32 bits
		return -1;
	}

	return nValue;
}

bool Gif::ReadImageData(IFile& file, GifAnimFrame& frame)
{
	const int32 nPixelCount = frame.m_nWidth * frame.m_nHeight;

	ASSERT(frame.m_aPixelBuffer.GetSize() == 0);
	if( !frame.m_aPixelBuffer.SetSize(nPixelCount))
	{
		// Error in file
		m_nLastError = notEnoughMemory;
		return false;
	}
	
	// More direct access to the pixels
	uint8* const pPixelBuffer = frame.m_aPixelBuffer.GetPtr();

	// Pixel to be written index
	m_nOutputIndex = 0;
	
	// This field meaning is tricky. The value is 2 for 2 colors pictures, otherwise it gives
	// the number of bits per pixel
	uint8 nLzwMinimumCodeSize;
	if( !file.Read8(nLzwMinimumCodeSize))
	{
		m_nLastError = uncompleteFile;
		return false;
	}
	if( nLzwMinimumCodeSize > 12)
	{
		m_nLastError = errBadLzwMinimumCodeSize;
		return false;
	}
	m_nCodeSize = nLzwMinimumCodeSize;
	
	// The LZW algorithm used in GIF is slightly modified.
	// It adds two special codes : "clear code" & "end of information"
	const int32 nClearCode = (1 << m_nCodeSize);
	const int32 nEndOfInformationCode = (1 << m_nCodeSize) + 1;
	
	InitTable();
	
	// Compressed data are made of blocks with 255 bytes maximum in each of them.
	// The anBlocks array contains one block of compressed data.
	// The last bits of the previous block may be used with the first bits of the current block, in
	// order to form a complete code. That's why we allocate two more bytes. The two first bytes
	// of anBlock are the two last bytes of the previous block.
	// Two bytes allow to store 16 bits, which is enough in the worst cases (a 12 bits code overlapping
	// several bytes).
	uint8 anBlock[2 + 256];
	Memory::Zero(anBlock, sizeof(anBlock));
	
	// Position in bits in the data block
	// We start at 16 because the first two bytes are from the previous block.
	int32 nBitPosition = 16;

	// The last code read
	int32 nLastCodeWord = -1;

	// The last character written
	uint8 nCurrentFirstChar = 0;

	// We loop on all the blocks
	while(true)
	{
		uint8 nBlockSize;
		if( !file.Read8(nBlockSize))
		{
			m_nLastError = uncompleteFile;
			return false;
		}
		
		if( nBlockSize == 0 )
			break;
		
		// Read the compressed block
		if( file.Read( anBlock + 2, nBlockSize) != nBlockSize)
		{
			m_nLastError = uncompleteFile;
			return false;
		}
		
		const int32 nMaxBitSize = (nBlockSize + 2) * 8;

		// Uncompress the block
		while( (nBitPosition + m_nCompressionSize) <= nMaxBitSize)
		{
			int32 nCurrentCodeWord = ReadBits(anBlock, m_nCompressionSize, nBitPosition);
			
			if( nCurrentCodeWord == -1)
			{
				// Error in file
				m_nLastError = errBadEntryInStringTable;
				return false;
			}

			// Does the read code has a special meaning ?
			if( nCurrentCodeWord == nEndOfInformationCode )
			{
				// Yes, end of the compressed data
				return true;
			}
			else if( nCurrentCodeWord == nClearCode )
			{
				// Yes, wants to re-initialize the table
				InitTable();
				nLastCodeWord = -1;
			}
			else
			{
				// Is the code in the string table ?
				if( nLastCodeWord == -1)
				{
					// First code read
					nCurrentFirstChar = OutputString(nCurrentCodeWord, pPixelBuffer);
					nLastCodeWord = nCurrentCodeWord;
				}
				else if( nCurrentCodeWord < m_nNextEntry)
				{
					// The code IS in the table
					
					nCurrentFirstChar = OutputString(nCurrentCodeWord, pPixelBuffer);
					
					if( m_nNextEntry < 4096)
					{
						// Create a new entry, where the string is 
						// nLastCodeWord + the first character of the string of nCurrentCodeWord
					
						m_aStringTable[m_nNextEntry].nPrevious = int16(nLastCodeWord);
						m_aStringTable[m_nNextEntry].nValue = nCurrentFirstChar;
						m_aStringTable[m_nNextEntry].nLength = m_aStringTable[nLastCodeWord].nLength + 1;
						m_nNextEntry++;
					
						nLastCodeWord = nCurrentCodeWord;
						nCurrentCodeWord = -1;
					}
				}
				else
				{
					// The code is NOT in the table
					
					if( m_nNextEntry < 4096)
					{
						// Create a new entry
						// From the old string + the first character of the old string
						m_aStringTable[m_nNextEntry].nPrevious = int16(nLastCodeWord);
						m_aStringTable[m_nNextEntry].nValue = nCurrentFirstChar;
						m_aStringTable[m_nNextEntry].nLength = m_aStringTable[nLastCodeWord].nLength + 1;
					
						OutputString(m_nNextEntry, pPixelBuffer);
					
						m_nNextEntry++;
					
						nLastCodeWord = nCurrentCodeWord;
					}
				}

				// Check if the codes length change
				// Warning : the length does not increase if the last code is 4095
				// Note that we test with 4096 because we just incremented m_nNextEntry
				if( m_nNextEntry == (1 << m_nCompressionSize) && m_nNextEntry != 4096)
					m_nCompressionSize++;
			}
		}
		// End of decompression for one block

		// Copy the two last bytes to the first bytes
		anBlock[0] = anBlock[0 + nBlockSize];
		anBlock[1] = anBlock[1 + nBlockSize];

		// Initialize the buffer position
		// "16" because of the 2 special first bytes
		nBitPosition = 16 - (nMaxBitSize - nBitPosition);
	}
	return true;
}

bool Gif::UninterlaceBuffer(GifAnimFrame& frame)
{
	const int32 nWidth = frame.m_nWidth;
	const int32 nHeight = frame.m_nHeight;

	const int32 nBufferSize = nWidth * nHeight;
	ByteArray aTmpBuffer;
	if( !aTmpBuffer.SetSize(nBufferSize))
	{
		m_nLastError = notEnoughMemory;
		return false;
	}

	uint8* const pTmpBuffer = aTmpBuffer.GetPtr();
	uint8* const pPixelBuffer = frame.m_aPixelBuffer.GetPtr();
	Memory::Copy(pTmpBuffer, pPixelBuffer, nBufferSize);

	// Pass 1
	uint8* pSource = pTmpBuffer;
	uint8* pDest = pPixelBuffer;
	for(int32 i = 0; i < m_nHeight; i += 8)
	{
		Memory::Copy(pDest, pSource, nWidth);
		pSource += nWidth;
		pDest += 8 * nWidth;
	}

	// Pass 2
	pDest = pPixelBuffer + 4 * nWidth;
	for(int32 i = 4; i < m_nHeight; i += 8)
	{
		Memory::Copy(pDest, pSource, nWidth);
		pSource += nWidth;
		pDest += 8 * nWidth;
	}

	// Pass 3
	pDest = pPixelBuffer + 2 * nWidth;
	for(int32 i = 2; i < m_nHeight; i += 4)
	{
		Memory::Copy(pDest, pSource, nWidth);
		pSource += nWidth;
		pDest += 4 * nWidth;
	}

	// Pass 4
	pDest = pPixelBuffer + 1 * nWidth;
	for(int32 i = 1; i < m_nHeight; i += 2)
	{
		Memory::Copy(pDest, pSource, nWidth);
		pSource += nWidth;
		pDest += 2 * nWidth;
	}
	return true;
}

String Gif::GetLastErrorString() const
{
	switch(m_nLastError)
	{
	case Gif::errNotAGifFile:
		return L"Not a GIF file";

	case Gif::errBadEntryInStringTable:
		return L"Bad entry in string table (file may by corrupted)";
	
	case Gif::errBadGifVersion:
		return L"Bad GIF version, should be 87a or 89a";

	case Gif::errBlockTerminatorNotFound:
		return L"Block terminator not found (0x00)";
	
	case Gif::errBadLzwMinimumCodeSize:
		return L"Bad LZW minimum code size";
	
	case Gif::errBadImageRect:
		return L"Bad image rect";
	}

	return ImageType::GetLastErrorString();
}

void Gif::CopyFramePixels(const GifAnimFrame& frame, bool bRestore)
{
	const uint8* pSrc = frame.m_aPixelBuffer.GetPtr();
	if( bRestore )
	{
		//pSrc = frame.m_aRestoreBuffer.GetPtr();
	}

	const int32 nFrameWidth = frame.m_nWidth;
	const int32 nFrameHeight = frame.m_nHeight;

	uint8* pDst = m_aPixelBuffer.GetPtr();

	// Small rect copy into big rect
	pDst += (frame.m_info.m_nOffsetY * m_nWidth) + frame.m_info.m_nOffsetX;

	// Transparency checking is done here
	const int16 nTransparentIndex = frame.m_nTransparentIndex;

	for(int32 iY = 0; iY < nFrameHeight; ++iY)
	{
		for(int32 iX = 0; iX < nFrameWidth; ++iX)
		{
			int16 nIndex = pSrc[iX];
			if( nIndex != nTransparentIndex )
			{
				pDst[iX] = uint8(nIndex);
			}
		}
		pDst += m_nWidth;
		pSrc += nFrameWidth;
	}
}

uint8 Gif::GetClearIndex()
{
	// How does GIF defines a transparent background ?
	// The old Netscape navigator used this method:
	// if a Graphic Control Extension block is present and its transparency is turned on,
	// then the background will be transparent
	// Use the transparent color
	
	// Use the transparent color of the current frame
	uint8 nClearIndex = m_nBackgroundColorIndex;
	if( m_aFrames.GetSize() >= 1 )
	{
		int16 nTransIndex = m_aFrames[m_nCurrentFrame].m_nTransparentIndex;
		if( nTransIndex >= 0 )
		{
			nClearIndex = uint8(nTransIndex);
		}
	}

	return nClearIndex;
}

void Gif::ClearBackground()
{
	Memory::Set(m_aPixelBuffer.GetPtr(), GetClearIndex(), m_nWidth * m_nHeight);
}

void Gif::ClearBackgroundRect(int nLeft, int nTop, int nWidth, int nHeight)
{
	uint8 nClearIndex = GetClearIndex();

	uint8* pDst = m_aPixelBuffer.GetPtr();

	// Small rect copy into big rect
	pDst += (nTop * m_nWidth) + nLeft;

	for(int32 iY = 0; iY < nHeight; ++iY)
	{
		for(int32 iX = 0; iX < nWidth; ++iX)
		{
			pDst[iX] = nClearIndex;
		}
		pDst += m_nWidth;
	}
}

