///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#ifndef CHUSTD_GIF_H
#define CHUSTD_GIF_H

#include "ImageType.h"

namespace chustd {

// Manages an image in Gif format
class Gif : public ImageType
{
public:
	/////////////////////////////////////////////////////////////
	virtual bool LoadFromFile(IFile& file);
	virtual PixelFormat GetPixelFormat() const;
	virtual const Palette* GetPalette() const;

	virtual String GetLastErrorString() const;
	/////////////////////////////////////////////////////////////

	virtual bool IsInterlaced() const;

	virtual int32 GetFrameCount() const;
	virtual int32 GetFrameIndex() const;
	virtual void  SetFrameIndex(int32 nFrameIndex);
	virtual const FrameInfo* GetFrameInfo();

	int32 GetLoopCount() const;

	Gif();
	virtual ~Gif();

	enum
	{
		errNotAGifFile = firstErrEnumDerived, // Not a Gif file (mismatch in header)
		errBadEntryInStringTable,
		errBadGifVersion,
		errBlockTerminatorNotFound,
		errBadLzwMinimumCodeSize,
		errBadImageRect
	};

	static bool IsGif(IFile& file);
	

protected:
	int32  m_nOutputIndex;

	// A Local Color Table supersedes a Global Color
	// The global color table is used if no local color table is found
	Palette m_GlobalColorTable;
	
	enum Version
	{
		gifverNotSet = 0,
		gifver87a = 1,
		gifver89a = 2
	};

	Version m_eVersion;
	
	uint8 m_nBackgroundColorIndex;
	
	enum GifDisposalMethod
	{
		gdmNotSpecified = 0,
		gdmDoNotDispose = 1,
		gdmRestoreToBkColor = 2,
		gdmRestoreToPrevious = 3
	};

	int16 m_nTransparentColorIndex; // Transparent color index for next frame
	bool  m_bTransparencyUsed;      // true if one of the frames use transparency
	uint8 m_nDisposalMethod;        // Disposal method for next frame
	int32 m_nDelayTime;             // Delay time for next frame
	
	///////////////////////////////////////////////
	int32 m_nLoopCount;    // Animated Gif. 0 = infinite
	
	///////////////////////////////////////////////
	// Decoding
	struct LzwCode
	{
		int32 nLength;
		int16 nPrevious;
		uint8 nValue;
	};

	LzwCode m_aStringTable[4096];
	int32 m_nNextEntry;

	int32 m_nCodeSize;
	int32 m_nCompressionSize; // One code size in bits (Changes during uncompression)
	
	class GifAnimFrame
	{
	public:
		ByteArray m_aPixelBuffer;
		int32 m_nWidth;
		int32 m_nHeight;
		Palette* m_pPalette;

		bool m_bInterlaced;
		int16 m_nTransparentIndex;

		FrameInfo m_info;

		GifAnimFrame()
		{
			m_nWidth = 0;
			m_nHeight = 0;
			m_pPalette = null;
			m_bInterlaced = false;
			m_nTransparentIndex = -1;
		}

		~GifAnimFrame()
		{
			delete m_pPalette;
		}
	};

	Array<GifAnimFrame> m_aFrames;
	int32 m_nCurrentFrame;

protected:
	void Initialize();

	bool ReadLogicalScreenDescriptor(IFile& file);
	
	bool ReadExtensionBlock(IFile& file);
	bool ReadImageDescriptor(IFile& file, GifAnimFrame& frame);
	bool ReadImageData(IFile& file, GifAnimFrame& frame);
	
	void InitTable();
	uint8 OutputString(int32 nIndex, uint8* pPixelBuffer);
	int32 ReadBits(uint8* pBuffer, int32 nLength, int32& nBitPosition);
	bool UninterlaceBuffer(GifAnimFrame& frame);

	bool ReadTerminator(IFile& file);
	bool ReadColorTable(IFile& file, Palette& pal);

	void CopyFramePixels(const GifAnimFrame& frame, bool bRestore);
	void ClearBackground();
	void ClearBackgroundRect(int nLeft, int nTop, int nWidth, int nHeight);
	void AdjustPalette(GifAnimFrame& frame);
	uint8 GetClearIndex();

	static FrameInfo::Disposal GenDisposalFromGifDisposal(uint8 nGifDisposal);
};

} // namespace chustd

#endif // ndef CHUSTD_GIF_H
