///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "Master.h"

extern "C" {

#if defined(_M_IX86)

////////////////////////////////////////////////////////////
// (u)int64 Shift Left (signed or unsigned are identical)
// in:  edx:eax : value to shift
// in:  cl : bit count
// out: edx:eax : result
// note: cl is destroyed
__declspec(naked)
int64 _allshl()
{
	__asm
	{
		sub cl, 32
		jb  Ok64
		// 32 or more!
		xchg eax, edx
		xor  eax, eax

		sub cl, 32
		jb Ok64
		// 64 or more!
		xor edx, edx
		ret
Ok64:
		and cl, 31
		shld edx, eax, cl
		shl  eax, cl
        ret
	}
}

////////////////////////////////////////////////////////////
// int64 Signed Shift Right
// in:  edx:eax : value to shift
// in:  cl : bit count
// out: edx:eax : result
// note: cl is destroyed
__declspec(naked)
int64 _allshr()
{
	__asm
	{
		sub cl, 32
		jb  Ok64
		// 32 or more!
		xchg eax, edx    // could do mov eax, edx but this takes one byte less
		cdq              // extend sign from eax to edx

		sub cl, 32
		jb Ok64
		// 64 or more!
		mov  eax, edx
		ret
Ok64:
		and cl, 31
		shrd eax, edx, cl
		sar  edx, cl
        ret
	}
}

// uint64 Unsigned Shift Right
// in:  edx:eax : value to shift
// in:  cl : bit count
// out: edx:eax : result
// note: cl is destroyed
__declspec(naked)
uint64 _aullshr()
{
	__asm
	{
		sub cl, 32
		jb  Ok64
		// 32 or more!
		xchg eax, edx    // FEDCBA98:76543210 =>  76543210:FEDCBA98
		xor  edx, edx    // 76543210:FEDCBA98 =>  00000000:FEDCBA98

		sub cl, 32
		jb Ok64
		// 64 or more!
		mov  eax, edx
		ret
Ok64:
		and cl, 31
		shrd eax, edx, cl
		shr  edx, cl
        ret
	}
}

////////////////////////////////////////////////////////////
// (u)int64 multiplication
// in: 1st pushed on stack : multiplier ((u)int64)
// in: 2st pushed on stack : multiplicand ((u)int64)
// out: edx:eax : result (multiplier * multiplicand)
// uses: ecx
// note: arguments are popped out of the stack
__declspec(naked)
int64 _allmul()
{
	__asm
	{
		// A:B * C:D
		mov  eax, dword ptr [esp +  8] // eax = A
		mov  ecx, dword ptr [esp + 12] // ecx = D
		mul  ecx                       // eax = low(A * D), ecx = D
		xchg eax, ecx                  // eax = D,  ecx = low(A * D)
		
		mov  eax, dword ptr [esp +  4] // eax = B
		mul  dword ptr [esp + 16]      // eax = low (B * C), ecx = low (A * D)
		add  ecx, eax                  // ecx = low (B * C + A * D)

		mov  eax, dword ptr [esp +  4] // eax = B
		mul  dword ptr [esp + 12]      // edx:eax = B * D
		add  edx, ecx                  // edx:eax = result

        ret 16 // 2 * int64 = 16 bytes
	}
}



// Stack frame (ebp) offsets for the operands of division operations
// Division is A:B / C:D
enum EOffset
{
	offsB = 0 ,
	offsA = 4 ,
	offsD = 8 ,
	offsC = 12
};

////////////////////////////////////////////////////////////
// uint64 Unsigned Divide with remainder
// Helper function for all division related functions
//
// in: ebp : base address for the parameters in the stack :
//         => 1st pushed on stack : divisor (uint64)
//         => 2st pushed on stack : dividend (uint64)
// out: edx:eax : quotient (dividend / divisor)
// out: ebx:ecx : remainder (dividend % divisor)

__declspec(naked)
uint64 Udiv64()
{
	__asm
	{
		// A:B / C:D
		mov  ecx, [ebp + offsC]
		test ecx, ecx
		jnz  udiv64
		
		// div 64 by 32
		xor  edx, edx
		mov  eax, [ebp + offsA]
		mov  ecx, [ebp + offsD]
		div  ecx                  // eax = A[high] / D = (A / D)[high]; edx = A[high] % D
		push eax                  // Keep track of (A / D)[high]
		mov  eax, [ebp + offsB]   // edx:eax = [A%D]:B
		div  ecx                  // eax = (A / D)[low]
		xor  ebx, ebx             // ebx = (A % D)[high]
		mov  ecx, edx             // ebx:ecx = (A % D)		
		pop  edx                  // edx:eax = (A / D)
		ret

	udiv64:
		mov  ebx, [ebp + offsD]   // ecx:ebx = C:D
		mov  eax, [ebp + offsB]
		mov  edx, [ebp + offsA]   // edx:eax = A:B
	
	decalLoop:
		// Right shift edx:eax and ecx:ebx until ecx:ebx has 32 significant bits
		shr  edx, 1
		rcr  eax, 1
		shrd ebx, ecx, 1
		shr  ecx, 1
		jnz  decalLoop
		div  ebx                 // Perform division on reduced operands. 

		// Now eax is either our result or result + 1
		// To check this : multiply with the divisor, the result must be <= dividend
		mov  ecx, eax                 // ecx = res
		mul  dword ptr [ebp + offsD]  // edx:eax = res * D
		mov  ebx, edx                 // ebx     = (res * D)[high]
		push eax                      // push (res * D)[low]
		mov  eax, [ebp + offsC]
		mul  ecx                      // edx:eax = res * C
		add  eax, ebx
		adc  edx, 0                   // edx:eax = res * C + (res * D)[high]
		pop  edx
		xchg edx, eax                 // edx:eax = res * (C:D)
		mov  ebx, [ebp + offsA]
		jnz  resBiased                // overflow

		cmp  edx, ebx
		ja   resBiased
		jb   resOk64

		cmp  eax, [ebp + offsB]
		jbe  resOk64
	resBiased:
		sub  eax, [ebp + offsD]
		sbb  edx, [ebp + offsC]       // edx:eax = (A:B / C:D) * C:D = A:B - (A:B % C:D)
		dec  ecx                      // 0:ecx = (A:B / C:D)

	resOk64:
		// Here edx:eax = div * C:D => Our result is A:B - edx:eax
		xchg eax, ecx                // eax = div, edx:ecx = A:B - (A:B % C:D)
		sub  ebx, edx
		mov  edx, [ebp + offsB]
		sub  edx, ecx                // edx = (A:B - div * C:D)[low]
		sbb  ebx, 0                  // ebx:edx = (A:B - div * C:D) = (A:B % C:D)
		mov  ecx, edx                // ebx:ecx = (A:B % C:D)
		xor  edx, edx                // ebx:ecx = (A:B % C:D), edx:eax = (A:B / C:D)
		ret
	}
}


////////////////////////////////////////////////////////////
// uint64 Unsigned Divide
// in: 1st pushed on stack : divisor (uint64)
// in: 2st pushed on stack : dividend (uint64)
// out: edx:eax : quotient (dividend / divisor)
// uses: ecx
// note: arguments are popped out of the stack
__declspec(naked)
uint64 _aulldiv()
{
	__asm
	{
		push ebp
		lea  ebp, [esp + 8]
		push ebx       // save ebx (this function should not modify it)
		call Udiv64
		pop ebx
		pop ebp
		ret 16
	}
 }

////////////////////////////////////////////////////////////
// uint64 Unsigned Remainder
// in: 1st pushed on stack : divisor (uint64)
// in: 2st pushed on stack : dividend (uint64)
// out: edx:eax : remainder (dividend % divisor)
// uses: ecx
// note: arguments are popped out of the stack
__declspec(naked)
uint64 _aullrem()
{
	__asm
	{
		push ebp
		lea  ebp, [esp + 8]
		push ebx       // save ebx (this function should not modify it)
		call Udiv64
		mov  edx, ebx  // Remainder is returned as ebx:ecx
		mov  eax, ecx
		pop ebx
		pop ebp
		ret 16
	}
}

////////////////////////////////////////////////////////////
// int64 Signed Divide
// in: 1st pushed on stack : divisor (int64)
// in: 2st pushed on stack : dividend (int64)
// out: edx:eax : quotient (dividend / divisor)
// uses: ecx
// note: arguments are popped out of the stack
__declspec(naked)
int64 _alldiv()
{
	__asm
	{
		push ebp
		lea ebp, [esp + 8]

		// A:B / C:D
		// Compute the div with the absolute values, and change sign if necessary
		// Result sign is the sign of the division
		push ebx
		mov  eax, [ebp + offsB]
		mov  edx, [ebp + offsA]
		mov  ebx, [ebp + offsD]
		mov  ecx, [ebp + offsC]

		xor  edx, ecx  // sign flag is set if result is negative
		pushfd         // push that on stack
		xor  edx, ecx  // restore edx
		jns  NoAbs1
		neg  edx
		neg  eax
		sbb  edx, 0
	NoAbs1:
		test ecx, ecx
		jns  NoAbs2
		neg  ecx
		neg  ebx
		sbb  ecx, 0
	NoAbs2:
		mov  [ebp + offsB], eax   // We'll use those values from memory during the calculation
		mov  [ebp + offsA], edx
		mov  [ebp + offsD], ebx
		mov  [ebp + offsC], ecx

		call Udiv64

		popfd
		jns NoNegate
		neg  edx                  // Negate edx:eax
		neg  eax
		sbb  edx, 0
	NoNegate:
		pop  ebx
		pop  ebp
		ret  16
	}
}

////////////////////////////////////////////////////////////
// int64 Signed Remainder
// in: 1st pushed on stack : divisor (int64)
// in: 2st pushed on stack : dividend (int64)
// out: edx:eax : remainder (dividend % divisor)
// uses: ecx
// note: arguments are popped out of the stack
__declspec(naked)
int64 _allrem()
{
	__asm
	{
		push ebp
		lea  ebp, [esp + 8]
		// A:B / C:D
		// Compute the div with the absolute values, and change sign if necessary
		// Result sign is the sign of the dividend
		push ebx
		mov  eax, [ebp + offsB]
		mov  edx, [ebp + offsA]
		mov  ebx, [ebp + offsD]
		mov  ecx, [ebp + offsC]
		
		test edx, edx  // sign flag is set if result is negative
		pushfd
		jns  NoAbs1
		neg  edx
		neg  eax
		sbb  edx, 0
	NoAbs1:
		test ecx, ecx
		jns  NoAbs2
		neg  ecx
		neg  ebx
		sbb  ecx, 0
	NoAbs2:
		mov  [ebp + offsB], eax   // We'll use those values from memory during the calculation
		mov  [ebp + offsA], edx
		mov  [ebp + offsD], ebx
		mov  [ebp + offsC], ecx

		call Udiv64
		mov  eax, ecx
		mov  edx, ebx

		popfd
		jns  NoNegate
		neg  edx
		neg  eax
		sbb  edx, 0
	NoNegate:
		pop  ebx
		pop  ebp
		ret  16
	}
}

//////////////////////////////////////////////////////////////////////////////////
// Visual C++ 8 specific
////////////////////////////////////////////////////////////
// uint64 Divide and Remainder
// in: 1st pushed on stack : divisor (uint64)
// in: 2st pushed on stack : dividend (uint64)
// out: edx:eax : quotient (dividend / divisor)
// out: ebx:ecx : remainder (dividend % divisor)
// note: arguments are popped out of the stack
__declspec(naked)
uint64 _aulldvrm()
{
	__asm
	{
		push ebp
		lea  ebp, [esp + 8]
		call Udiv64
		pop  ebp
		ret  16
	}
}

void cpplayer_aulldvrm(uint64 /*nBig*/, uint64 /*nDivisor*/, uint64* pnQuotient, uint64* pnRemainder)
{
	__asm
	{
		mov eax, [ebp + 8 + 0] // a_l
		mov edx, [ebp + 8 + 4] // a_h
		mov ecx, [ebp + 8 + 8] // b_l
		mov ebx, [ebp + 8 + 12] // b_h
		push ebx // h
		push ecx // l
		push edx // h
		push eax // l
		call _aulldvrm

		push edi

		mov edi, [pnQuotient]
		mov [edi], eax
		mov [edi + 4], edx

		mov edi, [pnRemainder]
		mov [edi], ecx
		mov [edi + 4], ebx

		pop edi
	}
}

////////////////////////////////////////////////////////////
// int64 Divide and Remainder
// in: 1st pushed on stack : divisor (int64)
// in: 2st pushed on stack : dividend (int64)
// out: edx:eax : quotient (dividend / divisor)
// out: ebx:ecx : remainder (dividend % divisor)
// note: arguments are popped out of the stack
__declspec(naked)
int64 _alldvrm()
{
	// a  b
	__asm
	{
		push ebp
		lea  ebp, [esp + 8]

		// A:B / C:D
		// Compute the div with the absolute values, and change sign if necessary
		mov  eax, [ebp + offsB]
		mov  edx, [ebp + offsA]
		mov  ebx, [ebp + offsD]
		mov  ecx, [ebp + offsC]

		xor  edx, ecx  // sign flag is set if quotient is negative
		pushfd         // push that on stack
		xor  edx, ecx  // restore edx + sign flag is set if remainder is negative
		pushfd         // push that on stack
		jns  NoAbs1
		neg  edx
		neg  eax
		sbb  edx, 0
	NoAbs1:
		test ecx, ecx
		jns  NoAbs2
		neg  ecx
		neg  ebx
		sbb  ecx, 0
	NoAbs2:
		mov  [ebp + offsB], eax   // We'll use those values from memory during the calculation
		mov  [ebp + offsA], edx
		mov  [ebp + offsD], ebx
		mov  [ebp + offsC], ecx

		call Udiv64

		popfd         // Must Negate remainder?
		jns  NoNegateRem
		neg  ebx
		neg  ecx
		sbb  ebx, 0
	NoNegateRem:
		popfd         // Must Negate quotient?
		jns  NoNegateQuot
		neg  edx
		neg  eax
		sbb  edx, 0
	NoNegateQuot:
		pop  ebp
		ret  16
	}
}

void cpplayer_alldvrm(int64 /*nBig*/, int64 /*nDivisor*/, int64* pnQuotient, int64* pnRemainder)
{
	__asm
	{
		mov eax, [ebp + 8 + 0] // a_l
		mov edx, [ebp + 8 + 4] // a_h
		mov ecx, [ebp + 8 + 8] // b_l
		mov ebx, [ebp + 8 + 12] // b_h
		push ebx // h
		push ecx // l
		push edx // h
		push eax // l
		call _alldvrm

		push edi

		mov edi, [pnQuotient]
		mov [edi], eax
		mov [edi + 4], edx

		mov edi, [pnRemainder]
		mov [edi], ecx
		mov [edi + 4], ebx

		pop edi
	}
}

#endif // defined(_M_IX86)

//////////////////////////////////////////////////////////////////////////////////
} // extern "C"
