///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "Master.h"
#include "IFile.h"

#include "Array.h"

using namespace chustd;
//////////////////////////////////////////////////////////////////////
IFile::IFile()
{

}

IFile::~IFile()
{

}

//////////////////////////////////////////////////////////////////////
uint16 IFile::Swap16(uint16 nValue)
{
	union UConvert
	{
		uint16 n;
		struct { uint8 n0, n1; } bytes;
	};

	UConvert conv = { nValue };
	uint16 n2 = uint16( (conv.bytes.n0 << 8) | conv.bytes.n1);

	return n2;
}

uint32 IFile::Swap32(uint32 nValue)
{
	union UConvert
	{
		uint32 n;
		struct { uint8 n0, n1, n2, n3; } bytes;
	};

	UConvert conv = { nValue };
	uint32 n2 = (conv.bytes.n0 << 24) | (conv.bytes.n1 << 16)
		| (conv.bytes.n2 << 8) | conv.bytes.n3;

	return n2;
}

uint64 IFile::Swap64(uint64 nValue)
{
	union UConvert
	{
		uint64 n;
		struct { uint32 n0, n1; } nWords32;
	};

	UConvert conv = { nValue };
	uint32 n0 = conv.nWords32.n0;
	uint32 n1 = conv.nWords32.n1;
	conv.nWords32.n0 = Swap32(n1);
	conv.nWords32.n1 = Swap32(n0);
	return conv.n;
}

float32 IFile::SwapFloat32(float32 fValue)
{
	union UConvert
	{
		float32 f;
		struct { uint8 nHi, nMidHi, nMidLo, nLo; } bytes;
		uint32 n;
	};

	UConvert conv = { fValue };
	conv.n = (conv.bytes.nHi << 24) | (conv.bytes.nMidHi << 16)
		| (conv.bytes.nMidLo << 8) | conv.bytes.nLo;

	return conv.f;
}

float64 IFile::SwapFloat64(float64 fValue)
{
	union UConvert
	{
		float64 f;
		struct { uint32 n0, n1; } nWords32;
		uint64 n;
	};

	UConvert conv = { fValue };
	uint32 n0 = conv.nWords32.n0;
	uint32 n1 = conv.nWords32.n1;
	conv.nWords32.n0 = Swap32(n1);
	conv.nWords32.n1 = Swap32(n0);
	return conv.f;
}

void IFile::Swap16(uint16* paValues, int32 nCount)
{
	uint8* paBytes = (uint8*) paValues;
	const int32 nByteCount = nCount * 2;

	for(int32 i = 0; i < nByteCount; i += 2)
	{
		uint8 nByteA = paBytes[i];
		uint8 nByteB = paBytes[i + 1];
		
		paBytes[i + 1] = nByteA;
		paBytes[i] = nByteB;
	}
}

void IFile::Swap32(uint32* paValues, int32 nCount)
{
	for(int32 i = 0; i < nCount; ++i)
	{
		paValues[i] = Swap32(paValues[i]);
	}
}

//////////////////////////////////////
bool IFile::ShouldSwapBytes() const
{
	ByteOrder eFileByteOrder = GetByteOrder();

	if( eFileByteOrder != chustd::k_ePlatformByteOrder)
	{
		// We are trying to store little endian data on a big endian machine
		// Or
		// We are trying to store big endian data on a little endian machine
		// We should swap bytes !
		return true;
	}

	// No need to swap

	// We are trying to store little endian data on a little endian machine
	// Or
	// We are trying to store big endian data on a big endian machine

	return false;
}

//////////////////////////////////////
bool IFile::Read8(uint8& nValue)
{
	return Read(&nValue, 1) == 1;
}

bool IFile::Read16(uint16& nValue)
{
	bool bRet = Read(&nValue, 2) == 2;
	if( ShouldSwapBytes() )
	{
		nValue = Swap16(nValue);
	}
	return bRet;
}

bool IFile::Read32(uint32& nValue)
{
	bool bRet = Read(&nValue, 4) == 4;
	if( ShouldSwapBytes() )
	{
		nValue = Swap32(nValue);
	}
	return bRet;
}

bool IFile::Read64(uint64& nValue)
{
	bool bRet = Read(&nValue, 8) == 8;
	if( ShouldSwapBytes() )
	{
		nValue = Swap64(nValue);
	}
	return bRet;
}

bool IFile::Read8(int8& nValue)
{
	return Read(&nValue, 1) == 1;
}

bool IFile::Read16(int16& nValue)
{
	bool bRet = Read(&nValue, 2) == 2;
	if( ShouldSwapBytes() )
	{
		nValue = Swap16(nValue);
	}
	return bRet;
}

bool IFile::Read32(int32& nValue)
{
	bool bRet = Read(&nValue, 4) == 4;
	if( ShouldSwapBytes() )
	{
		nValue = Swap32(nValue);
	}
	return bRet;
}

bool IFile::Read64(int64& nValue)
{
	bool bRet = Read(&nValue, 8) == 8;
	if( ShouldSwapBytes() )
	{
		nValue = Swap64(nValue);
	}
	return bRet;
}

bool IFile::Read32(float32& fValue)
{
	return Read32((uint32&) fValue);
}

bool IFile::Read64(float64& fValue)
{
	return Read64((uint64&) fValue);
}

bool IFile::Read8(bool& bValue)
{
	uint8 nValue = 0;
	bool bRet = Read8(nValue);
	bValue = nValue != 0;
	return bRet;
}

//////////////////////////////////////
bool IFile::Write8(uint8 nValue)
{
	return Write(&nValue, 1) == 1;
}

bool IFile::Write16(uint16 nValue)
{
	if( ShouldSwapBytes())
	{
		nValue = Swap16(nValue);
	}
	return Write(&nValue, 2) == 2;
}

bool IFile::Write32(uint32 nValue)
{
	if( ShouldSwapBytes())
	{
		nValue = Swap32(nValue);
	}
	return Write(&nValue, 4) == 4;
}

bool IFile::Write64(uint64 nValue)
{
	if( ShouldSwapBytes())
	{
		nValue = Swap64(nValue);
	}
	return Write(&nValue, 8) == 8;
}

bool IFile::Write8(int8 nValue)
{
	return Write(&nValue, 1) == 1;
}

bool IFile::Write16(int16 nValue)
{
	if( ShouldSwapBytes())
	{
		nValue = Swap16(nValue);
	}
	return Write(&nValue, 2) == 2;
}

bool IFile::Write32(int32 nValue)
{
	if( ShouldSwapBytes())
	{
		nValue = Swap32(nValue);
	}
	return Write(&nValue, 4) == 4;
}

bool IFile::Write64(int64 nValue)
{
	if( ShouldSwapBytes())
	{
		nValue = Swap64(nValue);
	}
	return Write(&nValue, 8) == 8;
}

bool IFile::Write32(float32 fValue)
{
	union FloatAsInt32 { float32 f; int32 n; } fai = { fValue };
	return Write32(fai.n);

}

bool IFile::Write64(float64 fValue)
{
	union FloatAsInt64 { float64 f; int64 n; } fai = { fValue };
	return Write64(fai.n);
}

bool IFile::Write8(bool bValue)
{
	uint8 nValue = bValue ? 0x01 : 0x00;
	return Write8(nValue);
}

bool IFile::WriteString(const String& strValue)
{
	const int32 nLength = strValue.GetLength();
	const int32 nByteCount = nLength * 2;
	return Write(strValue.GetBuffer(), nByteCount) == nByteCount;
}

bool IFile::WriteStringLine(const String& strValue)
{
	if( !WriteString(strValue) )
	{
		return false;
	}
	return WriteString(L"\n");
}

bool IFile::WriteString(const String& strValue, const TextEncoding& te)
{
	ByteArray anContent = strValue.ToBytes(te, false); // Don't add ending 0

	const int nSize = anContent.GetSize();
	return Write(anContent.GetPtr(), nSize) == nSize;
}

bool IFile::WriteStringLine(const String& strValue, const TextEncoding& te)
{
	if( !WriteString(strValue, te) )
	{
		return false;
	}
	return WriteString(L"\n", te);
}

///////////////////////

