///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "Master.h"
#include "ImageType.h"
#include "File.h"

//////////////////////////////////////////////////////////////////////
using namespace chustd;
//////////////////////////////////////////////////////////////////////

ImageType::ImageType()
{
	m_nWidth = 0;
	m_nHeight = 0;
	m_nLastError = 0;
}

ImageType::~ImageType()
{

}

bool Palette::AllAlphasAreOpaque() const
{
	for(int32 i = 0; i < m_nCount; ++i) 
	{
		if( m_aColors[i].GetAlpha() != 0)
			return false;
	}
	return true;
}

void Palette::SetAlphaFullOpaque()
{
	for(int32 i = 0; i < m_nCount; ++i) 
	{
		m_aColors[i].SetAlpha(255);
	}
}

////////////////////////////////////////////////////
bool ImageType::IsIndexed(PixelFormat pf)
{
	return pf == pf1bppIndexed || pf == pf2bppIndexed || pf == pf4bppIndexed || pf == pf8bppIndexed;
}

int32 ImageType::SizeofPixelInBits(PixelFormat ePixelFormat)
{
	switch(ePixelFormat)
	{
	case pfUnknown:
		return 0;
		
	case pf1bppGrayScale:
		return 1;
	case pf2bppGrayScale:
		return 2;
	case pf4bppGrayScale:
		return 4;
	case pf8bppGrayScale:
		return 8;
	case pf16bppGrayScale:
		return 16;
		
	case pf16bppGrayScaleAlpha:
		return 16;
	case pf32bppGrayScaleAlpha:
		return 32;

	case pf1bppIndexed:
		return 1;
	case pf2bppIndexed:
		return 2;
	case pf4bppIndexed:
		return 4;
	case pf8bppIndexed:
		return 8;
		
	case pf16bppArgb1555:
	case pf16bppArgb4444:
	case pf16bppRgb555:
	case pf16bppRgb565:
		return 16;

	case pf24bppRgb:
	case pf24bppBgr:
		return 24;

	case pf32bppRgba:
	case pf32bppBgra:
		return 32;

	case pf48bppRgb:
		return 48;

	case pf64bppRgba:
		return 64;

	}
	ASSERT(0);
	return 0;
}

int32 ImageType::ComputeByteWidth(PixelFormat epf, int32 nWidth)
{
	const int32 nSizeofPixelInBits = ImageType::SizeofPixelInBits(epf);

	// Size of a row in bits
	const int32 nBitWidth = nWidth * nSizeofPixelInBits;
	
	// Round up for size of a row in bytes
	const int32 nAddWidth = ((nBitWidth & 7) != 0) ? 1 : 0;
	const int32 nByteWidth = nBitWidth / 8 + nAddWidth;

	return nByteWidth;
}
//////////////////////////////////////////////////////////////////////////////


String ImageType::GetLastErrorString() const
{
	switch(m_nLastError)
	{
	case ioErr:
		return L"Error accessing to the file";
	case badFileFormat:
		return L"Bad File Format";

	case notEnoughMemory:
		return L"Not enough memory";

	case uncompleteFile:
		return L"Uncomplete file";
	}
	return L"Unknown error";
}

bool ImageType::Load(const String& strFileName)
{
	File file;
	if( !file.Open(strFileName))
	{
		m_nLastError = ioErr;
		return false;
	}
	return LoadFromFile(file);
}

int32 ImageType::GetWidth() const
{
	return m_nWidth;
}

int32 ImageType::GetHeight() const
{
	return m_nHeight;
}

void* ImageType::GetBuffer() const
{
	return m_aPixelBuffer.GetPtr();
}

void ImageType::FreeBuffer()
{
	m_aPixelBuffer.RemoveAll();
}

void ImageType::FlipVertical()
{
	PixelFormat epf = GetPixelFormat();
	if( epf == pfUnknown)
	{
		return;
	}
	const int32 nBytesPerRow = ImageType::ComputeByteWidth(epf, m_nWidth);
	uint8* pComponents = (uint8*) GetBuffer();
	const int32 nHeightDiv2 = m_nHeight / 2;

	for(int iY = 0; iY < nHeightDiv2; ++iY)
	{
		uint8* pSrc = pComponents + (iY * nBytesPerRow);
		uint8* pDst = pComponents + ((m_nHeight - 1 - iY) * nBytesPerRow);

		for(int iByte = 0; iByte < nBytesPerRow; ++iByte)
		{
			uint8 nSwap = pDst[iByte];
			pDst[iByte] = pSrc[iByte];
			pSrc[iByte] = nSwap;
		}
	}
}

void ImageType::SetAlphaFullOpaque()
{
	PixelFormat epf = GetPixelFormat();
	if( epf == pfUnknown)
	{
		return;
	}

	const int32 nPixelCount = m_nWidth * m_nHeight;
	if( epf == pf32bppRgba || epf == pf32bppBgra)
	{
		uint8* pDst = (uint8*) GetBuffer();
		const uint8 nAlpha = 255;
		for(int i = 0; i < nPixelCount; ++i)
		{
			pDst[3] = nAlpha;
			pDst += 4;
		}

	}
}

bool ImageType::IsAnimated() const
{
	return GetFrameCount() >= 2;
}

int32 ImageType::GetLastError() const
{ 
	return m_nLastError; 
}

bool ImageType::IsIndexed() const 
{ 
	return IsIndexed(GetPixelFormat()); 
}

int32 ImageType::GetFrameCount() const
{
	return 0;
}

int32 ImageType::GetFrameIndex() const
{
	return 0;
}

void  ImageType::SetFrameIndex(int32 nFrameIndex)
{
	return;
}

const FrameInfo* ImageType::GetFrameInfo()
{
	return null;
}
