///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#ifndef CHUSTD_IMAGETYPE_H
#define CHUSTD_IMAGETYPE_H

#include "File.h"
#include "FixArray.h"

#include "Surface32.h"

namespace chustd {

/////////////////////////////
class Palette
{
public:
	bool AllAlphasAreOpaque() const;
	void SetAlphaFullOpaque();

public:
	int32 m_nCount;        // Number of color in the palette
	chustd::FixArray<Color32, 256> m_aColors; // Array of colors
};

/////////////////////////////
class FrameInfo
{
public:
	// What to do with the destination buffer before rendering the next frame
	enum Disposal
	{
		dispNone = 0,              // Same as APNG_DISPOSE_OP_NONE
		dispClearToTransBlack = 1, // Same as APNG_DISPOSE_OP_BACKGROUND
		dispRestoreToPrevious = 2, // Same as APNG_DISPOSE_OP_PREVIOUS
		
		dispClearToBkColor, // May become deprecated as browsers interpret that as dispClearToTransBlack
	};

	Disposal m_disposal;
	int32    m_nDelayMs;    // How much time the frame must be displayed (milliseconds)
	int32    m_nOffsetX;    // Offset of the frame in the image
	int32    m_nOffsetY;    // Offset of the frame in the image
	uint8    m_nClearIndex; // In palette

	FrameInfo()
	{
		m_nDelayMs = 0;
		m_disposal = dispNone;
		m_nOffsetX = 0;
		m_nOffsetY = 0;
		m_nClearIndex = 0;
	}
};

/////////////////////////////
// Abstract class for image file type classes (png, gif, jpeg, tga, bmp)
class ImageType
{
public:
	// Possible errors for m_nLastError
	enum
	{
		noErr = 0,		// No error
		ioErr,			// Error accessing to the file
		badFileFormat,	// Bad file format
		notEnoughMemory,
		uncompleteFile,
		firstErrEnumDerived = 0x0100,
	};

	///////////////////////////////////////////////
	static bool IsIndexed(PixelFormat pf);
	static int32 SizeofPixelInBits(PixelFormat epf);
	static int32 ComputeByteWidth(PixelFormat epf, int32 nWidth);
	///////////////////////////////////////////////

	friend class ImageLoader;

public:
	bool Load(const chustd::String& strFileName);

	int32 GetWidth() const;
	int32 GetHeight() const;

	void* GetBuffer() const;
	void  FreeBuffer();

	bool  IsAnimated() const;
	int32 GetLastError() const;
	bool  IsIndexed() const;

	ImageType();

	///////////////////////////////////////////////////////
	// Must be implemented in the derived class
	virtual bool  LoadFromFile(IFile& file) = 0;
	virtual PixelFormat GetPixelFormat() const = 0;
	virtual const Palette* GetPalette() const = 0;
	///////////////////////////////////////////////////////

	///////////////////////////////////////////////////////
	// Would be nice to be implemented in the derived class, calling the base version on return
	virtual String GetLastErrorString() const;
	///////////////////////////////////////////////////////

	///////////////////////////////////////////////////////
	// Implemented if needed in the derived class
	//
	// Returns true if one RGB color or one Grey value is used as a transparent color
	// Note: this does not apply for palettized pictures as the transparent information
	// is already given by the palette structure (in the alpha field for each color).
	virtual bool IsSimpleTransparencyUsed() const { return false; }

	virtual uint16 GetGreyTransIndex() const { return 0; }
	virtual void GetTransIndexes(uint16& nRed, uint16& nGreen, uint16& nBlue) const
	{ nRed = 0; nGreen = 0; nBlue = 0; }

	// Animation
	virtual int32 GetFrameCount() const;
	virtual int32 GetFrameIndex() const;
	virtual void  SetFrameIndex(int32 nFrameIndex);
	virtual const FrameInfo* GetFrameInfo();
	
	///////////////////////////////////////////////////////
	
protected:
	int32 m_nWidth;
	int32 m_nHeight;
	int32 m_nLastError; // Last error number

	ByteArray m_aPixelBuffer;
	
protected:
	virtual ~ImageType();

	void FlipVertical();
	void SetAlphaFullOpaque();
};

} // namespace chustd

#endif // ndef CHUSTD_IMAGETYPE_H
