///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "Master.h"
#include "MemIniFile.h"
#include "FormatType.h"
#include "StringBuilder.h"
#include "File.h"
#include "TextEncoding.h"

using namespace chustd;

//////////////////////////////////////////////////////////////////////
void MemIniFile::Section::Dump(StringBuilder& sb, bool bCrLf) const
{
	static const wchar* pszRealEndLine = L"\n";
	static const wchar* pszDosEndLine = L"\r\n";
	const wchar* const pszEndLine = bCrLf ? pszDosEndLine : pszRealEndLine;

	sb += L"[";
	sb += m_strName;
	sb += L"]";
	sb += pszEndLine;

	const int32 nValueCount = m_apValues.GetSize();
	for(int32 i = 0; i < nValueCount; ++i)
	{
		const Value* pValue = m_apValues[i];

		sb += pValue->m_strName;
		sb += L" = ";
		sb += pValue->m_strValue;
		sb += pszEndLine;
	}
}

int32 MemIniFile::Section::GetArrayIndex(const String& strValueName) const
{
	const uint16 nFirstCodeUnit = strValueName.GetAt(0);
	
	int32 nIndex = -1;

	// Simple letter ?
	if( uint16('a') <= nFirstCodeUnit && nFirstCodeUnit <= uint16('z'))
	{
		nIndex = nFirstCodeUnit - uint16('a');
	}
	else if( uint16('A') <= nFirstCodeUnit && nFirstCodeUnit <= uint16('Z'))
	{
		nIndex = nFirstCodeUnit - uint16('A');
	}
	else
	{
		// Last array cell for variable names which do not start with a letter
		nIndex = 26;
	}

	return nIndex;
}

void MemIniFile::Section::SetValue(const String& strName, const String& strValue)
{
	if( strName.IsEmpty())
		return;

	const int32 nIndex = GetArrayIndex(strName);
	Array<Value*>& apValues = m_aapValues[nIndex];

	String strUpperName = strName.ToUpperCase();

	// Check if the value is not already there
	for(int32 i = 0; i < apValues.GetSize(); ++i)
	{
		if( apValues[i]->m_strUpperName == strUpperName)
		{
			// Found !
			apValues[i]->m_strValue = strValue;
			return;
		}
	}

	// The value does not exist in the section, we create it
	Value* pValue = new Value;
	pValue->m_strUpperName = strUpperName;
	pValue->m_strName = strName;
	pValue->m_strValue = strValue;

	// Add it to the linear array
	m_apValues.Add(pValue);

	// Add it to the map
	apValues.Add(pValue);
}

bool MemIniFile::Section::GetValue(const String& strName, String& strValue) const
{
	if( strName.IsEmpty())
		return false;

	const int32 nIndex = GetArrayIndex(strName);
	const Array<Value*>& apValues = m_aapValues[nIndex];

	String strUpperName = strName.ToUpperCase();

	// Check if the value is not already there
	for(int32 i = 0; i < apValues.GetSize(); ++i)
	{
		if( apValues[i]->m_strUpperName == strUpperName)
		{
			// Found !
			strValue = apValues[i]->m_strValue;
			return true;
		}
	}
	return false;
}

const MemIniFile::ValuePtrArray& MemIniFile::Section::GetValues() const
{
	return m_apValues;
}

//////////////////////////////////////////////////////////////////////

MemIniFile::MemIniFile()
{
	m_pCurrentSection = null;
}

MemIniFile::~MemIniFile()
{

}

bool MemIniFile::Load(const String& strFilePath)
{
	File file;
	if( !file.Open(strFilePath))
	{
		return false;
	}

	ByteArray aContent = file.GetContent();
	file.Close();

	// Convert the file content into a string
	String strContent;

	// Byte order mark found for UTF-8 ?
	if( aContent.GetSize() >= 3 
		&& aContent[0] == 0xEF && aContent[1] == 0xBB && aContent[2] == 0xBF)
	{
		// Byte order mark found, remove it before parsing

		// Create a smaller array from the content array
		ByteArrayBox bab( aContent.GetPtr() + 3, aContent.GetSize() - 3);

		strContent = String::FromBytes(bab.ToByteArray(), TextEncoding::Utf8());
	}
	else
	{
		strContent = String::FromBytes(aContent, TextEncoding::Utf8());
	}

	StringArray astr = strContent.SplitByEndlines();

	m_aSections.RemoveAll();
	m_aSections.EnsureCapacity(16);
	m_pCurrentSection = null;

	// Find the sections
	const int32 nStringCount = astr.GetSize();
	for(int32 iLine = 0; iLine < nStringCount; ++iLine)
	{
		String str = astr[iLine].Trim();
		if( !str.IsEmpty())
		{
			const int32 nLength = str.GetLength();
			if( str.GetAt(0) == ';')
			{
				// Line comment
			}
			else if( str.GetAt(0) == '[' && str.GetAt(nLength - 1) == ']')
			{
				String strSectionName(str.GetBuffer() + 1, nLength - 2);

				SetSection(strSectionName);
			}
			else
			{
				// Une valeur
				int32 nFoundAt = str.Find(L"=", 0);
				if( nFoundAt >= 0)
				{
					if( m_pCurrentSection != null)
					{
						String strValueName = str.Left(nFoundAt).Trim();
						String strValue = str.Mid(nFoundAt + 1).Trim();
						
						// Remove any end of line comment
						const int32 nFirstLength = strValue.GetLength();
						for(int32 iChar = 0; iChar < nFirstLength; ++iChar)
						{
							uint16 c = strValue.GetAt(iChar);
							if( c == ';')
							{
								strValue = strValue.Left(iChar).Trim();
								break;
							}
						}

						m_pCurrentSection->SetValue(strValueName, strValue);
					}
				}
			}
		}
	}

	m_pCurrentSection = null;
	m_strCurrentSection.Empty();

	return true;
}

bool MemIniFile::Dump(const String& strFilePath, bool bCrLf)
{
	// Put byte order mark
	StringBuilder sb;
	sb += L"\xFEFF"; // Will generate 0xEF, 0xBB, 0xBF as UTF-8

	static const wchar* pszRealEndLine = L"\n";
	static const wchar* pszDosEndLine = L"\r\n";
	const wchar* const pszEndLine = bCrLf ? pszDosEndLine : pszRealEndLine;

	StringBuilder sbTopComment;
	if( !m_strCommentLine1.IsEmpty())
	{
		sbTopComment += L"; ";
		sbTopComment += m_strCommentLine1;
		sbTopComment += pszEndLine;
	}

	if( !m_strCommentLine2.IsEmpty())
	{
		sbTopComment += L"; ";
		sbTopComment += m_strCommentLine2;
		sbTopComment += pszEndLine;
	}
	
	if( !sbTopComment.IsEmpty())
	{
		sbTopComment += pszEndLine; // Put an additional endline
		sb += sbTopComment.ToString();
	}

	const int32 nSectionCount = m_aSections.GetSize();
	for(int32 iSection = 0; iSection < nSectionCount; ++iSection)
	{
		m_aSections[iSection].Dump(sb, bCrLf);

		// One empty line for clearer file, but only if we are not dumping the last section
		// because we do not want two enlines at the end of the file
		if( iSection < (nSectionCount - 1))
		{
			sb += pszEndLine;
		}
	}

	File file;
	if( !file.Open(strFilePath, File::modeWrite))
	{
		return false;
	}

	ByteArray aResult = sb.ToString().ToBytes(TextEncoding::Utf8());
	int32 nWritten = file.Write(aResult.GetPtr(), aResult.GetSize());
	file.Close();

	return nWritten == aResult.GetSize();
}

bool MemIniFile::SetSection(const String& strSection)
{
	if( m_pCurrentSection != null)
	{
		// If we are lucky we are just doing a fast pointer comparison
		if( m_strCurrentSection == strSection)
			return true;
	}

	m_strCurrentSection = strSection;

	String strUpperSection = strSection.ToUpperCase();
	
	m_pCurrentSection = null;

	// Find the section (maybe it already exists)
	for(int32 i = 0; i < m_aSections.GetSize(); ++i)
	{
		if( m_aSections[i].m_strUpperName == strUpperSection)
		{
			// Yep !
			m_pCurrentSection = &m_aSections[i];
			break;
		}
	}

	if( m_pCurrentSection == null)
	{
		// Create the section
		int32 nPos = m_aSections.Add();
		m_pCurrentSection = &m_aSections[nPos];
		m_pCurrentSection->m_strName = strSection;
		m_pCurrentSection->m_strUpperName = strUpperSection;

		return false; // The section did not exist
	}
	return true;
}

bool MemIniFile::GetString(const String& strName, String& strValue) const
{
	if( m_pCurrentSection == null)
		return false;

	return m_pCurrentSection->GetValue(strName, strValue);
}

bool MemIniFile::GetInt32(const String& strName, int32& nValue) const
{
	if( m_pCurrentSection == null)
		return false;

	String strValue;
	if( !m_pCurrentSection->GetValue(strName, strValue))
		return false;

	return strValue.ToInt32(nValue);
}

bool MemIniFile::GetBool(const String& strName, bool& bValue) const
{
	if( m_pCurrentSection == null)
		return false;

	String strValue;
	if( !m_pCurrentSection->GetValue(strName, strValue))
		return false;

	return strValue.ToBool(bValue);
}

void MemIniFile::SetString(const String& strName, const String& strValue)
{
	if( m_pCurrentSection == null)
		return;

	m_pCurrentSection->SetValue(strName, strValue);	
}

void MemIniFile::SetInt32(const String& strName, int32 nValue)
{
	if( m_pCurrentSection == null)
		return;

	String strValue = String::FromInt(nValue);
	m_pCurrentSection->SetValue(strName, strValue);
}

void MemIniFile::SetBool(const String& strName, bool bValue)
{
	if( m_pCurrentSection == null)
		return;

	String strValue = bValue ? L"1" : L"0";
	m_pCurrentSection->SetValue(strName, strValue);
}

const MemIniFile::SectionArray& MemIniFile::GetSections() const
{
	return m_aSections;
}

void MemIniFile::SetCommentLine1(const String& strCommentLine1)
{
	m_strCommentLine1 = strCommentLine1;
}

void MemIniFile::SetCommentLine2(const String& strCommentLine2)
{
	m_strCommentLine2 = strCommentLine2;
}

