///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "Master.h"
#include "Memory.h"
#include "Array.h"
#include "MemoryAllocator.h"

#undef Alloc64

using namespace chustd;

#ifdef WIN32

//////////////////////////////////////////////////
class Win32PageAllocator : public IPageAllocator
{
public:
	Win32PageAllocator()
	{
		m_hProcessHeap = ::GetProcessHeap();
	}

	virtual void* Alloc(int32 nCount)
	{
		return HeapAlloc(m_hProcessHeap, 0, nCount * k_nPageSize);
	}

	virtual void Free(void* pFirst, int32 /*nCount*/)
	{
		BOOL bOk = HeapFree(m_hProcessHeap, 0, pFirst);
		VERIFY(bOk);
	}

private:
	HANDLE m_hProcessHeap;
};
//////////////////////////////////////////////////

#elif CHUOS

//////////////////////////////////////////////////
class ChuOSPageAllocator : public IPageAllocator
{
public:
	virtual void* Alloc(int32 nCount)
	{
		return chuos_AllocPages(nCount);
	}

	virtual void Free(void* pFirst, int32 nCount)
	{
		chuos_FreePages(pFirst, nCount);
	}
};
//////////////////////////////////////////////////

#endif

MemoryAllocator* g_pMemAlloc = null;

uint32 Memory::GetAllocated()
{
	return g_pMemAlloc->GetAllocated();
}

uint32 Memory::GetPeak()
{
	return g_pMemAlloc->GetPeak();
}

class DebugMemoryVisitor : public MemoryVisitor
{
public:
	virtual void VisitBlock(bool /*bValid*/, void* pAddr, int32 nSize, const char* pszFileName, int nLine)
	{
		ChustdDebugPrintLeak(pAddr, nSize, pszFileName, nLine);
	}
};

void Memory::DebugPrintBlocks()
{
	DebugMemoryVisitor dmv;
	g_pMemAlloc->EnumBlocks(dmv);
}

//////////////////////////////////////////////////////
void* Memory::Alloc64(int32 nByteCount, const char* pszFileName, int32 nLine)
{
	ASSERT(nByteCount >= 0);

	int32 nChunkCount = Memory::ByteCountToInt64Count(nByteCount);
	int32 nNewByteCount = 8 * (nChunkCount);

	void* pObject = g_pMemAlloc->Alloc(nNewByteCount, pszFileName, nLine);
	return pObject;
}

void* Memory::Alloc64(int32 nByteCount)
{
   return Memory::Alloc64(nByteCount, null, 0);
}

void Memory::Free64(void* pData)
{
	g_pMemAlloc->Free(pData);
}

#undef new

void* __cdecl operator new(size_t nSize)
{
	return g_pMemAlloc->Alloc( int32(nSize), null, 0);
}

void __cdecl operator delete(void* pBuffer)
{
	g_pMemAlloc->Free(pBuffer);
}

void* __cdecl operator new(size_t nSize, const char* pszFileName, int nLine)
{
	return g_pMemAlloc->Alloc( int32(nSize), pszFileName, nLine);
}

void  __cdecl operator delete(void* pBuffer, const char* /*pszFile*/, int /*nLine*/)
{
	g_pMemAlloc->Free(pBuffer);
}

void Memory::InitAllocator()
{
	// Create the memory allocator from a static buffer
	// Keep control on construction by constructing the object here
	static uint8 aBufferP[32];
	static uint8 aBufferM[736];

	const int nMASize = sizeof(MemoryAllocator);

	ASSERT( sizeof(aBufferM) >= nMASize);

	IPageAllocator*  pPageAlloc = (IPageAllocator*)  aBufferP;
	MemoryAllocator* pMemAlloc  = (MemoryAllocator*) aBufferM;
	
#ifdef WIN32
	::new(pPageAlloc) Win32PageAllocator;
#elif CHUOS
	::new(pPageAlloc) ChuOSPageAllocator;
#endif

	:: new(pMemAlloc) MemoryAllocator;

	pMemAlloc->SetPageAllocator(pPageAlloc);
	g_pMemAlloc = pMemAlloc;
}
/////////////////////////////////////////////////////////////////////////////
bool Memory::Is16BitsAligned(const void* p)
{
	uint8* pNull = null;
	uint8* p8 = (uint8*) p;
	uint32 nDiff = uint32(p8 - pNull);
	return (nDiff & 0x0001) == 0;
}

bool Memory::Is32BitsAligned(const void* p)
{
	uint8* pNull = null;
	uint8* p8 = (uint8*) p;
	uint32 nDiff = uint32(p8 - pNull);
	return (nDiff & 0x0003) == 0;
}

bool Memory::Is64BitsAligned(const void* p)
{
	uint8* pNull = null;
	uint8* p8 = (uint8*) p;
	uint32 nDiff = uint32(p8 - pNull);
	return (nDiff & 0x0007) == 0;
}

void Memory::Move(void* pDst, const void* pSrc, int32 nCount)
{
	if( pDst < pSrc)
	{
		Copy(pDst, pSrc, nCount);
	}
	else
	{
		// My poor cache is gonna be killed here
		uint8* pDst8 = (uint8*) pDst;
		uint8* pSrc8 = (uint8*) pSrc;

		for(int32 i = nCount - 1; i >= 0; --i)
		{
			pDst8[i] = pSrc8[i];
		}
	}
}

void Memory::Move16(void* pDst, const void* pSrc, int32 nCount)
{
	ASSERT( Is16BitsAligned(pDst));
	ASSERT( Is16BitsAligned(pSrc));

	if( pDst < pSrc)
	{
		Copy16(pDst, pSrc, nCount);
	}
	else
	{
		// My poor cache is gonna be killed here
		uint16* pDst16 = (uint16*) pDst;
		uint16* pSrc16 = (uint16*) pSrc;

		for(int32 i = nCount - 1; i >= 0; --i)
		{
			pDst16[i] = pSrc16[i];
		}
	}
}

void Memory::Copy(void* pDst, const void* pSrc, int32 nByteCount)
{
	if( Is32BitsAligned(pDst) && Is32BitsAligned(pSrc) && (nByteCount & 0x03) == 0 )
	{
		Copy32(pDst, pSrc, nByteCount / 4);
		return;
	}

	uint8* pDst8 = (uint8*) pDst;
	uint8* pSrc8 = (uint8*) pSrc;

	for(int32 i = 0; i < nByteCount; ++i)
	{
		pDst8[i] = pSrc8[i];
	}
}

void Memory::Copy16(void* pDst, const void* pSrc, int32 nCount)
{
	ASSERT( Is16BitsAligned(pDst));
	ASSERT( Is16BitsAligned(pSrc));

	uint16* pDst16 = (uint16*) pDst;
	uint16* pSrc16 = (uint16*) pSrc;

	for(int32 i = 0; i < nCount; ++i)
	{
		pDst16[i] = pSrc16[i];
	}
}

void Memory::Copy32(void* pDst, const void* pSrc, int32 nCount)
{
	ASSERT( Is32BitsAligned(pDst));
	ASSERT( Is32BitsAligned(pSrc));

	uint32* pDst32 = (uint32*) pDst;
	uint32* pSrc32 = (uint32*) pSrc;

	for(int32 i = 0; i < nCount; ++i)
	{
		pDst32[i] = pSrc32[i];
	}
}

void Memory::Copy64(void* pDst, const void* pSrc, int32 nCount)
{
	ASSERT( Is64BitsAligned(pDst));
	ASSERT( Is64BitsAligned(pSrc));

	uint64* pDst64 = (uint64*) pDst;
	uint64* pSrc64 = (uint64*) pSrc;

	for(int32 i = 0; i < nCount; ++i)
	{
		pDst64[i] = pSrc64[i];
	}
}

void Memory::Copy64From8(void* pDst, const void* pSrc, int32 nByteCount)
{
	ASSERT( Is64BitsAligned(pDst));
	
	uint8* pDst8 = (uint8*) pDst;
	uint8* pSrc8 = (uint8*) pSrc;

	int32 i = 0;
	for(; i < nByteCount - 7; i += 8)
	{
		pDst8[i] = pSrc8[i];
		pDst8[i + 1] = pSrc8[i + 1];
		pDst8[i + 2] = pSrc8[i + 2];
		pDst8[i + 3] = pSrc8[i + 3];
		pDst8[i + 4] = pSrc8[i + 4];
		pDst8[i + 5] = pSrc8[i + 5];
		pDst8[i + 6] = pSrc8[i + 6];
		pDst8[i + 7] = pSrc8[i + 7];
	}

	for(; i < nByteCount; ++i)
	{
		pDst8[i] = pSrc8[i];
	}
}

void Memory::Copy64From16(void* pDst, const void* pSrc, int32 nUInt16Count)
{
	ASSERT( Is64BitsAligned(pDst));
	ASSERT( Is16BitsAligned(pSrc));
	
	uint16* pDst16 = (uint16*) pDst;
	uint16* pSrc16 = (uint16*) pSrc;

	int32 i = 0;
	for(; i < nUInt16Count - 3; i += 4)
	{
		pDst16[i] = pSrc16[i];
		pDst16[i + 1] = pSrc16[i + 1];
		pDst16[i + 2] = pSrc16[i + 2];
		pDst16[i + 3] = pSrc16[i + 3];
	}

	for(; i < nUInt16Count; ++i)
	{
		pDst16[i] = pSrc16[i];
	}
}

void Memory::Swap(void* pDst, void* pSrc, int32 nCount)
{
	uint8* pDst8 = (uint8*) pDst;
	uint8* pSrc8 = (uint8*) pSrc;

	for(int32 i = 0; i < nCount; ++i)
	{
		uint8 nTmp = pDst8[i];
		pDst8[i] = pSrc8[i];
		pSrc8[i] = nTmp;
	}
}

void Memory::Zero(void* pDst, int32 nByteCount)
{
	if( Is16BitsAligned(pDst) && (nByteCount & 0x0001) == 0 )
	{
		Zero16(pDst, nByteCount / 2);
		return;
	}

	uint8* pDst8 = (uint8*) pDst;

	int32 nLastByte = nByteCount & 0x01;
	nByteCount -= nLastByte;

	for(int32 i = 0; i < nByteCount; i += 2)
	{
		pDst8[0] = 0;
		pDst8[1] = 0;
		pDst8 += 2;
	}

	if( nLastByte == 0)
		return;
	
	pDst8[0] = 0;
}

void Memory::Zero16(void* pDst, int32 nCount)
{
	ASSERT( Is16BitsAligned(pDst));

	uint16* pDst16 = (uint16*) pDst;

	for(int32 i = 0; i < nCount; ++i)
	{
		pDst16[i] = 0;
	}
}

void Memory::Zero32(void* pDst, int32 nCount)
{
	ASSERT( Is32BitsAligned(pDst));

	uint32* pDst32 = (uint32*) pDst;

	for(int32 i = 0; i < nCount; ++i)
	{
		pDst32[i] = 0;
	}
}

void Memory::Set(void* pDst, uint8 nValue, int32 nByteCount)
{
	if( Is16BitsAligned(pDst) && (nByteCount & 0x0001) == 0 )
	{
		uint16 nValue16 = MAKE16(nValue, nValue);
		Set16(pDst, nValue16, nByteCount / 2);
		return;
	}

	uint8* pDst8 = (uint8*) pDst;

	int32 nLastByte = nByteCount & 0x01;
	nByteCount -= nLastByte;
	for(int32 i = 0; i < nByteCount; i += 2)
	{
		pDst8[0] = nValue;
		pDst8[1] = nValue;
		pDst8 += 2;
	}

	if( nLastByte == 0)
		return;
	
	pDst8[0] = nValue;
}

void Memory::Set16(void* pDst, uint16 nValue, int32 nCount)
{
	ASSERT( Is16BitsAligned(pDst));

	uint16* pDst16 = (uint16*) pDst;

	for(int32 i = 0; i < nCount; ++i)
	{
		pDst16[i] = nValue;
	}
}

void Memory::Set32(void* pDst, uint32 nValue, int32 nCount)
{
	ASSERT( Is32BitsAligned(pDst));

	uint32* pDst32 = (uint32*) pDst;

	for(int32 i = 0; i < nCount; ++i)
	{
		pDst32[i] = nValue;
	}
}

bool Memory::IsEqual(const void* pSrc0, const void* pSrc1, int32 nByteCount)
{
	uint8* pSrc08 = (uint8*) pSrc0;
	uint8* pSrc18 = (uint8*) pSrc1;

	for(int32 i = 0; i < nByteCount; ++i)
	{
		if( pSrc08[i] != pSrc18[i])
			return false;
	}
	return true;
}

bool Memory::IsEqual32(const void* pSrc0, const void* pSrc1, int32 nCount)
{
	ASSERT( Is32BitsAligned(pSrc0));
	ASSERT( Is32BitsAligned(pSrc1));

	uint32* pSrc032 = (uint32*) pSrc0;
	uint32* pSrc132 = (uint32*) pSrc1;

	for(int32 i = 0; i < nCount; ++i)
	{
		if( pSrc032[i] != pSrc132[i])
			return false;
	}
	return true;
}

