///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "Master.h"
#include "MemoryAllocator.h"
#include "Math.h"
#include "Memory.h"
#include "Array.h" // For placement new

namespace chustd {

class ZoneTypeDirectory
{
public:
	ZoneTypeDirectory* m_pNext;

public:
	virtual bool  IsFull() const = 0;
	virtual bool  IsEmpty() const = 0;
	virtual void* AllocZone(IPageAllocator* pPageAlloc, int32 nManagedSizePower, int32 nSize) = 0;
	virtual bool  TryToFree(IPageAllocator* pPageAlloc, int32 nManagedSizePower, void* pMem, int32 nSize) = 0;
	virtual void  EnumBlocks(MemoryVisitor& memvis, int32 nManagedSizePower) = 0;
};

} // namespace chustd

////////////////////////////////////////////////////////////
using namespace chustd;
////////////////////////////////////////////////////////////

//////////
struct MemCouple
{
	void* pBuffer;
	int32 nSize; // En octets
};


//////////
// Les grosses allocations sont gres par cette classe
class BigZonesDirectory : public ZoneTypeDirectory
{
public:
	virtual bool IsFull() const
	{
		return m_nCoupleCount == GetMaxCoupleCount();
	}

	virtual bool IsEmpty() const
	{
		return m_nCoupleCount == 0;
	}

	static int32 GetMaxCoupleCount()
	{
		const int32 nCount = (k_nPageSize - sizeof(BigZonesDirectory)) / sizeof(MemCouple);
		return nCount;
	}

	void AddCouple(void* pMem, int32 nSize)
	{
		// Recherche une case de libre
		const int32 nMaxCoupleCount = GetMaxCoupleCount();
		for(int32 i = 0; i < nMaxCoupleCount; ++i)
		{
			if( m_aCouples[i].pBuffer == null)
			{
				// Libre
				m_aCouples[i].pBuffer = pMem;
				m_aCouples[i].nSize = nSize;
				break;
			}
		}
		m_nCoupleCount++;
	}

	virtual void* AllocZone(IPageAllocator* pPageAlloc, int32 /*nManagedSizePower*/, int32 nSize)
	{
		// Allocation de la grosse zone
		const int32 nBlockCount = Math::DivCeil(nSize, k_nPageSize);
		void* pMem = pPageAlloc->Alloc(nBlockCount);
		if( pMem == null)
		{
			return null;
		}

		AddCouple(pMem, nSize);
		return pMem;
	}

	// Returns true upon success
	virtual bool TryToFree(IPageAllocator* pPageAlloc, int32 /*nManagedSizePower*/, void* pMem, int32 nSize)
	{
		const int32 nPageCount = Math::DivCeil(nSize, k_nPageSize);

		const int32 nMaxCoupleCount = GetMaxCoupleCount();
		for(int32 i = 0; i < nMaxCoupleCount; ++i)
		{
			if( m_aCouples[i].pBuffer == pMem)
			{
				if( m_aCouples[i].nSize != nSize)
				{
					// Damaged
					ASSERT(0);
				}

				pPageAlloc->Free(pMem, nPageCount);

				// Libre
				m_aCouples[i].pBuffer = null;
				m_aCouples[i].nSize = 0;
				
				m_nCoupleCount--;
				return true;
			}
		}
		return false;
	}

	virtual void EnumBlocks(MemoryVisitor& memvis, int32 /*nManagedSizePower*/)
	{
		const int32 nMaxCoupleCount = GetMaxCoupleCount();
		for(int32 i = 0; i < nMaxCoupleCount; ++i)
		{
			MemCouple& mc = m_aCouples[i];
			if( mc.pBuffer != null && mc.nSize != 0)
			{
				void* pBlock = mc.pBuffer;
				MemoryAllocator::VisitBlock(pBlock, memvis);
			}
		}
	}

private:
	int32 m_nCoupleCount;
	MemCouple m_aCouples[1];
};

/////////////////////////////////////////////////////////////////////////////////////////////////////
// Les petites allocations
class SmallZonesDirectory : public ZoneTypeDirectory
{
private:
	struct Entry
	{
		///////////////////////
		void* m_pPage;
		int32 m_nNext;
		uint8 m_aStack[1]; // From 1 to 256 in size. 256 * 32 bytes = 8192 bytes
	
		///////////////////////
		uint8* AllocZone(IPageAllocator* pPageAlloc, int32 nManagedSizePower)
		{
			const int32 nZoneCountPerPage = k_nPageSize >> nManagedSizePower;

			if( m_pPage == null)
			{
				m_pPage = pPageAlloc->Alloc(1);
				if( m_pPage == null)
				{
					return null;
				}
				
				// Prepare the stack
				for(uint8 i = 0; i < nZoneCountPerPage; ++i)
				{
					m_aStack[i] = i;
				}
				m_nNext = nZoneCountPerPage;
			}

			// Pop stack
			if( m_nNext == 0)
			{
				// Stack empty
				return null;
			}

			--m_nNext;
			uint8 nLocalZoneIndex = m_aStack[m_nNext];
			
			uint8* pPage = (uint8*)m_pPage;
			uint8* pZone = pPage + (nLocalZoneIndex << nManagedSizePower);
			
			return pZone;
		}
	
		void FreeZone(IPageAllocator* pPageAlloc, uint8* pZone, int32 nManagedSizePower)
		{
			const int32 nZoneCountPerPage = k_nPageSize >> nManagedSizePower;
			int32 nLocalZoneIndex = int32(pZone - (uint8*)m_pPage) >> nManagedSizePower;
			
			ASSERT(nLocalZoneIndex >= 0);
			ASSERT(nLocalZoneIndex < nZoneCountPerPage);
			ASSERT(nLocalZoneIndex < 256); // Something is wrong in the algorithm if this assert fails

			// Push to stack
			m_aStack[m_nNext] = (uint8) nLocalZoneIndex;
			m_nNext++;

			if( m_nNext == nZoneCountPerPage)
			{
				pPageAlloc->Free(m_pPage, 1);
				m_pPage = null;
			}
		}
	};

public:
	/////////////////////////////////////////////////////////////////////////////////////////
	// Ajoute une nouvelle entre
	// Retourne l'adresse de la zone alloue, null sinon
	virtual void* AllocZone(IPageAllocator* pPageAlloc, int32 nManagedSizePower, int32 /*nSize*/)
	{
		if( IsFull())
			return null;

		uint8* paEntriesRaw = (uint8*) m_aEntries;
		const int32 nEntrySize = m_nEntrySize;

		// Test every entry this directory can hold
		const int32 nMaxEntryCount = m_nMaxEntryCount;
		for(int32 iEntry = 0; iEntry < nMaxEntryCount; ++iEntry)
		{
			Entry* pEntry = (Entry*) paEntriesRaw;
			
			//const int32 nZoneCountPerPage = k_nPageSize >> nManagedSizePower;
			uint8* pResult = pEntry->AllocZone(pPageAlloc, nManagedSizePower);
			if( pResult != null)
			{
				// Found
				m_nCurrentZoneCount++;
				return pResult;
			}

			paEntriesRaw += nEntrySize;
		}
		return null;
	}

	// Dsalloue une zone
	virtual bool TryToFree(IPageAllocator* pPageAlloc, int32 nManagedSizePower, void* pMem, int32 nSize)
	{
		ASSERT( !IsEmpty()); // TODO : think : make a real test ?
		ASSERT(nSize <= (1 << nManagedSizePower));

		uint8* paEntriesRaw = (uint8*) m_aEntries;
		const int32 nEntrySize = m_nEntrySize;

		// Find the entry that owns the zone
		const int32 nLimit = k_nPageSize - int32(1 << nManagedSizePower);
		const int32 nMaxEntryCount = m_nMaxEntryCount;
		for(int32 iEntry = 0; iEntry < nMaxEntryCount; ++iEntry)
		{
			Entry* pEntry = (Entry*) paEntriesRaw;

			uint8* pMinAddr = (uint8*)pEntry->m_pPage;
			uint8* pMaxAddr = pMinAddr + nLimit;

			if( pMinAddr <= pMem && pMem <= pMaxAddr)
			{
				// Entry found
				pEntry->FreeZone(pPageAlloc, (uint8*) pMem, nManagedSizePower);
				m_nCurrentZoneCount--;
				return true;
			}

			paEntriesRaw += nEntrySize;
		}
		return false;
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	virtual bool IsFull() const
	{
		return m_nCurrentZoneCount == m_nMaxZoneCount;
	}

	virtual bool IsEmpty() const
	{
		return m_nCurrentZoneCount == 0;
	}
	/////////////////////////////////////////////////////////////////////////////////////////

	/////////////////////////////////////////////////////////////////////////////////////////
	SmallZonesDirectory(int32 nManagedSizePower)
	{
		// Combien de descripteurs possibles dans ce directory ?
		const int32 nFreeByteCount = k_nPageSize - (sizeof(SmallZonesDirectory) - sizeof(Entry));
		
		const int32 nZonesPerEntry = k_nPageSize >> nManagedSizePower;
		//const int32 nZonePointerSize = sizeof(void*);

		// pPage + AllocCount + Bitmap
		const int32 nEntrySize = sizeof(uint8*) + 4 + nZonesPerEntry;
		m_nEntrySize = int16(nEntrySize);

		const int32 nMaxEntryCount = nFreeByteCount / nEntrySize;
		
		ASSERT(nMaxEntryCount <= 0x7fff);

		m_nMaxEntryCount = int16(nMaxEntryCount);
		m_nMaxZoneCount = int16(nMaxEntryCount * nZonesPerEntry);
		m_nCurrentZoneCount = 0;
	}

	virtual void EnumBlocks(MemoryVisitor& memvis, int32 nManagedSizePower)
	{
		// Dtermine la page  laquelle appartient la zone
		
		uint8* paEntriesRaw = (uint8*) m_aEntries;
		const int32 nEntrySize = m_nEntrySize;

		// Recherche l'entre correspondant  la mmoire  librer
		const int32 nMaxEntryCount = m_nMaxEntryCount;
		for(int32 iEntry = 0; iEntry < nMaxEntryCount; ++iEntry)
		{
			Entry* pEntry = (Entry*) paEntriesRaw;
			
			if( pEntry->m_pPage != null)
			{
				// Build a zone bitmap from stack
				uint8 aFreeIndex[256];
				Memory::Set(aFreeIndex, 1, 256);

				for(int32 iStack = pEntry->m_nNext - 1; iStack >= 0; --iStack)
				{
					uint8 nFreeZoneIndex = pEntry->m_aStack[iStack];
					aFreeIndex[nFreeZoneIndex] = 0;
				}

				const int32 nZoneCountPerBlock = k_nPageSize >> nManagedSizePower;
				for(int32 iZone = 0; iZone < nZoneCountPerBlock; ++iZone)
				{
					if( aFreeIndex[iZone] != 0)
					{
						// Une zone alloue
						uint8* pBlock = (uint8*)pEntry->m_pPage + (iZone << nManagedSizePower);
						MemoryAllocator::VisitBlock(pBlock, memvis);
					}
				}
			}

			paEntriesRaw += nEntrySize;
		}
	}

private:
	int16 m_nMaxEntryCount;
	int16 m_nMaxZoneCount;
	int16 m_nCurrentZoneCount;
	int16 m_nEntrySize;
	Entry m_aEntries[1];

	// More memory lying here
};

//////////////////////////////////////////////////////////////////
// Gestionnaire de zones alloues d'une taille donne
// Gre les rpertoires de zones
ZoneTypeDirManager::ZoneTypeDirManager()
{
	m_nManagedSizePower = -1;
	m_pOnes = null;
	m_pFullOnes = null;
}

////////////////////////////////////////////////////////////////////////////////
bool ZoneTypeDirManager::FreeZone(IPageAllocator* pPageAlloc, void* pMem, int32 nSize)
{
	// Only one thread at a time
	TmpLock tmplock(m_criticalsection);

	// Find the directory that holds the memory block pointed by pMem

	////////////////////////////////////////////////
	// Search into the full directories
	ZoneTypeDirectory* pDirPrevious = null;
	ZoneTypeDirectory* pDir = m_pFullOnes;
	while(true)
	{
		if( pDir == null)
			break;
 
		if( pDir->TryToFree(pPageAlloc, m_nManagedSizePower, pMem, nSize))
		{
			// Memory block found and deallocated
			// This directory is not full anymore, we put it in the list of notfull ones
			
			// 1) Relink
			if( pDirPrevious != null)
			{
				pDirPrevious->m_pNext = pDir->m_pNext;
			}
			else
			{
				m_pFullOnes = pDir->m_pNext;
			}

			// 2) Insert into notfull ones
			pDir->m_pNext = m_pOnes;
			m_pOnes = pDir;
			return true;
		}
		pDirPrevious = pDir;
		pDir = pDir->m_pNext;
	}

	////////////////////////////////////////////////
	// Search into the notfull directories
	pDirPrevious = null;
	pDir = m_pOnes;

	while(true)
	{
		if( pDir == null)
			break;

		if( pDir->TryToFree(pPageAlloc, m_nManagedSizePower, pMem, nSize))
		{
			// Memory block found and deallocated
			if( pDir->IsEmpty())
			{
				// It was also the last block used in that directory
				// so we can free that directory
				
				// 1) Relink
				if( pDirPrevious != null)
				{
					pDirPrevious->m_pNext = pDir->m_pNext;
				}
				else
				{
					m_pOnes = pDir->m_pNext;
				}

				// Free the page used for the directory
				pPageAlloc->Free(pDir, 1);
			}
			return true;
		}
		pDirPrevious = pDir;
		pDir = pDir->m_pNext;
	}
	
	return false;
}

////////////////////////////////////////////////////////////////////////////////
void* ZoneTypeDirManager::AllocZone(IPageAllocator* pPageAlloc, int32 nSize)
{
	// Only one thread at a time
	TmpLock tmplock(m_criticalsection);

	// Ajoute une entre dans l'index du gestionnaire des petites zones concern
	ZoneTypeDirectory* pDir = GetDirectoryToUse(pPageAlloc);
	if( pDir == null)
		return null;
	
	void* pResult = pDir->AllocZone(pPageAlloc, m_nManagedSizePower, nSize);
	if( pResult == null)
		return null;

	// Doit-on faire passer le rpertoire dans la liste des pleins ?
	// a sera fait  la prochaine allocation dans GetDirectoryToUse
	return pResult;
}

// Get a directory with some room for an allocation
ZoneTypeDirectory* ZoneTypeDirManager::GetDirectoryToUse(IPageAllocator* pPageAlloc)
{
	if( m_pOnes == null)
	{
		// No directory allocated
		ZoneTypeDirectory* pDir = NewDirectory(pPageAlloc, m_nManagedSizePower);
		if( pDir == null)
			return null;

		m_pOnes = pDir;
		return pDir;
	}

	while(true)
	{
		// Enough room ?
		ZoneTypeDirectory* pDir = m_pOnes;
		if( pDir == null)
		{
			// Needs to allocate a new directory
			break;
		}

		if( !pDir->IsFull())
		{
			// Yes, enough room in this directory
			return pDir;
		}
		
		// Not enough room in the head directory, we have to move it to the fullones list

		// Notfullones relinking
		m_pOnes = pDir->m_pNext;

		// Fullones relinking
		ZoneTypeDirectory* pPreviousFullHead = m_pFullOnes;
		m_pFullOnes = pDir;
		m_pFullOnes->m_pNext = pPreviousFullHead;
	}

	// Allocate a new directory for the notfullones list

	ZoneTypeDirectory* pDirNew = NewDirectory(pPageAlloc, m_nManagedSizePower);
	if( pDirNew == null)
		return null;

	// Notfullones linking
	m_pOnes = pDirNew;

	return pDirNew;
}

ZoneTypeDirectory* ZoneTypeDirManager::NewDirectory(IPageAllocator* pPageAlloc, int32 nManagedSize)
{
	void* pMem = pPageAlloc->Alloc(1);
	Memory::Zero32(pMem, k_nPageSize / 4);
	ZoneTypeDirectory* pDir = (ZoneTypeDirectory*) pMem;
	ConstructDirectory(pDir, nManagedSize);

	return pDir;
}

void ZoneTypeDirManager::EnumBlocks(MemoryVisitor& memvis)
{
	ZoneTypeDirectory* pDir = m_pFullOnes;
	while(true)
	{
		if( pDir == null)
			break;

		pDir->EnumBlocks(memvis, m_nManagedSizePower);
		pDir = pDir->m_pNext;
	}

	pDir = m_pOnes;
	while(true)
	{
		if( pDir == null)
			break;

		pDir->EnumBlocks(memvis, m_nManagedSizePower);
		pDir = pDir->m_pNext;
	}
}

//////////////////////////////////////////////////////////
// Placement new
#undef new

void BigZonesDirManager::ConstructDirectory(ZoneTypeDirectory* pDir, int32 /*nManagedSize*/)
{
	BigZonesDirectory* pBigDir = (BigZonesDirectory*) pDir;
	::new(pBigDir) BigZonesDirectory;
}

void SmallZonesDirManager::ConstructDirectory(ZoneTypeDirectory* pDir, int32 nManagedSize)
{
	SmallZonesDirectory* pSmallDir = (SmallZonesDirectory*) pDir;
	::new(pSmallDir) SmallZonesDirectory(nManagedSize);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Main class for memory allocation
//
// The MemoryAllocator manages:
//          DirManagers that manage:
//                           Directories that manage:
//                                          MemoryZones
const uint8  k_nFence8 = 0xfd;
const uint32 k_nFence32 = 0xfdfdfdfd;

MemoryAllocator::MemoryAllocator()
{
	for(int i = k_nMinZoneSizePower; i <= k_nPageSizePower; ++i)
	{
		m_aSmallZones[i - k_nMinZoneSizePower].m_nManagedSizePower = i;
	}

	m_pPageAlloc = null;

	m_nAllocated = 0;
	m_nPeak = 0;
}

void MemoryAllocator::SetPageAllocator(IPageAllocator* pPageAlloc)
{
	m_pPageAlloc = pPageAlloc;
}

// 8 bytes
struct MemBlockHeader
{
	uint32 nUserSize;
	uint32 nFence;
};

// 2*8 bytes on 32 bits platform, 3*8 on 64 bits platform with 64 bits pointers
struct MemBlockHeaderDbg
{
	const char* pszFileName;
	union { void* pAlign; uint32 nLine; }; // Force padding here for 64 bits platform

	uint32 nUserSize;
	uint32 nFence;
};

// Allocates memory
void* MemoryAllocator::Alloc(int32 nUserSize, const char* pszFileName, int nLine)
{
	// The requested amount cannot be negative (though it can be 0)
	if( nUserSize < 0)
	{
		ASSERT(0); // Warn the user
		return null;
	}

	// Start computing the needed internal size
	int32 nInternalSize = nUserSize + 1; // +1 is for the tail fence byte

	if( pszFileName == null)
	{
		nInternalSize += sizeof(MemBlockHeader);
	}
	else
	{
		nInternalSize += sizeof(MemBlockHeaderDbg);
	}

	// Get the appropriate directory manager
	void* pBuffer = null;
	int32 nZoneAllocatorType = GetAllocatorType(nInternalSize);

	if( nZoneAllocatorType < 0)
	{
		pBuffer = m_bigzones.AllocZone(m_pPageAlloc, nInternalSize);
	}
	else
	{
		pBuffer = m_aSmallZones[nZoneAllocatorType].AllocZone(m_pPageAlloc, 0);
	}

	if( pBuffer == null)
	{
		return null;
	}

	void* pUser;
	if( pszFileName == null)
	{
		MemBlockHeader* pHeader = (MemBlockHeader*) pBuffer;
		pHeader->nUserSize = nUserSize;
		pHeader->nFence = k_nFence32;
		pUser = pHeader + 1;
	}
	else
	{
		MemBlockHeaderDbg* pHeader = (MemBlockHeaderDbg*) pBuffer;
		pHeader->pszFileName = pszFileName;
		pHeader->nLine = nLine;
		pHeader->nUserSize = nUserSize | 0x80000000; // Debug flag
		pHeader->nFence = k_nFence32;
		pUser = pHeader + 1;
	}

	uint8* pUser8 = (uint8*) pUser;
	pUser8[nUserSize] = k_nFence8;

	m_nAllocated += nUserSize;

	if( m_nAllocated > m_nPeak)
	{
		m_nPeak = m_nAllocated;
	}
	return pUser;
}

void MemoryAllocator::Free(void* pUserAddr)
{
	if( pUserAddr == null)
		return;

	MemBlockHeader* pHeader = (MemBlockHeader*) pUserAddr;
	pHeader -= 1;
	
	int32 nUserSize = pHeader->nUserSize;
	
	uint8* pInternal;
	int32 nInternalSize;
	const char* pszFileName = null;
	int nLine = 0;
	
	if( nUserSize & 0x80000000)
	{
		// A filename & line number is provided
		MemBlockHeaderDbg* pHeaderDbg = (MemBlockHeaderDbg*) pUserAddr;
		pHeaderDbg -= 1;

		nUserSize &= 0x7fffffff; // Remove the flag
	
		pszFileName = pHeaderDbg->pszFileName;
		nLine = (int) pHeaderDbg->nLine;

		pInternal = (uint8*) pHeaderDbg;
		nInternalSize = sizeof(MemBlockHeaderDbg);
	}
	else
	{
		pInternal = (uint8*) pHeader;
		nInternalSize = sizeof(MemBlockHeader);
	}

	nInternalSize += nUserSize + 1; // + 1 for the tail fence byte

	if( pHeader->nFence != k_nFence32)
	{
		// Damaged
		VERIFYMSG(0, "Memory damaged at beginning of block");
		return;
	}

	uint8* pUser8 = (uint8*) pUserAddr;
	if( pUser8[nUserSize] != k_nFence8)
	{
		// Damaged
		VERIFYMSG(0, "Memory damaged at end of block");
		return;
	}

	int32 nZoneAllocatorType = GetAllocatorType(nInternalSize);
	if( nZoneAllocatorType < 0)
	{
		bool bFreeOk = m_bigzones.FreeZone(m_pPageAlloc, pInternal, nInternalSize);

		// Warn the user, the pointer was not found, the address is invalid
		VERIFYMSG(bFreeOk, "Invalid address for big block");
	}
	else
	{
		bool bFreeOk = m_aSmallZones[nZoneAllocatorType].FreeZone(m_pPageAlloc, pInternal, 0);
		
		// Warn the user, the pointer was not found, the address is invalid
		VERIFYMSG(bFreeOk, "Invalid address for small block");
	}

	m_nAllocated -= nUserSize;
}

// Find the most suitable allocator
// Returns -1 if the bigzone allocator should be use
// Returns >= 0 for the smallzone allocator index
int32 MemoryAllocator::GetAllocatorType(int32 nInternalSize)
{
	if( nInternalSize > k_nPageSize)
	{
		// Too big, use the big zones allocator
		return -1;
	}
	
	if( nInternalSize <= k_nMinZoneSize)
	{
		// Too small, use the first small zones allocator
		return 0;
	}

	// Use an intermediate small zone allocator
	int32 nSizePow2 = Math::NextPowerOf2(nInternalSize);
	int32 nPower = Math::Log2(nSizePow2);
	nPower -= k_nMinZoneSizePower;
	return nPower;
}

uint32 MemoryAllocator::GetAllocated() const
{
	return m_nAllocated;
}

uint32 MemoryAllocator::GetPeak() const
{
	return m_nPeak;
}

void MemoryAllocator::EnumBlocks(MemoryVisitor& memvis)
{
	foreach(m_aSmallZones, i)
	{
		ZoneTypeDirManager& ztm = m_aSmallZones[i];
		ztm.EnumBlocks(memvis);
	}

	m_bigzones.EnumBlocks(memvis);
}

// static
void MemoryAllocator::VisitBlock(void* pBlock, MemoryVisitor& memvis)
{
	void* pUserAddr = null;
	int32 nUserSize = 0;
	const char* pszFileName = null;
	int nLine = 0;

	bool bValid = false;

	MemBlockHeader* pHeader = (MemBlockHeader*) pBlock;
	if( (pHeader->nUserSize & 0x80000000) == 0 && pHeader->nFence == k_nFence32 )
	{
		// A normal block
		pszFileName = null;
		nLine = 0;
		pUserAddr = pHeader + 1;

		bValid = true;
	}
	else
	{
		// A debug block ?
		MemBlockHeaderDbg* pHeaderDbg = (MemBlockHeaderDbg*) pBlock;
		if( (pHeaderDbg->nUserSize & 0x80000000) != 0 && pHeaderDbg->nFence == k_nFence32)
		{
			pszFileName = pHeaderDbg->pszFileName;
			nLine = pHeaderDbg->nLine;
			pUserAddr = pHeaderDbg + 1;
			nUserSize &= 0x7fffffff;

			bValid = true;
		}
	}
	
	if( bValid)
	{
		uint8* pBuf8 = (uint8*) pUserAddr;
		bValid = pBuf8[nUserSize] == k_nFence8;
	}
	memvis.VisitBlock(bValid, pUserAddr, nUserSize, pszFileName, nLine);
}
