///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#ifndef CHUSTD_MEMORYALLOCATOR_H
#define CHUSTD_MEMORYALLOCATOR_H

#include "FixArray.h"
#include "CriticalSection.h"

namespace chustd {

///////////////////////////////////////////////////////////////////////////////
class IPageAllocator
{
public:
	virtual void* Alloc(int32 nCount) = 0;
	virtual void Free(void* pFirst, int32 nCount) = 0;
};

///////////////////////////////////////////////////////////////////////////////	
const int32 k_nPageSizePower = 12; // 4096 bytes
const int32 k_nPageSize = 1 << k_nPageSizePower;
const int32 k_nPageSizeMask = k_nPageSize - 1;

// Smallest allocatable internal size (internal size = user size + zone overhead)
const int32 k_nMinZoneSizePower = 5; // 32 bytes
const int32 k_nMinZoneSize = 1 << k_nMinZoneSizePower; // 32 bytes

//////////////////////////////////////////////////////////
class MemoryVisitor
{
public:
	virtual void VisitBlock(bool bValid, void* pUserAddr, int32 nUserSize, const char* pszFileName, int nLine) = 0;
};
///////////////////////////////////////////////////////////////////////////////
class ZoneTypeDirectory;

class ZoneTypeDirManager
{
public:
	int32 m_nManagedSizePower; // -1 = big zones

private:
	ZoneTypeDirectory* m_pOnes;     // List of directories with still room inside
	ZoneTypeDirectory* m_pFullOnes; // List of full directories

	CriticalSection m_criticalsection; // Only one thread at a time can allocate/free for a given zone size

public:
	ZoneTypeDirManager();

	bool FreeZone(IPageAllocator* pPageAlloc, void* pMem, int32 nSize);
	void* AllocZone(IPageAllocator* pPageAlloc, int32 nSize);

	void EnumBlocks(MemoryVisitor& memvis);

private:
	ZoneTypeDirectory* GetDirectoryToUse(IPageAllocator* pPageAlloc);
	ZoneTypeDirectory* NewDirectory(IPageAllocator* pPageAlloc, int32 nManagedSizePower);

protected:
	virtual void ConstructDirectory(ZoneTypeDirectory* pDir, int32 nManagedSizePower) = 0;
};

//////////////////////////////////////////////////////////
class BigZonesDirManager : public ZoneTypeDirManager
{
protected:
	virtual void ConstructDirectory(ZoneTypeDirectory* pDir, int32 nManagedSizePower);
};

//////////////////////////////////////////////////////////
class SmallZonesDirManager : public ZoneTypeDirManager
{
protected:
	virtual void ConstructDirectory(ZoneTypeDirectory* pDir, int32 nManagedSizePower);
};

//////////////////////////////////////////////////////////
class MemoryAllocator
{
public:
	void  SetPageAllocator(IPageAllocator* pPageAlloc);
	void* Alloc(int32 nUserSize, const char* pszFileName, int nLine);
	void  Free(void* pUserAddr);

	uint32 GetAllocated() const;
	uint32 GetPeak() const;

	void EnumBlocks(MemoryVisitor& memvis);

	MemoryAllocator();

	static void VisitBlock(void* pBlock, MemoryVisitor& memvis);

private:
	// Les gestionnaires de blocs allous dont la taille varie de 16  PaseSize/2 octets
	chustd::FixArray<SmallZonesDirManager, (k_nPageSizePower - k_nMinZoneSizePower) + 1> m_aSmallZones;
	
	// Le gestionnaire des gros blocs
	BigZonesDirManager m_bigzones; // [k_nBlockSize+1 .. oo]

	IPageAllocator* m_pPageAlloc;

	uint32 m_nAllocated;
	uint32 m_nPeak;

private:
	int32 GetAllocatorType(int32 nInternalSize);
};

} // namespace chustd

#endif // ndef CHUSTD_MEMORYALLOCATOR_H
