///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "Master.h"
#include "Png.h"
#include "File.h"
#include "Math.h"

//////////////////////////////////////////////////////////////////////
using namespace chustd;

static const uint8 k_PngSignature[8] = {'\x89','P','N','G','\r','\n','\x1a','\n'};

//////////////////////////////////////////////////////////////////////
void Png::HeaderInfo::SwapBytes()
{
	nWidth = IFile::Swap32(nWidth);
	nHeight = IFile::Swap32(nHeight);
}

//////////////////////////////////////////////////////////////////////
Png::Png()
{
	Initialize();
}

Png::~Png()
{

}

// Initialize members
// The initialization is made outside the constructor so we can reset the
// instance in order to load several images with the same instance
void Png::Initialize()
{
	m_nWidth = 0;
	m_nHeight = 0;
	m_nLastError = 0;

	m_nSizeofPixel = 0;
	m_nSizeofPixelInBits = 0;
	m_nUncompressedDataSize = 0;
	m_nByteWidth = 0;

	m_bFirstIdat = true;
	m_nOutputOffset = 0;

	m_bChunkFound_gAMA = false;
	m_nGamma = 100000;

	m_bChunkFound_tRNS = false;
	m_nTransGrey = 0;
	m_nTransRed = 0;
	m_nTransGreen = 0;
	m_nTransBlue = 0;
	
	m_bChunkFound_bKGD = false;
	m_nBkIndex = 0;
	m_nBkGrey = 0;
	m_nBkRed = 0;
	m_nBkGreen = 0;
	m_nBkBlue = 0;

	chustd::Memory::Zero(&m_IHDR, sizeof(m_IHDR));
}

bool Png::IsPng(IFile& file)
{
	const int32 nSignSize = sizeof(k_PngSignature);
	uint8 aRead[nSignSize];
	
	if( file.Read(aRead, nSignSize) != nSignSize)
	{
		return false;
	}

	if( !chustd::Memory::IsEqual(k_PngSignature, aRead, nSignSize))
	{
		return false;
	}
	return true;
}

bool Png::WriteSignature(IFile& file)
{
	int32 nWritten = file.Write(k_PngSignature, 8);
	return nWritten == 8;
}

// Loads a png picture from an open file
// Returns false if the loading failed
bool Png::LoadFromFile(IFile& file)
{
	// Reset the object
	FreeBuffer();
	Initialize();


	//////////////////////////////////////////////////////
	// Read the header first
	if( !IsPng(file))
	{
		m_nLastError = errNotAPngFile;
		return false;
	}

	ChunkedFile cd(file);

	// Current position is the start of a chunk collection

	bool bOkHandled = true;
	bool bIENDFound = false;

	while( !bIENDFound)
	{
		bool bChunkPresent = cd.BeginChunkRead();
		if( !bChunkPresent)
		{
			// No more chunk
			break;
		}

		switch(cd.m_nChunk)
		{
		case chunk_IHDR:
			bOkHandled = Handle_IHDR(file, cd.m_nSizeof);
			break;
		case chunk_gAMA:
			bOkHandled = Handle_gAMA(file, cd.m_nSizeof);
			break;
		case chunk_PLTE:
			bOkHandled = Handle_PLTE(file, cd.m_nSizeof);
			break;
		case chunk_tRNS:
			bOkHandled = Handle_tRNS(file, cd.m_nSizeof);
			break;
		case chunk_IDAT:
			bOkHandled = Handle_IDAT(file, cd.m_nSizeof);
			break;
		case chunk_bKGD:
			bOkHandled = Handle_bKGD(file, cd.m_nSizeof);
			break;

		case chunk_IEND:
			bIENDFound = true;
			break;

		default:
			break;
		}
		
		if( !bOkHandled)
		{
			// Error while reading a chunk content in one of the Handle_* function
			// m_nLastError is correctly set by one of those functions
			break;
		}

		if( !cd.EndChunkRead())
		{
			// Unexpected end of stream, CRC not present or chunk mishandled (read too much)
			m_nLastError = uncompleteFile;
			bOkHandled = false;
			break;
		}
	}

	m_CompressedBuffer.RemoveAll();

	if( bOkHandled && !bIENDFound)
	{
		// End of stream with no damaged chunk but IEND not found
		m_nLastError = errNoIENDFound;
		
		// We set bOkHandled to false so we perform the needed on-error cleanup
		bOkHandled = false;
	}
	
	if( !bOkHandled)
	{
		// An error occurred, we clean the object except the lasterror field
		m_ZStream.InflateEnd();

		int32 nLastError = m_nLastError;
		FreeBuffer();
		m_nLastError = nLastError;
		return false;
	}
	
	bool bTerminateOk = TerminateJob();
	if( !bTerminateOk)
	{
		// An error occurred, we clean the object except the lasterror field
		// (TerminateJob fills m_nLastError correctly)

		int32 nLastError = m_nLastError;
		FreeBuffer();
		m_nLastError = nLastError;
		return false;
	}

	return true;
}

bool Png::Handle_IHDR(IFile& file, uint32 nDataSizeof)
{
	// Png IHDR is 13 bytes
	if( nDataSizeof != 13)
	{
		m_nLastError = errBadHeaderSize;
		return false;
	}

	ASSERT( sizeof(HeaderInfo) >= 13);

	if( file.Read(&m_IHDR, nDataSizeof) != int32(nDataSizeof))
	{
		m_nLastError = uncompleteFile;
		return false;
	}
	
	if( file.ShouldSwapBytes())
	{
		m_IHDR.SwapBytes();
	}

	// Fill parent class attributes
	m_nWidth = m_IHDR.nWidth;
	m_nHeight = m_IHDR.nHeight;

	// Maximum size is 2**25, this lets us 7 bits to do computations on width and height
	const int32 nMaxWidth = 1 << 25;
	if( m_IHDR.nWidth <= 0 || m_IHDR.nWidth >= nMaxWidth
		|| m_IHDR.nHeight <= 0 || m_IHDR.nHeight >= nMaxWidth)
	{
		// We limit the picture size to a sensible maximum size so we can detect
		// bad data in the file
		m_nLastError = errBadPicSize;
		return false;
	}

	if( m_IHDR.nFilterMethod != 0)
	{
		// The only filtering method supported is "Adaptative"
		// (none, sub, up, average, paeth)
		m_nLastError = errBadFilterMethod;
		return false;
	}

	if( m_IHDR.nInterlaceMethod > 1)
	{
		m_nLastError = errBadInterlaceMethod;
		return false;
	}

	PixelFormat epf = GetPixelFormat();
	m_nSizeofPixelInBits = ImageType::SizeofPixelInBits(epf);

	// Compute the width of the image in bytes
	m_nByteWidth = ImageType::ComputeByteWidth(epf, m_IHDR.nWidth);

	// Computes the size of a pixel in bytes
	m_nSizeofPixel = (m_nSizeofPixelInBits + 7 ) / 8;

	//////////////////////////////////////////////////////////////////////
	// Allocate the image buffer

	return AllocateImageBuffer(m_IHDR.nInterlaceMethod == 1);
}

// Allocates the image buffer
// bInterlacedBuffer: true if the buffer will hold interlaced data
// Returns false if allocation failed
bool Png::AllocateImageBuffer(bool bInterlacedBuffer)
{
	// One extra byte is used in order to store the filtering sub-code
	const int32 nRowByteCount = m_nByteWidth + 1;

	if( bInterlacedBuffer)
	{
		// The raw data size is bigger is interlacing is used
		m_nUncompressedDataSize = ComputeInterlacedSize(m_IHDR.nWidth, m_IHDR.nHeight, m_nSizeofPixelInBits);
	}
	else
	{
		m_nUncompressedDataSize = nRowByteCount * m_IHDR.nHeight;
	}

	if( !m_aPixelBuffer.SetSize(m_nUncompressedDataSize))
	{
		m_nLastError = notEnoughMemory;
		return false;
	}
	return true;
}

// Gets gamma information
bool Png::Handle_gAMA(IFile& file, uint32 nDataSizeof)
{
	m_bChunkFound_gAMA = true;

	if( nDataSizeof < 4)
	{
		m_nLastError = errNotEnoughDataInGammaChunk;
		return false;
	}
	
	if( !file.Read32(m_nGamma))
	{
		m_nLastError = uncompleteFile;
		return false;
	}

	return true;
}

// Gets bKGD information
bool Png::Handle_bKGD(IFile& file, uint32 nDataSizeof)
{
	m_bChunkFound_bKGD = true;

	if( m_IHDR.nColorType == 0x03)
	{
		// Indexed color
		if( nDataSizeof < 1)
		{
			m_nLastError = errNotEnoughDataInBkChunk;
			return false;
		}

		if( !file.Read8(m_nBkIndex))
		{
			m_nLastError = uncompleteFile;
			return false;
		}
		return true;
	}
	
	if( m_IHDR.nColorType == 0x00 || m_IHDR.nColorType == 0x04)
	{
		// Grey or Grey+Alpha
		if( nDataSizeof < 2)
		{
			m_nLastError = errNotEnoughDataInBkChunk;
			return false;
		}

		if( !file.Read16(m_nBkGrey))
		{
			m_nLastError = uncompleteFile;
			return false;
		}
		return true;
	}

	if( m_IHDR.nColorType == 0x02 || m_IHDR.nColorType == 0x06)
	{
		if( nDataSizeof < 6)
		{
			m_nLastError = errNotEnoughDataInBkChunk;
			return false;
		}
		
		// TrueColor or TrueColor+Alpha
		if( !(file.Read16(m_nBkRed) && file.Read16(m_nBkGreen) && file.Read16(m_nBkBlue)) )
		{
			m_nLastError = uncompleteFile;
			return false;
		}
		return true;
	}

	// Other type should not exist and be detected when handling the IHDR chunk
	return true;
}

bool Png::Handle_tRNS(IFile& file, uint32 nDataSizeof)
{
	m_bChunkFound_tRNS = true;

	if( m_IHDR.nColorType == 3)
	{
		// Indexed color image
		// Note : there can be less alpha values than palette entries
		// In this case, the remaining entries use 255 for their alpha value
		uint8 aAlphaValues[256];

		int32 nColorCount = nDataSizeof;
		if( nColorCount > 256)
		{
			nColorCount = 256;
		}
		else if( nColorCount < 0)
		{
			nColorCount = 0;
		}

		file.Read( aAlphaValues, nColorCount);

		for(int32 i = 0; i < nColorCount; ++i)
		{
			m_PLTE.m_aColors[i].SetAlpha( aAlphaValues[i]);
		}
	}
	else if( m_IHDR.nColorType == 0)
	{
		// Greyscale
		if( !file.Read16(m_nTransGrey))
		{
			m_nLastError = uncompleteFile;
			return false;
		}
	}
	else if( m_IHDR.nColorType == 2)
	{
		// True color without alpha
		if( !(file.Read16(m_nTransRed) && file.Read16(m_nTransGreen) && file.Read16(m_nTransBlue)) )
		{
			m_nLastError = uncompleteFile;
			return false;
		}
	}
	return true;
}

// Gets palette information
bool Png::Handle_PLTE(IFile& file, uint32 nDataSizeof)
{
	int32 nColorCount = nDataSizeof / 3;
	if( nColorCount > 256)
	{
		nColorCount = 256;
	}
	m_PLTE.m_nCount = nColorCount;

	uint8 aTmpPalette[256 * 3];
	file.Read( aTmpPalette, nColorCount * 3);

	int32 iTmpColor = 0;

	for(int32 i = 0; i < m_PLTE.m_nCount; ++i)
	{
		const uint8 r = aTmpPalette[iTmpColor + 0];
		const uint8 g = aTmpPalette[iTmpColor + 1];
		const uint8 b = aTmpPalette[iTmpColor + 2];
		iTmpColor += 3;

		m_PLTE.m_aColors[i].SetRgb(r, g, b);
	}
	return true;
}

// Gets image data information
bool Png::Handle_IDAT(IFile& file, uint32 nDataSizeof)
{
	if( m_bFirstIdat)
	{
		m_bFirstIdat = false;

		// The first time we meet an IDAT chunk, we initialize the zstream structure
		m_ZStream.SetBuffers(null, 0, null, 0);
		
		const int nZErr = m_ZStream.InflateInit();
		if( nZErr != ZStream::Z_OK)
		{
			m_nLastError = errInflateErr;
			return false;
		}
		m_nOutputOffset = 0;
	}

	const uint32 nCompressedSize = nDataSizeof;
	m_CompressedBuffer.SetSize(nCompressedSize);

	uint32 nRead = file.Read(m_CompressedBuffer.GetPtr(), nCompressedSize);
	if( nRead != nCompressedSize)
	{
		m_nLastError = errUnexpectedEndOfFileInIDAT;
		return false;
	}
	
	const uint32 nAvailableOut = m_nUncompressedDataSize - m_nOutputOffset;

	uint8* pIn  = m_CompressedBuffer.GetPtr();
	uint8* pOut = m_aPixelBuffer.GetPtr() + m_nOutputOffset;
	m_ZStream.SetBuffers(pIn, nCompressedSize, pOut, nAvailableOut);
	
	int nZRet2 = m_ZStream.Inflate(ZStream::Z_SYNC_FLUSH);
	
	if( nZRet2 == ZStream::Z_OK)
	{
	
	}
	else if( nZRet2 == ZStream::Z_STREAM_END)
	{
	
	}
	else
	{
		m_nLastError = errInflateErr;
		return false;
	}

	uint32 nUncompressedSize = nAvailableOut - m_ZStream.GetOutAvailable();
	m_nOutputOffset += nUncompressedSize;

	return true;
}

PixelFormat Png::GetPixelFormat() const
{
	// To the bottom : 
	//   default: grey, flags: 0x1=palette used, 0x2=color used, 0x4=alpha used
	// To the left :
	//   Bit depth
	static const PixelFormat aPF[7][6] =
	{
		{ pf1bppGrayScale,    pf2bppGrayScale,    pf4bppGrayScale,    pf8bppGrayScale,      pf16bppGrayScale  },
		{ pfUnknown,  pfUnknown,  pfUnknown,  pfUnknown,    pfUnknown },
		{ pfUnknown,  pfUnknown,  pfUnknown,  pf24bppRgb,       pf48bppRgb   },
		{ pf1bppIndexed, pf2bppIndexed, pf4bppIndexed, pf8bppIndexed,   pfUnknown },
		{ pfUnknown,  pfUnknown,  pfUnknown,  pf16bppGrayScaleAlpha, pf32bppGrayScaleAlpha  },
		{ pfUnknown,  pfUnknown,  pfUnknown,  pfUnknown,    pfUnknown },
		{ pfUnknown,  pfUnknown,  pfUnknown,  pf32bppRgba,  pf64bppRgba  }
	};

	PixelFormat pf = pfUnknown;
	const int32 nDown = m_IHDR.nColorType;
	const int32 nLeft = Math::Log2( m_IHDR.nBitDepth );

	if(nDown >= 0 && nDown <= 6 && nLeft >= 0 && nLeft <= 4)
	{
		pf = aPF[nDown][nLeft];
	}

	return pf;
}

bool Png::GetIHDRFeaturesFromPixelFormat(PixelFormat epf, uint8& nColorType, uint8& nBitsPerComponent)
{
	switch(epf)
	{
	case pfUnknown:
		return false;
		
	case pf1bppGrayScale:      // 1bit per pixel
		nColorType = cfGrey;
		nBitsPerComponent = 1;
		return true;

	case pf2bppGrayScale:       // 2bits per pixel
		nColorType = cfGrey;
		nBitsPerComponent = 2;
		return true;

	case pf4bppGrayScale:       // 4bits per pixel
		nColorType = cfGrey;
		nBitsPerComponent = 4;
		return true;

	case pf8bppGrayScale:       // 8bits per pixel
		nColorType = cfGrey;
		nBitsPerComponent = 8;
		return true;

	case pf16bppGrayScale:      // 16bits per pixel
		nColorType = cfGrey;
		nBitsPerComponent = 16;
		return true;

	case pf16bppGrayScaleAlpha:  // GGAA
		nColorType = cfAlpha;
		nBitsPerComponent = 8;
		return true;

	case pf32bppGrayScaleAlpha: // GGGGAAAA
		nColorType = cfAlpha;
		nBitsPerComponent = 16;
		return true;

	case pf1bppIndexed:    // Indexed 1bit per pixel
		nColorType = cfColor | cfPalette;
		nBitsPerComponent = 1;
		return true;

	case pf2bppIndexed:    // Indexed 2bits per pixel
		nColorType = cfColor | cfPalette;
		nBitsPerComponent = 2;
		return true;

	case pf4bppIndexed:    // Indexed 4bits per pixel
		nColorType = cfColor | cfPalette;
		nBitsPerComponent = 4;
		return true;

	case pf8bppIndexed:    // Indexed 8bits per pixel
		nColorType = cfColor | cfPalette;
		nBitsPerComponent = 8;
		return true;

	case pf24bppRgb:        // RRGGBB
		nColorType = cfColor;
		nBitsPerComponent = 8;
		return true;

	case pf48bppRgb:       // RRRRGGGGBBBB 
		nColorType = cfColor;
		nBitsPerComponent = 16;
		return true;

	case pf32bppRgba:   // RRGGBBAA
		nColorType = cfColor | cfAlpha;
		nBitsPerComponent = 8;
		return true;

	case pf32bppBgra:   // BBGGRRAA
		nColorType = cfColor | cfAlpha;
		nBitsPerComponent = 8;
		return true;

	case pf64bppRgba:  // RRRRGGGGBBBBAAAA	
		nColorType = cfColor | cfAlpha;
		nBitsPerComponent = 16;
		return true;
	}
	
	return false;
}


const Palette* Png::GetPalette() const
{
	return &m_PLTE;
}

// Finalizes uncompression and unfiltering
bool Png::TerminateJob()
{	
	int nZRet = m_ZStream.InflateEnd();
	if( nZRet != ZStream::Z_OK)
	{
		m_nLastError = errInflateErr;
		return false;
	}

	if( m_IHDR.nInterlaceMethod == 1)
	{
		return UnfilterAndUninterlace();
	}

	// No interlacing, just unfilter
	if( !UnfilterBlock(m_aPixelBuffer.GetPtr(), m_IHDR.nHeight, m_nByteWidth, m_nSizeofPixel))
	{
		m_nLastError = errBadFilterType;  // Unknown filtering type
		return false;
	}
	return true;
}

// One of the Adaptative unfiltering method
// a = left, b = top, c = top-left
uint8 Png::PaethPredictor(uint8 a, uint8 b, uint8 c)
{
	const int32 p = a + b - c;  // Initial estimation
	
	// Compute distances
	int32 pa = p - a;
	int32 pb = p - b;
	int32 pc = p - c;
	
	if( pa < 0 )
	{
		pa = -pa;
	}

	if( pb < 0 )
	{
		pb = -pb;
	}

	if( pc < 0 )
	{
		pc = -pc;
	}

	if( pa <= pb && pa <= pc )
	{
		return a;
	}

	if( pb <= pc )
	{
		return b;
	}
	return c;
}

bool Png::UnfilterAndUninterlace()
{
	// Keep the previous data to be used as input in the unfiltering process
	ByteArray aOldBuffer;
	m_aPixelBuffer.MoveTo(aOldBuffer);
	
	// m_aPixelBuffer is now empty, we reallocate it as it will be used as output in the unfiltering process
	// m_nUncompressedDataSize is modified too
	if( !AllocateImageBuffer(false))
	{
		return false;
	}

	uint8* const pInterlaced = aOldBuffer.GetPtr();
	const int32  nInterlacedSize = aOldBuffer.GetSize(); // Used in debug mode

	uint8* const pResult = m_aPixelBuffer.GetPtr();
	const int32  nResultSize = m_nUncompressedDataSize;

	static const int32 aStartingRow[7] = { 0, 0, 4, 0, 2, 0, 1 };
	static const int32 aStartingCol[7] = { 0, 4, 0, 2, 0, 1, 0 };
	static const int32 aRowIncrement[7] = { 8, 8, 8, 4, 4, 2, 2 };
	static const int32 aColIncrement[7] = { 8, 8, 4, 4, 2, 2, 1 };

	///////////////////////////////////////////////////////////////////////////
	// Masks arrays for pixel sizes < 8
	static const uint8 aMask4[2] = { 0xf0, 0x0f };
	static const uint8 aShift4[2] = { 4, 0 };

	static const uint8 aMask2[4] = { 0xc0, 0x30, 0x0c, 0x03 };
	static const uint8 aShift2[4] = { 6, 4, 2, 0 };

	static const uint8 aMask1[8] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };
	static const uint8 aShift1[8] = { 7, 6, 5, 4, 3, 2, 1, 0 };
	
	const uint8* paMask = null;
	const uint8* paShift = null;
	int32 nDstShift = 0;

	if( m_nSizeofPixelInBits < 8)
	{
		Memory::Zero(pResult, nResultSize);
		if( m_nSizeofPixelInBits == 4)
		{
			paMask = aMask4;
			paShift = aShift4;
			nDstShift = 1;
		}
		else if( m_nSizeofPixelInBits == 2)
		{
			paMask = aMask2;
			paShift = aShift2;
			nDstShift = 2;
		}
		else if( m_nSizeofPixelInBits == 1)
		{
			paMask = aMask1;
			paShift = aShift1;
			nDstShift = 3;
		}
	}
	///////////////////////////////////////////////////////////////////////////
	const int32 nWidth = m_IHDR.nWidth;
	const int32 nHeight = m_IHDR.nHeight;

	// Current byte position in the interlaced buffer
	int32 nSrcIndex = 0;
	
	// Sub index for pixel < 8 bits
	int32 nSrcSubIndex = 0;
	int32 nMaxSrcSubIndex = (8 / m_nSizeofPixelInBits) - 1;
	
	for(int8 iPass = 0; iPass < 7; ++iPass)
	{
		//////////////////////////////////////////////////////////////////////////////
		// Unfilter the pass-pixel-block
		int32 nLocalRowCount, nPixelBytesPerRow;
		GetPassSize(iPass, nLocalRowCount, nPixelBytesPerRow, nWidth, nHeight, m_nSizeofPixelInBits);

		if( !UnfilterBlock(pInterlaced + nSrcIndex, nLocalRowCount, nPixelBytesPerRow, m_nSizeofPixel))
		{
			m_nLastError = errBadFilterType;  // Unknown filtering type
			return false;
		}
		//////////////////////////////////////////////////////////////////////////////
		if( nLocalRowCount == 0 || nPixelBytesPerRow == 0)
		{
			// Empty pass
			continue;
		}

		// Unfilter the pass-pixel-block
		int32 nRow = aStartingRow[iPass];
		while(nRow < nHeight)
		{
			int32 nCol = aStartingCol[iPass];
			if( nCol >= nWidth)
			{
				// End of row reached
				break;
			}

			// Some data is available for this pass,
			const int32 nDstRowOffset = nRow * m_nByteWidth;
			do
			{
				ASSERT(nSrcIndex < nInterlacedSize);
				if( m_nSizeofPixelInBits > 8)
				{
					const int32 nDstOffset = nDstRowOffset + nCol * m_nSizeofPixel;
					for(int32 i = 0; i < m_nSizeofPixel; ++i)
					{
						pResult[nDstOffset + i] = pInterlaced[nSrcIndex];
						nSrcIndex++;
					}
				}
				else if( m_nSizeofPixelInBits == 8)
				{
					pResult[nDstRowOffset + nCol] = pInterlaced[nSrcIndex];
					nSrcIndex++;
				}
				else if( m_nSizeofPixelInBits < 8)
				{
					////////////////////////////////////
					// Extract the index value
					uint8 nByte = pInterlaced[nSrcIndex];
					nByte &= paMask[nSrcSubIndex];
					nByte >>= paShift[nSrcSubIndex];
					////////////////////////////////////
					
					////////////////////////////////////
					int32 nDstOffset = nDstRowOffset + (nCol >> nDstShift);
					int32 nDstSubIndex = nCol & nMaxSrcSubIndex;
					nByte <<= paShift[nDstSubIndex];

					pResult[nDstOffset] |= nByte;

					nSrcSubIndex++;

					if( nSrcSubIndex > nMaxSrcSubIndex)
					{
						// End of byte
						nSrcIndex++;
						nSrcSubIndex = 0;
					}
				}
				
				nCol += aColIncrement[iPass];
			}
			while(nCol < nWidth);

			if( nSrcSubIndex != 0)
			{
				// Padding
				nSrcIndex++;
				nSrcSubIndex = 0; // Reset the sub pixel index
			}

			nRow += aRowIncrement[iPass];
		}
		// Jump to the next pass bytes (only if pass not empy)
		// (nLocalRowCount == number of filtering bytes for this pass)
		// so we jump over the bytes which gave the sub-filtering info
		nSrcIndex += nLocalRowCount;
	}

	return true;
}

int32 Png::ComputeInterlacedSize(int32 nWidth, int32 nHeight, int32 nSizeofPixelInBits)
{
	int32 nTotalByteCount = 0;

	for(int8 iPass = 0; iPass < 7; ++iPass)
	{
		int32 nRowCount, nPixelBytesPerRow;
		GetPassSize(iPass, nRowCount, nPixelBytesPerRow, nWidth, nHeight, nSizeofPixelInBits);

		if( nPixelBytesPerRow > 0)
		{
			// + 1 for each row to count the filtering byte
			nTotalByteCount += nRowCount * (nPixelBytesPerRow + 1);
		}
	}
	return nTotalByteCount;
}

// Gets the size of a block for a given pass
void Png::GetPassSize(int8 iPass, int32& nRowCount, int32& nPixelBytesPerRow,
					   int32 nWidth, int32 nHeight, int32 nSizeofPixelInBits)
{
	ASSERT(0 <= iPass && iPass <= 6);

	static const int32 anAdders[8] =   { 7, 7, 3, 3, 1, 1, 0, 0 };
	static const int32 anDividers[8] = { 8, 8, 8, 4, 4, 2, 2, 1 };

	int32 nPixelsPerRow = (nWidth + anAdders[iPass + 1]) / anDividers[iPass + 1];
	int32 nPixelBitsPerRow = nPixelsPerRow * nSizeofPixelInBits;
	
	nPixelBytesPerRow = nPixelBitsPerRow / 8 + ((nPixelBitsPerRow % 8) > 0 ? 1 : 0);

	nRowCount = (nHeight + anAdders[iPass]) / anDividers[iPass];
}

// pBlock : intput/ouput
// input format : each row starts with a filter type byte
// Fxxxxxx
// Fxxxxxx
// Fxxxxxx
// output format : linear version of the picture, without filter type byte
// yyyyyy
// yyyyyy
// yyyyyy
// xxx
// nPixelBytesPerRow : number of bytes for one row of the picture, NOT counting the filter type byte
// nBytesPerPixel : rounded to 1 if the pixel depth is less than 8
bool Png::UnfilterBlock(uint8* pBlock, int32 nRowCount, int32 nPixelBytesPerRow, int32 nBytesPerPixel)
{
	if( nPixelBytesPerRow == 0)
		return true;

	// pRow points on the start of the filtered row
	// pUnfilteredRow points on the start of the unfiltered row
	// Then, for each row, pUnfilteredRow loses one byte in comparison with pRow

	uint8* pRow = pBlock;
	uint8* pUnfilteredRow = pBlock;

	// Create constant data for the loop
	const int32 nSizeofPixel = nBytesPerPixel;
	const int32 nByteWidth = nPixelBytesPerRow;

	for(int32 iRow = 0; iRow < nRowCount; ++iRow)
	{
		// Initialize the index (iByte) of the current byte in the row
		int32 iByte = 0;
		
		// Get the filtering method used for this row
		const uint8 nMethod = pRow[0];
				
		// Jump over the filtering method byte
		pRow++;
		
		switch(nMethod)
		{
		case 0:
			// Method "None"
			{
				int32 nCounter = nByteWidth;
				do
				{
					const uint8 nPixel = pRow[0];
					pRow++;
					
					pUnfilteredRow[0] = nPixel;
					pUnfilteredRow++;
					
				}
				while(--nCounter > 0);
			}
			break;
		case 1:
			// Method "Sub"
			for(;iByte < nSizeofPixel; iByte++)
			{
				pUnfilteredRow[0] = pRow[0];
				pUnfilteredRow++;
				pRow++;
			}
			for(;iByte < nByteWidth; iByte++)
			{
				uint8 nSub = pRow[0];
				uint8 nLeft = pUnfilteredRow[0 - nSizeofPixel];
				uint8 nRes = uint8(nSub + nLeft);
				pUnfilteredRow[0] = nRes;
				pUnfilteredRow++;
				pRow++;
			}
			break;
		case 2:
			// Method "Up"
			if( iRow == 0)
			{
				for(;iByte < nByteWidth; iByte++)
				{
					uint8 nUp = pRow[0];
					uint8 nAbove = 0;
					uint8 nRes = uint8(nUp + nAbove);
					pUnfilteredRow[0] = nRes;
					pUnfilteredRow++;
					pRow++;
				}
			}
			else
			{
				for(;iByte < nByteWidth; iByte++)
				{
					uint8 nUp = pRow[0];
					uint8 nAbove = pUnfilteredRow[0 - nByteWidth];
					uint8 nRes = uint8(nUp + nAbove);
					pUnfilteredRow[0] = nRes;
					pUnfilteredRow++;
					pRow++;
				}
			}
			break;
		case 3:
			// Method "Average"
			if( iRow == 0)
			{
				for(; iByte < nSizeofPixel; iByte++)
				{
					uint8 nAverage = pRow[0];
					uint8 nLeft = 0;
					uint8 nAbove = 0;
					uint8 nRes = uint8(nAverage + (nLeft + nAbove) / 2);
					pUnfilteredRow[0] = nRes;
					pUnfilteredRow++;
					pRow++;
				}
				for(;iByte < nByteWidth; iByte++)
				{
					uint8 nAverage = pRow[0];
					uint8 nLeft = pUnfilteredRow[0 - nSizeofPixel];
					uint8 nAbove = 0;
					uint8 nRes = uint8(nAverage + (nLeft + nAbove) / 2);
					pUnfilteredRow[0] = nRes;
					pUnfilteredRow++;
					pRow++;
				}
			}
			else
			{
				for(; iByte < nSizeofPixel; iByte++)
				{
					uint8 nAverage = pRow[0];
					uint8 nLeft = 0;
					uint8 nAbove = pUnfilteredRow[0 - nByteWidth];
					uint8 nRes = uint8(nAverage + (nLeft + nAbove) / 2);
					pUnfilteredRow[0] = nRes;
					pUnfilteredRow++;
					pRow++;
				}
				for(;iByte < nByteWidth; iByte++)
				{
					uint8 nAverage = pRow[0];
					uint8 nLeft = pUnfilteredRow[0 - nSizeofPixel];
					uint8 nAbove = pUnfilteredRow[0 - nByteWidth];
					uint8 nRes = uint8(nAverage + (nLeft + nAbove) / 2);
					pUnfilteredRow[0] = nRes;
					pUnfilteredRow++;
					pRow++;
				}
			}
			break;
		case 4:
			// Method "Paeth"
			if( iRow == 0)
			{
				for(;iByte < nSizeofPixel; iByte++)
				{
					uint8 nPaeth = pRow[0];
					uint8 nLeft = 0;
					uint8 nAbove = 0;
					uint8 nUpperLeft = 0;
					uint8 nRes = uint8(nPaeth + PaethPredictor(nLeft, nAbove, nUpperLeft));
					pUnfilteredRow[0] = nRes;
					pUnfilteredRow++;
					pRow++;
				}
				for(;iByte < nByteWidth; iByte++)
				{
					uint8 nPaeth = pRow[0];
					uint8 nLeft = pUnfilteredRow[0 - nSizeofPixel];
					uint8 nAbove = 0;
					uint8 nUpperLeft = 0;
					uint8 nRes = uint8(nPaeth + PaethPredictor(nLeft, nAbove, nUpperLeft));
					pUnfilteredRow[0] = nRes;
					pUnfilteredRow++;
					pRow++;
				}
			}
			else
			{
				for(;iByte < nSizeofPixel; iByte++)
				{
					uint8 nPaeth = pRow[0];
					uint8 nLeft = 0;
					uint8 nAbove = pUnfilteredRow[0 - nByteWidth];
					uint8 nUpperLeft = 0;
					uint8 nRes = uint8(nPaeth + PaethPredictor(nLeft, nAbove, nUpperLeft));
					pUnfilteredRow[0] = nRes;
					pUnfilteredRow++;
					pRow++;
				}
				for(;iByte < nByteWidth; iByte++)
				{
					uint8 nPaeth = pRow[0];
					uint8 nLeft = pUnfilteredRow[0 - nSizeofPixel];
					uint8 nAbove = pUnfilteredRow[0 - nByteWidth];
					uint8 nUpperLeft = pUnfilteredRow[0 - nSizeofPixel-nByteWidth];
					uint8 nRes = uint8(nPaeth + PaethPredictor(nLeft, nAbove, nUpperLeft));
					pUnfilteredRow[0] = nRes;
					pUnfilteredRow++;
					pRow++;
				}
			}
			break;
		default:
			return false;
		}
	}
	return true;
}

bool Png::Dump(const String& strFileName, const PngDumpSettings& ds)
{
	File file;
	if( !file.Open(strFileName, File::modeWrite))
		return false;

	return Dump(file, ds);
}

// pBlock points on a buffer which already has room for the sub-filtering byte info given
// at the beginning of each row.
void Png::FilterBlock(uint8* const pBlock, int32 nRowCount, int32 nPixelBytesPerRow, int32 nBytesPerPixel)
{
	// Add the byte size for the filtering type
	const int32 nTotalBytesPerRow = nPixelBytesPerRow + 1;
	///////////////////////////////////////////////////////////////////////////

	//const int32 nFilteredBufSize = nFilteredRowSize * nRowCount;

	// Allocate buffers for 5 possible filtered rows
	ByteArray rbFilteredRows;
	rbFilteredRows.SetSize(nPixelBytesPerRow * 5);

	uint8* const pFilteredRows = rbFilteredRows.GetPtr();

	uint8* const pDstNone = pFilteredRows;
	uint8* const pDstSub = pDstNone + nPixelBytesPerRow;
	uint8* const pDstUp = pDstSub + nPixelBytesPerRow;
	uint8* const pDstAverage = pDstUp + nPixelBytesPerRow;
	uint8* const pDstPaeth = pDstAverage + nPixelBytesPerRow;

	uint8* const apTmpRows[5] = {pDstNone, pDstSub, pDstUp, pDstAverage, pDstPaeth };
	
	////////////////////////////////////////////////////////////////////////
	// The filtering process must be bottom-up
	// TODO : this kills the CPU cache we should sometimes prefetch some data to speed-up the process

	uint8* pRow = pBlock + nTotalBytesPerRow * nRowCount - nTotalBytesPerRow;

	for(int32 iRow = nRowCount - 1; iRow >= 0; --iRow)
	{
		uint8* pMethod = pRow;

		// Jump over the method byte
		pRow++;

		// Filtering method used
		uint8 nMethod = 0;

		///////////////////////////////////////
		// Try all of the 5 sub-filter types //
		///////////////////////////////////////

		// Filtering = None (0)
		for(int32 iByte = 0; iByte < nPixelBytesPerRow; ++iByte)
		{
			const uint8 nByteCurrent = pRow[iByte];
			const uint8 nNone = nByteCurrent;
			
			pDstNone[iByte] = nNone;
		}

		// Try other types of filtering

		// Filtering = Sub (1)
		for(int32 iByte = 0; iByte < nPixelBytesPerRow; ++iByte)
		{
			uint8 nByteLeft;
			if( iByte < nBytesPerPixel)
			{
				nByteLeft = 0;
			}
			else
			{
				nByteLeft = pRow[iByte - nBytesPerPixel];
			}

			const uint8 nByteCurrent = pRow[iByte];
			const uint8 nSub = uint8(nByteCurrent - nByteLeft);
			
			pDstSub[iByte] = nSub;
		}

		// Filtering = Up (2)
		for(int32 iByte = 0; iByte < nPixelBytesPerRow; ++iByte)
		{
			uint8 nByteUp;
			if( iRow == 0)
			{
				nByteUp = 0;
			}
			else
			{
				nByteUp = pRow[-nTotalBytesPerRow + iByte];
			}

			const uint8 nByteCurrent = pRow[iByte];
			const uint8 nUp = uint8(nByteCurrent - nByteUp);
			
			pDstUp[iByte] = nUp;
		}

		// Filtering = Average (3)
		for(int32 iByte = 0; iByte < nPixelBytesPerRow; ++iByte)
		{
			uint8 nByteUp;
			if( iRow == 0)
			{
				nByteUp = 0;
			}
			else
			{
				nByteUp = pRow[-nTotalBytesPerRow + iByte];
			}

			uint8 nByteLeft;
			if( iByte < nBytesPerPixel)
			{
				nByteLeft = 0;
			}
			else
			{
				nByteLeft = pRow[iByte - nBytesPerPixel];
			}

			const uint8 nByteCurrent = pRow[iByte];
			const uint8 nAverage = uint8(nByteCurrent - (nByteLeft + nByteUp) / 2);
			
			pDstAverage[iByte] = nAverage;
		}

		// Filtering = Paeth (4)
		for(int32 iByte = 0; iByte < nPixelBytesPerRow; ++iByte)
		{
			uint8 nByteUp;
			if( iRow == 0)
			{
				nByteUp = 0;
			}
			else
			{
				nByteUp = pRow[-nTotalBytesPerRow + iByte];
			}

			uint8 nByteLeft;
			if( iByte < nBytesPerPixel)
			{
				nByteLeft = 0;
			}
			else
			{
				nByteLeft = pRow[iByte - nBytesPerPixel];
			}

			uint8 nByteUpLeft;
			if( iByte < nBytesPerPixel || iRow == 0)
			{
				nByteUpLeft = 0;
			}
			else
			{
				nByteUpLeft = pRow[-nTotalBytesPerRow + iByte - nBytesPerPixel];
			}

			const uint8 nByteCurrent = pRow[iByte];
			const uint8 nPaeth = uint8(nByteCurrent - PaethPredictor(nByteLeft, nByteUp, nByteUpLeft));
			
			pDstPaeth[iByte] = nPaeth;
		}

		//////////////////////////////////////////////////////////////////////////////////
		// Choose the best filtered row of the 5

		uint32 nSmallest = MAX_UINT32;
		nMethod = 0; // Index of the method to use. If all have a sum of 0, choose the method 0

		for(uint8 i = 0; i < 5; ++i)
		{
			uint32 nSum = 0;
			for(int32 iByte = 0; iByte < nPixelBytesPerRow; ++iByte)
			{
				uint8 nByte = apTmpRows[i][iByte];

				if( nByte >= 128)
				{
					nByte = uint8(256 - nByte);
				}
			
				nSum += nByte;
			}

			if( nSum < nSmallest)
			{
				nSmallest = nSum;
				nMethod = i;
			}
		}

		///////////////////////////////////////////////////////////////////////
		// Copy the wanted filtered row to the result buffer
		pMethod[0] = nMethod;
		Memory::Copy(pRow, apTmpRows[nMethod], nPixelBytesPerRow);

		pRow -= (1 + nTotalBytesPerRow);
	}
	////////////////////////////////////////////////////////////////////////
}

// pSrc : pointer to a linear buffer of pixels
void Png::InterlaceAndFilter(uint8* pDst, const uint8* pSrc,
							  const int32 nSrcWidth, const int32 nSrcHeight,
							  const int32 nSrcPixelBytesPerRow,
							  int32 nSizeofPixelInBits, bool bDoFiltering)
{
	static const int32 aStartingRow[7] =  { 0, 0, 4, 0, 2, 0, 1 };
	static const int32 aRowIncrement[7] = { 8, 8, 8, 4, 4, 2, 2 };

	static const int32 aStartingCol[7] =  { 0, 4, 0, 2, 0, 1, 0 };
	static const int32 aColIncrement[7] = { 8, 8, 4, 4, 2, 2, 1 };

	///////////////////////////////////////////////////////////////////////////
	// Masks arrays for pixel sizes < 8
	static const uint8 aMask4[2] = { 0xf0, 0x0f };
	static const uint8 aShift4[2] = { 4, 0 };

	static const uint8 aMask2[4] = { 0xc0, 0x30, 0x0c, 0x03 };
	static const uint8 aShift2[4] = { 6, 4, 2, 0 };

	static const uint8 aMask1[8] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };
	static const uint8 aShift1[8] = { 7, 6, 5, 4, 3, 2, 1, 0 };
	
	const uint8* paMask = null;
	const uint8* paShift = null;
	int32 nSrcShift = 0;

	if( nSizeofPixelInBits < 8)
	{
		if( nSizeofPixelInBits == 4)
		{
			paMask = aMask4;
			paShift = aShift4;
			nSrcShift = 1;
		}
		else if( nSizeofPixelInBits == 2)
		{
			paMask = aMask2;
			paShift = aShift2;
			nSrcShift = 2;
		}
		else if( nSizeofPixelInBits == 1)
		{
			paMask = aMask1;
			paShift = aShift1;
			nSrcShift = 3;
		}
	}

	///////////////////////////////////////////////////////////////////////////
	// The Png file is made of several sub-images, going more and more detailed

	const int32 nSizeofPixelInBytes = (nSizeofPixelInBits + 7 ) / 8;
	int32 nDstIndex = 0; // The byte index in the destination buffer
	
	// Sub index for pixel < 8 bits
	int32 nDstSubIndex = 0;
	int32 nMaxDstSubIndex = (8 / nSizeofPixelInBits) - 1;

	for(int8 iPass = 0; iPass < 7; ++iPass)
	{
		//////////////////////////////////////////////////////////////////////////////
		uint8* const pDstPass = pDst + nDstIndex;

		int32 nLocalRowCount, nLocalPixelBytesPerRow;
		GetPassSize(iPass, nLocalRowCount, nLocalPixelBytesPerRow, nSrcWidth, nSrcHeight, nSizeofPixelInBits);
	
		//////////////////////////////////////////////////////////////////////////////
		if( nLocalRowCount == 0 || nLocalPixelBytesPerRow == 0)
		{
			// Empty pass
			continue;
		}

		if( nSizeofPixelInBits < 8)
		{
			// As for small bit depth we perform binary OR to set the pixels, the destination buffer
			// must be previously filled with 0s
			Memory::Zero(pDstPass, (nLocalPixelBytesPerRow + 1) * nLocalRowCount);
		}

		// Prepare the pass-pixel-block by getting pixels around from the original buffer
		int32 nRow = aStartingRow[iPass];
		while(nRow < nSrcHeight)
		{
			int32 nCol = aStartingCol[iPass];
			if( nCol >= nSrcWidth)
			{
				// End of row reached
				break;
			}

			// Insert an empty byte for the sub-filtering value
			// (We will filter the whole pass buffer later)
			pDst[nDstIndex] = 0;
			nDstIndex++;

			const int32 nSrcRowOffset = nRow * nSrcPixelBytesPerRow;
			do
			{
				if( nSizeofPixelInBits > 8)
				{
					const int32 nSrcOffset = nSrcRowOffset + nCol * nSizeofPixelInBytes;
					for(int32 i = 0; i < nSizeofPixelInBytes; ++i)
					{
						pDst[nDstIndex] = pSrc[nSrcOffset + i];
						nDstIndex++;
					}
				}
				else if( nSizeofPixelInBits == 8)
				{
					pDst[nDstIndex] = pSrc[nSrcRowOffset + nCol];
					nDstIndex++;
				}
				else if( nSizeofPixelInBits < 8)
				{
					////////////////////////////////////////////////////
					// Extract the pixel from the source
					int32 nSrcOffset = nSrcRowOffset + (nCol >> nSrcShift);
					uint8 nByte = pSrc[nSrcOffset];
					int32 nSrcSubIndex = nCol & nMaxDstSubIndex;
					
					nByte &= paMask[nSrcSubIndex];
					nByte >>= paShift[nSrcSubIndex];
										
					// nByte now contains the source pixel value
					////////////////////////////////////////////////////

					nByte <<= paShift[nDstSubIndex];

					pDst[nDstIndex] |= nByte;

					nDstSubIndex++;

					if( nDstSubIndex > nMaxDstSubIndex)
					{
						// End of byte
						nDstIndex++;
						nDstSubIndex = 0;
					}
				}
				
				nCol += aColIncrement[iPass];
			}
			while(nCol < nSrcWidth);

			if( nDstSubIndex != 0)
			{
				// Padding
				nDstIndex++;
				nDstSubIndex = 0; // Reset the sub pixel index
			}

			nRow += aRowIncrement[iPass];
		}
		
		// End of the pass !
		if( bDoFiltering)
		{
			FilterBlock(pDstPass, nLocalRowCount, nLocalPixelBytesPerRow, nSizeofPixelInBytes);
		}
	}
}


bool Png::Dump(IFile& fileDst, const PngDumpSettings& ds)
{
	const uint8* pSrc = (const uint8*) ds.m_pBuffer;
	const int32 nWidth = ds.m_nWidth;
	const int32 nHeight = ds.m_nHeight;
	PixelFormat epf = ds.m_epf;
	const Palette* pPalette = ds.m_pPalette;

	uint8 nColorType;
	uint8 nBitsPerComponent;
	if( !GetIHDRFeaturesFromPixelFormat(epf, nColorType, nBitsPerComponent))
		return false;
	
	if( (nColorType & cfPalette) && pPalette == null)
		return false;

	const int32 nPixelBytesPerRow = ImageType::ComputeByteWidth(epf, nWidth);
	const int32 nSrcSize = nPixelBytesPerRow * nHeight;

	/////////////////////////////////////
	const int32 nBitsPerPixel = SizeofPixelInBits(epf);

	int nStrategy = ZStream::Z_DEFAULT_STRATEGY;
	if( ds.m_nZLibStrategy == PngDumpSettings::zlibStrategyGuess)
	{
		// Guess which strategy we should use
		if( ds.m_nFiltering != 0)
		{
			// Usually achieves better compression with filtered images
			if( nBitsPerPixel <= 8 )
			{
				// But for low depths only
				nStrategy = ZStream::Z_FILTERED;
			}
		}
	}
	else
	{
		if( ds.m_nZLibStrategy == PngDumpSettings::zlibStrategyFilter)
		{
			nStrategy = ZStream::Z_FILTERED;
		}
	}

	////////////////////////////////////////////////////////////
	ByteArray rbSwapped;
	if( epf == pf32bppBgra)
	{
		rbSwapped.SetSize(nSrcSize);
		uint8* pNewBuffer = rbSwapped.GetPtr();

		// Should swap bytes as PNG does not support BGRA8
		// We convert the pixels into RGBA8 format
		const int32 nPixelCount = nSrcSize / 4;
		for(int32 iCopy = 0; iCopy < nPixelCount; ++iCopy)
		{
			uint8 b = pSrc[4 * iCopy + 0];
			uint8 g = pSrc[4 * iCopy + 1];
			uint8 r = pSrc[4 * iCopy + 2];
			uint8 a = pSrc[4 * iCopy + 3];
			
			pNewBuffer[4 * iCopy + 0] = r;
			pNewBuffer[4 * iCopy + 1] = g;
			pNewBuffer[4 * iCopy + 2] = b;
			pNewBuffer[4 * iCopy + 3] = a;
		}

		pSrc = pNewBuffer;
		epf = pf32bppRgba;
	}
	////////////////////////////////////////////////////////////

	const int32 nBytesPerPixel = (nBitsPerPixel + 7 ) / 8;

	// If special sub-filtering is not requested do not try other methods
	bool bDoFiltering = false;
	if( ds.m_nFiltering != 0)
	{
		bDoFiltering = true;
	}
	
	// The buffer which will be given to the zlib routines
	ByteArray abBufferToCompress;

	int32 nBufferToCompressSize = 0;
	uint8* pBufferToCompress = null;

	if( ds.m_bInterlace)
	{
		nBufferToCompressSize = ComputeInterlacedSize(nWidth, nHeight, nBitsPerPixel);
		if( !abBufferToCompress.SetSize(nBufferToCompressSize))
		{
			// Not enough memory
			return false;
		}

		pBufferToCompress = abBufferToCompress.GetPtr();
		InterlaceAndFilter(pBufferToCompress, pSrc, nWidth, nHeight, nPixelBytesPerRow, nBitsPerPixel, bDoFiltering);
	}
	else
	{
		/////////////////////////////////////////////////////////////
		// Create a new buffer from the image with a blank byte at each
		// row beginning (the sub-filtering method)

		nBufferToCompressSize = nSrcSize + nHeight;
		if( !abBufferToCompress.SetSize(nBufferToCompressSize))
		{
			// Not enough memory
			return false;
		}
		pBufferToCompress = abBufferToCompress.GetPtr();

		uint8* pCurNewByte = pBufferToCompress;
		const uint8* pCurSrcByte = pSrc;

		for(int32 iRow = 0; iRow < nHeight; ++iRow)
		{
			pCurNewByte[0] = 0;
			pCurNewByte++;

			for(int32 iByte = 0; iByte < nPixelBytesPerRow; ++iByte)
			{
				pCurNewByte[0] = pCurSrcByte[0];
				pCurSrcByte++;
				pCurNewByte++;
			}
		}

		////////////////////////////////////////////////////////////////
		// Apply filtering
		if( bDoFiltering)
		{
			FilterBlock(pBufferToCompress, nHeight, nPixelBytesPerRow, nBytesPerPixel);
		}
	}

	// ZLib documentation about compress() :
	// Upon entry, destLen is the total size of the destination buffer,
	// which must be at least 0.1% larger than sourceLen plus 12 bytes.
	uint32 nCompressedBufferSize = 12 + (nBufferToCompressSize * 3) / 2;
	
	ByteArray rbCompressedBuffer;
	rbCompressedBuffer.SetSize(nCompressedBufferSize);
	uint8* const pCompressedBuffer = rbCompressedBuffer.GetPtr();

	const int32 nCompressionLevel = ds.m_nZLibCompressionLevel;

	int nMaxWindowBits = 15;
	int nMemLevel = 8;

	// Sometimes it improves compression to use settings for low memory platforms,
	// that's why we have this option
	if( ds.m_nZLibWindowBitsAndMem == PngDumpSettings::zlibWindowBitsAndMemLow )
	{
		nMaxWindowBits = 14;
		nMemLevel = 6;
	}

	int32 nRet = Compress(pCompressedBuffer, &nCompressedBufferSize,
						pBufferToCompress, nBufferToCompressSize,
						nCompressionLevel, nStrategy, nMaxWindowBits, nMemLevel);

	// Not needed anymore
	abBufferToCompress.RemoveAll();

	if( nRet != 0)
	{
		// Error in compression
		return false;
	}
	/////////////////////////////////////////////////////////////////////////
	// Send the data to the stream
	
	// Start to send the png signature
	if( fileDst.Write(k_PngSignature, 8) != 8)
		return false;

	// First chunk : the header	
	ChunkedFile file(fileDst);
	
	// 0 = deflate/inflate
	uint8 nCompressionMethod = 0;
	
	// 0 = Adaptative
	uint8 nFilterMethod = 0;
	
	// 1 = Adam7
	uint8 nInterlaceMethod = ds.m_bInterlace ? uint8(1) : uint8(0);
	
	file.BeginChunkWrite(chunk_IHDR);
	
	HeaderInfo hi;
	hi.nWidth = nWidth;
	hi.nHeight = nHeight;
	hi.nBitDepth = nBitsPerComponent;
	hi.nColorType = nColorType;
	hi.nCompressionMethod = nCompressionMethod;
	hi.nFilterMethod = nFilterMethod;
	hi.nInterlaceMethod = nInterlaceMethod;
	
	if( file.ShouldSwapBytes())
	{
		hi.SwapBytes();
	}
	
	const int32 nHeaderSize = 13;
	if( file.Write(&hi, nHeaderSize) != nHeaderSize)
	{
		return false;
	}

	if( !file.EndChunkWrite())
	{
		return false;
	}

	if( nColorType & cfPalette)
	{
		// Palettized picture
		const int32 nColorCount = pPalette->m_nCount;

		if( !(0 <= nColorCount && nColorCount <= 256))
		{
			// Bad number of colors
			return false;
		}

		/////////////////////////////////////////////////////////////////////////
		// Write the palette colors
		file.BeginChunkWrite(chunk_PLTE);
		
		uint8 aTmpPalette[256 * 3];
		int32 iTmpColor = 0;
		for(int32 i = 0; i < nColorCount; ++i)
		{
			uint8 r, g, b, a;
			pPalette->m_aColors[i].ToRgba(r, g, b, a);

			aTmpPalette[iTmpColor + 0] = r;
			aTmpPalette[iTmpColor + 1] = g;
			aTmpPalette[iTmpColor + 2] = b;
			iTmpColor += 3;
		}

		file.Write( aTmpPalette, nColorCount * 3);

		if( !file.EndChunkWrite())
			return false;
	
		/////////////////////////////////////////////////////////////////////////
		// Check if we need to write a tRNS chunk
		// We don't if all alphas are 255
		// Note : an external png optimizer should sort the palette colors by their
		// alpha values so the lowest alpha is found first
		
		int32 nAlphaCount = 0;
		for(int32 i = nColorCount - 1; i >= 0; --i)
		{
			if( pPalette->m_aColors[i].GetAlpha() != 255)
			{
				nAlphaCount = i + 1;
				break;
			}
		}

		if( nAlphaCount > 0)
		{
			// tRNS chunk needed
			uint8 aAlphas[256];
			for(int32 i = 0; i < nAlphaCount; ++i)
			{
				aAlphas[i] = pPalette->m_aColors[i].GetAlpha();
			}

			file.BeginChunkWrite(chunk_tRNS);
			
			file.Write(aAlphas, nAlphaCount);

			if( !file.EndChunkWrite())
				return false;
		}
	}
	else if( nColorType == 0)
	{
		// Grey
		if( ds.m_bUseTransparentColor)
		{
			file.BeginChunkWrite(chunk_tRNS);
			file.Write16(ds.m_nTransGrey);
			
			if( !file.EndChunkWrite())
				return false;
		}
	}
	else if( nColorType == 2)
	{
		// True colors
		if( ds.m_bUseTransparentColor)
		{
			file.BeginChunkWrite(chunk_tRNS);
			file.Write16(ds.m_nTransRed);
			file.Write16(ds.m_nTransGreen);
			file.Write16(ds.m_nTransBlue);

			if( !file.EndChunkWrite())
				return false;
		}
	}

	////////////////////////////////////////////////////
	// bKGD chunk
	if( ds.m_bUseBackgroundColor)
	{
		if( nColorType == 0x03)
		{
			// Indexed color
			file.BeginChunkWrite(chunk_bKGD);
			file.Write8(ds.m_nBkIndex);
			if( !file.EndChunkWrite())
				return false;
		}
		else if( nColorType == 0x00 || nColorType == 0x04)
		{
			// Grey or Grey+Alpha
			file.BeginChunkWrite(chunk_bKGD);
			file.Write16(ds.m_nBkGrey);
			if( !file.EndChunkWrite())
				return false;
		}
		else if( nColorType == 0x02 || nColorType == 0x06)
		{
			// TrueColor or TrueColor+Alpha
			file.BeginChunkWrite(chunk_bKGD);
			file.Write16(ds.m_nBkRed);
			file.Write16(ds.m_nBkGreen);
			file.Write16(ds.m_nBkBlue);
			if( !file.EndChunkWrite())
				return false;
		}
	}
	////////////////////////////////////////////////////

	// Next chunk : image compressed data
	file.BeginChunkWrite(chunk_IDAT);
	
	file.Write(pCompressedBuffer, nCompressedBufferSize);
	
	if( !file.EndChunkWrite())
		return false;
	
	
	// Last chunk : end chunk
	
	file.BeginChunkWrite(chunk_IEND);
	if( !file.EndChunkWrite())
		return false;
	
	return true;
}

int Png::Compress(uint8* pDest, uint32* pDestLen, const uint8* pSource, uint32 nSourceLen, 
				  int nLevel, int nStrategy, int nWindowBits, int nMemLevel)
{
	ASSERT(nStrategy == ZStream::Z_DEFAULT_STRATEGY || nStrategy == ZStream::Z_FILTERED);

	ZStream stream;
	stream.SetBuffers(pSource, nSourceLen, pDest, *pDestLen);
	
	int err;
	
	uint8 nMethod = ZStream::Z_DEFLATED;

	err = stream.DeflateInit2(nLevel, nMethod, nWindowBits, nMemLevel, nStrategy);
	if( err != ZStream::Z_OK)
		return err;
	
	err = stream.Deflate(ZStream::Z_FINISH);
	if( err != ZStream::Z_STREAM_END)
	{
		stream.DeflateEnd();
		return err == ZStream::Z_OK ? ZStream::Z_BUF_ERROR : err;
	}
	*pDestLen = stream.GetOutTotalRead();
	
	err = stream.DeflateEnd();
	return err;
}

String Png::GetLastErrorString() const
{
	switch(m_nLastError)
	{
	case Png::errNotAPngFile:
		return L"Not a PNG file";

	case Png::errBadHeaderSize:
		return L"Invalid header size";
		
	case Png::errBadColorType:
		return L"Invalid or not supported color format";

	case Png::errBadFilterMethod:
		return L"Invalid or unsupported filtering method";
		
	case Png::errBadFilterType:
		return L"Bad filtering type";

	case Png::errInflateErr:
		return L"Decompression error (zlib : " + String::FromAsciiSZ(m_ZStream.GetLastError()) + L")";

	case Png::errBadPicSize:
		return L"Invalid image size";
	
	case Png::errNoIENDFound:
		return L"No IEND chunk found at the end of the file data (file corrupted)";

	case Png::errBadInterlaceMethod:
		return L"Unsupported interlacing method (Adam7 only)";

	case Png::errUnexpectedEndOfFileInIDAT:
		return L"Unexpected end of file reached while reading IDAT chunk";
	
	case Png::errNotEnoughDataInGammaChunk:
		return L"Not enough data in gAMA chunk";

	case Png::errNotEnoughDataInBkChunk:
		return L"Not enough data in bKGD chunk";
		////////////////////////////////////////////////////////
	}

	return ImageType::GetLastErrorString();
}
