///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#ifndef CHUSTD_PNG_H
#define CHUSTD_PNG_H

#include "ImageType.h"
#include "ChunkedFile.h"

#include "ZStream.h"

namespace chustd {

class PngDumpSettings
{
public:
	const void*	m_pBuffer;
	const Palette*	m_pPalette;
	int32		m_nWidth;
	int32		m_nHeight;
	PixelFormat	m_epf;

	enum Onk
	{ 
		zlibStrategyGuess = 0x00,   // Default
		zlibStrategyDefault = 0x01,
		zlibStrategyFilter = 0x02,
	
		zlibWindowBitsAndMemHigh = 0x00, // Default
		zlibWindowBitsAndMemLow = 0x10   // Can sometimes improve compression
	};

	uint8       m_nZLibCompressionLevel; // [1..9], default = 6
	uint8       m_nZLibStrategy;
	uint8       m_nZLibWindowBitsAndMem;
	
	bool        m_bUseTransparentColor; // Set to true to use the following members
	uint16      m_nTransGrey;
	uint16      m_nTransRed;
	uint16      m_nTransGreen;
	uint16      m_nTransBlue;

	bool        m_bUseBackgroundColor;
	uint8       m_nBkIndex;
	uint16      m_nBkGrey;
	uint16      m_nBkRed;
	uint16      m_nBkGreen;
	uint16      m_nBkBlue;
	
	// 0 = Adaptative filtering with "None" set for each scanline
	// 1 = Adaptative filtering with various sub-method set for each scanline
	uint8       m_nFiltering;    // Default = 1
	bool        m_bInterlace;    // true to save an interlaced png

	PngDumpSettings()
	{
		m_pBuffer = null;
		m_pPalette = null;
		m_nWidth = 0;
		m_nHeight = 0;
		m_epf = pfUnknown;
		m_nZLibCompressionLevel = 6;
		m_nZLibStrategy = zlibStrategyGuess;
		m_nZLibWindowBitsAndMem = zlibWindowBitsAndMemHigh;
		
		m_bUseTransparentColor = false;
		m_nTransGrey = 0;
		m_nTransRed = 0;
		m_nTransGreen = 0;
		m_nTransBlue = 0;

		m_bUseBackgroundColor = false;
		m_nBkIndex = 0;
		m_nBkGrey = 0;
		m_nBkRed = 0;
		m_nBkGreen = 0;
		m_nBkBlue = 0;
		
		m_nFiltering = 1;
		m_bInterlace = false;
	}
};

//////////////////////////////////////////////////////////////////////////////////////
// This class is used to read and write PNG pictures
class Png : public ImageType
{
public:
	// Possible errors for m_nLastError
	enum
	{
		errNotAPngFile = firstErrEnumDerived, // Not a png file (mismatch in header)
		errBadHeaderSize,      // Invalid header size
		errBadColorType,       // Invalid or not supported color format
		errBadFilterMethod,    // Invalid or unsupported filtering method
		errBadFilterType,      // Bad filtering type
		errInflateErr,         // Decompression error (zlib)
		errBadInterlaceMethod, // Unsupported interlacing method (Adam7 only)
		errBadPicSize,         // Invalid image size
		errNoIENDFound,        // No IEND chunk found at the end of the stream
		errUnexpectedEndOfFileInIDAT,
		errNotEnoughDataInGammaChunk,
		errNotEnoughDataInBkChunk
	};

	enum
	{
		chunk_IHDR = 'IHDR',
		chunk_PLTE = 'PLTE',
		chunk_IDAT = 'IDAT',
		chunk_IEND = 'IEND',
		chunk_gAMA = 'gAMA',
		chunk_tRNS = 'tRNS',
		chunk_bKGD = 'bKGD'
	};

public:
	/////////////////////////////////////////////////
	virtual bool LoadFromFile(IFile& file);
	virtual PixelFormat GetPixelFormat() const;
	virtual const Palette* GetPalette() const;
	/////////////////////////////////////////////////

	virtual bool IsSimpleTransparencyUsed() const { return m_bChunkFound_tRNS; }
	virtual uint16 GetGreyTransIndex() const { return m_nTransGrey; }
	virtual void GetTransIndexes(uint16& nRed, uint16& nGreen, uint16& nBlue) const
	{ nRed = m_nTransRed; nGreen = m_nTransGreen; nBlue = m_nTransBlue; }

	virtual String GetLastErrorString() const;

	/////////////////////////////////////////////////////////////////////////////////////
	static bool Dump(const String& strFileName, const PngDumpSettings& ds);
	static bool Dump(IFile& file, const PngDumpSettings& ds);

	/////////////////////////////////////////////////////////////////////////////////////
	bool	HasBackgroundColor() const { return m_bChunkFound_bKGD; }
	uint8	GetBkIndex() const { return m_nBkIndex; }
	uint16	GetBkGrey() const { return m_nBkGrey; }
	uint16	GetBkRed() const { return m_nBkRed; }
	uint16	GetBkGreen() const { return m_nBkGreen; }
	uint16	GetBkBlue() const { return m_nBkBlue; }

	bool	HasGamma() const { return m_bChunkFound_gAMA; }
	uint32	GetGamma() const { return m_nGamma; }

	virtual bool IsInterlaced() const { return m_IHDR.nInterlaceMethod != 0; }

	static bool IsPng(IFile& file);
	static bool WriteSignature(IFile& file);

	Png();
	virtual ~Png();
	

protected:
	// Header strucutre
	struct HeaderInfo
	{
		uint32 nWidth;
		uint32 nHeight;
		uint8  nBitDepth;   // Bit count for each RGB component or for each index
		uint8  nColorType;  // Color flags:
	                       // default: grey, flags: 0x1=palette used, 0x2=color used, 0x4=alpha used
		uint8  nCompressionMethod; // 0=deflate/inflate (zlib)
		uint8  nFilterMethod;      // 0=Adaptatif: 5 types=(none,sub,up,average,paeth)

		uint8  nInterlaceMethod;   // 0=no interlace, 1=Adam7
		uint8  nDummy1, nDummy2,nDummy3; // For 32 bits alignement

		void SwapBytes();
	};
	
	// nColorType flags
	enum { cfGrey = 0, cfPalette = 1, cfColor = 2, cfAlpha = 4 };
	//////////////////////////////////////////////////////////
	HeaderInfo m_IHDR;
	
	Palette m_PLTE;  // Palette information

	int32 m_nUncompressedDataSize;	// Buffer size

	int32 m_nSizeofPixel;		// Pixel size in bytes; if nbbpp < 8 => 1 byte
	int32 m_nSizeofPixelInBits;	// Pixel size in bits
	
	int32 m_nByteWidth;			// Size of one row, not counting the filter byte

	/////////////////////////////////////////////////////////////////////////////////
	bool	m_bChunkFound_tRNS; // True if m_nTransGrey, Red, Green or Blue is present
	uint16	m_nTransGrey;   // Transparent grey level
	uint16	m_nTransRed;
	uint16	m_nTransGreen;
	uint16	m_nTransBlue;
	/////////////////////////////////////////////////////////////////////////////////

	/////////////////////////////////////////////////////////////////////////////////
	bool	m_bChunkFound_gAMA;      // true if gamma is activated
	uint32	m_nGamma;       // Gamma * 100000
	/////////////////////////////////////////////////////////////////////////////////

	/////////////////////////////////////////////////////////////////////////////////
	bool	m_bChunkFound_bKGD;
	uint8	m_nBkIndex;
	uint16	m_nBkGrey;
	uint16	m_nBkRed;
	uint16	m_nBkGreen;
	uint16	m_nBkBlue;
	/////////////////////////////////////////////////////////////////////////////////

	bool      m_bFirstIdat;
	ZStream   m_ZStream;
	uint32    m_nOutputOffset;
	ByteArray m_CompressedBuffer;

protected:
	void Initialize();

	static inline uint8 PaethPredictor(uint8 a, uint8 b, uint8 c);

	bool Handle_IDAT(IFile& file, uint32 nDataSizeof);
	bool Handle_PLTE(IFile& file, uint32 nDataSizeof);
	bool Handle_IHDR(IFile& file, uint32 nDataSizeof);
	bool Handle_gAMA(IFile& file, uint32 nDataSizeof);
	bool Handle_tRNS(IFile& file, uint32 nDataSizeof);
	bool Handle_bKGD(IFile& file, uint32 nDataSizeof);

	bool AllocateImageBuffer(bool bInterlacedBuffer);
	bool TerminateJob();
	
	bool UnfilterAndUninterlace();
	
	static bool UnfilterBlock(uint8* pBlock, int32 nRowCount, int32 nPixelBytesPerRow, int32 nBytesPerPixel);
	
	static void GetPassSize(int8 iPass, int32& nRowCount, int32& nPixelBytesPerRow,
		int32 nWidth, int32 nHeight, int32 nSizeofPixelInBits);
	
	static int32 ComputeInterlacedSize(int32 nWidth, int32 nHeight, int32 nSizeofPixelInBits);
	static bool GetIHDRFeaturesFromPixelFormat(PixelFormat epf, uint8& nColorType, uint8& nBitDepthPerComponent);
	static void FilterBlock(uint8* const pBlock, int32 nRowCount, int32 nPixelBytesPerRow, int32 nBytesPerPixel);

	static void InterlaceAndFilter(uint8* pDst, const uint8* pSrc,
		const int32 nSrcWidth, const int32 nSrcHeight, const int32 nSrcPixelBytesPerRow,
		int32 nSizeofPixelInBits, bool bDoFiltering);

	// ZLib compression tuned for PNG
	static int Compress(uint8* pDest, uint32* pDestLen, const uint8* pSource, uint32 nSourceLen, 
		int nLevel, int nStrategy, int nWindowBits, int nMemLevel);
};

} // namespace chustd;

#endif // ndef CHUSTD_PNG_H
