///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "Master.h"
#include "ReadOnlySurface32.h"
#include "ImageType.h"
#include "Png.h"

//////////////////////////////////////////////////////////////////////
using namespace chustd;
//////////////////////////////////////////////////////////////////////
Color32 Color32::black(0, 0, 0);
Color32 Color32::white(255, 255, 255);
Color32 Color32::red(255, 0, 0);
Color32 Color32::green(0, 255, 0);
Color32 Color32::blue(0, 0, 255);
Color32 Color32::yellow(255, 255, 0);
Color32 Color32::cyan(0, 255, 255);
Color32 Color32::magenta(255, 0, 255);
//////////////////////////////////////////////////////////////////////

bool ReadOnlySurface32::Create(int32 nWidth, int32 nHeight)
{
	return super::Create(nWidth, nHeight, nWidth, 32);
}

void ReadOnlySurface32::Create(const ReadOnlySurface32& surfSource, const Rectangle& rect)
{
	SetBufferHelper(surfSource, rect);
}

bool ReadOnlySurface32::Create(const ImageType& it)
{
	bool bOk = Create(it.GetWidth(), it.GetHeight());
	if( !bOk)
	{
		return false;
	}

	return UpdatePixels(it);
}

bool ReadOnlySurface32::UpdatePixels(const ImageType& it)
{
	const int32 nWidth = GetWidth();
	const int32 nHeight = GetHeight();
	const int32 nPixelCount = nWidth * nHeight;

	if( it.GetWidth() != nWidth || it.GetHeight() != nHeight )
	{
		// Cannot update, not the same rect
		return false;
	}

	uint8* pSrc = (uint8*) it.GetBuffer();
	if( pSrc == null)
	{
		return false;
	}

	PixelFormat pfSrc = it.GetPixelFormat();

	uint8* pDst = (uint8*) GetBuffer();
	PixelFormat pfDst = Color32::GetInternalFormat();

	ASSERT(pfDst == pf32bppRgba || pfDst == pf32bppBgra);

	uint8 iR, iG, iB, iA;
	if( pfDst == pf32bppBgra)
	{
		iB = 0;
		iG = 1;
		iR = 2;
		iA = 3;
	}
	else
	{
		// RR GG BB AA
		iR = 0;
		iG = 1;
		iB = 2;
		iA = 3;
	}

	bool bSimpleTransparencyUsed = it.IsSimpleTransparencyUsed();
	
	if( pfSrc == pf24bppBgr)
	{
		// TGA stores true color images this way
		for(int32 i = nPixelCount - 1; i >= 0; --i)
		{
			pDst[iR] = pSrc[2]; // R
			pDst[iG] = pSrc[1]; // G
			pDst[iB] = pSrc[0]; // B
			pDst[iA] = 255;
			
			pSrc += 3;
			pDst += 4;
		}
	}
	else if( pfSrc == pf24bppRgb)
	{
		if( bSimpleTransparencyUsed)
		{
			// One RGB color is 100% transparent
			uint16 nRed, nGreen, nBlue;
			it.GetTransIndexes(nRed, nGreen, nBlue);

			if( pfDst == pf32bppBgra)
			{
				for(int32 i = nPixelCount - 1; i >= 0; --i)
				{
					pDst[2] = pSrc[0]; // R
					pDst[1] = pSrc[1]; // G
					pDst[0] = pSrc[2]; // B
					
					pDst[3] = ( pSrc[0] == nRed && pSrc[1] == nGreen && pSrc[2] == nBlue) ? uint8(0) : uint8(255);
					
					pSrc += 3;
					pDst += 4;
				}
			}
			else
			{
				// pf32bppRgba
				for(int32 i = nPixelCount - 1; i >= 0; --i)
				{
					pDst[0] = pSrc[0]; // R
					pDst[1] = pSrc[1]; // G
					pDst[2] = pSrc[2]; // B
					
					pDst[3] = ( pSrc[0] == nRed && pSrc[1] == nGreen && pSrc[2] == nBlue) ? uint8(0) : uint8(255);
					
					pSrc += 3;
					pDst += 4;
				}
			}
		}
		else
		{
			// No simple transparency
			if( pfDst == pf32bppBgra)
			{
				for(int32 i = nPixelCount - 1; i >= 0; --i)
				{
					pDst[2] = pSrc[0]; // R
					pDst[1] = pSrc[1]; // G
					pDst[0] = pSrc[2]; // B
					pDst[3] = 255;
					
					pSrc += 3;
					pDst += 4;
				}
			}
			else
			{
				// pfRgba8
				for(int32 i = nPixelCount - 1; i >= 0; --i)
				{
					pDst[0] = pSrc[0]; // R
					pDst[1] = pSrc[1]; // G
					pDst[2] = pSrc[2]; // B
					pDst[3] = 255;
					
					pSrc += 3;
					pDst += 4;
				}
			}
		}
	}
	else if( pfSrc == pf48bppRgb)
	{
		// Conversion rules :
		// Add 1 if the low byte is greater to 127
		// aa00 --> aa
		// aa7f --> aa
		// aa80 --> ab
		// aaff --> ab

		// Saturate if the upper byte is 255
		// ff7f --> ff
		// ff80 --> ff
		// ffff --> ff

		if( bSimpleTransparencyUsed)
		{
			// One RGB color is 100% transparent
			uint16 nRed, nGreen, nBlue;
			it.GetTransIndexes(nRed, nGreen, nBlue);

			for(int32 i = 0; i < nPixelCount; i++)
			{
				uint32 r = uint16( (pSrc[0] << 8) | pSrc[1]);
				uint32 g = uint16( (pSrc[2] << 8) | pSrc[3]);
				uint32 b = uint16( (pSrc[4] << 8) | pSrc[5]);
				pSrc += 6;

				uint8 a = (r == nRed && g == nGreen && b == nBlue) ? uint8(0) : uint8(255);

				r = (r >> 8) + ((r >> 7) & 1);
				r -= (r >> 8);

				g = (g >> 8) + ((g >> 7) & 1);
				g -= (g >> 8);

				b = (b >> 8) + ((b >> 7) & 1);
				b -= (b >> 8);

				pDst[iR] = uint8(r);
				pDst[iG] = uint8(g);
				pDst[iB] = uint8(b);
				pDst[iA] = uint8(a);
				pDst+= 4;
			}
		}
		else
		{
			for(int32 i = 0; i < nPixelCount; i++)
			{
				uint32 r = uint16( (pSrc[0] << 8) | pSrc[1]);
				uint32 g = uint16( (pSrc[2] << 8) | pSrc[3]);
				uint32 b = uint16( (pSrc[4] << 8) | pSrc[5]);
				pSrc += 6;

				r = (r >> 8) + ((r >> 7) & 1);
				r -= (r >> 8);

				g = (g >> 8) + ((g >> 7) & 1);
				g -= (g >> 8);

				b = (b >> 8) + ((b >> 7) & 1);
				b -= (b >> 8);

				pDst[iR] = uint8(r);
				pDst[iG] = uint8(g);
				pDst[iB] = uint8(b);
				pDst[iA] = 255;
				pDst += 4;
			}
		}
	}
	else if( pfSrc == pf32bppBgra)
	{
		for(int32 i = 0; i < nPixelCount; i++)
		{
			const uint8 r = pSrc[2];
			const uint8 g = pSrc[1];
			const uint8 b = pSrc[0];
			const uint8 a = pSrc[3];
			pSrc += 4;

			pDst[iR] = r;
			pDst[iG] = g;
			pDst[iB] = b;
			pDst[iA] = a;
			pDst+= 4;
		}
	}
	else if( pfSrc == pf32bppRgba)
	{
		for(int32 i = 0; i < nPixelCount; i++)
		{
			const uint8 r = pSrc[0];
			const uint8 g = pSrc[1];
			const uint8 b = pSrc[2];
			const uint8 a = pSrc[3];
			pSrc += 4;

			pDst[iR] = r;
			pDst[iG] = g;
			pDst[iB] = b;
			pDst[iA] = a;
			pDst+= 4;
		}
	}
	else if( pfSrc == pf64bppRgba)
	{
		// Conversion rules :
		// Add 1 if the low byte is greater to 127
		// aa00 --> aa
		// aa7f --> aa
		// aa80 --> ab
		// aaff --> ab

		// Saturate if the upper byte is 255
		// ff7f --> ff
		// ff80 --> ff
		// ffff --> ff

		for(int32 i = 0; i < nPixelCount; i++)
		{
			uint32 r = uint16( (pSrc[0] << 8) | pSrc[1]);
			uint32 g = uint16( (pSrc[2] << 8) | pSrc[3]);
			uint32 b = uint16( (pSrc[4] << 8) | pSrc[5]);
			uint32 a = uint16( (pSrc[6] << 8) | pSrc[7]);
			pSrc += 8;

			r = (r >> 8) + ((r >> 7) & 1);
			r -= (r >> 8);

			g = (g >> 8) + ((g >> 7) & 1);
			g -= (g >> 8);

			b = (b >> 8) + ((b >> 7) & 1);
			b -= (b >> 8);

			a = (a >> 8) + ((a >> 7) & 1);
			a -= (a >> 8);

			pDst[iR] = uint8(r);
			pDst[iG] = uint8(g);
			pDst[iB] = uint8(b);
			pDst[iA] = uint8(a);
			pDst += 4;
		}
	}
	else if( pfSrc == pf8bppIndexed)
	{
		const Palette* pPalette = it.GetPalette();
		if( pPalette == null)
		{
			// Wanted a palette
			return false;
		}

		const uint32* pColors = (const uint32*) &pPalette->m_aColors[0];
		
		// Colors are stored in the same way than Surface32, so we directly copy
		// the colors into the destination buffer
		uint32* pDst32 = (uint32*) pDst;
		for(int32 i = 0; i < nPixelCount; ++i)
		{
			pDst32[i] = pColors[ pSrc[i] ];
		}
		///////////////////////////////////////////////////////////////
	}
	else if( pfSrc == pf8bppGrayScale)
	{
		//////////////////////////////////////////////////
		// Set the alpha values
		uint8 aAlphas[256];
		int32 nTransparentIndex = -1;
		if( it.IsSimpleTransparencyUsed())
		{
			nTransparentIndex = it.GetGreyTransIndex();
		}

		for(int32 i = 0; i < 256; ++i)
		{
			if( i == nTransparentIndex)
			{
				aAlphas[i] = 0;
			}
			else
			{
				aAlphas[i] = 255;
			}
		}
		//////////////////////////////////////////////////

		for(int32 i = 0; i < nPixelCount; i++)
		{
			const uint8 nIndex = pSrc[i];
			
			pDst[iR] = nIndex;
			pDst[iG] = nIndex;
			pDst[iB] = nIndex;
			pDst[iA] = aAlphas[nIndex];
			pDst+= 4;
		}
	}
	else if( pfSrc == pf16bppRgb555)
	{
		// Conversion rules :
		// Reuse the 3 most significant bits for the lower part of the 8 bit
		// 00000 --> 00000000
		// 00011 --> 00011000
		// 00111 --> 00111001
		// 01111 --> 01111011
		// 11111 --> 11111111
		
		for(int32 i = 0; i < nPixelCount; i++)
		{
			uint16 nPixel = uint16( (pSrc[0] << 8) | pSrc[1]);
			pSrc += 2;

			uint8 r = uint8( (nPixel & 0x7c00) >> 10 );
			uint8 g = uint8( (nPixel & 0x03e0) >> 5 );
			uint8 b = uint8( (nPixel & 0x001f) );

			r = uint8( (r << 3) + ((r >> 2) & 0x07));
			g = uint8( (g << 3) + ((g >> 2) & 0x07));
			b = uint8( (b << 3) + ((b >> 2) & 0x07));

			pDst[iR] = uint8(r);
			pDst[iG] = uint8(g);
			pDst[iB] = uint8(b);
			pDst[iA] = uint8(255);
			pDst += 4;
		}
	}
	else if( pfSrc == pf16bppRgb565)
	{
		// Conversion rules :
		// Reuse the 3 most significant bits for the lower part of the 8 bit
		// 00000 --> 00000000
		// 00011 --> 00011000
		// 00111 --> 00111001
		// 01111 --> 01111011
		// 11111 --> 11111111
		
		for(int32 i = 0; i < nPixelCount; i++)
		{
			uint16 nPixel = uint16( (pSrc[0] << 8) | pSrc[1]);
			pSrc += 2;

			uint8 r = uint8( (nPixel & 0xf800) >> 11 );
			uint8 g = uint8( (nPixel & 0x07e0) >> 5 );
			uint8 b = uint8( (nPixel & 0x001f) );

			r = uint8( (r << 3) + ((r >> 2) & 0x07));
			g = uint8( (g << 2) + ((g >> 1) & 0x03));
			b = uint8( (b << 3) + ((b >> 2) & 0x07));

			pDst[iR] = uint8(r);
			pDst[iG] = uint8(g);
			pDst[iB] = uint8(b);
			pDst[iA] = uint8(255);
			pDst += 4;
		}
	}
	else if( pfSrc == pf16bppArgb1555)
	{
		// Conversion rules :
		// Reuse the 3 most significant bits for the lower part of the 8 bit
		// 00000 --> 00000000
		// 00011 --> 00011000
		// 00111 --> 00111001
		// 01111 --> 01111011
		// 11111 --> 11111111
		
		for(int32 i = 0; i < nPixelCount; i++)
		{
			uint16 nPixel = uint16( (pSrc[0] << 8) | pSrc[1]);
			pSrc += 2;

			// Force a signed shift for alpha so we get either 0x00 or 0xFF
			uint8 a = uint8(int16(nPixel & 0x8000) >> 15);

			uint8 r = uint8( (nPixel & 0x7c00) >> 10 );
			uint8 g = uint8( (nPixel & 0x03e0) >> 5 );
			uint8 b = uint8( (nPixel & 0x001f) );

			a = 255;
			r = uint8( (r << 3) + ((r >> 2) & 0x07));
			g = uint8( (g << 3) + ((g >> 2) & 0x07));
			b = uint8( (b << 3) + ((b >> 2) & 0x07));

			pDst[iR] = uint8(r);
			pDst[iG] = uint8(g);
			pDst[iB] = uint8(b);
			pDst[iA] = uint8(a);
			pDst += 4;
		}
	}
	else if( pfSrc == pf16bppGrayScale)
	{
		// Conversion rules :
		// Add 1 if the low byte is greater to 127
		// aa00 --> aa
		// aa7f --> aa
		// aa80 --> ab
		// aaff --> ab

		// Saturate if the upper byte is 255
		// ff7f --> ff
		// ff80 --> ff
		// ffff --> ff

		if( bSimpleTransparencyUsed)
		{
			// One grey index is 100% transparent
			uint16 nTransparentIndex = it.GetGreyTransIndex();

			for(int32 i = 0; i < nPixelCount; i++)
			{
				uint32 g = uint16( (pSrc[0] << 8) | pSrc[1]);
				pSrc += 2;

				uint8 a = (g == nTransparentIndex) ? uint8(0) : uint8(255);

				g = (g >> 8) + ((g >> 7) & 1);
				g -= (g >> 8);

				pDst[iR] = uint8(g);
				pDst[iG] = uint8(g);
				pDst[iB] = uint8(g);
				pDst[iA] = uint8(a);
				pDst += 4;
			}
		}
		else
		{
			for(int32 i = 0; i < nPixelCount; i++)
			{
				uint32 g = uint16( (pSrc[0] << 8) | pSrc[1]);
				pSrc += 2;

				g = (g >> 8) + ((g >> 7) & 1);
				g -= (g >> 8);

				pDst[iR] = uint8(g);
				pDst[iG] = uint8(g);
				pDst[iB] = uint8(g);
				pDst[iA] = 255;
				pDst += 4;
			}
		}
	}
	else if( pfSrc == pf16bppGrayScaleAlpha)
	{
		for(int32 i = 0; i < nPixelCount; i++)
		{
			const uint8 nIndex = pSrc[0];
			const uint8 nAlpha = pSrc[1];
			pSrc += 2;
			
			pDst[iR] = nIndex;
			pDst[iG] = nIndex;
			pDst[iB] = nIndex;
			pDst[iA] = nAlpha;
			pDst += 4;
		}
	}
	else if( pfSrc == pf32bppGrayScaleAlpha)
	{
		for(int32 i = 0; i < nPixelCount; i++)
		{
			uint32 g = uint16( (pSrc[0] << 8) | pSrc[1]);
			uint32 a = uint16( (pSrc[2] << 8) | pSrc[3]);
			pSrc += 4;
			
			// Gray part
			g = (g >> 8) + ((g >> 7) & 1);
			g -= (g >> 8);

			// Alpha part
			a = (a >> 8) + ((a >> 7) & 1);
			a -= (a >> 8);

			pDst[iR] = uint8(g);
			pDst[iG] = uint8(g);
			pDst[iB] = uint8(g);
			pDst[iA] = uint8(a);
			pDst += 4;
		}
	}
	else if( pfSrc == pf1bppIndexed || pfSrc == pf2bppIndexed || pfSrc == pf4bppIndexed)
	{
		int32 nBitPerPix = ImageType::SizeofPixelInBits(pfSrc);

		int32 nMask = ~((~0L) << nBitPerPix);

		const Palette* pPalette = it.GetPalette();
		if( pPalette == null)
		{
			// Wanted a palette
			return false;
		}

		int32 nShift = 8 - nBitPerPix;
		for(int32 y = 0; y < nHeight; ++y)
		{
			for(int32 x = 0; x < nWidth; ++x)
			{
				const uint8 nByte = pSrc[0];
						
				int32 nIndex = (nByte >> nShift) & nMask;

				uint8 r, g, b, a;
				pPalette->m_aColors[nIndex].ToRgba(r, g, b, a);

				nShift -= nBitPerPix;
				if( nShift < 0)
				{
					nShift = 8 - nBitPerPix;
					pSrc += 1;
				}

				pDst[iR] = r;
				pDst[iG] = g;
				pDst[iB] = b;
				pDst[iA] = a;
				pDst += 4;
			}
			
			// Some padding bits remain
			if( nShift != 8 - nBitPerPix)
			{
				nShift = 8 - nBitPerPix;
				pSrc += 1;
			}
		}
	}
	else if( pfSrc == pf1bppGrayScale || pfSrc == pf2bppGrayScale || pfSrc == pf4bppGrayScale)
	{
		// Greyscale image

		int32 nBitsPerPix = ImageType::SizeofPixelInBits(pfSrc);

		//////////////////////////////////////////////////////////////////////////////////////////
		// Create the table to convert the indexes stored with less than 8 bits into 8 bits values
		const int32 nColCount = 1 << nBitsPerPix;
		const int32 nBitBlockCount = 8 / nBitsPerPix;
		uint8 aConvertTable[256];
		
		for(int32 iCol = 0; iCol < nColCount; ++iCol)
		{
			uint8 nGrey = 0;
			for(int32 iBit = 0; iBit < nBitBlockCount; ++iBit)
			{
				nGrey <<= nBitsPerPix;
				nGrey |= iCol;
			}

			aConvertTable[iCol] = nGrey;
		}

		//////////////////////////////////////////////////
		// Set the alpha values
		uint8 aAlphas[256];
		int32 nTransparentIndex = -1;
		if( it.IsSimpleTransparencyUsed())
		{
			nTransparentIndex = it.GetGreyTransIndex();
		}

		for(int32 i = 0; i < nColCount; ++i)
		{
			if( i == nTransparentIndex)
			{
				aAlphas[i] = 0;
			}
			else
			{
				aAlphas[i] = 255;
			}
		}
		//////////////////////////////////////////////////

		//////////////////////////////////////////////////////////////////////////////////////////

		int32 nMask = ~((~0L) << nBitsPerPix);

		int32 nShift = 8 - nBitsPerPix;
		for(int32 y = 0; y < nHeight; ++y)
		{
			for(int32 x = 0; x < nWidth; ++x)
			{
				const uint8 nByte = pSrc[0];
						
				int32 nIndex = (nByte >> nShift) & nMask;
				const uint8 nGrey = aConvertTable[nIndex];

				nShift -= nBitsPerPix;
				if( nShift < 0)
				{
					nShift = 8 - nBitsPerPix;
					pSrc += 1;
				}

				pDst[iR] = nGrey;
				pDst[iG] = nGrey;
				pDst[iB] = nGrey;
				pDst[iA] = aAlphas[nIndex];
				pDst += 4;
			}
			
			// Some padding bits remain
			if( nShift != 8 - nBitsPerPix)
			{
				nShift = 8 - nBitsPerPix;
				pSrc += 1;
			}
		}
	}
	else
	{
		// Unsupported pixel format
		return false;
	}
	
	return true;
}

bool ReadOnlySurface32::DumpAsPng(const String& strFileName)
{
	PngDumpSettings ds;
	ds.m_epf = Color32::GetInternalFormat();
	ds.m_pBuffer = GetBuffer();
	ds.m_nWidth = GetWidth();
	ds.m_nHeight = GetHeight();
	
	return Png::Dump(strFileName, ds);
}
