///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#ifndef CHUSTD_READONLYSURFACE32_H
#define CHUSTD_READONLYSURFACE32_H

#include "SurfaceBase.h"

namespace chustd {

enum PixelFormat
{
	pfUnknown,
	
	pf1bppGrayScale,
	pf2bppGrayScale,
	pf4bppGrayScale,
	pf8bppGrayScale,
	pf16bppGrayScale,

	pf16bppGrayScaleAlpha,
	pf32bppGrayScaleAlpha,

	pf1bppIndexed,
	pf2bppIndexed,
	pf4bppIndexed,
	pf8bppIndexed,

	pf16bppArgb1555,
	pf16bppArgb4444,
	pf16bppRgb555,
	pf16bppRgb565,

	pf24bppRgb,
	pf24bppBgr,

	pf32bppRgba,
	pf32bppBgra,

	pf48bppRgb,

	pf64bppRgba
};

// Color32 stores bytes in a compatible way with Surface32
// Native color format is plateform dependant. For instance pf32bppBgra is the
// best suitable in Windows because it is compatible with DIBs.
struct Color32
{
public:
	static Color32 black;
	static Color32 white;
	static Color32 red;
	static Color32 blue;
	static Color32 green;
	static Color32 yellow;
	static Color32 cyan;
	static Color32 magenta;

public:
	inline void ToRgb(uint8& r, uint8& g, uint8& b) const
	{
		uint32 nColor = m_nValue;

		b = uint8((nColor >> 0) & 0x00ff);
		g = uint8((nColor >> 8) & 0x00ff);
		r = uint8((nColor >> 16) & 0x00ff);
	}

	inline void ToRgb(uint32& r, uint32& g, uint32& b) const
	{
		uint32 nColor = m_nValue;

		b = uint32((nColor >> 0) & 0x00ff);
		g = uint32((nColor >> 8) & 0x00ff);
		r = uint32((nColor >> 16) & 0x00ff);
	}

	inline void ToRgba(uint8& r, uint8& g, uint8& b, uint8& a) const
	{
		uint32 nColor = m_nValue;

		b = uint8((nColor >> 0) & 0x00ff);
		g = uint8((nColor >> 8) & 0x00ff);
		r = uint8((nColor >> 16) & 0x00ff);
		a = uint8((nColor >> 24) & 0x00ff);
	}

	inline void ToRgba(uint32& r, uint32& g, uint32& b, uint32& a) const
	{
		uint32 nColor = m_nValue;

		b = ((nColor >> 0) & 0x00ff);
		g = ((nColor >> 8) & 0x00ff);
		r = ((nColor >> 16) & 0x00ff);
		a = ((nColor >> 24) & 0x00ff);
	}

	Color32()
	{
		m_nValue = 0xff000000;
	}

	Color32(uint32 r, uint32 g, uint32 b, uint32 a = 255)
	{
		m_nValue = (a << 24) | (r << 16) | (g << 8) | b;
	}

	void SetRgb(uint8 r, uint8 g, uint8 b)
	{
		m_nValue = (m_nValue & 0xff000000) | (r << 16) | (g << 8) | b;
	}

	void SetAlpha(uint8 a)
	{
		m_nValue = (m_nValue & 0x00ffffff) | (a << 24);
	}

	uint8 GetAlpha() const
	{
		return uint8(m_nValue >> 24);
	}

	bool IsEqualRgb(Color32 other)
	{
		return (m_nValue & 0x00ffffff) == (other.m_nValue & 0x00ffffff);
	}

	Color32& operator = (const Color32& color)
	{
		m_nValue = color.m_nValue;
		return *this;
	}

	bool operator == (const Color32& color)
	{
		return (m_nValue == color.m_nValue);
	}

	// Get the native pixel format. May change depending on the plateform.
	// (Windows = pf32bppBgra)
	static PixelFormat GetInternalFormat()
	{
		if( k_ePlatformByteOrder == boLittleEndian)
			return pf32bppBgra;

		return pf32bppRgba;
	}

	static Color32 From16bppRgb555(uint16 nPixel)
	{
		uint8 r = uint8( (nPixel & 0x7c00) >> 10 );
		uint8 g = uint8( (nPixel & 0x03e0) >> 5 );
		uint8 b = uint8( (nPixel & 0x001f) );

		r = uint8( (r << 3) + ((r >> 2) & 0x07));
		g = uint8( (g << 3) + ((g >> 2) & 0x07));
		b = uint8( (b << 3) + ((b >> 2) & 0x07));

		return Color32(r, g, b, 255);
	}
public:
	union
	{
		uint32 m_nValue; // Win32-x86: BGRA in memory, ARGB in register
		struct Bytes
		{
			uint8 b, g, r, a;
		};
	};
};

class ImageType;
class String;

class ReadOnlySurface32 : public SurfaceBase
{
	typedef SurfaceBase super;
public:
	void Create(const ReadOnlySurface32& surfSource, const Rectangle& rect);
	bool Create(int32 nWidth, int32 nHeight);
	bool Create(const ImageType& it);

	bool UpdatePixels(const ImageType& it);

	bool DumpAsPng(const String& strFileName);

protected:
	ReadOnlySurface32(class SurfaceBaseImpl* p) : SurfaceBase(p) {}
	bool UpdatePixels(int32 nWidth, int32 nHeight, void* pPixels, PixelFormat pf);
};

} // namespace chustd

#endif // ndef CHUSTD_READONLYSURFACE32_H
