///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "Master.h"
#include "ScanType.h"

#include "String.h"
//////////////////////////////////////////////////////////////////////////////////////////////////
namespace chustd {

/////////////////////////////////////////////////////////////////////////////////////////////////
// ScanType for int32
bool IsBaseDigit(wchar c, int32 nBase, int32& nDigit)
{
	if( nBase > 16)
		return false;

	if( nBase <= 10)
	{
		const wchar nMaxDigit = wchar('0' + (nBase - 1));
		if( '0' <= c && c <= nMaxDigit)
		{
			nDigit = c - '0';
			return true;
		}
		return false;
	}

	if( '0' <= c && c <= '9')
	{
		nDigit = c - '0';
		return true;
	}

	if( 'a' <= c && c <= 'f')
	{
		nDigit = (c - 'a') + 10;
		return true;
	}

	if( 'A' <= c && c <= 'F')
	{
		nDigit = (c - 'A') + 10;
		return true;
	}
	return false;
}

bool ScanType(const wchar* psz, int32& nAdvance, char cScanArg, int32* pResult)
{
	const wchar* pszStart = psz + nAdvance;
	
	int32 nBase = 10;
	if( cScanArg == 'x' || cScanArg == 'X')
	{
		nBase = 16;
	}
	else if( cScanArg == 'b' || cScanArg == 'B')
	{
		nBase = 2;
	}

	enum { stateWhitespace, stateSignOrDigit, stateDigit };
	int nState = stateWhitespace;

	int32 nResult = 0;
	bool bPositive = true;
	int32 nDigitCount = 0;

	int32 iChar = 0;
	while(true)
	{
		wchar c = pszStart[iChar];
		if( nState == stateWhitespace)
		{
			if( !String::IsBlankChar(c))
			{
				nState = stateSignOrDigit;
			}
		}

		if( nState == stateSignOrDigit)
		{
			if( c == '+')
			{
				// Ok
			}
			else if( c == '-' || c == 0x2212)
			{
				bPositive = false;
			}
			else
			{
				nState = stateDigit;
			}
		}

		if( nState == stateDigit)
		{
			int32 nDigit;
			if( IsBaseDigit(c, nBase, nDigit))
			{
				nResult = nResult * nBase + nDigit;
				nDigitCount++;
			}
			else
			{
				// End of parsing
				break;
			}
		}
		iChar++;
	}

	if( nDigitCount == 0)
	{
		// Error
		return false;
	}
	
	if( !bPositive)
	{
		nResult = -nResult;
	}

	nAdvance += iChar;
	*pResult = nResult;

	return true;
}

// ScanType for int16
bool ScanType(const wchar* psz, int32& nAdvance, char cScanArg, int16* pResult)
{
	// Use the int32 version of ScanType
	int32 nPreResult = 0;
	if( !ScanType(psz, nAdvance, cScanArg, &nPreResult))
		return false;

	*pResult = int16(nPreResult);
	return true;
}

// ScanType for int8
bool ScanType(const wchar* psz, int32& nAdvance, char cScanArg, int8* pResult)
{
	// Use the int32 version of ScanType
	int32 nPreResult = 0;
	if( !ScanType(psz, nAdvance, cScanArg, &nPreResult))
		return false;

	*pResult = int8(nPreResult);
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////
// ScanType for uint32
bool ScanType(const wchar* psz, int32& nAdvance, char cScanArg, uint32* pResult)
{
	const wchar* pszStart = psz + nAdvance;
	
	int32 nBase = 10;
	if( cScanArg == 'x' || cScanArg == 'X')
	{
		nBase = 16;
	}
	else if( cScanArg == 'b' || cScanArg == 'B')
	{
		nBase = 2;
	}

	enum { stateWhitespace, statePlusOrDigit, stateDigit };
	int nState = stateWhitespace;

	int32 nResult = 0;
	int32 nDigitCount = 0;

	int32 iChar = 0;
	while(true)
	{
		wchar c = pszStart[iChar];
		if( nState == stateWhitespace)
		{
			if( !String::IsBlankChar(c))
			{
				nState = statePlusOrDigit;
			}
		}

		if( nState == statePlusOrDigit)
		{
			if( c == '+')
			{
				// Ok
			}
			else
			{
				nState = stateDigit;
			}
		}

		if( nState == stateDigit)
		{
			int32 nDigit;
			if( IsBaseDigit(c, nBase, nDigit))
			{
				nResult = nResult * nBase + nDigit;
				nDigitCount++;
			}
			else
			{
				// End of parsing
				break;
			}
		}
		iChar++;
	}

	if( nDigitCount == 0)
	{
		// Error
		return false;
	}
	
	nAdvance += iChar;
	*pResult = nResult;

	return true;
}

// ScanType for uint16
bool ScanType(const wchar* psz, int32& nAdvance, char cScanArg, uint16* pResult)
{
	// Use the uint32 version of ScanType
	uint32 nPreResult = 0;
	if( !ScanType(psz, nAdvance, cScanArg, &nPreResult))
		return false;

	*pResult = uint16(nPreResult);
	return true;
}

// ScanType for uint8
bool ScanType(const wchar* psz, int32& nAdvance, char cScanArg, uint8* pResult)
{
	// Use the uint32 version of ScanType
	uint32 nPreResult = 0;
	if( !ScanType(psz, nAdvance, cScanArg, &nPreResult))
		return false;

	*pResult = uint8(nPreResult);
	return true;
}

//////////////////////////////////////////////////////////////////////////////////////////////
// ScanType for float 64
// TODO: Lame, lame code, we need something better here
bool ScanType(const wchar* psz, int32& nAdvance, char /*cScanArg*/, float64* pResult)
{
	int32 nIntPart = 0;
	if( !ScanType(psz, nAdvance, 0, &nIntPart))
		return false;
		
	float64 fFrac = 0;
	if( psz[nAdvance] == L'.')
	{
		uint32 nFracPart = 0;
		int32 nFracPartLength = 0;

		nAdvance++;
		nFracPartLength = nAdvance;

		if( !ScanType(psz, nAdvance, 0, &nFracPart))
			return false;

		nFracPartLength = nAdvance - nFracPartLength;

		int32 nFracDiv = 1;
		for(int32 i = 0; i < nFracPartLength; ++i)
		{
			nFracDiv *= 10;
		}
		fFrac = float64(nFracPart);
		fFrac /= float64(nFracDiv);
	}
		
	float64 fResult = float64(nIntPart);
	fResult += fFrac;

	*pResult = fResult;

	return true;
}

// ScanType for float32
bool ScanType(const wchar* psz, int32& nAdvance, char cScanArg, float32* pResult)
{
	// Use the double version of ScanType
	float64 dResult = 0;
	if( !ScanType(psz, nAdvance, cScanArg, &dResult))
		return false;

	*pResult = float32(dResult);
	return true;
}

/////////////////////////////////////////////////////////////////////////////////////////////////
// ScanType for bool
bool ScanType(const wchar* psz, int32& nAdvance, char cScanArg, bool* pResult)
{
	// Use the uint8 version of ScanType
	uint8 nPreResult = 0;
	if( !ScanType(psz, nAdvance, cScanArg, &nPreResult))
		return false;

	*pResult = nPreResult == 0 ? false : true;
	return true;
}

/////////////////////////////////////////////////////////////////////////////////////////////////
bool ScanType(const wchar* psz, int32& nAdvance, char /*cScanArg*/, String* pResult)
{
	int32 iChar = nAdvance;
	while(true)
	{
		wchar c = psz[iChar];
		if( c == 0 || c  == ' ' || c == '\t' || c == '\r' || c == '\n')
			break;
		++iChar;
	}

	const int32 nCharCount = iChar - nAdvance;
	pResult->SetBuffer(psz + nAdvance, nCharCount);

	nAdvance += nCharCount;

	return true;
}
/////////////////////////////////////////////////////////////////////////////////////////////////

} // namespace chustd

