///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "Master.h"
#include "Socket.h"
#include "String.h"
#include "IFile.h"
#include "StringBuilder.h"
#include "FormatType.h"

using namespace chustd;

//////////////////////////////////////////////////////////////////////
int Socket::ms_nSocketStartupCount = 0;
bool Socket::ms_bNetworkWorking = true;
//////////////////////////////////////////////////////////////////////

Socket::Socket()
{
#ifdef WIN32
	if( ms_nSocketStartupCount == 0)
	{
		WSADATA wsaData;
		int nRet = WSAStartup(MAKEWORD( 2, 2 ), &wsaData);
		ms_bNetworkWorking = ( nRet == 0);
	}

	ms_nSocketStartupCount++;
#elif CHUOS
#endif

	m_nIDSocket = INVALID_SOCKET;
	m_nType = typeUnknown;
}

Socket::~Socket()
{
	Close();
	
	ms_nSocketStartupCount--;
	if( ms_nSocketStartupCount <= 0)
	{
#ifdef WIN32
		WSACleanup();
#endif
	}
}

bool Socket::Open(Type eType)
{
	if( !ms_bNetworkWorking)
		return false;

	if( m_nType != typeUnknown)
	{
		// Socket already open, close it before we continue
		Close();
	}

#ifdef WIN32
	// Socket creation
	int nBkSocketType = 0;
	if( eType == typeTcp)
	{
		nBkSocketType = SOCK_STREAM;
	}
	else if( eType == typeUdp)
	{
		nBkSocketType = SOCK_DGRAM;
	}
	else
	{
		return false;
	}

	const SOCKET_ID nIDSocket = socket(AF_INET, nBkSocketType, 0);
	if( nIDSocket == INVALID_SOCKET)
		return false;

	m_nIDSocket = int64(nIDSocket);
	m_nType = uint8(eType);
	return true;

#elif CHUOS
	return false;
#endif
	
}

bool Socket::IsOpen() const
{
	return m_nType != typeUnknown;
}

void Socket::Close()
{
	if( m_nIDSocket != INVALID_SOCKET)
	{
#ifdef WIN32
		closesocket( SOCKET_ID(m_nIDSocket)); // Windows
#elif CHUOS

#else
		close(m_nIDSocket); // Other
#endif
	}

	m_nIDSocket = INVALID_SOCKET;
	m_nType = typeUnknown;
}


int32 Socket::Receive(uint8* pBuffer, int32 nBufferSize, bool bBlockUntilSizeReached)
{
	if( !IsOpen())
		return 0;

#ifdef WIN32
	if( !bBlockUntilSizeReached)
	{
		return recv( SOCKET_ID(m_nIDSocket), (char*)pBuffer, nBufferSize, 0);
	}

	int32 nReceived = 0; // Number of bytes received
	do
	{
		const int nRet = recv(SOCKET_ID(m_nIDSocket), (char*)pBuffer + nReceived, nBufferSize - nReceived, 0);
		if( nRet <= 0)
		{
			break;
		}
		nReceived += nRet;
	}while(nReceived < nBufferSize);
	return nReceived;

#elif CHUOS
	return 0;
#endif

}

int32 Socket::Send(const uint8* pBuffer, int32 nBufferSize, bool bBlockUntilSizeReached)
{
	if( !IsOpen())
		return 0;

#ifdef WIN32
	if( !bBlockUntilSizeReached)
	{
		return send(SOCKET_ID(m_nIDSocket), (const char*)pBuffer, nBufferSize, 0);
	}

	int32 nSent = 0; // Number of bytes sent
	do
	{
		int nRet = send(SOCKET_ID(m_nIDSocket), (const char*)(pBuffer + nSent), nBufferSize - nSent, 0);
		if( nRet <= 0)
		{
			if( bBlockUntilSizeReached && IsSendBufferFlooded())
			{
				// We should retry sending
				bool bOk = BlockUntilSendPossible();
				if( !bOk)
				{
					// An error occured while waiting for sending authorization
					return 0;
				}
			}
			else
			{
				// Real error
				return 0;
			}
		}
		else
		{
			nSent += nRet;
		}
	}while( nSent < nBufferSize);
	
	return nSent;
#elif CHUOS
	return 0;
#endif
}

// Returns true if a send operation could not be completed because of a too small TCP/IP send buffer
bool Socket::IsSendBufferFlooded() const
{
#ifdef WIN32
	int nRet = WSAGetLastError();

	if( nRet == WSAEWOULDBLOCK)
	{
		// Yep, send buffer is too small, we should retry sending
		// Note : this error can occurs on some Windows configuration even if the socket
		// has not been set for asynchronous write operations (WSAAsyncSelect)
		return true;
	}
#elif CHUOS
	return false;

#else
	if( errno == EWOULDBLOCK)
	{
		// Non blocking socket
		return true;
	}
#endif

	// Other error, does not concern send buffer flooding
	return false;
}

bool Socket::BlockUntilSendPossible()
{
	ASSERT(m_nIDSocket != INVALID_SOCKET);

#ifdef WIN32
	fd_set fdsWrite;

	FD_ZERO(&fdsWrite); // Clear the fd_set
	
	// int on Unix, uint on Windows, so warning here
	SOCKET_ID nSocketId = SOCKET_ID(m_nIDSocket);
	FD_SET(nSocketId, &fdsWrite); // Add the socket ID to the socket set
 
	int nRet = select(0, NULL, &fdsWrite, NULL, NULL);
	if( nRet <= 0)
		return false;
#elif CHUOS
#endif
	return true;
}

int64 Socket::GetSocketID() const
{
	return m_nIDSocket;
}

bool Socket::StringToIpAddress(const String& strIpAddress, uint32& nIpAddress)
{
	const int32 nStringLength = strIpAddress.GetLength();
	const wchar* psz = strIpAddress.GetBuffer();

	/////////////////////////////////////////////////////////////////////
	// First pass, count dot and store dots positions
	int32 aDotIndex[3];
	int32 nDotCount = 0;

	for(int32 i = 0; i < nStringLength; i++)
	{
		wchar c = psz[i];
		if( c == '.')
		{
			if( nDotCount == 3)
			{
				// Too many dots
				return false;
			}
			aDotIndex[nDotCount] = i;
			nDotCount++;
		}
	}

	if( nDotCount != 3)
	{
		// Not enough dots
		return false;
	}
	/////////////////////////////////////////////////////////////////////
	const int32 nLength0 = aDotIndex[0];
	if( nLength0 > 3 || nLength0 <= 0)
		return false;

	const int32 nLength1 = (aDotIndex[1] - aDotIndex[0]) - 1;
	if( nLength1 > 3 || nLength1 <= 0)
		return false;

	const int32 nLength2 = (aDotIndex[2] - aDotIndex[1]) - 1;
	if( nLength2 > 3 || nLength2 <= 0)
		return false;
	
	const int32 nLength3 = (nStringLength - aDotIndex[2]) - 1;
	if( nLength3 > 3 || nLength3 <= 0)
		return false;

	int32 aPieceLengths[4] = { nLength0, nLength1, nLength2, nLength3 };

	uint32 nResult = 0; // IP Address, to be copied into nIpAddress later

	for(int32 iPiece = 0; iPiece < 4; iPiece++)
	{
		const int32 nPieceLength = aPieceLengths[iPiece];

		uint32 nVal = 0;
		for(int32 iDigit = 0; iDigit < nPieceLength; iDigit++)
		{
			const uint16 c = psz[iDigit];
			if( '0' <= c && c <= '9')
			{
				const int32 nDigit = c - '0';
				
				nVal *= 10;
				nVal += nDigit;
			}
			else
			{
				// The character is not a digit
				return false;
			}
		}
		
		if( nVal > 255)
		{
			// Each piece is 8 bits so > 255 is not acceptable
			return false;
		}

		// Copy Piece value
		nResult <<= 8;
		nResult |= nVal;

		// Jump over last piece + dot
		psz += nPieceLength + 1;
	}
	
	nIpAddress = nResult;
	return true;
}

String Socket::IpAddressToString(uint32 nIpAddress)
{
	uint8 nA = uint8( (nIpAddress >> 24) & 0x00ff);
	uint8 nB = uint8( (nIpAddress >> 16) & 0x00ff);
	uint8 nC = uint8( (nIpAddress >> 8) & 0x00ff);
	uint8 nD = uint8( (nIpAddress) & 0x00ff);

	StringBuilder sb;

	wchar szBuffer[12];

	FormatInt32(nA, szBuffer);
	sb += szBuffer;
	sb += '.';

	FormatInt32(nB, szBuffer);
	sb += szBuffer;
	sb += '.';

	FormatInt32(nC, szBuffer);
	sb += szBuffer;
	sb += '.';

	FormatInt32(nD, szBuffer);
	sb += szBuffer;

	return sb.ToString();
}
