///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "Master.h"
#include "Sort.h"
#include "Memory.h"
#include "Math.h"

using namespace chustd;


void Sort::ByteSortAtt(uint32* paElements, uint32* paSortAttributs, int32 nSize)
{
	uint32* paTmpElements = new uint32[nSize];
	uint32* paTmpSortAttributs = new uint32[nSize];

	ByteSortAtt(paElements, paSortAttributs, paTmpElements, paTmpSortAttributs, nSize);

	delete[] paTmpSortAttributs;
	delete[] paTmpElements;
}

void Sort::ByteSortAtt(uint32* paElements, uint32* paSortAttributs, 
		uint32* paNewElements, uint32* paNewSortAttributs, int32 nSize)
{
	uint32 aOccurrences[256];
	
	// 4 passes for each byte of the uint32
	for( int8 nShift = 0 ; nShift < 4 ; nShift++ )
	{
		Memory::Zero32(aOccurrences, 256);
			
		for( int32 i = 0; i < nSize; i++ )
		{
			const uint8 nByte = Byte( paSortAttributs[i], nShift );
			aOccurrences[nByte]++;
		}

		// Compute final positions in the sorted array
		uint32 nPosition = 0;
		for( int32 i = 0; i < 256; i++ )
		{			
			const uint32 nIncrement = aOccurrences[i];
			aOccurrences[i] = nPosition;
			nPosition += nIncrement;
		}

		for( int32 i = 0; i < nSize; i++ )
		{			
			const uint8 nByte = Byte( paSortAttributs[i], nShift );
			nPosition = aOccurrences[nByte];
			aOccurrences[nByte]++;

			paNewSortAttributs[nPosition] = paSortAttributs[i];
			paNewElements[nPosition] = paElements[i];
		}

		// Swap array pointers
		// Final array is valid because we have an even number of passes
		uint32* pTmp = paSortAttributs;
		paSortAttributs = paNewSortAttributs;
		paNewSortAttributs = pTmp;

		uint32* pTmpElements = paElements;
		paElements = paNewElements;
		paNewElements = pTmpElements;
	}
}

void Sort::BubbleSort(uint32* pa, int32 nSize)
{
	for( int32 i = 0; i < nSize; ++i )
	{
		bool bHasSwapped = false;

		for( int32 j = 0; j < nSize - 1 - i; ++j )
		{
			if ( pa[j] > pa[j + 1] )
			{
				bHasSwapped = true;
				
				// Swap
				const uint32 nTmp = pa[j + 1];
				pa[j + 1] = pa[j];
				pa[j] = nTmp;
			}
			
			if ( !bHasSwapped )
				return;
		}
	}
}

void Sort::ByteSortAtt(uint32* paElements, uint8* paSortAttributs, int32 nSize)
{
	uint32* paTmpElements = new uint32[nSize];

	uint32 aOccurrences[256];
	
	Memory::Zero32(aOccurrences, 256);
		
	for( int32 i = 0; i < nSize; i++ )
	{
		const uint8 nByte = paSortAttributs[i];
		aOccurrences[nByte]++;
	}

	// Compute final positions in the sorted array
	uint32 nPosition = 0;
	for( int32 i = 0; i < 256; i++ )
	{			
		const uint32 nIncrement = aOccurrences[i];
		aOccurrences[i] = nPosition;
		nPosition += nIncrement;
	}

	for( int32 i = 0; i < nSize; i++ )
	{			
		const uint8 nByte = paSortAttributs[i];
		nPosition = aOccurrences[nByte];
		aOccurrences[nByte]++;

		paTmpElements[nPosition] = paElements[i];
	}

	Memory::Copy32(paElements, paTmpElements, nSize);

	delete[] paTmpElements;
}

void Sort::ByteSort(const uint8* pIn, uint8* pOut, int32 nSize)
{
	uint32 aOccurrences[256];
	
	// Step 1 : clear the occurrence array
	Memory::Zero32(aOccurrences, 256);
	
	// Step 2 : count the occurrence of each element
	for(int32 i = 0; i < nSize; ++i)
	{
		const uint8 nByte = pIn[i];
		aOccurrences[nByte]++;
	}

	// Step 3 : compute final positions in the sorted array
	uint32 nPosition = 0;
	for(int32 i = 0; i < 256; ++i)
	{			
		const uint32 nIncrement = aOccurrences[i];
		aOccurrences[i] = nPosition;
		nPosition += nIncrement;
	}
	
	for(int32 i = 0; i < nSize; ++i)
	{			
		const uint8 nByte = pIn[i];
		nPosition = aOccurrences[nByte];
		aOccurrences[nByte]++;

		pOut[nPosition] = pIn[i];
	}
}

void Sort::ByteSort(uint32* pIn, uint32* pOut, int32 nSize)
{
	uint32 aOccurrences[256];
	
	// 4 passes for each byte of the uint32
	for( int8 nShift = 0 ; nShift < 4 ; nShift++ )
	{
		Memory::Zero32(aOccurrences, 256);
			
		for( int32 i = 0; i < nSize; i++ )
		{
			const uint8 nByte = Byte( pIn[i], nShift );
			aOccurrences[nByte]++;
		}

		// Compute final positions in the sorted array
		uint32 nPosition = 0;
		for( int32 i = 0; i < 256; i++ )
		{			
			const uint32 nIncrement = aOccurrences[i];
			aOccurrences[i] = nPosition;
			nPosition += nIncrement;
		}

		for( int32 i = 0; i < nSize; i++ )
		{			
			const uint8 nByte = Byte( pIn[i], nShift );
			nPosition = aOccurrences[nByte];
			aOccurrences[nByte]++;

			pOut[nPosition] = pIn[i];
		}

		// Swap array pointers
		// Final array is valid because we have an even number of passes
		Math::Swap(pIn, pOut);
	}
}

// The uint32 are stored and managed in Little Endian, even on big endian machines
// Final result in pOut1
// pIn and pOut2 can point on the same data if pIn data can be overwritten
void Sort::ByteSortLittleEndian(const uint32* pIn, uint32* pOut1, uint32* pOut2, int32 nSize)
{
	uint32 aOccurrences[256];
	
	uint32* pOut = pOut1;

	// 4 passes for each byte of the uint32
	for(int iPass = 0; iPass < 4; ++iPass)
	{
		Memory::Zero32(aOccurrences, 256);

		int nShift;
		if( k_ePlatformByteOrder == boBigEndian )
		{
			// Shift in reverse order
			nShift = 4 - iPass;
		}
		else
		{
			nShift = iPass;
		}
			
		for( int32 i = 0; i < nSize; i++ )
		{
			const uint8 nByte = Byte( pIn[i], nShift );
			aOccurrences[nByte]++;
		}

		// Compute final positions in the sorted array
		uint32 nPosition = 0;
		for( int32 i = 0; i < 256; i++ )
		{			
			const uint32 nIncrement = aOccurrences[i];
			aOccurrences[i] = nPosition;
			nPosition += nIncrement;
		}

		for( int32 i = 0; i < nSize; i++ )
		{			
			const uint8 nByte = Byte( pIn[i], nShift );
			nPosition = aOccurrences[nByte];
			aOccurrences[nByte]++;

			pOut[nPosition] = pIn[i];
		}

		// Swap pointers for next byte
		if( (iPass & 0x01) == 0)
		{
			pIn = pOut1;
			pOut = pOut2;
		}
		else
		{
			pIn = pOut2;
			pOut = pOut1;
		}
	}
}
