///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "Master.h"
#include "String.h"
#include "StringData.h"
#include "UnicodeCaseMapping.h"
#include "Array.h"
#include "ScanType.h"
#include "FormatType.h"
#include "TextEncoding.h"
#include "StringBuilder.h"

using namespace chustd;

String::String()
{
	m_pszBuffer = StringData::GetNullInstance()->GetBuffer();
}

String::String(const char* psz)
{
	if( psz == null || psz[0] == 0 )
	{
		m_pszBuffer = StringData::GetNullInstance()->GetBuffer();
	}
	else
	{
		int32 nLength = String::SZLength(psz);
		uint16* pszNew = StringData::CreateInstance(nLength);
		if( pszNew == null )
		{
			// Raise exception ?
			m_pszBuffer = StringData::GetNullInstance()->GetBuffer();
		}
		else
		{
			for(int32 i = 0; i < nLength; ++i)
			{
				pszNew[i] = psz[i];
			}
			m_pszBuffer = pszNew;
		}
	}
}

String::String(const wchar* psz)
{
	if( psz == null || psz[0] == 0 )
	{
		m_pszBuffer = StringData::GetNullInstance()->GetBuffer();
	}
	else
	{
		const int32 nLength = SZLength(psz);
		m_pszBuffer = StringData::CreateInstance(nLength);
		
		// Note : no need for final zero as CreateInstance already did it for us
		Memory::Copy64From16(m_pszBuffer, psz, nLength);
	}
}

String::String(const wchar* psz, int32 nLength)
{
	if( psz == null )
	{
		m_pszBuffer = StringData::GetNullInstance()->GetBuffer();
	}
	else
	{
		m_pszBuffer = StringData::CreateInstance(nLength);
		
		// Note : no need for final zero as CreateInstance already did it for us
		Memory::Copy64From16(m_pszBuffer, psz, nLength);
	}
}

String::String(const String& str)
{
	m_pszBuffer = str.m_pszBuffer;
	ASSERT(m_pszBuffer != null);
	
	StringData* pData = GetData();
	if( pData != StringData::GetNullInstance() )
	{
		pData->AddRef();
	}
}

String::~String()
{
	Unref();
}

//////////////////////////////////////////////////////////////////////
void String::Empty()
{
	Unref();
	m_pszBuffer = StringData::GetNullInstance()->GetBuffer();
}


void String::SetBuffer(const wchar* pBuffer, int32 nLength)
{
	Unref();

	if( pBuffer != null && nLength != 0 )
	{
		m_pszBuffer = StringData::CreateInstance(nLength);
		Memory::Copy16(m_pszBuffer, pBuffer, nLength);
		m_pszBuffer[nLength] = 0;
	}
	else
	{
		m_pszBuffer = StringData::GetNullInstance()->GetBuffer();
	}
}

String& String::operator = (const String& str)
{
	if( this == &str )
		return *this;

	Unref();

	m_pszBuffer = str.m_pszBuffer;
	
	StringData* pData = GetData();
	if( pData != StringData::GetNullInstance() )
	{
		pData->AddRef();
	}

	return *this;
}

String& String::operator = (uint32 nCodePoint)
{
	Unref();

	int32 nLength = (nCodePoint > 0x0000ffff) ? 2 : 1;
	m_pszBuffer = StringData::CreateInstance(nLength);
		
	if( nLength == 1 )
	{
		m_pszBuffer[0] = uint16(nCodePoint);
	}
	else
	{
		// TODO
		ASSERT(0);

		uint16 nSurrogateHi = 0;
		uint16 nSurrogateLo = 0;

		m_pszBuffer[0] = nSurrogateHi;
		m_pszBuffer[1] = nSurrogateLo;
	}
	return *this;
}
//////////////////////////////////////////////////////////////////////
int32 String::DiffPtr(const void* pA, const void* pB, int32 nCharSize)
{
	const uint8* pA8 = (uint8*) pA;
	const uint8* pB8 = (uint8*) pB;

	if( sizeof(void*) > 4 )
	{
		// Check overflow
		const int32 nMaxLength = 0x7fffffff;
		int64 nLength64 = int64(pA8 - pB8);
		nLength64 >>= (nCharSize - 1);

		if( nLength64 > 0x7fffffff )
		{
			ASSERT(0); // Something is bad in the machinery
			return nMaxLength;
		}
		return int32(nLength64);
	}

	int32 nLength = int32(pA8 - pB8) >> (nCharSize - 1);
	return nLength;
}

int32 String::SZLength(const wchar* psz)
{
	const wchar* pszStart = psz;
	while( *psz != 0 )
	{
		psz++;
	}
	return DiffPtr(psz, pszStart, 2);
}

int32 String::SZLength(const char* psz)
{
	const char* pszStart = psz;
	while( *psz != 0 )
	{
		psz++;
	}
	return DiffPtr(psz, pszStart, 1);
}

//////////////////////////////////////////////////////////////////////
// Unrefs the StringData instance.
// If the reference counter reachs 0 the StringData instance is deleted
void String::Unref()
{
	StringData* pData = GetData();
	if( pData == StringData::GetNullInstance() )
		return;

	// Should always passed, otherwise some memory trashing occured or one of the String 
	// functions is buggy
	ASSERT( pData->GetRef() > 0);

	pData->Release();
	if( pData->GetRef() <= 0 )
	{
		// The instance owns the data
		pData->Delete();
		m_pszBuffer = StringData::GetNullInstance()->GetBuffer();
	}
}

///////////////////////////////////////////////////////////////////////////////////////
int32 String::GetLength() const
{
	return GetData()->GetLength();
}

uint16 String::GetAt(int32 nIndex) const
{
	ASSERT(0 <= nIndex && nIndex < GetLength());

	return m_pszBuffer[nIndex];
}

///////////////////////////////////////////////////////////////////////////////////////
// Gets a StringData pointer from the StringData sz buffer of the String instance
const chustd::StringData* String::GetData() const
{
	return chustd::StringData::GetInstance(m_pszBuffer);
}

chustd::StringData* String::GetData()
{
	return chustd::StringData::GetInstance(m_pszBuffer);
}


String String::ToUpperCase() const
{
	const uint16* const * const pPages = UnicodeCaseMapping::GetUcs2ToUpperPages();
	const uint32* const pPage10428 = UnicodeCaseMapping::GetUcs4ToUpperPage10428();

	StringBuilder sb;
	int32 nCodeUnitIndex = 0;
	while(true)
	{
		uint32 nCodePoint = GetCodePoint(nCodeUnitIndex);
		if( nCodePoint == 0xffffffff )
		{
			break;
		}

		if( nCodePoint <= 0x0000ffff )
		{
			// Use UCS-2 pages
			uint8 nPage = uint8(nCodePoint >> 8);
			uint8 nSubCode = uint8(nCodePoint & 0x00ff);

			const uint16* const pPage = pPages[nPage];
			if( pPage != null && pPage[nSubCode] != 0 )
			{
				nCodePoint = pPage[nSubCode];
			}
			else
			{
				// Find unconditionnal 1:N mapping
				const uint16* pMulti = UnicodeCaseMapping::CodeUnitToUpperMulti( uint16(nCodePoint));
				if( pMulti != null )
				{
					for(int iNew = 0; iNew < 3; ++iNew)
					{
						if( pMulti[iNew] == 0 )
						{
							break;
						}
						sb += pMulti[iNew];
					}
					
					// Reset the main loop, as 1:N is a special case
					continue;
				}
			}
		}
		else
		{
			int32 nPage10428Index = nCodePoint - 0x10428;
			if( 0 <= nPage10428Index && nPage10428Index <= 0x27 )
			{
				nCodePoint = pPage10428[nPage10428Index];
			}
		}
		sb += nCodePoint;
	}
	return sb.ToString();
}


String String::ToLowerCase() const
{
	const uint16* const * const pPages = UnicodeCaseMapping::GetUcs2ToLowerPages();
	const uint32* const pPage10400 = UnicodeCaseMapping::GetUcs4ToLowerPage10400();

	StringBuilder sb;
	int32 nCodeUnitIndex = 0;
	while(true)
	{
		uint32 nCodePoint = GetCodePoint(nCodeUnitIndex);
		if( nCodePoint == 0xffffffff )
		{
			break;
		}

		if( nCodePoint <= 0x0000ffff )
		{
			// Use UCS-2 pages
			uint8 nPage = uint8(nCodePoint >> 8);
			uint8 nSubCode = uint8(nCodePoint & 0x00ff);

			const uint16* const pPage = pPages[nPage];
			if( pPage != null && pPage[nSubCode] != 0 )
			{
				nCodePoint = pPage[nSubCode];
			}
			else
			{
				// Find unconditionnal 1:N mapping
				const uint16* pMulti = UnicodeCaseMapping::CodeUnitToLowerMulti( uint16(nCodePoint));
				if( pMulti != null )
				{
					for(int iNew = 0; iNew < 3; ++iNew)
					{
						if( pMulti[iNew] == 0 )
						{
							break;
						}
						sb += pMulti[iNew];
					}
					
					// Reset the main loop, as 1:N is a special case
					continue;
				}
			}
		}
		else
		{
			int32 nPage10400Index = nCodePoint - 0x10400;
			if( 0 <= nPage10400Index && nPage10400Index <= 0x27 )
			{
				nCodePoint = pPage10400[nPage10400Index];
			}
		}
		sb += nCodePoint;
	}
	return sb.ToString();
}

//////////////////////////////////////////////////////////////////////
// Global friend functions
namespace chustd
{

const String operator + (const String& str1, const String& str2)
{
	const int32 nDataLength1 = str1.GetData()->GetLength();
	const int32 nDataLength2 = str2.GetData()->GetLength();
	
	if( nDataLength1 == 0 )
	{
		return str2;
	}
	if( nDataLength2 == 0 )
	{
		return str1;
	}

	uint16* pszNew = StringData::CreateInstance(nDataLength1 + nDataLength2);
	
	if( nDataLength1 > 0 )
	{
		// Copy first part
		const int32 nChunkCount1 = StringData::GetUsedChunkCount(nDataLength1);
		Memory::Copy64(pszNew, str1.m_pszBuffer, nChunkCount1);
	}

	// Copy second part
	// + 1 to copy the final 0 which might not be at the right place after a FastUInt32Copy
	Memory::Copy16(pszNew + nDataLength1, str2.m_pszBuffer, (nDataLength2 + 1));

	String str;
	str.m_pszBuffer = pszNew;
	return str;
}

const String operator + (const String& str1, const wchar* psz)
{
	if( psz == null || psz[0] == 0 )
		return str1;

	const int32 nDataLength1 = str1.GetData()->GetLength();
	const int32 nPszLength = String::SZLength(psz);
	
	uint16* pszNew = StringData::CreateInstance(nDataLength1 + nPszLength);

	if( nDataLength1 > 0 )
	{
		// Copy first part
		const int32 nChunkCount1 = StringData::GetUsedChunkCount(nDataLength1);
		Memory::Copy64(pszNew, str1.m_pszBuffer, nChunkCount1);
	}
	
	// Copy second part
	// + 1 to copy the final 0 which might not be at the right place after a FastUInt32Copy
	Memory::Copy16(pszNew + nDataLength1, psz, (nPszLength + 1));

	String str;
	str.m_pszBuffer = pszNew;
	return str;
}

const String operator + (const wchar* psz, const String& str2)
{
	if( psz == null || psz[0] == 0 )
		return str2;

	const int32 nPszLength = String::SZLength(psz);
	const int32 nDataLength2 = str2.GetData()->GetLength();

	uint16* pszNew = StringData::CreateInstance(nPszLength + nDataLength2);

	Memory::Copy16(pszNew, psz, nPszLength);
	Memory::Copy16(pszNew + nPszLength, str2.m_pszBuffer, nDataLength2);

	String str;
	str.m_pszBuffer = pszNew;
	return str;
}

} // namespace chustd

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool String::IsBlankChar(uint32 nCodePoint)
{
	return (nCodePoint == ' ' || nCodePoint == '\t' || nCodePoint == '\n' || nCodePoint == '\r');
}

// http://www.w3.org/TR/2004/REC-xml-20040204/
bool String::IsBaseChar(uint32 nCodePoint)
{
	static const SCharRange pg00[] = 
	{ 
		{0x0041, 0x005A}, {0x0061, 0x007A}, {0x00C0, 0x00D6}, {0x00D8, 0x00F6}, {0x00F8, 0x00FF}, {0}	
	};
	static const SCharRange pg01[] = 
	{
		{0x0100, 0x0131}, {0x0134, 0x013E}, {0x0141, 0x0148}, {0x014A, 0x017E}, {0x0180, 0x01C3},
		{0x01CD, 0x01F0}, {0x01F4, 0x01F5}, {0x01FA, 0x0217}, {0}
	};
	static const SCharRange pg02[] =
	{
		{0x0250, 0x02A8}, {0x02BB, 0x02C1}, {0}
	};
	static const SCharRange pg03[] =
	{
		{0x0386, 0x0386}, {0x0388, 0x038A}, {0x038C, 0x038C}, {0x038E, 0x03A1}, {0x03A3, 0x03CE},
		{0x03D0, 0x03D6}, {0x03DA, 0x03DA}, {0x03DC, 0x03DC}, {0x03DE, 0x03DE}, {0x03E0, 0x03E0},
		{0x03E2, 0x03F3}, {0}
	};
	static const SCharRange pg04[] = 
	{
		{0x0401, 0x040C}, {0x040E, 0x044F}, {0x0451, 0x045C}, {0x045E, 0x0481}, {0x0490, 0x04C4},
		{0x04C7, 0x04C8}, {0x04CB, 0x04CC}, {0x04D0, 0x04EB}, {0x04EE, 0x04F5}, {0x04F8, 0x04F9}, {0}
	};
	static const SCharRange pg05[] = 
	{
		{0x0531, 0x0556}, {0x0559, 0x0559}, {0x0561, 0x0586}, {0x05D0, 0x05EA}, {0x05F0, 0x05F2}, {0}
	};
	static const SCharRange pg06[] = 
	{
		{0x0621, 0x063A}, {0x0641, 0x064A}, {0x0671, 0x06B7}, {0x06BA, 0x06BE}, {0x06C0, 0x06CE},
		{0x06D0, 0x06D3}, {0x06D5, 0x06D5}, {0x06E5, 0x06E6}, {0}
	};
	static const SCharRange pg09[] = 
	{
		{0x0905, 0x0939}, {0x093D, 0x093D}, {0x0958, 0x0961}, {0x0985, 0x098C}, {0x098F, 0x0990},
		{0x0993, 0x09A8}, {0x09AA, 0x09B0}, {0x09B2, 0x09B2}, {0x09B6, 0x09B9}, {0x09DC, 0x09DD},
		{0x09DF, 0x09E1}, {0x09F0, 0x09F1}, {0}
	};
	static const SCharRange pg0A[] =
	{
		{0x0A05, 0x0A0A}, {0x0A0F, 0x0A10}, {0x0A13, 0x0A28}, {0x0A2A, 0x0A30}, {0x0A32, 0x0A33}, {0x0A35, 0x0A36},
		{0x0A38, 0x0A39}, {0x0A59, 0x0A5C}, {0x0A5E, 0x0A5E}, {0x0A72, 0x0A74}, {0x0A85, 0x0A8B}, {0x0A8D, 0x0A8D},
		{0x0A8F, 0x0A91}, {0x0A93, 0x0AA8},	{0x0AAA, 0x0AB0}, {0x0AB2, 0x0AB3}, {0x0AB5, 0x0AB9}, {0x0ABD, 0x0ABD},
		{0x0AE0, 0x0AE0}, {0}
	};
	static const SCharRange pg0B[] =
	{
		{0x0B05, 0x0B0C}, {0x0B0F, 0x0B10}, {0x0B13, 0x0B28}, {0x0B2A, 0x0B30}, {0x0B32, 0x0B33}, {0x0B36, 0x0B39},
		{0x0B3D, 0x0B3D}, {0x0B5C, 0x0B5D}, {0x0B5F, 0x0B61}, {0x0B85, 0x0B8A}, {0x0B8E, 0x0B90}, {0x0B92, 0x0B95},
		{0x0B99, 0x0B9A}, {0x0B9C, 0x0B9C}, {0x0B9E, 0x0B9F}, {0x0BA3, 0x0BA4}, {0x0BA8, 0x0BAA}, {0x0BAE, 0x0BB5},
		{0x0BB7, 0x0BB9}, {0}
	};
	static const SCharRange pg0C[] =
	{
		{0x0C05, 0x0C0C}, {0x0C0E, 0x0C10}, {0x0C12, 0x0C28}, {0x0C2A, 0x0C33}, {0x0C35, 0x0C39}, {0x0C60, 0x0C61},
		{0x0C85, 0x0C8C}, {0x0C8E, 0x0C90}, {0x0C92, 0x0CA8}, {0x0CAA, 0x0CB3}, {0x0CB5, 0x0CB9}, {0x0CDE, 0x0CDE},
		{0x0CE0, 0x0CE1}, {0}
	};
	static const SCharRange pg0D[] =
	{
		{0x0D05, 0x0D0C}, {0x0D0E, 0x0D10}, {0x0D12, 0x0D28}, {0x0D2A, 0x0D39}, {0x0D60, 0x0D61}, {0}
	};

	static const SCharRange pg0E[] =
	{
		{0x0E01, 0x0E2E}, {0x0E30, 0x0E30}, {0x0E32, 0x0E33}, {0x0E40, 0x0E45}, {0x0E81, 0x0E82}, {0x0E84, 0x0E84},
		{0x0E87, 0x0E88}, {0x0E8A, 0x0E8A}, {0x0E8D, 0x0E8D}, {0x0E94, 0x0E97}, {0x0E99, 0x0E9F}, {0x0EA1, 0x0EA3},
		{0x0EA5, 0x0EA5}, {0x0EA7, 0x0EA7}, {0x0EAA, 0x0EAB}, {0x0EAD, 0x0EAE}, {0x0EB0, 0x0EB0}, {0x0EB2, 0x0EB3},
		{0x0EBD, 0x0EBD}, {0x0EC0, 0x0EC4}, {0}
	};
	static const SCharRange pg0F[] =
	{
		{0x0F40, 0x0F47}, {0x0F49, 0x0F69}, {0}
	};
	static const SCharRange pg10[] =
	{
		{0x10A0, 0x10C5}, {0x10D0, 0x10F6}, {0}
	};
	static const SCharRange pg11[] =
	{
		{0x1100, 0x1100}, {0x1102, 0x1103}, {0x1105, 0x1107}, {0x1109, 0x1109}, {0x110B, 0x110C}, {0x110E, 0x1112},
		{0x113C, 0x113C}, {0x113E, 0x113E}, {0x1140, 0x1140}, {0x114C, 0x114C}, {0x114E, 0x114E}, {0x1150, 0x1150},
		{0x1154, 0x1155}, {0x1159, 0x1159}, {0x115F, 0x1161}, {0x1163, 0x1163}, {0x1165, 0x1165}, {0x1167, 0x1167},
		{0x1169, 0x1169}, {0x116D, 0x116E}, {0x1172, 0x1173}, {0x1175, 0x1175}, {0x119E, 0x119E}, {0x11A8, 0x11A8},
		{0x11AB, 0x11AB}, {0x11AE, 0x11AF}, {0x11B7, 0x11B8}, {0x11BA, 0x11BA}, {0x11BC, 0x11C2}, {0x11EB, 0x11EB},
		{0x11F0, 0x11F0}, {0x11F9, 0x11F9}, {0}
	};
	static const SCharRange pg1E[] =
	{
		{0x1E00, 0x1E9B}, {0x1EA0, 0x1EF9}, {0}
	};
	static const SCharRange pg1F[] =
	{
		{0x1F00, 0x1F15}, {0x1F18, 0x1F1D}, {0x1F20, 0x1F45}, {0x1F48, 0x1F4D}, {0x1F50, 0x1F57}, {0x1F59, 0x1F59},
		{0x1F5B, 0x1F5B}, {0x1F5D, 0x1F5D}, {0x1F5F, 0x1F7D}, {0x1F80, 0x1FB4}, {0x1FB6, 0x1FBC}, {0x1FBE, 0x1FBE},
		{0x1FC2, 0x1FC4}, {0x1FC6, 0x1FCC}, {0x1FD0, 0x1FD3}, {0x1FD6, 0x1FDB}, {0x1FE0, 0x1FEC}, {0x1FF2, 0x1FF4},
		{0x1FF6, 0x1FFC}, {0}
	};
	static const SCharRange pg21[] =
	{
		{0x2126, 0x2126}, {0x212A, 0x212B}, {0x212E, 0x212E}, {0x2180, 0x2182}, {0}
	};
	static const SCharRange pg30[] =
	{
		{0x3041, 0x3094}, {0x30A1, 0x30FA}, {0}
	};
	static const SCharRange pg31[] =
	{
		{0x3105, 0x312C}, {0}
	};
	
	static const SCharRange* const aPages[] = 
	{
		pg00, pg01, pg02, pg03, pg04, pg05, pg06, null,
		null, pg09, pg0A, pg0B, pg0C, pg0D, pg0E, pg0F,

		pg10, pg11,	null, null, null, null, null, null,
		null, null, null, null, null, null, pg1E, pg1F,
		
		null, null, null, null,	null, null, null, null,
		null, null, null, null,	null, null, null, null,

		pg30, pg31
	};
	
	uint8 nHi = uint8(nCodePoint >> 8);
	if( nHi > 0x31 )
	{
		if( 0xAC00 <= nCodePoint && nCodePoint <= 0xD7A3 )
			return true;

		return false;
	}

	// Use the pages table
	const SCharRange* const paPage = aPages[nHi];
	if( paPage == null )
		return false;
	
	int32 iRange = 0;
	while( paPage[iRange].nMin != 0 )
	{
		if( paPage[iRange].nMin <= nCodePoint && nCodePoint <= paPage[iRange].nMax )
			return true;

		++iRange;
	}

	return false;
}

bool String::IsIdeographic(uint32 nCodePoint)
{
	if( 0x4E00 <= nCodePoint && nCodePoint <= 0x9FA5 )
		return true;
	
	if( nCodePoint == 0x3007 )
		return true;
	
	if( 0x3021 <= nCodePoint && nCodePoint <= 0x3029 )
		return true;

	return false;
}

bool String::IsCombiningChar(uint32 nCodePoint)
{
	if( nCodePoint < 0x0300 || nCodePoint > 0x309A )
		return false;

	static const SCharRange aRanges0[] = 
	{
		{0x0300, 0x0345}, {0x0360, 0x0361}, {0x0483, 0x0486}, {0x0591, 0x05A1}, {0x05A3, 0x05B9},
		{0x05BB, 0x05BD}, {0x05BF, 0x05BF},
		{0x05C1, 0x05C2}, {0x05C4, 0x05C4},
		{0x064B, 0x0652}, {0x0670, 0x0670},
		{0x06D6, 0x06DC},
		{0x06DD, 0x06DF},
		{0x06E0, 0x06E4},
		{0x06E7, 0x06E8},
		{0x06EA, 0x06ED},
		{0x0901, 0x0903}, {0x093C, 0x093C}, 
		{0x093E, 0x094C}, {0x094D, 0x094D}, 
		{0x0951, 0x0954}, 
		{0x0962, 0x0963}, 
		{0x0981, 0x0983}, {0x09BC, 0x09BC}, {0x09BE, 0x09BE}, {0x09BF, 0x09BF}, 
		{0x09C0, 0x09C4},
		{0x09C7, 0x09C8}, 
		{0x09CB, 0x09CD}, {0x09D7, 0x09D7}, 
		{0x09E2, 0x09E3}, {0x0A02, 0x0A02}, {0x0A3C, 0x0A3C}, {0x0A3E, 0x0A3E}, {0x0A3F, 0x0A3F},
		{0}
	};

	if( nCodePoint <= 0x0A3F )
	{
		int32 iRange = 0;
		while( aRanges0[iRange].nMin != 0 )
		{
			if( aRanges0[iRange].nMin <= nCodePoint && nCodePoint <= aRanges0[iRange].nMax )
				return true;

			++iRange;
		}
		return false;
	}

	static const SCharRange aRanges1[] =
	{
		{0x0A40, 0x0A42}, 
		{0x0A47, 0x0A48}, 
		{0x0A4B, 0x0A4D}, 
		{0x0A70, 0x0A71}, 
		{0x0A81, 0x0A83}, {0x0ABC, 0x0ABC},
		{0x0ABE, 0x0AC5}, 
		{0x0AC7, 0x0AC9}, 
		{0x0ACB, 0x0ACD}, 
		{0x0B01, 0x0B03}, {0x0B3C, 0x0B3C}, 
		{0x0B3E, 0x0B43},
		{0x0B47, 0x0B48}, 
		{0x0B4B, 0x0B4D}, 
		{0x0B56, 0x0B57}, 
		{0x0B82, 0x0B83}, 
		{0x0BBE, 0x0BC2},
		{0x0BC6, 0x0BC8}, 
		{0x0BCA, 0x0BCD}, {0x0BD7, 0x0BD7},
		{0x0C01, 0x0C03}, 
		{0x0C3E, 0x0C44}, 
		{0x0C46, 0x0C48},
		{0x0C4A, 0x0C4D}, 
		{0x0C55, 0x0C56}, 
		{0x0C82, 0x0C83}, 
		{0x0CBE, 0x0CC4}, 
		{0x0CC6, 0x0CC8},
		{0x0CCA, 0x0CCD}, 
		{0x0CD5, 0x0CD6},
		{0}
	};

	if( nCodePoint <= 0x0CD6 )
	{
		int32 iRange = 0;
		while( aRanges1[iRange].nMin != 0 )
		{
			if( aRanges1[iRange].nMin <= nCodePoint && nCodePoint <= aRanges1[iRange].nMax )
				return true;

			++iRange;
		}
		return false;
	}

	static const SCharRange aRanges2[] =
	{
		{0x0D02, 0x0D03}, 
		{0x0D3E, 0x0D43}, 
		{0x0D46, 0x0D48}, 
		{0x0D4A, 0x0D4D}, {0x0D57, 0x0D57}, {0x0E31, 0x0E31}, 
		{0x0E34, 0x0E3A}, 
		{0x0E47, 0x0E4E}, {0x0EB1, 0x0EB1}, 
		{0x0EB4, 0x0EB9},
		{0x0EBB, 0x0EBC}, 
		{0x0EC8, 0x0ECD}, 
		{0x0F18, 0x0F19}, {0x0F35, 0x0F35}, {0x0F37, 0x0F37}, {0x0F39, 0x0F39}, {0x0F3E, 0x0F3E}, {0x0F3F, 0x0F3F}, 
		{0x0F71, 0x0F84}, 
		{0x0F86, 0x0F8B}, 
		{0x0F90, 0x0F95}, {0x0F97, 0x0F97}, 
		{0x0F99, 0x0FAD},
		{0x0FB1, 0x0FB7}, {0x0FB9, 0x0FB9}, 
		{0x20D0, 0x20DC}, {0x20E1, 0x20E1}, 
		{0x302A, 0x302F}, {0x3099, 0x3099}, {0x309A, 0x309A},
		{0}
	};

	int32 iRange = 0;
	while( aRanges2[iRange].nMin != 0 )
	{
		if( aRanges2[iRange].nMin <= nCodePoint && nCodePoint <= aRanges2[iRange].nMax )
			return true;

		++iRange;
	}
	return false;
}

bool String::IsDigit(uint32 nCodePoint)
{
	if( nCodePoint <= 255 )
	{
		if( 0x0030 <= nCodePoint && nCodePoint <= 0x0039 )
			return true;
		return false;
	}
	
	if( 0x0660 <= nCodePoint && nCodePoint <= 0x0669 )
		return true;
	if( 0x06F0 <= nCodePoint && nCodePoint <= 0x06F9 )
		return true;
	if( 0x0966 <= nCodePoint && nCodePoint <= 0x096F )
		return true;
	if( 0x09E6 <= nCodePoint && nCodePoint <= 0x09EF )
		return true;
	if( 0x0A66 <= nCodePoint && nCodePoint <= 0x0A6F )
		return true;
	if( 0x0AE6 <= nCodePoint && nCodePoint <= 0x0AEF )
		return true;
	if( 0x0B66 <= nCodePoint && nCodePoint <= 0x0B6F )
		return true;
	if( 0x0BE7 <= nCodePoint && nCodePoint <= 0x0BEF )
		return true;
	if( 0x0C66 <= nCodePoint && nCodePoint <= 0x0C6F )
		return true;
	if( 0x0CE6 <= nCodePoint && nCodePoint <= 0x0CEF )
		return true;
	if( 0x0D66 <= nCodePoint && nCodePoint <= 0x0D6F )
		return true;
	if( 0x0E50 <= nCodePoint && nCodePoint <= 0x0E59 )
		return true;
	if( 0x0ED0 <= nCodePoint && nCodePoint <= 0x0ED9 )
		return true;
	if( 0x0F20 <= nCodePoint && nCodePoint <= 0x0F29 )
		return true;
	
	return false;
}

bool String::IsExtenderChar(uint32 nCodePoint)
{
	if( nCodePoint == 0x00B7 || nCodePoint == 0x02D0 
		|| nCodePoint == 0x02D1 || nCodePoint == 0x0387 || nCodePoint== 0x0640
		|| nCodePoint == 0x0E46 || nCodePoint == 0x0EC6 || nCodePoint == 0x3005 )
		return true;
	
	if( (0x3031 <= nCodePoint && nCodePoint <= 0x3035) 
		|| (0x309D <= nCodePoint && nCodePoint <= 0x309E) 
		|| (0x30FC <= nCodePoint && nCodePoint <= 0x30FE))
		return true;

	return false;
}

bool String::IsBinaryEqual(const wchar* psz) const
{
	const int32 nThisLength = GetData()->GetLength();
	const int32 nLength = (psz != null) ? SZLength(psz) : 0;
	
	if( nThisLength == 0 && nLength == 0 )
		return true;

	if( nThisLength != nLength )
		return false;

	if( Memory::Is32BitsAligned(psz) )
	{	
		// Multiple of 4
		
		// First we compare each block
		int32 i = 0;
		for(; i < nLength - 3; i += 4)
		{
			const uint32* pBuf1 = (uint32*) (m_pszBuffer + i);
			const uint32* pBuf2 = (uint32*) (psz + i);

			if( pBuf1[0] != pBuf2[0] )
			{
				return false;
			}
		}
		
		// Then we compare the remaining bytes
		for(; i < nLength; ++i)
		{
			const uint16* pBuf1 = (uint16*) (m_pszBuffer + i);
			const uint16* pBuf2 = (uint16*) (psz + i);

			if( pBuf1[0] != pBuf2[0] )
			{
				return false;
			}
		}
		return true;
	}

	for(int32 i = 0; i < nLength; ++i)
	{
		const uint16* pBuf1 = (uint16*) (m_pszBuffer + i);
		const uint16* pBuf2 = (uint16*) (psz + i);

		if( pBuf1[0] != pBuf2[0] )
		{
			return false;
		}
	}
	return true;
}

bool String::IsBinaryEqual(const String& str) const
{
	// Fast test
	const uint16* pBuf1 = m_pszBuffer;
	const uint16* pBuf2 = str.m_pszBuffer;

	if( pBuf1 == pBuf2 )
		return true;

	const int32 nLength1 = GetData()->GetLength();
	const int32 nLength2 = str.GetData()->GetLength();

	if( nLength1 != nLength2 )
		return false;
	
	// Comparison one chunk (which contains several characters) per iteration
	const int32 nChunkCount = StringData::GetUsedChunkCount(nLength1);

	// Note : we do not loop up to the final uint64 because the unused chars may be
	// different from the two strings
	// Example : 
	// String1 : ABCDEFGH IJ0ZZZZZ
	// String2 : ABCDEFGH KL0WWWWW
	// The two strings are equal but the unused chars (Z and W) are different
	
	const int32 nInc = 8 / sizeof(uint16);

	for(int32 i = 0; i < nChunkCount - 1; ++i)
	{
		const uint64 n1 = ((const uint64*) pBuf1)[0];
		const uint64 n2 = ((const uint64*) pBuf2)[0];
		if( n1 != n2 )
		{
			return false;
		}	
		pBuf1 += nInc;
		pBuf2 += nInc;
	}

	// Final characters
	while( pBuf1[0] != 0 )
	{
		if( pBuf1[0] != pBuf2[0] )
		{
			return false;
		}
		pBuf1++;
		pBuf2++;
	}

	return true;
}

bool String::IsBinaryEqualPartially(int32 nStartAt, const String& str) const
{
	// Fast test
	const uint16* pBuf1 = m_pszBuffer + nStartAt;
	const uint16* pBuf2 = str.m_pszBuffer;

	if( pBuf1 == pBuf2 )
	{
		return true;
	}

	const int32 nLength1 = GetData()->GetLength();
	const int32 nLength2 = str.GetData()->GetLength();

	if( nStartAt + nLength2 > nLength1 )
		return false;
	
	// TODO: Darken, please use your powerful mind to code 
	// a chunk optimized version of this one

	// Don't check \0 at end of str !
	for( int i = 0; i < nLength2; ++i )
	{
		if( pBuf1[0] != pBuf2[0] )
		{
			return false;
		}
		pBuf1++;
		pBuf2++;
	}

	return true;	
}

bool String::StartsWith(const String& str) const
{
	return IsBinaryEqualPartially(0, str);
}

bool String::EndsWith(const String& str) const
{
	const int nStartAt = GetLength() - str.GetLength();
	return IsBinaryEqualPartially(nStartAt, str);
}


String String::Right(int32 nCount) const
{
	if( nCount < 0 )
	{
		nCount = 0;
	}

	const int32 nLength = GetData()->GetLength();
	if( nCount > nLength )
	{
		nCount = nLength;
	}

	uint16* pszNew = StringData::CreateInstance(nCount);
	const uint16* const pszSource = m_pszBuffer + (nLength - nCount);
	Memory::Copy16(pszNew, pszSource, nCount);

	String strOut;
	strOut.m_pszBuffer = pszNew;
	return strOut;
}

String String::Left(int32 nCount) const
{
	if( nCount < 0 )
	{
		nCount = 0;
	}

	const int32 nLength = GetData()->GetLength();
	if( nCount > nLength )
	{
		nCount = nLength;
	}

	uint16* pszNew = StringData::CreateInstance(nCount);
	const uint16* const pszSource = m_pszBuffer;
	
	Memory::Copy16(pszNew, pszSource, nCount);

	String strOut;
	strOut.m_pszBuffer = pszNew;
	return strOut;
}

String String::Mid(int32 nFirst, int32 nCount) const
{
	// The first position can be equal to or greater than the length of the string
	// In that case you get an empty string
	if( nFirst >= GetLength() )
	{
		return String();
	}

	if( nCount <= 0 )
	{
		return String();
	}

	const int32 nLength = GetData()->GetLength();
	if( nFirst + nCount > nLength )
	{
		nCount = nLength - nFirst;
	}

	uint16* pszNew = StringData::CreateInstance(nCount);
	const uint16* const pszSource = m_pszBuffer + nFirst;
	
	Memory::Copy16(pszNew, pszSource, nCount);

	String strOut;
	strOut.m_pszBuffer = pszNew;
	return strOut;
}

bool String::ToInt32(int32& nVal, char cArg) const
{
	if( IsEmpty() )
		return false;

	const uint16* pszThis = GetData()->GetBuffer();
	int32 nAdvance = 0;
	
	bool bOk = ScanType( (wchar*)pszThis, nAdvance, cArg, &nVal);
	if( !bOk )
	{
		return false;
	}

	if( nAdvance < GetData()->GetLength() )
	{
		return false;
	}

	return true;
}

bool String::ToBool(bool& bVal) const
{
	int32 nVal = 0;
	bool bRet = ToInt32(nVal);
	if( !bRet )
	{
		return false;
	}
	bVal = nVal != 0;
	return true;
}

bool String::ToFloat(float32& fVal, bool bStrict) const
{
	float64 fVal64 = 0;
	bool bRes = ToFloat(fVal64, bStrict);
	fVal = float32(fVal64);
	return bRes;
}

bool String::ToFloat(float64& fVal, bool bStrict) const
{
	if( IsEmpty() )
		return false;

	const uint16* pszThis = GetData()->GetBuffer();
	int32 nAdvance = 0;
	bool bOk = ScanType( (wchar*)pszThis, nAdvance, 0, &fVal);
	if( !bOk )
		return false;

	if( bStrict && nAdvance < GetData()->GetLength() )
		return false;

	return true;
}


StringArray String::SplitByEndlines() const
{
	StringArray astrResults;
	astrResults.SetAsReturnValue();

	Int32Array anResults;
	const int32 nLength = GetData()->GetLength();
	if( nLength == 0 )
		return astrResults;

	const uint16* pszThis = GetData()->GetBuffer();
	
	int32 nCutStartAt = 0; // Next string first character index
	int32 iCutStopAt = 0;  // Current character index

	// Prepare slot for the first string
	anResults.Add(0);

	for(; iCutStopAt < nLength; iCutStopAt++)
	{
		uint16 c = pszThis[iCutStopAt];
		uint16 cNext = pszThis[iCutStopAt + 1];

		if( (c == '\n') || (c == '\r' && cNext != '\n') )
		{
			// Carriage return made of only one character : \n or \r
			anResults.Add(iCutStopAt);
			
			nCutStartAt = iCutStopAt + 1;
			anResults.Add(nCutStartAt);

		}
		else if( c == '\r' && cNext == '\n' )
		{
			// Carriage return made of two characters : \r\n
			anResults.Add(iCutStopAt);

			iCutStopAt++; // Jump over '\n' in the loop
			nCutStartAt = iCutStopAt + 1;
			anResults.Add(nCutStartAt);
		}
	}

	anResults.Add(iCutStopAt);

	const int32 nRangeCount = anResults.GetSize();
	ASSERT( (nRangeCount & 1) == 0); // Should be even

	astrResults.SetSize(nRangeCount / 2);
	for(int32 i = 0; i < nRangeCount; i += 2)
	{
		int32 nStart = anResults[i];
		int32 nStop = anResults[i + 1];

		astrResults[i / 2].SetBuffer( (wchar*)(pszThis + nStart), nStop - nStart);
	}

	return astrResults;
}


StringArray String::Split(uint32 nCodePoint) const
{
	StringArray astrResults;
	astrResults.SetAsReturnValue();

	const int32 nLength = GetData()->GetLength();
	if( nLength == 0 )
		return astrResults;

	uint16 nFirst, nNext;
	int32 nCodeUnitCount = CodePointToCodeUnits(nCodePoint, nFirst, nNext);
	
	if( nCodeUnitCount == 0 )
		return astrResults;
	
	const uint16* pszThis = GetData()->GetBuffer();
	int32 nPreviousCut = 0;

	for(int32 i = 0; i < nLength; ++i)
	{
		uint16 nValue = pszThis[i];
		if( nValue == nFirst )
		{
			// Found a unit !
			int32 nCutIndex = i;
			if( nNext != 0 )
			{
				if( i < (nLength - 1) && pszThis[i + 1] == nNext )
				{
					// Ok
					++i; // Jump over to count the second code unit
				}
				else
				{
					nCutIndex = -1; // Do not cut
				}
			}

			if( nCutIndex >= 0 )
			{
				int32 nPieceSize = nCutIndex - nPreviousCut;
				String strPiece = Mid(nPreviousCut, nPieceSize);
				astrResults.Add(strPiece);

				nPreviousCut = i + 1;
			}
		}
	}

	// Get the last piece
	int32 nCutIndex = nLength;
	int32 nPieceSize = nCutIndex - nPreviousCut;
	String strPiece = Mid(nPreviousCut, nPieceSize);
	astrResults.Add(strPiece);

	return astrResults;
}

StringArray String::Split(const String& str) const
{
	StringArray astrResults;
	astrResults.SetAsReturnValue();

	const int32 nLength = GetData()->GetLength();
	if( nLength == 0 )
		return astrResults;

	int32 nStartAt = 0;
	
	while(true)
	{
		int32 nPos = Find(str, nStartAt);
		if( nPos < 0 )
		{
			break;
		}

		String strLeft = Mid(nStartAt, nPos - nStartAt);
		astrResults.Add(strLeft);

		nStartAt = nPos + str.GetLength();
	}

	// Get the last piece
	if( nStartAt <= nLength )
	{
		String strLast = Mid(nStartAt);
		astrResults.Add(strLast);
	}

	return astrResults;
}

ByteArray String::ToBytes(const chustd::TextEncoding& te, bool bAddEndingZero) const
{
	ByteArray aBytes = te.StringToBytes(*this);

	if( bAddEndingZero )
	{
		aBytes.Add(0);
	}
	aBytes.SetAsReturnValue();
	return aBytes;
}


// static
String String::FromBytes(const ByteArray& aBytes, const chustd::TextEncoding& te)
{
	return te.BytesToString(aBytes);
}

// Finds a substring in the current string
// Returns >= 0 if found, -1 otherwise
int32 String::Find(const String& strToFind, int32 nStartAt) const
{
	ASSERT(nStartAt >= 0);

	const int32 nLength = GetData()->GetLength();

	if( nStartAt >= nLength )
		return -1;

	const int32 nToFindLength = strToFind.GetLength();
	if( nToFindLength > nLength || nToFindLength == 0 )
		return -1;
	
	const int32 nLastPos = nLength - nToFindLength + 1;
	
	const uint16* const pszToFind = strToFind.m_pszBuffer;
	const uint16* const pszBuffer = m_pszBuffer;
	for(int32 i = nStartAt; i < nLastPos; i++)
	{
		int32 j = 0;
		for(; j < nToFindLength; j++)
		{
			if( pszToFind[j] != pszBuffer[i + j] )
				break;
		}
		if( j == nToFindLength )
		{
			// Found !
			return i;
		}
	}
	return -1;
}


// Returns the index of the first occurrence of a one of the characters defined in strChars
// Returns >= 0 if found, -1 otherwise
int32 String::FindOneOf(const String& strChars, int32 nStartAt = 0) const
{
	const int32 nLength = GetData()->GetLength();
	const int32 nCharCount = strChars.GetLength();
	if( nCharCount == 0 )
	{
		return -1;
	}
	
	const uint16* const pszBuffer = m_pszBuffer;
	const uint16* const pszChars = strChars.m_pszBuffer;

	for(int32 i = nStartAt; i < nLength; ++i)
	{
		uint16 nSrc = pszBuffer[i];
		for(int32 iChar = 0; iChar < nCharCount; ++iChar)
		{
			if( nSrc == pszChars[iChar] )
			{
				return i;
			}
		}
	}
	return -1;
}

// Replaces every occurrence of strSrc with strNew
// TODO: can be optimized by removing too many memory allocations
String String::ReplaceAll(const String& strWhat, const String& strWith) const
{
	Int32Array aPositions;

	const int32 nLengthWhat = strWhat.GetLength();
	const int32 nLengthWith = strWith.GetLength();

	int32 nStartAt = 0;
	while(true)
	{			
		// Find the string to replace
		int32 nFoundAt = Find( strWhat, nStartAt);
		if( nFoundAt < 0 )
		{
			break;
		}
		aPositions.Add(nFoundAt);

		nStartAt = nFoundAt + nLengthWhat;
	}
	const int32 nReplaceCount = aPositions.GetSize();
	if( nReplaceCount == 0 )
	{
		return *this;
	}

	const int32 nOldLength = GetLength();
	const int32 nNewLength = GetLength() - (nReplaceCount * (nLengthWhat - nLengthWith));
	
	// Create a new StringData for our own use
	uint16* const pszNewDataBuffer = StringData::CreateInstance(nNewLength);
	
	uint16* pSrc = m_pszBuffer;
	uint16* pDst = pszNewDataBuffer;

	const uint16* pszWith = strWith.GetData()->GetBuffer();

	int32 nSrcStart = 0;
	for(int32 i = 0; i < nReplaceCount; ++i)
	{
		int32 nSrcStop = aPositions[i];
		int32 nSegmentSize = nSrcStop - nSrcStart;

		Memory::Copy(pDst, pSrc + nSrcStart, nSegmentSize * sizeof(uint16));
		pDst += nSegmentSize;

		Memory::Copy(pDst, pszWith, nLengthWith * sizeof(uint16));
		pDst += nLengthWith;

		nSrcStart = nSrcStop + nLengthWhat;
	}

	// Copy last segment
	int32 nSegmentSize = nOldLength - nSrcStart;
	Memory::Copy(pDst, pSrc + nSrcStart, nSegmentSize * sizeof(uint16));
	
	String strOut;
	strOut.m_pszBuffer = pszNewDataBuffer;
	return strOut;
}

// Returns a copy of the string without both leading and trailing blank chars
String String::Trim() const
{
	const int32 nLength = GetLength();
	const uint16* pszThis = GetData()->GetBuffer();

	int32 iStart = 0;
	for(; iStart < nLength; ++iStart)
	{
		uint16 nChar = uint16(pszThis[iStart]);
		if( !IsBlankChar(nChar) )
			break;
	}

	int32 iStop = nLength - 1;
	for(; iStop >= iStart; --iStop)
	{
		uint16 nChar = uint16(pszThis[iStop]);
		if( !IsBlankChar(nChar) )
			break;
	}

	if( iStart == 0 && iStop == nLength - 1 )
	{
		// Just returns a copy
		return *this;
	}

	int32 nNewLength = (iStop - iStart) + 1;

	uint16* pszNew = StringData::CreateInstance(nNewLength);
	Memory::Copy16(pszNew, pszThis + iStart, nNewLength);

	String strOut;
	strOut.m_pszBuffer = pszNew;
	return strOut;
}

// Returns the number of code units assigned
// Returns 0 if nCodePoint is out of range
int32 String::CodePointToCodeUnits(uint32 nCodePoint, uint16& nUtf16First, uint16& nUtf16Next)
{
	if( nCodePoint < 0x0000FFFF )
	{
		nUtf16First = uint16(nCodePoint);
		nUtf16Next = 0;
		return 1;
	}
	else if( nCodePoint <= 0x0010FFFF )
	{
		// U+10000..U+10FFFF stored in 
		// U+D800..U+DBFF (high surrogate)
		// U+DC00..U+DFFF (low surrogate)
		uint16 nHi = uint16( (nCodePoint - 0x10000) / 0x0400 + 0xD800);
		uint16 nLo = uint16( (nCodePoint - 0x10000) % 0x0400 + 0xDC00);

		nUtf16First = nHi;
		nUtf16Next = nLo;
		return 2;
	}
	
	return 0;
}

// Returns 0xffffffff upon error
uint32 String::GetCodePoint(int32& nStart) const
{
	const int32 nLength = GetLength();

	uint32 nInvalidCodePoint = 0xffffffff;

	ASSERT(nStart >= 0);
	
	if( !(0 <= nStart && nStart < nLength) )
	{
		return nInvalidCodePoint;
	}

	uint16 nHi = m_pszBuffer[nStart];
	nStart += 1;

	// A surrogate ?
	if( nHi < 0xD800 || nHi > 0xDFFF )
	{
		// Nop, standard code point
		return nHi;
	}

	if( 0xDBFF < nHi )
	{
		// Error ! Low surrogate used first, this is an error
		return nInvalidCodePoint;
	}

	/////////////////////////////////////////////
	// Get the second surrogate (low)
	if( nStart >= nLength )
	{
		// Error ! Unexpected end of string
		return nInvalidCodePoint;
	}
	uint16 nLo = m_pszBuffer[nStart];
	nStart += 1;

	// A low surrogate ?
	if( !(0xDBFF < nLo && nLo <= 0xDFFF) )
	{
		// Error ! We expected a low surrogate
		return nInvalidCodePoint;
	}

	uint32 nCodePoint = 0x10000 + ((nHi - 0xD800) << 10) + (nLo - 0xDC00);
	return nCodePoint;
}

// static
String String::FromInt(int32 nVal, char cArg)
{
	wchar szBuffer[40];

	if( cArg == 'x' )
	{
		FormatUInt32Hex( uint32(nVal), szBuffer);
	}
	else if( cArg == 'b' )
	{
		FormatUInt32Bin( uint32(nVal), szBuffer);
	}
	else
	{
		FormatInt32(nVal, szBuffer);
	}

	return String(szBuffer);
}

String String::FromInt(int32 nVal, char cArg, int8 nWidth, wchar cFill)
{
	wchar szBuffer[40];

	if( cArg == 'x' )
	{
		FormatUInt32Hex( uint32(nVal), szBuffer, nWidth, cFill);
	}
	else if( cArg == 'b' )
	{
		FormatUInt32Bin( uint32(nVal), szBuffer, nWidth, cFill);
	}
	else
	{
		FormatInt32(nVal, szBuffer, nWidth, cFill);
	}

	return String(szBuffer);
}

// static
String String::FromInt64(int64 nVal)
{
	wchar szBuffer[40];

	FormatInt64(nVal, szBuffer);

	return String(szBuffer);
}

String String::FromAsciiSZ(const char* psz)
{
	return String(psz);
}

wchar* String::GetUnsafeBuffer(int32 nWantedLength)
{
	Unref();

	uint16* pszNew = StringData::CreateInstance(nWantedLength);
	if( pszNew == null )
	{
		m_pszBuffer = StringData::GetNullInstance()->GetBuffer();
		return null;
	}
	
	m_pszBuffer = pszNew;
	return (wchar*)pszNew;
}

String String::InsertAt(int32 nIndex, const String& str) const
{
	if( nIndex < 0 )
	{
		ASSERT(0);
		return String();
	}

	const int32 nLength = GetLength();

	if( nIndex > nLength )
	{
		ASSERT(0);
		return String();
	}

	// nIndex == nLength is allowed and is equivalent to an append

	const uint16* pszThis = GetData()->GetBuffer();
		
	const uint16* pszOther = str.GetData()->GetBuffer();
	const int32 nOtherLength = str.GetLength();

	uint16* pszNew = StringData::CreateInstance(nLength + nOtherLength);
	if( pszNew == null )
	{
		return String();
	}
	
	Memory::Copy16(pszNew, pszThis, nIndex);
	Memory::Copy16(pszNew + nIndex, pszOther, nOtherLength);
	Memory::Copy16(pszNew + nIndex + nOtherLength, pszThis + nIndex, nLength - nIndex);

	String strOut;
	strOut.m_pszBuffer = pszNew;
	return strOut;
}

String String::Pad(int32 nLength, uint32 nCodePoint) const
{
	const int32 nThisLength = GetLength();
	if( nLength <= nThisLength )
		return *this;

	uint16 nUtf16First, nUtf16Next;
	int32 nCodeUnitCount = CodePointToCodeUnits(nCodePoint, nUtf16First, nUtf16Next);
	if( nCodeUnitCount == 0 )
	{
		return String(); // Failed
	}

	// TODO: should we increase length to avoid cutting a surrogate pair ?
	// TODO: should nLength be the code unit length or the code point length ?

	uint16* pszThis = m_pszBuffer;
	uint16* pszNew = StringData::CreateInstance(nLength);
	if( pszNew == null )
	{
		return String(); // Failed
	}
	
	Memory::Copy16(pszNew, pszThis, nThisLength);
	if( nCodeUnitCount == 1 )
	{
		Memory::Set16(pszNew + nThisLength, nUtf16First, nLength - nThisLength);
	}
	else
	{
		for(int32 i = 0; i < (nLength - nThisLength); ++i)
		{
			if( (i & 0x01) == 0 )
			{
				pszNew[nLength + i] = nUtf16First;
			}
			else
			{
				pszNew[nLength + i] = nUtf16Next;
			}
		}
	}

	String strOut;
	strOut.m_pszBuffer = pszNew;
	return strOut;
}
