///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#ifndef CHUSTD_STRING_H
#define CHUSTD_STRING_H

namespace chustd {

class StringData;
class ByteArray;
class TextEncoding;

class String;
class StringArray;

// Immutable UTF-16 string
class String  
{
public:
	String();
	String(const char* psz); // Encoding = latin-1
	String(const wchar* psz);
	String(const wchar* psz, int32 nLength);
	String(const String& str);
			
	~String();

	// Makes the string empty - Frees all characters 
	void Empty();

	// Returns true if the string is empty
	bool IsEmpty() const {	return ( m_pszBuffer[0] == 0); }

	// Gets the string length in number of UTF-16 code units
	int32 GetLength() const;

	static int32 SZLength(const wchar* psz);
	static int32 SZLength(const char* psz);

	// Gets a code unit at a specific position
	uint16 GetAt(int32 nIndex) const;

	// Gets a code point at a specific position and increment the index
	uint32 GetCodePoint(int32& nStart) const;

	// http://www.unicode.org/faq/casemap_charprop.html
	
	// Returns a lowercase version of the string
	String ToLowerCase() const;
	
	// Returns an uppercase version of the string
	String ToUpperCase() const;

	// Initializes a string with an external string
	String& operator = (const String& str);
	//String& operator = (const wchar* psz);
	String& operator = (uint32 nCodePoint);

	// Converts the string to an atomic type
	// cArg : 'x' for hexa, 'b' for binary
	bool ToInt32(int32& nVal, char cArg = 0) const;
	bool ToBool(bool& bVal) const;
	bool ToFloat(float32& fVal, bool bStrict=true) const; // bStrict : no garbage chars after the float
	bool ToFloat(float64& fVal, bool bStrict=true) const; // bStrict : no garbage chars after the float

	// Converts an int to a string
	// cArg : 'x' for hexa, 'b' for binary
	static String FromInt(int32 nVal, char cArg = 0);
	static String FromInt(int32 nVal, char cArg, int8 nWidth, wchar cFill = ' ');
	static String FromInt64(int64 nVal);

	bool IsBinaryEqual(const wchar* psz) const;
	bool IsBinaryEqual(const String& str) const;

	bool IsBinaryEqualPartially(int32 nStartAt, const String& str) const;

	bool StartsWith(const String& str) const;
	bool EndsWith(const String& str) const;

	// Conversion operator (never returns null)
	const wchar* GetBuffer() const { return (wchar*)m_pszBuffer; }

	// Sets the content of the string with a buffer
	// nLength gives the number of character (final 0 not included)
	void SetBuffer(const wchar* pBuffer, int32 nLength);

	// Gets the internal buffer
	wchar* GetUnsafeBuffer(int32 nWantedLength);

	// Gets a copy of the left, right or middle part of the current string
	String Left(int32 nCount) const;
	String Right(int32 nCount) const;
	String Mid(int32 nFirst, int32 nCount) const;
	String Mid(int32 nFirst) const { return Mid(nFirst, GetLength() - nFirst); }

	// Cats two strings
	friend const String operator + (const String& str1, const String& str2);
	friend const String operator + (const String& str1, const wchar* psz);
	friend const String operator + (const wchar* psz, const String& str2);

	static bool IsBlankChar(uint32 nCodePoint);
	static bool IsBaseChar(uint32 nCodePoint);
	static bool IsIdeographic(uint32 nCodePoint);
	static bool IsLetter(uint32 nCodePoint) { return IsBaseChar(nCodePoint) || IsIdeographic(nCodePoint); }

	static bool IsDigit(uint32 nCodePoint);
	static bool IsCombiningChar(uint32 nCodePoint);
	static bool IsExtenderChar(uint32 nCodePoint);

	// Splits a string by endlines : single \r and \n or \r\n couples
	StringArray SplitByEndlines() const;

	// Splits a string by a character
	StringArray Split(uint32 nCodePoint) const;

	// Splits a string by another string
	StringArray Split(const String& str) const;

	// Converts this String to an array of charset encoded characters
	ByteArray ToBytes(const chustd::TextEncoding& te, bool bAddEndingZero = false) const;

	// Creates a new string from an array of charset encoded characters
	static String FromBytes(const ByteArray& aBytes, const chustd::TextEncoding& te);
	
	// Creates a new string from a C string ended with 0
	static String FromAsciiSZ(const char* psz);

	// Returns the index of the first occurrence of a string
	// Returns -1 if nothing found
	int32 Find(const String& strToFind, int32 nStartAt) const;

	// Returns the index of the first occurrence of a one of the characters defined in strChars
	// Returns -1 if nothing found
	int32 FindOneOf(const String& strChars, int32 nStartAt) const;

	// Replaces all occurrences of a string with another string
	String ReplaceAll(const String& strWhat, const String& strWith) const;

	// Returns a copy of the string without both leading and trailing blank chars
	String Trim() const;

	// Returns a string right filled with characters up to nLength
	String Pad(int32 nLength, uint32 nCodePoint) const;

	// Inserts a string at a specific position
	String InsertAt(int32 nIndex, const String& str) const;

	// Converts an Unicode code point into one or two UTF-16 code units
	// If the code point can fit in one uint16 only, then nUtf16First is filled and nUtf16Next is set to 0
	// Returns the number of assigned code units
	static int32 CodePointToCodeUnits(uint32 nCodePoint, uint16& nUtf16First, uint16& nUtf16Next);

private:
	// Points actually on StringData::m_szBuffer
	// so we keep 99% compatibility with C strings :-)
	// This pointer is never null
	uint16* m_pszBuffer;

private:
	void Unref();

	// Gets a CStringData pointer from the StringData sz buffer of the String instance
	const chustd::StringData* GetData() const;
	chustd::StringData* GetData();

	// Computes the difference between two addresses, saturates and asserts if the length
	// found is greater to the max length (0x7fffffff)
	static int32 DiffPtr(const void* pA, const void* pB, int32 nCharSize);

private:
	struct SCharRange { uint16 nMin, nMax; };
};

inline bool operator == (const String& str, const wchar* psz)
{
	return str.IsBinaryEqual(psz);
}
inline bool operator == (const String& str1, const String& str2)
{
	return str1.IsBinaryEqual(str2);
}

inline bool operator != (const String& str, const wchar* psz)
{
	return !str.IsBinaryEqual(psz);
}
inline bool operator != (const String& str1, const String& str2)
{
	return !str1.IsBinaryEqual(str2);
}

} // namespace chustd

#endif // ndef CHUSTD_STRING_H
