///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#ifndef CHUSTD_STRINGDATA_H
#define CHUSTD_STRINGDATA_H

// Private header

#include "Memory.h"
#include "Math.h"

namespace chustd {

class StringData
{
public:
	static uint16* CreateInstance(int32 nWantedLength);
	static uint16* CreateInstance(StringData* pData);

	uint16* GetBuffer() { return m_szBuffer; }
	const uint16* GetBuffer() const { return m_szBuffer; }
	int32 GetLength() const { return m_nLength; }
	void AddRef() { ASSERT(m_nRef != MAX_INT32); m_nRef++;}
	void Release() { ASSERT(m_nRef != MAX_INT32); m_nRef--;}
	int32 GetRef() const { return m_nRef; }

	void SetLength(int32 nLength) { ASSERT(m_nRef != MAX_INT32); m_nLength = nLength; }
	void Delete() {	uint64* pThis = (uint64*) this;	Memory::Free64(pThis);	}

	// Returns the number of chunks allocated for the char buffer
	// Each chunk is an uint64
	static inline int32 GetAllocatedChunkCount(int32 nLength)
	{
		// Counting the ending 0
		const int32 nSizeInBytes = (nLength + 1) * sizeof(uint16);
		
		// Number of needed blocks. Each block is 64 bytes.
		const int32 nBlockCount = chustd::Math::DivCeil(nSizeInBytes, 64);
		
		// Number of needed chunks. Each chunk is 8 bytes.
		const int32 nChunkCount = nBlockCount * (64 / 8);
		
		return nChunkCount;
	}

	static inline int32 GetAllocatedCharCount(int32 nLength)
	{
		return GetAllocatedChunkCount(nLength) * 8;
	}

	static inline int32 GetUsedChunkCount(int32 nLength)
	{
		int32 nSizeInBytes = (nLength + 1) * sizeof(uint16);
		return Memory::ByteCountToInt64Count(nSizeInBytes);
	}

	static inline StringData* GetInstance(uint16* psz)
	{
		ASSERT(psz != null);
		uint64* pBuf64 = (uint64*) psz;
		pBuf64 -= 1;
		
		StringData* pInstance = (StringData*) pBuf64;
		
		ASSERT(pInstance->m_nRef >= 0);
		ASSERT(pInstance->m_nLength >= 0);

		return pInstance;
	}

	static StringData* GetNullInstance()
	{
		static int32 nullData[] = {0, MAX_INT32, 0};
		return (StringData*) nullData;
	}

public:
	// Only for ms_nullData. See StringTemplate.h
	StringData(int32) : m_nLength(0), m_nRef(MAX_INT32) { m_szBuffer[0] = 0; }
private:

	StringData() {}

private:
	int32 m_nLength;     // Character count (0 not included)
	int32 m_nRef;        // Reference counter for this StringData
	uint16 m_szBuffer[1]; // C string, always ends with a 0
};


} // namespace chustd

#endif // ndef CHUSTD_STRINGDATA_H
