///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "Master.h"
#include "Surface32.h"
#include "Surface8.h"
#include "String.h"
#include "Font.h"
#include "Math.h"

#include "SurfaceBaseImpl.h"
#include "Array.h"

namespace chustd {

class Surface32Impl : public chustd::SurfaceBaseImpl
{
public:
	Color32 m_colConstant;
	BlendAlphaMode m_eBlendAlphaMode;

	Surface32Impl() : m_colConstant(0, 0, 0), m_eBlendAlphaMode(bamUseDstAlpha)
	{

	}
};
}

//////////////////////////////////////////////////////////////////////
using namespace chustd;
//////////////////////////////////////////////////////////////////////
Surface32::Surface32() : ReadOnlySurface32(new Surface32Impl())
{

}

Color32* Surface32::GetBuffer() const
{
	return (Color32*) ReadOnlySurface32::GetBuffer();
}

void Surface32::Clear(Color32 col) const
{
	const int32 nPitch = GetPitch();
	const int32 nWidth = GetWidth();
	const int32 nHeight = GetHeight();
	const int32 nNextRowOffset = nPitch - nWidth; 

	Color32* pPixels = GetBuffer();

	for( int32 iY = 0; iY < nHeight; ++iY )
	{
		for( int32 iX = 0; iX < nWidth; ++iX )
		{
			pPixels[0] = col;
			pPixels++;
		}
		pPixels += nNextRowOffset;
	}
}

void Surface32::Fill(Color32 col) const
{
	FillRect(0, 0, GetWidth(), GetHeight(), col);
}

void Surface32::FillVerticalGradient(Color32 col1, Color32 col2) const
{
	FillVerticalGradient(Rectangle(0, 0, GetWidth(), GetHeight()), col1, col2);
}

void Surface32::FillVerticalGradient(const Rectangle& rectUser, Color32 col1, Color32 col2) const
{
	// Translate the destination according to the surface internal offset
	Rectangle rect = rectUser;
	DoOffset(rect);

	Rectangle rectSurface(0, 0, GetWidth(), GetHeight());
	Rectangle rectClipped(rect);
	rectClipped.ClipIn(rectSurface);
	
	//int32 nRectX = rect.m_ptTopLeft.m_nX;
	int32 nRectY = rect.m_ptTopLeft.m_nY;
	//int32 nRectWidth = rect.m_dimension.m_nWidth;
	int32 nRectHeight = rect.m_dimension.m_nHeight;
	//int32 nRectX2 = rect.GetX2();
	//int32 nRectY2 = rect.GetY2();

	int32 nClipX = rectClipped.m_ptTopLeft.m_nX;
	int32 nClipY = rectClipped.m_ptTopLeft.m_nY;
	int32 nClipWidth = rectClipped.m_dimension.m_nWidth;
	int32 nClipHeight = rectClipped.m_dimension.m_nHeight;
	int32 nClipX2 = rectClipped.GetX2();
	int32 nClipY2 = rectClipped.GetY2();

	if( nClipWidth == 0 || nClipHeight == 0 )
	{
		return;
	}

	const int32& nPitch = GetPitch();
	const int32 nNextRowOffset = nPitch - nClipWidth; 

	Color32* pPixels = GetBuffer() + rectClipped.m_ptTopLeft.ComputeOffset(nPitch);

	uint8 r1, g1, b1, a1;
	col1.ToRgba(r1, g1, b1, a1);
	
	uint8 r2, g2, b2, a2;
	col2.ToRgba(r2, g2, b2, a2);
	
	for( int32 iY = nClipY; iY <= nClipY2; ++iY )
	{

		const uint32 nGradientIndex = iY - nRectY;
		const uint32 nGradient = nGradientIndex * 255 / nRectHeight;
		const uint32 nOneLessGradient = 255 - nGradient;
		
		// Note : pour de la transparence parfaite
		// il faudrait diviser par 255 au lieu de 256
		const uint8 r_src = uint8(( (r1 * nOneLessGradient) + (r2 * nGradient)) >> 8);
		const uint8 g_src = uint8(( (g1 * nOneLessGradient) + (g2 * nGradient)) >> 8);
		const uint8 b_src = uint8(( (b1 * nOneLessGradient) + (b2 * nGradient)) >> 8);
		const uint8 a_src = uint8(( (a1 * nOneLessGradient) + (a2 * nGradient)) >> 8);

		if( a_src == 255 )
		{
			// Opaque
			for( int32 iX = nClipX; iX <= nClipX2; ++iX )
			{
				pPixels[0] = Color32(r_src, g_src, b_src, a_src);
				pPixels++;
			}
		}
		else if( a_src == 0 )
		{
			// No pixels to draw
			pPixels += nClipWidth;
		}
		else
		{
			// Alpha
			const uint32 nAlpha = a_src;
			const uint32 nOneLessAlpha = 255 - a_src;
			const uint32 nRedSrcMultAlpha = r_src * nAlpha;
			const uint32 nGreenSrcMultAlpha = g_src * nAlpha;
			const uint32 nBlueSrcMultAlpha = b_src * nAlpha;
			
			for( int32 iX = nClipX; iX <= nClipX2; ++iX )
			{
				Color32 colPixelDst = pPixels[0];
				uint32 r_dst, g_dst, b_dst, a_dst;
				colPixelDst.ToRgba(r_dst, g_dst, b_dst, a_dst);
				
				// Note : pour de la transparence parfaite
				// il faudrait diviser par 255 au lieu de 256
				const uint8 r = uint8(( nRedSrcMultAlpha + (r_dst * nOneLessAlpha)) >> 8);
				const uint8 g = uint8(( nGreenSrcMultAlpha + (g_dst * nOneLessAlpha)) >> 8);
				const uint8 b = uint8(( nBlueSrcMultAlpha + (b_dst * nOneLessAlpha)) >> 8);
				
				pPixels[0] = Color32(r, g, b, uint8(a_dst));
				pPixels++;
			}
		}
		pPixels += nNextRowOffset;
	}
}

void Surface32::FillHorizontalGradient(Color32 col1, Color32 col2) const
{
	Surface32::FillHorizontalGradient(Rectangle(0, 0, GetWidth(), GetHeight()), col1, col2);
}

void Surface32::FillHorizontalGradient(const Rectangle& rectUser, Color32 col1, Color32 col2) const
{
	// Translate the destination according to the surface internal offset
	Rectangle rect = rectUser;
	DoOffset(rect);

	Rectangle rectSurface(0, 0, GetWidth(), GetHeight());
	Rectangle rectClipped(rect);
	rectClipped.ClipIn(rectSurface);
	
	const int32 nRectX = rect.m_ptTopLeft.m_nX;
	//const int32 nRectY = rect.m_ptTopLeft.m_nY;
	const int32 nRectWidth = rect.m_dimension.m_nWidth;
	//const int32 nRectHeight = rect.m_dimension.m_nHeight;
	//int32 nRectX2 = rect.GetX2();
	//int32 nRectY2 = rect.GetY2();

	int32 nClipX = rectClipped.m_ptTopLeft.m_nX;
	int32 nClipY = rectClipped.m_ptTopLeft.m_nY;
	int32 nClipWidth = rectClipped.m_dimension.m_nWidth;
	int32 nClipHeight = rectClipped.m_dimension.m_nHeight;
	int32 nClipX2 = rectClipped.GetX2();
	int32 nClipY2 = rectClipped.GetY2();

	if( nClipWidth == 0 || nClipHeight == 0 )
	{
		return;
	}

	const int32& nPitch = GetPitch();
	const int32 nNextRowOffset = nPitch - nClipWidth;

	Color32* pPixels = GetBuffer() + rectClipped.m_ptTopLeft.ComputeOffset(nPitch);

	uint8 r1, g1, b1, a1;
	col1.ToRgba(r1, g1, b1, a1);
	
	uint8 r2, g2, b2, a2;
	col2.ToRgba(r2, g2, b2, a2);
	
	// Fill a gradient row
	Color32* acolorLine = (Color32* const) new uint32[nRectWidth];

	// Manage clipping with start color
	if( nRectX < 0)
	{
		int32 nDelta = -nRectX;
		int32 nMaxDelta = nRectWidth;
		if( nMaxDelta != 0)
		{
			r1 = uint8(r1 + ((r2 - r1) * nDelta) / nMaxDelta);
			g1 = uint8(g1 + ((g2 - g1) * nDelta) / nMaxDelta);
			b1 = uint8(b1 + ((b2 - b1) * nDelta) / nMaxDelta);
			a1 = uint8(a1 + ((a2 - a1) * nDelta) / nMaxDelta);
		}
	}

	for( int32 i = 0; i < nClipWidth; ++i )
	{
		//const int nIndex = i - nRectX;
		const uint32 nGradient = i * 255 / nClipWidth;
		const uint32 nOneLessGradient = 255 - nGradient;
		
		// Note : pour de la transparence parfaite
		// il faudrait diviser par 255 au lieu de 256
		const uint8 r_src = uint8(( (r1 * nOneLessGradient) + (r2 * nGradient)) >> 8);
		const uint8 g_src = uint8(( (g1 * nOneLessGradient) + (g2 * nGradient)) >> 8);
		const uint8 b_src = uint8(( (b1 * nOneLessGradient) + (b2 * nGradient)) >> 8);
		const uint8 a_src = uint8(( (a1 * nOneLessGradient) + (a2 * nGradient)) >> 8);

		acolorLine[i] = Color32(r_src, g_src, b_src, a_src);
	}

	uint8 r_src, g_src, b_src, a_src;

	for( int32 iY = nClipY; iY <= nClipY2; ++iY )
	{
		for( int32 iX = nClipX; iX <= nClipX2; ++iX )
		{
			acolorLine[iX - nClipX].ToRgba(r_src, g_src, b_src, a_src);
			
			if( a_src == 255 )
			{
				// Opaque
				pPixels[0] = Color32(r_src, g_src, b_src, a_src);
			}
			else if( a_src == 0 )
			{
				// No pixels to draw
			}
			else
			{
				// Alpha
				const uint32 nAlpha = a_src;
				const uint32 nOneLessAlpha = 255 - a_src;
				const uint32 nRedSrcMultAlpha = r_src * nAlpha;
				const uint32 nGreenSrcMultAlpha = g_src * nAlpha;
				const uint32 nBlueSrcMultAlpha = b_src * nAlpha;
				
				Color32 colPixelDst = pPixels[0];
				uint32 r_dst, g_dst, b_dst, a_dst;
				colPixelDst.ToRgba(r_dst, g_dst, b_dst, a_dst);
				
				// Note : pour de la transparence parfaite
				// il faudrait diviser par 255 au lieu de 256
				const uint8 r = uint8(( nRedSrcMultAlpha + (r_dst * nOneLessAlpha)) >> 8);
				const uint8 g = uint8(( nGreenSrcMultAlpha + (g_dst * nOneLessAlpha)) >> 8);
				const uint8 b = uint8(( nBlueSrcMultAlpha + (b_dst * nOneLessAlpha)) >> 8);
				
				pPixels[0] = Color32(r, g, b, uint8(a_dst));
			}
			pPixels++;
		}
		pPixels += nNextRowOffset;
	}

	delete[] acolorLine;
}

void Surface32::FillRect(const Rectangle& rectUser, Color32 col) const
{
	// Translate the destination according to the surface internal offset
	Rectangle rect2 = rectUser;
	DoOffset(rect2);

	Rectangle rectSurface(0, 0, GetWidth(), GetHeight());
	Rectangle rectClipped(rect2);
	rectClipped.ClipIn(rectSurface);
	
	int32 nClipX = rectClipped.m_ptTopLeft.m_nX;
	int32 nClipY = rectClipped.m_ptTopLeft.m_nY;
	int32 nClipWidth = rectClipped.m_dimension.m_nWidth;
	int32 nClipHeight = rectClipped.m_dimension.m_nHeight;
	int32 nClipX2 = rectClipped.GetX2();
	int32 nClipY2 = rectClipped.GetY2();

	if( nClipWidth == 0 || nClipHeight == 0 )
	{
		return;
	}

	const int32 nPitch = GetPitch();
	const int32 nNextRowOffset = nPitch - nClipWidth; 

	Color32* pPixels = GetBuffer() + rectClipped.m_ptTopLeft.ComputeOffset(nPitch);

	uint8 r_src, g_src, b_src, a_src;
	col.ToRgba(r_src, g_src, b_src, a_src);

	if( a_src == 0)
		return;

	if( a_src == 255 )
	{
		// Opaque
		for( int32 iY = nClipY; iY <= nClipY2; ++iY )
		{
			for( int32 iX = nClipX; iX <= nClipX2; ++iX )
			{
				pPixels[0] = col;
				pPixels++;
			}
			pPixels += nNextRowOffset;
		}
	}
	else
	{
		// Alpha
		const uint32 nAlpha = a_src;
		const uint32 nOneLessAlpha = 255 - a_src;
		const uint32 nRedSrcMultAlpha = r_src * nAlpha;
		const uint32 nGreenSrcMultAlpha = g_src * nAlpha;
		const uint32 nBlueSrcMultAlpha = b_src * nAlpha;

		for( int32 iY = nClipY; iY <= nClipY2; ++iY )
		{
			for( int32 iX = nClipX; iX <= nClipX2; ++iX )
			{
				Color32 colPixelDst = pPixels[0];
				uint32 r_dst, g_dst, b_dst, a_dst;
				colPixelDst.ToRgba(r_dst, g_dst, b_dst, a_dst);
				
				// Note : pour de la transparence parfaite
				// il faudrait diviser par 255 au lieu de 256
				const uint8 r = uint8(( nRedSrcMultAlpha + (r_dst * nOneLessAlpha)) >> 8);
				const uint8 g = uint8(( nGreenSrcMultAlpha + (g_dst * nOneLessAlpha)) >> 8);
				const uint8 b = uint8(( nBlueSrcMultAlpha + (b_dst * nOneLessAlpha)) >> 8);
				
				pPixels[0] = Color32(r, g, b, uint8(a_dst));
				pPixels++;
			}
			pPixels += nNextRowOffset;
		}
	}
}

void Surface32::FillRect(int32 nX, int32 nY, int32 nWidth, int32 nHeight, Color32 col) const
{
	FillRect(Rectangle(nX, nY, nWidth, nHeight), col);	
}


void Surface32::ClearRect(const Rectangle& rectUser, Color32 col) const
{
	// Translate the destination according to the surface internal offset
	Rectangle rect2 = rectUser;
	DoOffset(rect2);

	Rectangle rectSurface(0, 0, GetWidth(), GetHeight());
	Rectangle rectClipped(rect2);
	rectClipped.ClipIn(rectSurface);
	
	int32 nClipX = rectClipped.m_ptTopLeft.m_nX;
	int32 nClipY = rectClipped.m_ptTopLeft.m_nY;
	int32 nClipWidth = rectClipped.m_dimension.m_nWidth;
	int32 nClipHeight = rectClipped.m_dimension.m_nHeight;
	int32 nClipX2 = rectClipped.GetX2();
	int32 nClipY2 = rectClipped.GetY2();

	if( nClipWidth == 0 || nClipHeight == 0 )
	{
		return;
	}

	const int32 nPitch = GetPitch();
	const int32 nNextRowOffset = nPitch - nClipWidth; 

	Color32* pPixels = GetBuffer() + rectClipped.m_ptTopLeft.ComputeOffset(nPitch);

	for( int32 iY = nClipY; iY <= nClipY2; ++iY )
	{
		for( int32 iX = nClipX; iX <= nClipX2; ++iX )
		{
			pPixels[0] = col;
			pPixels++;
		}
		pPixels += nNextRowOffset;
	}
}

void Surface32::ClearRect(int32 nX, int32 nY, int32 nWidth, int32 nHeight, Color32 col) const
{
	ClearRect(Rectangle(nX, nY, nWidth, nHeight), col);	
}

void Surface32::SetBlendAlphaMode(BlendAlphaMode eMode)
{
	Surface32Impl* pImpl = (Surface32Impl*) m_pImpl;
	pImpl->m_eBlendAlphaMode = eMode;
}

// Blit d'une image 32 bits contenant vers le buffer
// R = ColR * SrcA + DstR * (1 - A)
// G = ColG * SrcA + DstG * (1 - A)
// B = ColB * SrcA + DstB * (1 - A)
// A = max( SrcA, DstA)
void Surface32::BlendSurface(const ReadOnlySurface32& surfSrc, int32 nDstX, int32 nDstY) const
{
	// Translate the destination according to the surface internal offset
	DoOffset(nDstX, nDstY);

	SurfaceBaseImpl sbiClip;
	if( !ClipSurface(surfSrc, nDstX, nDstY, sbiClip))
		return;

	const int32 nSrcPitch = sbiClip.m_nPitch;
	int32 nSrcWidth = sbiClip.m_nWidth;
	int32 nSrcHeight = sbiClip.m_nHeight;
	const uint32* pSrcPixels = (uint32*) sbiClip.m_pBuffer;

	const int32 nPitch = m_pImpl->m_nPitch;
	int32 nOffsetDst = nDstY * nPitch + nDstX;

	const int32 nNextRowSrc = nSrcPitch - nSrcWidth;
	const int32 nNextRowDst = nPitch - nSrcWidth;

	uint32* pDstPixels = (uint32*) GetBuffer() + nOffsetDst;

	for(int y = nSrcHeight - 1; y >= 0; --y)
	{
		for(int x = nSrcWidth - 1; x >= 0; --x)
		{
			uint32 nPixelSrc = pSrcPixels[0];
			const uint32 b_src = (nPixelSrc) & 0xff;
			const uint32 g_src = (nPixelSrc >> 8) & 0xff;
			const uint32 r_src = (nPixelSrc >> 16) & 0xff;
			
			const uint32 nAlpha = (nPixelSrc >> 24) & 0xff;
			const uint32 nOneLessAlpha = 255 - nAlpha;

			uint32 nPixelDst = pDstPixels[0];
			const uint32 b_dst = (nPixelDst) & 0xff;
			const uint32 g_dst = (nPixelDst >> 8) & 0xff;
			const uint32 r_dst = (nPixelDst >> 16) & 0xff;
			const uint32 a_dst = (nPixelDst >> 24) & 0xff;

			// Note : pour de la transparence parfaite
			// il faudrait diviser par 255 au lieu de 256
			const uint8 r = uint8(( (r_src * nAlpha) + (r_dst * nOneLessAlpha)) >> 8);
			const uint8 g = uint8(( (g_src * nAlpha) + (g_dst * nOneLessAlpha)) >> 8);
			const uint8 b = uint8(( (b_src * nAlpha) + (b_dst * nOneLessAlpha)) >> 8);
			
			uint32 a_final = a_dst > nAlpha ? a_dst : nAlpha;
			pDstPixels[0] = (a_final << 24) | (r << 16) | (g << 8) | b;

			pSrcPixels++;
			pDstPixels++;
		}
		pSrcPixels += nNextRowSrc;
		pDstPixels += nNextRowDst;
	}
}

// R = ColR * SrcA + DstR * (1 - A)
// G = ColG * SrcA + DstG * (1 - A)
// B = ColB * SrcA + DstB * (1 - A)
// A = DstA
void Surface32::BlendSurface(const ReadOnlySurface8& surfSrc, int32 nDstX, int32 nDstY, Color32 col) const
{
	// Translate the destination according to the surface internal offset
	DoOffset(nDstX, nDstY);

	SurfaceBaseImpl sbiClip;
	if( !ClipSurface(surfSrc, nDstX, nDstY, sbiClip))
		return;

	const int32 nSrcPitch = sbiClip.m_nPitch;
	int32 nSrcWidth = sbiClip.m_nWidth;
	int32 nSrcHeight = sbiClip.m_nHeight;
	const uint8* pSrcPixels = (uint8*) sbiClip.m_pBuffer;

	const int32 nPitch = m_pImpl->m_nPitch;
	int32 nOffsetDst = nDstY * nPitch + nDstX;

	const int32 nNextRowSrc = nSrcPitch - nSrcWidth;
	const int32 nNextRowDst = nPitch - nSrcWidth;

	uint32* pDst = (uint32*)GetBuffer() + nOffsetDst;

	uint32 r_src, g_src, b_src, a_src;
	col.ToRgba(r_src, g_src, b_src, a_src);

	for(int y = nSrcHeight - 1; y >= 0; --y)
	{
		for(int x = nSrcWidth - 1; x >= 0; --x)
		{
			uint32 nPixel = pDst[0];
			
			const uint32 b_dst = (nPixel) & 0xff;
			const uint32 g_dst = (nPixel >> 8) & 0xff;
			const uint32 r_dst = (nPixel >> 16) & 0xff;
			const uint32 a_dst = (nPixel >> 24) & 0xff;
			
			const uint32 nAlpha = pSrcPixels[0];
			const uint32 nOneLessAlpha = 255 - nAlpha;

			// Note : pour de la transparence parfaite
			// il faudrait diviser par 255 au lieu de 256
			const uint8 r = uint8(( (r_src * nAlpha) + (r_dst * nOneLessAlpha)) >> 8);
			const uint8 g = uint8(( (g_src * nAlpha) + (g_dst * nOneLessAlpha)) >> 8);
			const uint8 b = uint8(( (b_src * nAlpha) + (b_dst * nOneLessAlpha)) >> 8);
			
			pDst[0] = (a_dst << 24) | (r << 16) | (g << 8) | b;

			pSrcPixels++;
			pDst++;
		}
		pSrcPixels += nNextRowSrc;
		pDst += nNextRowDst;
	}
}

void Surface32::DrawString(const String& str, int32 nX, int32 nY, Color32 col, const Font& font) const
{
	if( !m_pImpl)
		return;

	int32 nPenX = nX;
	int32 nPenY = nY + font.GetAscent();

	const int32 nCharCount = str.GetLength();
	for(int i = 0; i < nCharCount; i++ )
	{
		int32 nOffsetX, nOffsetY, nAdvance;
		const Surface8& surfFont = font.RenderChar(str.GetAt(i), nOffsetX, nOffsetY, nAdvance);

		BlendSurface(surfFont, nPenX + nOffsetX, nPenY + nOffsetY, col);
				
		nPenX += nAdvance;
	}

}

void Surface32::DrawString(const chustd::String& str, const chustd::Rectangle& rect, Alignment eAlign,
						   Color32 col, const chustd::Font& font) const
{
	const int nAlign = int(eAlign);
	const int32 nRectX = rect.m_ptTopLeft.m_nX;
	const int32 nRectY = rect.m_ptTopLeft.m_nY;
	const int32 nRectWidth = rect.m_dimension.m_nWidth;
	const int32 nRectHeight = rect.m_dimension.m_nHeight;

	int32 nSplitWidth = nRectWidth;
	if( nAlign & alNoBreak)
	{
		// Split regarding endlines only
		nSplitWidth = -1;
	}
	chustd::StringArray aStrings = font.SplitString(str, nSplitWidth);

	const int32 nPieceCount = aStrings.GetSize();
	const int32 nLineHeight = font.GetLineHeight();
	const int32 nTotalHeight = nPieceCount * nLineHeight;
	//const int32 nAscent = font.GetAscent();
	//const int32 nDescent = font.GetSize() - nAscent;

	////////////////////////////////////////////////////////
	// Compute the vertical position
	int32 nStartTextY = 0;

	if( nAlign & alTop)
	{
		nStartTextY = nRectY;
	}
	else if( nAlign & alMiddle)
	{
		nStartTextY = nRectY + (nRectHeight - nTotalHeight) / 2;
	}
	else if( nAlign & alBottom)
	{
		nStartTextY = nRectY + nRectHeight - nTotalHeight;
	}
	else
	{
		ASSERT(0); // Bad param
		return;
	}
	////////////////////////////////////////////////////////
	
	for(int32 i = 0; i < nPieceCount; ++i)
	{
		const String& strPiece = aStrings[i];
		const int32 nPieceWidth = font.GetTextWidth(strPiece);

		int32 nStartTextX = 0;
		
		if( nAlign & alLeft)
		{
			nStartTextX = nRectX;
		}
		else if( nAlign & alCenter)
		{
			nStartTextX = nRectX + (nRectWidth - nPieceWidth) / 2;
		}
		else if( nAlign & alRight)
		{
			nStartTextX = nRectX + (nRectWidth - nPieceWidth);
		}
		else
		{
			ASSERT(0); // Bad param
			return;
		}

		DrawString(strPiece, nStartTextX, nStartTextY, col, font);
		nStartTextY += nLineHeight;
	}
}


// Note : no surface internal offset translation done in this function
static void Surface32_BlendPixel(Color32* pDst, Color32 col, uint8 nIntensity)
{
	uint32 r_src, g_src, b_src, a_src;
	col.ToRgba(r_src, g_src, b_src, a_src);
	
	uint32 nAlpha = a_src;
	nAlpha *= nIntensity;
	nAlpha /= 255;
	const uint32 nOneLessAlpha = 255 - nAlpha;

	uint32 r_dst, g_dst, b_dst, a_dst;
	pDst[0].ToRgba(r_dst, g_dst, b_dst, a_dst);

	// Note : pour de la transparence parfaite
	// il faudrait diviser par 255 au lieu de 256
	const uint8 r = uint8(( (r_src * nAlpha) + (r_dst * nOneLessAlpha)) >> 8);
	const uint8 g = uint8(( (g_src * nAlpha) + (g_dst * nOneLessAlpha)) >> 8);
	const uint8 b = uint8(( (b_src * nAlpha) + (b_dst * nOneLessAlpha)) >> 8);
	
	pDst[0] = Color32(r, g, b, uint8(a_dst));
}

void Surface32::DrawLine(Point pt1, Point pt2, Color32 col) const
{
	if( !m_pImpl)
		return;

	// Translate the destination according to the surface internal offset
	DoOffset(pt1);
	DoOffset(pt2);

	///////////////////////////////////////
	// Cohen-Sutherland rejecting

	// 0101 | 0100 | 0110
	// -----+------+------
	// 0001 | 0000 | 0010
	// -----+------+------
	// 1001 | 1000 | 1010
	
	int32 nX1 = pt1.m_nX;
	int32 nY1 = pt1.m_nY;

	int32 nX2 = pt2.m_nX;
	int32 nY2 = pt2.m_nY;

	int32 nMaxX = GetWidth() - 1;
	int32 nMaxY = GetHeight() - 1;

	uint8 nCode1 = 0;
	nCode1 |= nX1 < 0 ? 1 : 0;
	nCode1 |= nX1 > nMaxX ? 2 : 0;
	nCode1 |= nY1 < 0 ? 4 : 0;
	nCode1 |= nY1 > nMaxY ? 8 : 0;

	uint8 nCode2 = 0;
	nCode2 |= nX2 < 0 ? 1 : 0;
	nCode2 |= nX2 > nMaxX ? 2 : 0;
	nCode2 |= nY2 < 0 ? 4 : 0;
	nCode2 |= nY2 > nMaxY ? 8 : 0;

	if( (nCode1 & nCode2) != 0)
	{
		// Completely outside
		return;
	}
	///////////////////////////////////////

	///////////////////////////////////////
	int32 nDeltaX = (pt2.m_nX - pt1.m_nX);
	int32 nDeltaY = (pt2.m_nY - pt1.m_nY);

	int32 nAbsDeltaX = Math::Abs(nDeltaX);
	int32 nAbsDeltaY = Math::Abs(nDeltaY);

	// Quel axe on incrmente de un en un
	bool bXMajor;

	// Inverse les point si ncessaire pour rendre l'incrment positif
	if( nAbsDeltaX >= nAbsDeltaY)
	{
		bXMajor = true;

		if( nDeltaX < 0)
		{
			Math::Swap(nX1, nX2);
			Math::Swap(nY1, nY2);

			nDeltaY = -nDeltaY;
		}
	}
	else
	{
		bXMajor = false;

		if( nDeltaY < 0)
		{
			Math::Swap(nX1, nX2);
			Math::Swap(nY1, nY2);

			nDeltaX = -nDeltaX;
		}
	}

	////////////////////////////
	Color32* pDstPixels = (Color32*) GetBuffer();
	const int32 nPitch = m_pImpl->m_nPitch;
		
	///////////////////////////

	if( bXMajor)
	{
		// X augmente de un en un pendant que Y varie doucement
		int32 nIncr = ( nAbsDeltaX == 0 ) ? 0 : (nDeltaY << 16) / nAbsDeltaX;

		int32 nY = nY1 << 16;
		int32 nMaxY16 = nMaxY << 16;

		if( (nCode1 | nCode2) != 0)
		{
			// Ligne pas compltement  l'intrieur
		
			///////////////////////////////////////////////
			// Starting point
			if( nX1 < 0)
			{
				nY += (-nX1) * nIncr;
				nX1 = 0;
			}

			// Top
			if( nY < 0)
			{
				// Calcule le nouveau X de dpart
				int32 nAddToX = (-nY) / nIncr;

				nX1 += nAddToX;
				nY += nAddToX * nIncr;

				if( nY >= 0)
				{
					// On se met un petit peu en arrire en raison de l'anti-aliasing
					nX1 -= 1;
					nY -= nIncr;
				}
			}
			else if( nY > nMaxY16)
			{
				int32 nY1_16 = nY;

				// Calcule le nouveau X de fin
				int32 nDif = nY1_16 - nMaxY16;
				int32 nSubToX = (nDif) / nIncr;
				//int32 nRem = nDif % nIncr;

				nX1 -= nSubToX;
				nY1_16 -= nSubToX * nIncr;

				if( nY1_16 <= nMaxY16)
				{
					// On se met un petit peu en arrire en raison de l'anti-aliasing
					nX1 -= 1;
				}

				nY = nY1_16;
			}

			///////////////////////////////////////////////
			// Ending point

			int32 nY2_16 = nY2 << 16;
			if( nX2 > nMaxX)
			{
				int32 nDif = nX2 - nMaxX;

				nY2_16 -= (nDif) * nIncr;
				nX2 = nMaxX;
			}

			if( nY2_16 < 0)
			{
				// Calcule le nouveau X de dpart
				int32 nSubToX = (nY2_16) / nIncr;

				nX2 -= nSubToX;
				nY2_16 += nSubToX * nIncr;

				if( nY2_16 >= 0)
				{
					// On se met un petit peu en avant en raison de l'anti-aliasing
					nX1 += 1;
					nY2_16 += nIncr;
				}
			}
			else if( nY2_16 > nMaxY16)
			{
				// Calcule le nouveau X de fin
				int32 nDif = nY2_16 - nMaxY16;
				int32 nSubToX = (nDif) / nIncr;
				//int32 nRem = nDif % nIncr;

				nX2 -= nSubToX;
				nY2_16 -= nSubToX * nIncr;

				if( nY2_16 <= nMaxY16)
				{
					// On se met un petit peu en avant en raison de l'anti-aliasing
					nX2 += 1;
				}
			}

			nAbsDeltaX = nX2 - nX1;
		}

		// Trac
		for(int32 i = 0; i <= nAbsDeltaX; ++i)
		{
			int32 nOffset = (nY >> 16) * nPitch + (nX1 + i);

			uint8 nInt = uint8((nY & 0x0000ffff) >> 8);
			
			if( nY >= 0)
			{
				Surface32_BlendPixel(pDstPixels + nOffset, col, 255 - nInt);
			}
			
			if( nY < nMaxY16)
			{
				Surface32_BlendPixel(pDstPixels + nOffset + nPitch, col, nInt);
			}
			
			nY += nIncr;
		}
	}
	else
	{
		int32 nX = nX1 << 16;
		int32 nMaxX16 = nMaxX << 16;

		// Y augmente de un en un pendant que X varie doucement
		int32 nIncr = ( nAbsDeltaY == 0 ) ? 0 : (nDeltaX << 16) / nAbsDeltaY;

		for(int32 i = 0; i <= nAbsDeltaY; ++i)
		{
			int32 nOffset = (nY1 + i) * nPitch + (nX >> 16);

			uint8 nInt = uint8((nX & 0x0000ffff) >> 8);
			
			if( nX >= 0)
			{
				Surface32_BlendPixel(pDstPixels + nOffset, col, 255 - nInt);
			}

			if( nX < nMaxX16)
			{
				Surface32_BlendPixel(pDstPixels + nOffset + 1, col, nInt);
			}
			
			nX += nIncr;
		}
	}
}

void Surface32::DrawRect(const Rectangle& rect, Color32 col) const
{
	Point ptTL = rect.GetPointTopLeft();
	Point ptTR = rect.GetPointTopRight();
	Point ptBL = rect.GetPointBottomLeft();
	
	FillRect(ptTL.m_nX, ptTL.m_nY, rect.m_dimension.m_nWidth, 1, col);
	FillRect(ptTL.m_nX, ptTL.m_nY, 1, rect.m_dimension.m_nHeight, col);

	FillRect(ptTR.m_nX, ptTR.m_nY, 1, rect.m_dimension.m_nHeight, col);
	FillRect(ptBL.m_nX, ptBL.m_nY, rect.m_dimension.m_nWidth, 1, col);
}

void Surface32::DrawPixel(int32 nX, int32 nY, Color32 col) const
{
	if( !m_pImpl)
		return;

	// Translate the destination according to the surface internal offset
	DoOffset(nX, nY);

	const int32 nHeight = GetHeight();
	const int32 nWidth = GetWidth();
	const int32 nPitch = GetPitch();
	Color32* pDstPixels = (Color32*)GetBuffer();

	if( !(0 <= nX && nX < nWidth))
		return;

	if( !(0 <= nY && nY < nHeight))
		return;

	uint8 r, g, b, a;
	col.ToRgba(r, g, b, a);

	int32 nOffset = nY * nPitch + nX;

	if( a == 255)
	{
		// Opaque
		pDstPixels[nOffset] = col;
	}
	else
	{
		Surface32_BlendPixel(pDstPixels + nOffset, col, 255);
	}
}

void Surface32::CopySurface(const ReadOnlySurface32& surfSrc, int32 nDstX, int32 nDstY) const
{
	if( !m_pImpl)
		return;

	// Translate the destination according to the surface internal offset
	DoOffset(nDstX, nDstY);

	SurfaceBaseImpl sbiClip;
	if( !ClipSurface(surfSrc, nDstX, nDstY, sbiClip))
		return;

	const int32 nSrcPitch = sbiClip.m_nPitch;
	int32 nSrcWidth = sbiClip.m_nWidth;
	int32 nSrcHeight = sbiClip.m_nHeight;
	const uint32* pSrcPixels = (uint32*) sbiClip.m_pBuffer;

	const int32 nPitch = m_pImpl->m_nPitch;
	int32 nOffsetDst = nDstY * nPitch + nDstX;

	const int32 nNextRowSrc = nSrcPitch - nSrcWidth;
	const int32 nNextRowDst = nPitch - nSrcWidth;

	uint32* pDstPixels = (uint32*) GetBuffer() + nOffsetDst;

	for(int y = nSrcHeight - 1; y >= 0; --y)
	{
		for(int x = nSrcWidth - 1; x >= 0; --x)
		{
			pDstPixels[0] = pSrcPixels[0];

			pSrcPixels++;
			pDstPixels++;
		}
		pSrcPixels += nNextRowSrc;
		pDstPixels += nNextRowDst;
	}
}

void Surface32::MaskSurface(const ReadOnlySurface32& surfSrc, int32 nDstX, int32 nDstY) const
{
	if( !m_pImpl)
		return;

	// Translate the destination according to the surface internal offset
	DoOffset(nDstX, nDstY);

	SurfaceBaseImpl sbiClip;
	if( !ClipSurface(surfSrc, nDstX, nDstY, sbiClip))
		return;

	const int32 nSrcPitch = sbiClip.m_nPitch;
	int32 nSrcWidth = sbiClip.m_nWidth;
	int32 nSrcHeight = sbiClip.m_nHeight;
	const uint32* pSrcPixels = (uint32*) sbiClip.m_pBuffer;

	const int32 nPitch = m_pImpl->m_nPitch;
	int32 nOffsetDst = nDstY * nPitch + nDstX;

	const int32 nNextRowSrc = nSrcPitch - nSrcWidth;
	const int32 nNextRowDst = nPitch - nSrcWidth;

	uint32* pDstPixels = (uint32*) GetBuffer() + nOffsetDst;

	for(int y = nSrcHeight - 1; y >= 0; --y)
	{
		for(int x = nSrcWidth - 1; x >= 0; --x)
		{
			int32 nSrc = pSrcPixels[0];
			int32 nMask = (nSrc >> 31);
			nSrc &= nMask;
			
			int32 nDst = pDstPixels[0];
			nDst = (nDst & ~nMask) | nSrc;
			
			pDstPixels[0] = nDst;
			
			pSrcPixels++;
			pDstPixels++;
		}
		pSrcPixels += nNextRowSrc;
		pDstPixels += nNextRowDst;
	}
}

void Surface32::DrawHrzLine(int32 nX1, int32 nX2, int32 nY, Color32 col) const
{
	// Note : can be implemented by avoiding the call to FillRect and doing the drawing job itself

	int32 nDiff = (nX2 - nX1);
	if( nDiff < 0)
	{
		return;
	}

	FillRect(nX1, nY, (nX2 - nX1) + 1, 1, col);
}

void Surface32::DrawVrtLine(int32 nX, int32 nY1, int32 nY2, Color32 col) const
{
	// Note : can be implemented by avoiding the call to FillRect and doing the drawing job itself

	int32 nDiff = (nY2 - nY1);
	if( nDiff < 0)
	{
		return;
	}

	FillRect(nX, nY1, 1, (nY2 - nY1) + 1, col);
}

void Surface32::DrawRoundCorner(int32 nCornerX, int32 nCornerY, int32 nRadius, CornerType eType, Color32 col) const
{
	// 2 low bits : the corner position
	// 3rd bit : circle or disc
	int nType = int(eType);

	int nCorner = nType & 0x03;
	bool bFilled = (nType & ctFilled) != 0;

	int32 nCenterX;
	int32 nCenterY;

	if( nCorner == ctTopLeft)
	{
		 nCenterX = nCornerX + nRadius;
		 nCenterY = nCornerY + nRadius;
	}
	else if( nCorner == ctTopRight)
	{
		nCenterX = nCornerX - 1;
		nCenterY = nCornerY + nRadius;
	}
	else if( nCorner == ctBottomRight)
	{
		nCenterX = nCornerX - 1;
		nCenterY = nCornerY - 1;
	}
	else if( nCorner == ctBottomLeft)
	{
		nCenterX = nCornerX + nRadius;
		nCenterY = nCornerY - 1;
	}
	else
	{
		ASSERT(0);
		return;
	}

	int32 nX = 1;
	int32 nY = nRadius;
	int32 nM = 5 - 4 * nRadius;
	
	while( nX < nY)
	{
		if( !bFilled)
		{
			if( nCorner == ctTopLeft)
			{
				DrawPixel(nCenterX - nX, nCenterY - nY, col);
				DrawPixel(nCenterX - nY, nCenterY - nX, col);
			}
			else if( nCorner == ctTopRight)
			{
				DrawPixel(nCenterX + nX, nCenterY - nY, col);
				DrawPixel(nCenterX + nY, nCenterY - nX, col);
			}
			else if( nCorner == ctBottomRight)
			{
				DrawPixel(nCenterX + nX, nCenterY + nY, col);
				DrawPixel(nCenterX + nY, nCenterY + nX, col);
			}
			else if( nCorner == ctBottomLeft)
			{
				DrawPixel(nCenterX - nX, nCenterY + nY, col);
				DrawPixel(nCenterX - nY, nCenterY + nX, col);
			}
		}
		else
		{
			if( nCorner == ctTopLeft)
			{
				DrawVrtLine(nCenterX - nX, nCenterY - nY, (nCenterY - nX) - 1, col);
				DrawHrzLine(nCenterX - nY, nCenterX - nX, nCenterY - nX, col);
			}
			else if( nCorner == ctTopRight)
			{
				DrawVrtLine(nCenterX + nX, nCenterY - nY, (nCenterY - nX) - 1, col);
				DrawHrzLine(nCenterX + nX, nCenterX + nY, nCenterY - nX, col);
			}
			else if( nCorner == ctBottomRight)
			{
				DrawVrtLine(nCenterX + nX, (nCenterY + nX) + 1, nCenterY + nY, col);
				DrawHrzLine(nCenterX + nX, nCenterX + nY, nCenterY + nX, col);
			}
			else if( nCorner == ctBottomLeft)
			{
				DrawVrtLine(nCenterX - nX, (nCenterY + nX) + 1, nCenterY + nY, col);
				DrawHrzLine(nCenterX - nY, nCenterX - nX, nCenterY + nX, col);
			}
		}

		if( nM > 0)
		{
				nY = nY - 1;
				nM = nM - 8 * nY;
		}
		nX++;
		nM = nM + 8 * nX + 4;
	}

	// Last common pixel
	if( nCorner == ctTopLeft)
	{
		DrawPixel(nCenterX - nX, nCenterY - nY, col);
	}
	else if( nCorner == ctTopRight)
	{
		DrawPixel(nCenterX + nX, nCenterY - nY, col);
	}
	else if( nCorner == ctBottomRight)
	{
		DrawPixel(nCenterX + nX, nCenterY + nY, col);
	}
	else if( nCorner == ctBottomLeft)
	{
		DrawPixel(nCenterX - nX, nCenterY + nY, col);
	}
}

void Surface32::DrawRoundRect(const Rectangle& rect, int32 nRadius, Color32 col) const
{
	int32 nRectX = rect.m_ptTopLeft.m_nX;
	int32 nRectY = rect.m_ptTopLeft.m_nY;
	int32 nRectWidth = rect.m_dimension.m_nWidth;
	int32 nRectHeight = rect.m_dimension.m_nHeight;

	const int32 nInternalWidth = nRectWidth - nRadius * 2;
	const int32 nInternalHeight = nRectHeight - nRadius * 2;

	DrawRoundCorner(nRectX, nRectY, nRadius, ctTopLeft, col);
	DrawRoundCorner(nRectX + nRectWidth - nRadius, nRectY, nRadius, ctTopRight, col);
	DrawRoundCorner(nRectX + nRectWidth - nRadius, nRectY + nRectHeight - nRadius, nRadius, ctBottomRight, col);
	DrawRoundCorner(nRectX, nRectY + nRectHeight - nRadius, nRadius, ctBottomLeft, col);

	FillRect(nRectX + nRadius, nRectY, nInternalWidth, 1, col);
	FillRect(nRectX + nRadius, nRectY + nRectHeight - 1, nInternalWidth, 1, col);

	FillRect(nRectX, nRectY + nRadius, 1, nInternalHeight, col);
	FillRect(nRectX + nRectWidth - 1, nRectY + nRadius, 1, nInternalHeight, col);
}

void Surface32::FillRoundRect(const Rectangle& rect, int32 nRadius, Color32 col) const
{
	int32 nRectX = rect.m_ptTopLeft.m_nX;
	int32 nRectY = rect.m_ptTopLeft.m_nY;
	int32 nRectWidth = rect.m_dimension.m_nWidth;
	int32 nRectHeight = rect.m_dimension.m_nHeight;

	const int32 nInternalWidth = nRectWidth - nRadius * 2;
	const int32 nInternalHeight = nRectHeight - nRadius * 2;

	DrawRoundCorner(nRectX, nRectY, nRadius, ctTopLeftFilled, col);
	DrawRoundCorner(nRectX + nRectWidth - nRadius, nRectY, nRadius, ctTopRightFilled, col);
	DrawRoundCorner(nRectX + nRectWidth - nRadius, nRectY + nRectHeight - nRadius, nRadius, ctBottomRightFilled, col);
	DrawRoundCorner(nRectX, nRectY + nRectHeight - nRadius, nRadius, ctBottomLeftFilled, col);

	FillRect(nRectX + nRadius, nRectY, nInternalWidth, nRadius, col);
	FillRect(nRectX + nRadius, nRectY + nRectHeight - nRadius, nInternalWidth, nRadius, col);

	FillRect(nRectX, nRectY + nRadius, nRectWidth, nInternalHeight, col);
}

void Surface32::DrawRoundCornerAA(int32 nCornerX, int32 nCornerY, int32 nRadius, CornerType eType, Color32 col) const
{
	// 2 low bits : the corner position
	// 3rd bit : circle or disc
	int nType = int(eType);

	int nCorner = nType & 0x03;
	bool bFilled = (nType & ctFilled) != 0;

	int32 nCenterX;
	int32 nCenterY;

	if( nCorner == ctTopLeft)
	{
		 nCenterX = nCornerX + nRadius;
		 nCenterY = nCornerY + nRadius;
	}
	else if( nCorner == ctTopRight)
	{
		nCenterX = nCornerX - 1;
		nCenterY = nCornerY + nRadius;
	}
	else if( nCorner == ctBottomRight)
	{
		nCenterX = nCornerX - 1;
		nCenterY = nCornerY - 1;
	}
	else if( nCorner == ctBottomLeft)
	{
		nCenterX = nCornerX + nRadius;
		nCenterY = nCornerY - 1;
	}
	else
	{
		ASSERT(0);
		return;
	}

	const uint8 nConstAlpha = col.GetAlpha();
	const int32 nR2_16 = (nRadius * nRadius) << 16;

	int32 nX = 1;
	int32 nY = nRadius;
	int32 nT = 0;
	
	uint8 nLum1 = 0;
	uint8 nLum2 = 0;
	int32 nRes = 0;

	while(true)
	{
		int32 nX2_16 = (nX * nX) << 16;

		// NOTE : a lookup table for small values would be interesting
		int32 nDist256 = Math::IntegerSqrt(nR2_16 - nX2_16);

		int32 nDistRound256 = nDist256 & 0xffffff00;
		if( (nDist256 & 0x00ff) != 0)
		{
			nDistRound256 += 0x100;
		}
		nRes = nDistRound256 - nDist256;
		
		if( nX + 2 > nY)
			break;

		if( nRes < nT)
		{
			nY--;
		}

		nLum1 = uint8( (nConstAlpha * (256 - nRes)) >> 8);
		nLum2 = uint8( (nConstAlpha * nRes) >> 8);
		
		if( nCorner == ctTopLeft)
		{
			col.SetAlpha(nLum1);
			DrawPixel(nCenterX - nY, nCenterY - nX, col);
			DrawPixel(nCenterX - nX, nCenterY - nY, col);

			if( !bFilled)
			{
				// Side pixels
				int32 nX1 = nCenterX - nY + 1;
				int32 nX2 = nCenterX - nX;

				col.SetAlpha(nLum2);
				DrawPixel(nX1, nCenterY - nX, col);
				
				if( nX1 != nX2)
				{
					DrawPixel(nX2, nCenterY - nY + 1, col);
				}
			}
			else
			{
				col.SetAlpha(nConstAlpha);
				DrawHrzLine(nCenterX - nY + 1, nCenterX - nX, nCenterY - nX, col);
				DrawVrtLine(nCenterX - nX, nCenterY - nY + 1, nCenterY - nX - 1, col);
			}
		}
		else if( nCorner == ctTopRight)
		{
			col.SetAlpha(nLum1);
			DrawPixel(nCenterX + nY, nCenterY - nX, col);
			DrawPixel(nCenterX + nX, nCenterY - nY, col);

			if( !bFilled)
			{
				// Side pixels
				int32 nX1 = nCenterX + nY - 1;
				int32 nX2 = nCenterX + nX;

				col.SetAlpha(nLum2);
				DrawPixel(nX1, nCenterY - nX, col);
				
				if( nX1 != nX2)
				{
					DrawPixel(nX2, nCenterY - nY + 1, col);
				}
			}
			else
			{
				col.SetAlpha(nConstAlpha);
				DrawHrzLine(nCenterX + nX + 1, nCenterX + nY - 1, nCenterY - nX, col);
				DrawVrtLine(nCenterX + nX, nCenterY - nY + 1, nCenterY - nX, col);
			}
		}
		else if( nCorner == ctBottomRight)
		{
			col.SetAlpha(nLum1);
			DrawPixel(nCenterX + nY, nCenterY + nX, col);
			DrawPixel(nCenterX + nX, nCenterY + nY, col);
			
			if( !bFilled)
			{
				// Side pixels
				int32 nX1 = nCenterX + nY - 1;
				int32 nX2 = nCenterX + nX;

				col.SetAlpha(nLum2);
				DrawPixel(nX1, nCenterY + nX, col);
				
				if( nX1 != nX2)
				{
					DrawPixel(nX2, nCenterY + nY - 1, col);
				}
			}
			else
			{
				col.SetAlpha(nConstAlpha);
				DrawHrzLine(nCenterX + nX + 1, nCenterX + nY - 1, nCenterY + nX, col);
				DrawVrtLine(nCenterX + nX, nCenterY + nX, nCenterY + nY - 1, col);
			}
		}
		else if( nCorner == ctBottomLeft)
		{
			col.SetAlpha(nLum1);
			DrawPixel(nCenterX - nY, nCenterY + nX, col);
			DrawPixel(nCenterX - nX, nCenterY + nY, col);

			if( !bFilled)
			{
				// Side pixels
				int32 nX1 = nCenterX - nY + 1;
				int32 nX2 = nCenterX - nX;

				col.SetAlpha(nLum2);
				DrawPixel(nX1, nCenterY + nX, col);
				
				if( nX1 != nX2)
				{
					DrawPixel(nX2, nCenterY + nY - 1, col);
				}
			}
			else
			{
				col.SetAlpha(nConstAlpha);
				DrawHrzLine(nCenterX - nY + 1, nCenterX - nX - 1, nCenterY + nX, col);
				DrawVrtLine(nCenterX - nX, nCenterY + nX, nCenterY + nY - 1, col);
			}
		}

		nX++;
		nT = nRes;
	}
	
	if( nX != nY)
	{
		nY--;

		col.SetAlpha(nLum1);

		// Draw last common pixel
		if( nCorner == ctTopLeft)
		{
			DrawPixel(nCenterX - nY, nCenterY - nX, col);
		}
		else if( nCorner == ctTopRight)
		{
			DrawPixel(nCenterX + nY, nCenterY - nX, col);
		}
		else if( nCorner == ctBottomRight)
		{
			DrawPixel(nCenterX + nY, nCenterY + nX, col);
		}
		else if( nCorner == ctBottomLeft)
		{
			DrawPixel(nCenterX - nY, nCenterY + nX, col);
		}
	}
}

void Surface32::DrawRoundRectAA(const Rectangle& rect, int32 nRadius, Color32 col) const
{
	int32 nRectX = rect.m_ptTopLeft.m_nX;
	int32 nRectY = rect.m_ptTopLeft.m_nY;
	int32 nRectWidth = rect.m_dimension.m_nWidth;
	int32 nRectHeight = rect.m_dimension.m_nHeight;

	const int32 nInternalWidth = nRectWidth - nRadius * 2;
	const int32 nInternalHeight = nRectHeight - nRadius * 2;

	DrawRoundCornerAA(nRectX, nRectY, nRadius, ctTopLeft, col);
	DrawRoundCornerAA(nRectX + nRectWidth - nRadius, nRectY, nRadius, ctTopRight, col);
	DrawRoundCornerAA(nRectX + nRectWidth - nRadius, nRectY + nRectHeight - nRadius, nRadius, ctBottomRight, col);
	DrawRoundCornerAA(nRectX, nRectY + nRectHeight - nRadius, nRadius, ctBottomLeft, col);

	FillRect(nRectX + nRadius, nRectY, nInternalWidth, 1, col);
	FillRect(nRectX + nRadius, nRectY + nRectHeight - 1, nInternalWidth, 1, col);

	FillRect(nRectX, nRectY + nRadius, 1, nInternalHeight, col);
	FillRect(nRectX + nRectWidth - 1, nRectY + nRadius, 1, nInternalHeight, col);
}

void Surface32::FillRoundRectAA(const Rectangle& rect, int32 nRadius, Color32 col) const
{
	int32 nRectX = rect.m_ptTopLeft.m_nX;
	int32 nRectY = rect.m_ptTopLeft.m_nY;
	int32 nRectWidth = rect.m_dimension.m_nWidth;
	int32 nRectHeight = rect.m_dimension.m_nHeight;

	const int32 nInternalWidth = nRectWidth - nRadius * 2;
	const int32 nInternalHeight = nRectHeight - nRadius * 2;

	DrawRoundCornerAA(nRectX, nRectY, nRadius, ctTopLeftFilled, col);
	DrawRoundCornerAA(nRectX + nRectWidth - nRadius, nRectY, nRadius, ctTopRightFilled, col);
	DrawRoundCornerAA(nRectX + nRectWidth - nRadius, nRectY + nRectHeight - nRadius, nRadius, ctBottomRightFilled, col);
	DrawRoundCornerAA(nRectX, nRectY + nRectHeight - nRadius, nRadius, ctBottomLeftFilled, col);

	FillRect(nRectX + nRadius, nRectY, nInternalWidth, nRadius, col);
	FillRect(nRectX + nRadius, nRectY + nRectHeight - nRadius, nInternalWidth, nRadius, col);

	FillRect(nRectX, nRectY + nRadius, nRectWidth, nInternalHeight, col);
}
