///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "Master.h"
#include "SurfaceBase.h"
#include "Memory.h"
#include "Math.h"

//////////////////////////////////////////////////////////////////////
#include "SurfaceBaseImpl.h"
//////////////////////////////////////////////////////////////////////

using namespace chustd;

SurfaceBase::SurfaceBase()
{
	m_pImpl = new SurfaceBaseImpl();
}

SurfaceBase::SurfaceBase(SurfaceBaseImpl* pImpl)
{
	m_pImpl = pImpl;
}

SurfaceBase::~SurfaceBase()
{
	FreeBuffer();
	delete m_pImpl;
}

int32 SurfaceBase::GetWidth() const
{
	return m_pImpl->m_nWidth;
}

int32 SurfaceBase::GetHeight() const
{
	return m_pImpl->m_nHeight;
}

int32 SurfaceBase::GetPitch() const
{
	return m_pImpl->m_nPitch;
}

void* SurfaceBase::GetBuffer() const
{
	return m_pImpl->m_pBuffer;
}

Rectangle SurfaceBase::GetRectangle() const
{
	Rectangle rect;
	rect.m_ptTopLeft.m_nX = 0;
	rect.m_ptTopLeft.m_nY = 0;
	rect.m_dimension.m_nWidth = GetWidth();
	rect.m_dimension.m_nHeight = GetHeight();
	return rect;
}

// Creates a sub-surface from an existing surface
bool SurfaceBase::SetBufferHelper(const SurfaceBase& surfSource, const Rectangle& rect)
{
	const int32 nSourceX = 0;
	const int32 nSourceY = 0;
	const int32 nSourceWidth = surfSource.GetWidth();
	const int32 nSourceHeight = surfSource.GetHeight();
	const int32 nSourcePitch = surfSource.GetPitch();
	const int8  nSourceBitsPerPixel = surfSource.m_pImpl->m_nBitsPerPixel;

	// Small consistency check
	if( nSourceWidth > nSourcePitch )
	{
		ASSERT(false);
		return false;
	}

	// Clip the source rectangle against the source surface
	Rectangle rectSurface(nSourceX, nSourceY, nSourceWidth, nSourceHeight);
	Rectangle rectClipped(rect);
	rectClipped.Translate(surfSource.m_pImpl->m_nDstOffsetX, surfSource.m_pImpl->m_nDstOffsetY);
	Rectangle rectTranslated = rectClipped;

	rectClipped.ClipIn(rectSurface);

	// Unallocate if data were previously allocated
	FreeBuffer();

	//////////////////////////////////////////////////////////////////////////////////
	// Compute the destication offset in case the rectangle is outside of the source surface
	int32 nDstOffsetX = 0;
	int32 nDstOffsetY = 0;

	if( rectTranslated.m_ptTopLeft.m_nX < 0)
	{
		nDstOffsetX = rectTranslated.m_ptTopLeft.m_nX;
	}

	if( rectTranslated.m_ptTopLeft.m_nY < 0)
	{
		nDstOffsetY = rectTranslated.m_ptTopLeft.m_nY;
	}
	//////////////////////////////////////////////////////////////////////////////////
	const Point& ptSource = rectClipped.m_ptTopLeft;
	const Dimension& dimSource = rectClipped.m_dimension;

	int32 nBytesPerPixel = nSourceBitsPerPixel / 8;
	int32 nByteOffset = (ptSource.m_nY * nSourcePitch + ptSource.m_nX) * nBytesPerPixel;

	m_pImpl->m_pBuffer = (uint8*) surfSource.m_pImpl->m_pBuffer + nByteOffset;
	m_pImpl->m_nWidth = dimSource.m_nWidth;
	m_pImpl->m_nHeight = dimSource.m_nHeight;
	m_pImpl->m_nPitch = nSourcePitch;
	m_pImpl->m_nBitsPerPixel = nSourceBitsPerPixel;
	m_pImpl->m_nDstOffsetX = nDstOffsetX;
	m_pImpl->m_nDstOffsetY = nDstOffsetY;

	m_pImpl->m_bOwnBuffer = false;

	return true;
}

void SurfaceBase::FreeBuffer()
{
	if( !m_pImpl)
		return;

	if( m_pImpl->m_bOwnBuffer && m_pImpl->m_pBuffer != null)
	{
		//int32 nOffset = ((m_pImpl->m_nPitch * m_pImpl->m_nBitsPerPixel) / 8) * m_pImpl->m_nHeight;
		//uint8* pBuf = (uint8*) m_pImpl->m_pBuffer;

		chustd::Memory::Free64(m_pImpl->m_pBuffer);
	}
	m_pImpl->m_pBuffer = null;
}

void SurfaceBase::Empty()
{
	FreeBuffer();

	m_pImpl->m_nWidth = 0;
	m_pImpl->m_nHeight = 0;
	m_pImpl->m_nPitch = 0;
	m_pImpl->m_nBitsPerPixel = 0;

	m_pImpl->m_bOwnBuffer = false;
}

bool SurfaceBase::SetBuffer(void* pBuffer, int32 nWidth, int32 nHeight, int32 nPitch, int8 nBitsPerPixel)
{
	FreeBuffer();

	m_pImpl->m_pBuffer = pBuffer;
	m_pImpl->m_nWidth = nWidth;
	m_pImpl->m_nHeight = nHeight;
	m_pImpl->m_nPitch = nPitch;
	m_pImpl->m_nBitsPerPixel = nBitsPerPixel;

	m_pImpl->m_bOwnBuffer = false;

	return true;
}

bool SurfaceBase::Create(int32 nWidth, int32 nHeight, int32 nPitch, int8 nBitsPerPixel)
{
	FreeBuffer();

	const int32 nBytePitch = Math::DivCeil(nPitch * nBitsPerPixel, 8);
	m_pImpl->m_pBuffer = Memory::Alloc64(nBytePitch * nHeight);
	if( !m_pImpl->m_pBuffer)
	{
		return false;
	}

	m_pImpl->m_nWidth = nWidth;
	m_pImpl->m_nHeight = nHeight;
	m_pImpl->m_nPitch = nPitch;
	m_pImpl->m_nBitsPerPixel = nBitsPerPixel;

	m_pImpl->m_bOwnBuffer = true;

	return true;
}

// Returns true if visible
bool SurfaceBase::ClipSurface(const SurfaceBase& surfSrc, int32& nDstX, int32& nDstY, SurfaceBaseImpl& sbiOut) const
{
	ASSERT(m_pImpl);

	const int32 nDestWidth = m_pImpl->m_nWidth;
	const int32 nDestHeight = m_pImpl->m_nHeight;

	int32 nNewSrcWidth = surfSrc.GetWidth();
	int32 nDstPixelStartX = 0;
	
	// Clipping en X
	if( nDstX < 0)
	{
		// Le sprite sort  gauche
		// Est-ce qu'il est compltement sorti  gauche ?
		if( nDstX + nNewSrcWidth <= 0)
		{
			// Note : le 0 est inclus. Exemple
			// Sprite sort en -7, largeur de 7 : en dehors, mais l'addition donne 0
			// Oui, il sort, pas la peine de le dessiner
			return false;
		}

		// Une partie reste visible, on modifie les informations de dpart
		const int32 nOutWidth = 0 - nDstX;
		nDstPixelStartX = nOutWidth;
		nDstX = 0;
		nNewSrcWidth -= nOutWidth;
	}

	if( nDstX + nNewSrcWidth > nDestWidth)
	{
		// Le sprite sort  droite
		// Est-ce qu'il est compltement sorti  droite ?
		if( nDstX > nDestWidth)
		{
			// Oui, il sort, pas la peine de le dessiner
			return false;
		}

		// Une partie reste visible, on modifie les informations de dpart
		const int32 nVisibleWidth = nDestWidth - nDstX;
		nNewSrcWidth = nVisibleWidth;
	}

	int32 nNewSrcHeight = surfSrc.GetHeight();
	int32 nDstPixelStartY = 0;

	// Clipping en Y
	if( nDstY < 0)
	{
		// Le sprite sort en haut
		if( nDstY + nNewSrcHeight <= 0)
		{
			return false;
		}

		// Une partie reste visible, on modifie les informations de dpart
		const int32 nOutHeight = 0 - nDstY;
		nDstPixelStartY = nOutHeight;
		nDstY = 0;
		nNewSrcHeight -= nOutHeight;
	}

	if( nDstY + nNewSrcHeight > nDestHeight)
	{
		// Le sprite sort en bas
		if( nDstY > nDestHeight)
		{
			return false;
		}

		// Une partie reste visible, on modifie les informations de dpart
		int nVisibleHeight = nDestHeight - nDstY;
		nNewSrcHeight = nVisibleHeight;
	}

	SurfaceBase surfOut(&sbiOut);
	surfOut.SetBufferHelper(surfSrc, Rectangle(nDstPixelStartX, nDstPixelStartY, nNewSrcWidth, nNewSrcHeight));
	surfOut.m_pImpl = null;

	return true;
}

void SurfaceBase::DoOffset(int32& nX, int32& nY) const
{
	if( m_pImpl == null)
		return;

	nX += m_pImpl->m_nDstOffsetX;
	nY += m_pImpl->m_nDstOffsetY;
}

void SurfaceBase::DoOffset(Point& pt) const
{
	if( m_pImpl == null)
		return;

	pt.m_nX += m_pImpl->m_nDstOffsetX;
	pt.m_nY += m_pImpl->m_nDstOffsetY;
}

void SurfaceBase::DoOffset(chustd::Rectangle& rect) const
{
	if( m_pImpl == null)
		return;

	rect.m_ptTopLeft.m_nX += m_pImpl->m_nDstOffsetX;
	rect.m_ptTopLeft.m_nY += m_pImpl->m_nDstOffsetY;
}

void SurfaceBase::SetDstOffset(int32 nX, int32 nY)
{
	m_pImpl->m_nDstOffsetX = nX;
	m_pImpl->m_nDstOffsetY = nY;
}
