///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "Master.h"
#include "TextEncoding.h"
#include "StringBuilder.h"
#include "StringData.h"

using namespace chustd;

///////////////////////////////////////////////////////////////////////////////
Array<const TextEncoding*> TextEncoding::ms_apKnownEncodings;

const uint8 k_nReplaceChar = '?';

///////////////////////////////////////////////////////////////////////////////
// Globals as we want them to be all constructed so they are registered in ms_apKnownEncodings
static TextEncodingWindows1252 g_teWindows1252;
static TextEncodingIso8859_1   g_teIso8859_1;
static TextEncodingIso8859_15  g_teIso8859_15;
static TextEncodingUtf8        g_teIsoUtf8;
static TextEncodingUtf16       g_teIsoUtf16Le(boLittleEndian);
static TextEncodingUtf16       g_teIsoUtf16Be(boBigEndian);
static TextEncodingUtf32       g_teIsoUtf32Le(boLittleEndian);
static TextEncodingUtf32       g_teIsoUtf32Be(boBigEndian);
///////////////////////////////////////////////////////////////////////////////
const TextEncoding& TextEncoding::Windows1252()
{
	return g_teWindows1252;
}

const TextEncoding& TextEncoding::Iso8859_1()
{
	return g_teIso8859_1;
}

const TextEncoding& TextEncoding::Iso8859_15()
{
	return g_teIso8859_15;
}

const TextEncoding& TextEncoding::Utf8()
{
	return g_teIsoUtf8;
}

const TextEncoding& TextEncoding::Utf16Le()
{
	return g_teIsoUtf16Le;
}

const TextEncoding& TextEncoding::Utf16Be()
{
	return g_teIsoUtf16Be;
}

const TextEncoding& TextEncoding::Utf32Le()
{
	return g_teIsoUtf32Le;
}

const TextEncoding& TextEncoding::Utf32Be()
{
	return g_teIsoUtf32Be;
}

///////////////////////////////////////////////////////////////////////////////

TextEncoding::TextEncoding()
{
	ms_apKnownEncodings.Add(this);
}

TextEncoding::~TextEncoding()
{
	ms_apKnownEncodings.Remove(this);
}

const Array<const TextEncoding*>& TextEncoding::GetKnownEncodings()
{
	return ms_apKnownEncodings;	
}

String TextEncoding::BytesToString(const ByteArray& aBytes) const
{
	const int32 nSize = aBytes.GetSize();
	int32 nStart = 0;

	StringBuilder sb;
	while(true)
	{
		if( nStart >= nSize)
			break;

		uint32 nCodePoint = 0;
		ExtractStatus es = ExtractCodePoint(aBytes, nStart, nCodePoint);
		if( es != esNoError)
			break;

		sb += nCodePoint;
	}
	return sb.ToString();
}

ByteArray TextEncoding::StringToBytes(const String& str) const
{
	ByteArray aBytes;
	aBytes.SetAsReturnValue();
	aBytes.EnsureCapacity(str.GetLength());

	int32 nIndex = 0;
	while(true)
	{
		uint32 nCodePoint = str.GetCodePoint(nIndex);
		if( nCodePoint == 0xffffffff)
		{
			break;
		}
		
		InsertCodePoint(nCodePoint, aBytes);
	}
	return aBytes;
}

///////////////////////////////////////////////////////////////////////////////////////////
// Optimized version for 8 bits encodings
String TextEncoding8Bits::BytesToString(const ByteArray& aBytes) const
{
	const int32 nSize = aBytes.GetSize();
	int32 nStart = 0;

	String strOut;
	uint16* pDst = (uint16*) strOut.GetUnsafeBuffer(nSize);

	for(int32 i = 0; i < nSize; ++i)
	{
		uint32 nCodePoint = 0;
		ExtractStatus es = ExtractCodePoint(aBytes, nStart, nCodePoint);
		if( es != esNoError)
			break;

		// No known 8 bits encoding that encodes CodePoints > 65536,
		// but a text encoding class that does not follow this rule is 
		// still free to override this function and do specific handling
		pDst[i] = uint16(nCodePoint);
	}
	
	return strOut;
}

///////////////////////////////////////////////////////////////////////////////////////////
// windows-1252

// From 0x80 to 0x9f
static const uint16 TextEncodingWindows1252_aConvert[32] = 
{
	0x20AC,	// 0x80 EURO SIGN
    0x0081,	// 0x81 UNDEFINED
	0x201A,	// 0x82 SINGLE LOW-9 QUOTATION MARK
	0x0192,	// 0x83 LATIN SMALL LETTER F WITH HOOK
	0x201E,	// 0x84 DOUBLE LOW-9 QUOTATION MARK
	0x2026,	// 0x85 HORIZONTAL ELLIPSIS
	0x2020,	// 0x86 DAGGER
	0x2021,	// 0x87 DOUBLE DAGGER
	0x02C6,	// 0x88 MODIFIER LETTER CIRCUMFLEX ACCENT
	0x2030,	// 0x89 PER MILLE SIGN
	0x0160,	// 0x8A LATIN CAPITAL LETTER S WITH CARON
	0x2039,	// 0x8B SINGLE LEFT-POINTING ANGLE QUOTATION MARK
	0x0152,	// 0x8C LATIN CAPITAL LIGATURE OE
    0x008D,	// 0x8D UNDEFINED
	0x017D,	// 0x8E LATIN CAPITAL LETTER Z WITH CARON
    0x008F,	// 0x8F UNDEFINED
    0x0090,	// 0x90 UNDEFINED
	0x2018,	// 0x91 LEFT SINGLE QUOTATION MARK
	0x2019,	// 0x92 RIGHT SINGLE QUOTATION MARK
	0x201C,	// 0x93 LEFT DOUBLE QUOTATION MARK
	0x201D,	// 0x94 RIGHT DOUBLE QUOTATION MARK
	0x2022,	// 0x95 BULLET
	0x2013,	// 0x96 EN DASH
	0x2014,	// 0x97 EM DASH
	0x02DC,	// 0x98 SMALL TILDE
	0x2122,	// 0x99 TRADE MARK SIGN
	0x0161,	// 0x9A LATIN SMALL LETTER S WITH CARON
	0x203A,	// 0x9B SINGLE RIGHT-POINTING ANGLE QUOTATION MARK
	0x0153,	// 0x9C LATIN SMALL LIGATURE OE
    0x009D,	// 0x9D	UNDEFINED
	0x017E,	// 0x9E LATIN SMALL LETTER Z WITH CARON
	0x0178	// 0x9F LATIN CAPITAL LETTER Y WITH DIAERESIS
};

String TextEncodingWindows1252::GetName() const
{
	static const String strName = L"windows-1252";
	return strName;
}

TextEncoding::ExtractStatus TextEncodingWindows1252::ExtractCodePoint(const ByteArray& aBytes, int32& nStart, uint32& nCodePoint) const
{
	const int32 nSize = aBytes.GetSize();
	if( nStart >= nSize)
	{
		ASSERT(0); // Bad param, warn the user
		return esBadParam;
	}

	uint8 nValue = aBytes[nStart];
	nStart++;

	if( nValue < 0x0080 || nValue > 0x009f)
	{
		// Compatible value
		nCodePoint = nValue;
	}
	else
	{
		nValue -= 0x0080;
		nCodePoint = TextEncodingWindows1252_aConvert[nValue];
	}
	return esNoError;
}

void TextEncodingWindows1252::InsertCodePoint(uint32 nCodePoint, ByteArray& aBytes) const
{
	if( nCodePoint < 0x0080 || (0x009f < nCodePoint && nCodePoint <= 0x00ff))
	{
		// Compatible value
		aBytes.Add( uint8(nCodePoint));
		return;
	}

	// Look for our code point
	for(int32 i = 0; i < 32; ++i)
	{
		if( nCodePoint == TextEncodingWindows1252_aConvert[i])
		{
			// Found !
			uint32 nResult = 0x0080 + i;
			aBytes.Add( uint8(nResult));
			return;
		}
	}

	aBytes.Add(k_nReplaceChar);
}
///////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////
// ISO-8859-1

String TextEncodingIso8859_1::GetName() const
{
	static const String strName = L"ISO-8859-1";
	return strName;
}

TextEncoding::ExtractStatus TextEncodingIso8859_1::ExtractCodePoint(const ByteArray& aBytes, int32& nStart, uint32& nCodePoint) const
{
	const int32 nSize = aBytes.GetSize();
	if( nStart >= nSize)
	{
		ASSERT(0); // Bad param, warn the user
		return esBadParam;
	}

	const uint8 nValue = aBytes[nStart];
	nStart++;

	// Compatible value
	nCodePoint = nValue;
	return esNoError;
}

void TextEncodingIso8859_1::InsertCodePoint(uint32 nCodePoint, ByteArray& aBytes) const
{
	if( (0x0000 < nCodePoint && nCodePoint <= 0x00ff))
	{
		// Compatible value
		aBytes.Add( uint8(nCodePoint));
		return;
	}

	aBytes.Add(k_nReplaceChar);
}
///////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////
// ISO-8859-15

// From 0xA0 to 0xbf
static const wchar CharsetIso8859_15_aConvert[32] = 
{
	0x00A0,	// 0xA0 --NO-BREAK SPACE
	0x00A1,	// 0xA1 --INVERTED EXCLAMATION MARK
	0x00A2,	// 0xA2 --CENT SIGN
	0x00A3,	// 0xA3 --POUND SIGN
	0x20AC,	// 0xA4 EURO SIGN
	0x00A5,	// 0xA5 --YEN SIGN
	0x0160,	// 0xA6 LATIN CAPITAL LETTER S WITH CARON
	0x00A7,	// 0xA7 --SECTION SIGN
	0x0161,	// 0xA8 LATIN SMALL LETTER S WITH CARON
	0x00A9,	// 0xA9 --COPYRIGHT SIGN
	0x00AA,	// 0xAA --FEMININE ORDINAL INDICATOR
	0x00AB,	// 0xAB --LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
	0x00AC,	// 0xAC --NOT SIGN
	0x00AD,	// 0xAD --SOFT HYPHEN
	0x00AE,	// 0xAE --REGISTERED SIGN
	0x00AF,	// 0xAF --MACRON
	0x00B0,	// 0xB0 --DEGREE SIGN
	0x00B1,	// 0xB1 --PLUS-MINUS SIGN
	0x00B2,	// 0xB2 --SUPERSCRIPT TWO
	0x00B3,	// 0xB3 --SUPERSCRIPT THREE
	0x017D,	// 0xB4 LATIN CAPITAL LETTER Z WITH CARON
	0x00B5,	// 0xB5 --MICRO SIGN
	0x00B6,	// 0xB6 --PILCROW SIGN
	0x00B7,	// 0xB7 --MIDDLE DOT
	0x017E,	// 0xB8 LATIN SMALL LETTER Z WITH CARON
	0x00B9,	// 0xB9 --SUPERSCRIPT ONE
	0x00BA,	// 0xBA --MASCULINE ORDINAL INDICATOR
	0x00BB,	// 0xBB --RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
	0x0152,	// 0xBC LATIN CAPITAL LIGATURE OE
	0x0153,	// 0xBD LATIN SMALL LIGATURE OE
	0x0178,	// 0xBE LATIN CAPITAL LETTER Y WITH DIAERESIS
	0x00BF	// 0xBF --INVERTED QUESTION MARK
};

String TextEncodingIso8859_15::GetName() const
{
	static const String strName = L"ISO-8859-15";
	return strName;
}

TextEncoding::ExtractStatus TextEncodingIso8859_15::ExtractCodePoint(const ByteArray& aBytes, int32& nStart, uint32& nCodePoint) const
{
	const int32 nSize = aBytes.GetSize();
	if( nStart >= nSize)
	{
		ASSERT(0); // Bad param, warn the user
		return esBadParam;
	}

	uint8 nValue = aBytes[nStart];
	nStart++;

	if( nValue < 0x00a0 || nValue > 0x00bf)
	{
		// Compatible value
		nCodePoint = nValue;
	}
	else
	{
		nValue -= 0x00a0;
		nCodePoint = CharsetIso8859_15_aConvert[nValue];
	}
	return esNoError;
}

void TextEncodingIso8859_15::InsertCodePoint(uint32 nCodePoint, ByteArray& aBytes) const
{
	if( nCodePoint < 0x00a0 || (0x00bf < nCodePoint && nCodePoint <= 0x00ff))
	{
		// Compatible value
		aBytes.Add( uint8(nCodePoint));
		return;
	}

	// Look for our code point
	for(int32 i = 0; i < 32; ++i)
	{
		if( nCodePoint == CharsetIso8859_15_aConvert[i])
		{
			// Found !
			uint32 nResult = 0x00a0 + i;
			aBytes.Add( uint8(nResult));
			return;
		}
	}

	aBytes.Add(k_nReplaceChar);
}
///////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////
// UTF-8

String TextEncodingUtf8::GetName() const
{
	static const String strName = L"UTF-8";
	return strName;
}

TextEncoding::ExtractStatus TextEncodingUtf8::ExtractCodePoint(const ByteArray& aBytes, int32& nStart, uint32& nCodePoint) const
{
	const int32 nSize = aBytes.GetSize();
	if( nStart >= nSize)
	{
		return esBadParam;
	}

	int8 nByteSequence = 0;
	uint32 nChar = 0;

	const uint8* pSrc = aBytes.GetPtr();
	
	uint8 nByte = pSrc[nStart];
	nStart++;

	if( nByte <= 0x7F)
	{
		nCodePoint = nByte;
		return esNoError;
	}
	else if( (nByte & 0xE0) == 0xC0)
	{
		// 110xxxxx 10xxxxxx
		nChar = nByte & 0x1F;
		nByteSequence = 1;
	}
	else if( (nByte & 0xF0) == 0xE0)
	{
		// 1110xxxx 10xxxxxx 10xxxxxx
		nChar = nByte & 0x0F;
		nByteSequence = 2;
	}
	else if( (nByte & 0xF8) == 0xF0)
	{
		// 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
		nChar = nByte & 0x07;
		nByteSequence = 3;
	}
	else if( (nByte & 0xFC) == 0xF8)
	{
		// 111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
		nChar = nByte & 0x03;
		nByteSequence = 4;
	}
	else if( (nByte & 0xFE) == 0xFC)
	{
		// 1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
		nChar = nByte & 0x01;
		nByteSequence = 5;
	}
	else
	{
		// Bad byte sequence starter
		return esUtf8BadByteSequenceStarter;
	}

	if( nStart + nByteSequence > nSize)
	{
		return esUnexpectedEndOfString;
	}

	for(int32 i = 0; i < nByteSequence; ++i)
	{
		uint8 nByte = pSrc[nStart + i];

		// Remaining bytes
		if( (nByte & 0xC0) != 0x80)
		{
			// Bad byte in sequence
			return esUtf8BadByteInSequence;
		}

		nChar <<= 6;
		nChar |= (nByte & 0x3F);
	}

	// Ok
	nCodePoint = nChar;
	nStart += nByteSequence;

	return esNoError;
}

String TextEncodingUtf8::BytesToString(const ByteArray& aBytes) const
{
	const int32 nSize = aBytes.GetSize();
	int32 nStart = 0;
	
	Array<uint16> aTmp;
	if( !aTmp.SetSize(nSize + nSize / 4))
	{
		return String();
	}
	
	String strOut;
	for(;;)
	{
		uint16* pDst = aTmp.GetPtr();
		const int32 nDstSize = aTmp.GetSize();
		
		int32 iDstUnit = 0;
		ExtractStatus es = esNoError;

		for(; iDstUnit < nSize; ++iDstUnit)
		{
			uint32 nCodePoint = 0;
			es = ExtractCodePoint(aBytes, nStart, nCodePoint);
			if( es != esNoError)
			{
				break;
			}
			
			if( nCodePoint <= 0x0000ffff )
			{
				pDst[iDstUnit] = uint16(nCodePoint);
			}
			else
			{
				// Manage surrogate pairs
				uint16 nUnit0, nUnit1;
				int32 nCodeUnitCount = String::CodePointToCodeUnits(nCodePoint, nUnit0, nUnit1);
				if( nCodeUnitCount == 1 )
				{
					pDst[iDstUnit] = nUnit0;
				}
				else if( nCodeUnitCount == 2 )
				{
					pDst[iDstUnit] = nUnit0;
					iDstUnit++;
					pDst[iDstUnit] = nUnit1;
				}
			}
		}
		
		strOut = strOut + String((wchar*)(pDst), iDstUnit);
		if( es != esNoError )
		{
			break;
		}
	}
	return strOut;
}

void TextEncodingUtf8::InsertCodePoint(uint32 nCodePoint, ByteArray& aBytes) const
{
	if( nCodePoint <= 0x0000007F)
	{
		uint8 nByte0 = uint8(nCodePoint);
		aBytes.Add(nByte0);
	}
	else if( nCodePoint <= 0x000007FF)
	{
		// 110xxxxx 10xxxxxx (11 bits)
		uint8 nByte0 = uint8( (nCodePoint >> 6)     | 0x00C0);
		uint8 nByte1 = uint8( (nCodePoint & 0x003F) | 0x0080);

		aBytes.Add(nByte0);
		aBytes.Add(nByte1);
	}
	else if( nCodePoint <= 0x0000FFFF)
	{
		// 1110xxxx 10xxxxxx 10xxxxxx (16 bits)
		uint8 nByte0 = uint8( (nCodePoint >> 12)           | 0x00E0);
		uint8 nByte1 = uint8( ((nCodePoint >> 6) & 0x003F) | 0x0080);
		uint8 nByte2 = uint8( ((nCodePoint) & 0x003F)      | 0x0080);

		aBytes.Add(nByte0);
		aBytes.Add(nByte1);
		aBytes.Add(nByte2);
	}
	else if( nCodePoint <= 0x0010FFFF)
	{
		// 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx (21 bits)
		uint8 nByte0 = uint8(  (nCodePoint >> 18)           | 0x00F0);
		uint8 nByte1 = uint8( ((nCodePoint >> 12) & 0x003F) | 0x0080);
		uint8 nByte2 = uint8( ((nCodePoint >>  6) & 0x003F) | 0x0080);
		uint8 nByte3 = uint8( ((nCodePoint      ) & 0x003F) | 0x0080);

		aBytes.Add(nByte0);
		aBytes.Add(nByte1);
		aBytes.Add(nByte2);
		aBytes.Add(nByte3);
	}
	else
	{
		// Invalid value
		// Error
		return;
	}
}


///////////////////////////////////////////////////////////////////////////////////////////
// UTF-16

TextEncodingUtf16::TextEncodingUtf16(ByteOrder eByteOrder) : m_eByteOrder(eByteOrder)
{

}

String TextEncodingUtf16::GetName() const
{
	static const String strNameBe = L"UTF-16LE";
	static const String strNameLe = L"UTF-16BE";
	
	if( m_eByteOrder == boLittleEndian)
	{
		return strNameLe;
	}
	return strNameBe;
}

TextEncoding::ExtractStatus TextEncodingUtf16::ExtractCodePoint(const ByteArray& aBytes, int32& nStart, uint32& nCodePoint) const
{
	const int32 nSize = aBytes.GetSize();
	if( nStart >= nSize)
	{
		ASSERT(0); // Bad param, warn the user
		return esBadParam;
	}

	const uint8* pSrc = aBytes.GetPtr();

	// Is next byte out of bound ?
	if( nStart + 1 >= nSize)
	{
		return esUnexpectedEndOfString; // Unexpected end of string
	}

	uint8 nByte0 = pSrc[nStart];
	uint8 nByte1 = pSrc[nStart + 1];
	
	uint16 nHi = 0;
	if( m_eByteOrder == boBigEndian)
	{
		nHi = uint8((nByte0 << 8) | nByte1);
	}
	else
	{
		nHi = uint8((nByte1 << 8) | nByte0);
	}
	nStart += 2;

	// A surrogate ?
	if( nHi < 0xD800 || nHi > 0xDFFF)
	{
		// Nop, standard char
		nCodePoint = nHi;
		return esNoError;
	}

	if( 0xDBFF < nHi)
	{
		// Low surrogate used first, this is an error
		return esUtf16LowSurrogateFoundFirst;
	}

	/////////////////////////////////////////////
	// Get the second surrogate (low)
	if( nStart + 2 < nSize)
	{
		return esUnexpectedEndOfString; // Unexpected end of string
	}

	uint8 nByte2 = pSrc[nStart];
	uint8 nByte3 = pSrc[nStart + 1];
	
	uint16 nLo = 0;
	if( m_eByteOrder == boBigEndian)
	{
		nLo = uint8((nByte2 << 8) | nByte3);
	}
	else
	{
		nLo = uint8((nByte3 << 8) | nByte2);
	}
	nStart += 2;


	// A low surrogate ?
	if( !(0xDBFF < nLo && nLo <= 0xDFFF))
	{
		// Nop	
		return esUtf16SecondSurrogateNotFound;
	}

	nCodePoint = 0x10000 + ((nHi - 0xD800) << 10) + (nLo - 0xDC00);

	return esNoError;
}

void TextEncodingUtf16::InsertCodePoint(uint32 nCodePoint, ByteArray& aBytes) const
{
	uint16 nUtf16First, nUtf16Next;
	int32 nCodeUnitCount = String::CodePointToCodeUnits(nCodePoint, nUtf16First, nUtf16Next);
	if( nCodeUnitCount == 0)
		return;

	uint8 nHi = uint8(nUtf16First >> 8);
	uint8 nLo = uint8(nUtf16First & 0x00ff);

	if( m_eByteOrder == boBigEndian)
	{
		aBytes.Add(nHi);
		aBytes.Add(nLo);
	}
	else
	{
		aBytes.Add(nLo);
		aBytes.Add(nHi);
	}

	if( nCodeUnitCount == 1)
		return;

	uint8 nHi2 = uint8(nUtf16Next >> 8);
	uint8 nLo2 = uint8(nUtf16Next & 0x00ff);

	if( m_eByteOrder == boBigEndian)
	{
		aBytes.Add(nHi2);
		aBytes.Add(nLo2);
	}
	else
	{
		aBytes.Add(nLo2);
		aBytes.Add(nHi2);
	}

}


///////////////////////////////////////////////////////////////////////////////////////////
// UTF-32

TextEncodingUtf32::TextEncodingUtf32(ByteOrder eByteOrder) : m_eByteOrder(eByteOrder)
{

}

String TextEncodingUtf32::GetName() const
{
	static const String strNameBe = L"UTF-32LE";
	static const String strNameLe = L"UTF-32BE";
	
	if( m_eByteOrder == boLittleEndian)
	{
		return strNameLe;
	}
	return strNameBe;
}


////////////////////////////////////
TextEncoding::ExtractStatus TextEncodingUtf32::ExtractCodePoint(const ByteArray& aBytes, int32& nStart, uint32& nCodePoint) const
{
	const int32 nSize = aBytes.GetSize();
	if( nStart >= nSize)
	{
		ASSERT(0); // Bad param, warn the user
		return esBadParam;
	}

	if( (nStart + 4) >= nSize)
	{
		return esUnexpectedEndOfString;
	}

	uint8 nByte0 = aBytes[nStart];
	uint8 nByte1 = aBytes[nStart + 1];
	uint8 nByte2 = aBytes[nStart + 2];
	uint8 nByte3 = aBytes[nStart + 3];

	if( m_eByteOrder == boBigEndian)
	{
		nCodePoint = (nByte0 << 24) | (nByte1 << 16) | (nByte2 << 8) | nByte3;
	}
	else
	{
		nCodePoint = (nByte3 << 24) | (nByte2 << 16) | (nByte1 << 8) | nByte0;
	}

	nStart += 4;
	return esNoError;
}

void TextEncodingUtf32::InsertCodePoint(uint32 nCodePoint, ByteArray& aBytes) const
{
	uint8 nByte0 = uint8( (nCodePoint >> 24) & 0x00ff);
	uint8 nByte1 = uint8( (nCodePoint >> 16) & 0x00ff);
	uint8 nByte2 = uint8( (nCodePoint >> 8) & 0x00ff);
	uint8 nByte3 = uint8( (nCodePoint) & 0x00ff);

	if( m_eByteOrder == boBigEndian)
	{
		aBytes.Add(nByte0);
		aBytes.Add(nByte1);
		aBytes.Add(nByte2);
		aBytes.Add(nByte3);
	}
	else
	{
		aBytes.Add(nByte3);
		aBytes.Add(nByte2);
		aBytes.Add(nByte1);
		aBytes.Add(nByte0);
	}
}
