///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#ifndef CHUSTD_TEXTENCODING_H
#define CHUSTD_TEXTENCODING_H

#include "Array.h"

namespace chustd {

class TextEncoding
{
public:
	virtual String    BytesToString(const ByteArray& aBytes) const;
	ByteArray StringToBytes(const String& str) const;

	enum ExtractStatus
	{
		esNoError,
		esUnexpectedEndOfString,
		esBadParam,

		esUtf8BadByteSequenceStarter,
		esUtf8BadByteInSequence,

		esUtf16SecondSurrogateNotFound,
		esUtf16LowSurrogateFoundFirst,
	};

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Overridden by derived classes

	// Extracts a code point from an array of encoded bytes
	// aBytes : the array of encoded bytes
	// nStart : (in) index in aBytes where to start (out) index of the next code point in aBytes 
	// nCodePoint : (out) result
	// Returns esNoError upon success
	virtual ExtractStatus ExtractCodePoint(const ByteArray& aBytes, int32& nStart, uint32& nCodePoint) const = 0;

	// Inserts a code point at the end of an array of encoded bytes
	// nCodePoint : the code point to insert
	// aBytes : the array of encoded bytes
	virtual void          InsertCodePoint(uint32 nCodePoint, ByteArray& aBytes) const = 0;

	virtual String GetName() const = 0;
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	static const Array<const TextEncoding*>& GetKnownEncodings();

public:
	static const TextEncoding& Windows1252();
	static const TextEncoding& Iso8859_1();
	static const TextEncoding& Iso8859_15();
	static const TextEncoding& Utf8();
	static const TextEncoding& Utf16Le();
	static const TextEncoding& Utf16Be();
	static const TextEncoding& Utf32Le();
	static const TextEncoding& Utf32Be();

protected:
	TextEncoding();
	virtual ~TextEncoding();

private:
	static Array<const TextEncoding*> ms_apKnownEncodings;
};

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
class TextEncoding8Bits : public TextEncoding
{
public:
	virtual String    BytesToString(const ByteArray& aBytes) const;
};

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
class TextEncodingWindows1252 : public TextEncoding8Bits
{
public:
	virtual ExtractStatus ExtractCodePoint(const ByteArray& aBytes, int32& nStart, uint32& nCodePoint) const;
	virtual void          InsertCodePoint(uint32 nCodePoint, ByteArray& aBytes) const;
	virtual String GetName() const;
};

class TextEncodingIso8859_1 : public TextEncoding8Bits
{
public:
	virtual ExtractStatus ExtractCodePoint(const ByteArray& aBytes, int32& nStart, uint32& nCodePoint) const;
	virtual void          InsertCodePoint(uint32 nCodePoint, ByteArray& aBytes) const;
	virtual String GetName() const;
};

class TextEncodingIso8859_15 : public TextEncoding8Bits
{
public:
	virtual ExtractStatus ExtractCodePoint(const ByteArray& aBytes, int32& nStart, uint32& nCodePoint) const;
	virtual void          InsertCodePoint(uint32 nCodePoint, ByteArray& aBytes) const;
	virtual String GetName() const;
};

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
class TextEncodingUtf8 : public TextEncoding
{
public:
	virtual ExtractStatus ExtractCodePoint(const ByteArray& aBytes, int32& nStart, uint32& nCodePoint) const;
	virtual void          InsertCodePoint(uint32 nCodePoint, ByteArray& aBytes) const;
	virtual String GetName() const;

	static ExtractStatus ExtractChar(const ByteArray& aBytes, int32& nStartAt, uint32& nUcs4);
	
	virtual String BytesToString(const ByteArray& aBytes) const;
};

class TextEncodingUtf16 : public TextEncoding
{
public:
	TextEncodingUtf16(ByteOrder eByteOrder);

	virtual ExtractStatus ExtractCodePoint(const ByteArray& aBytes, int32& nStart, uint32& nCodePoint) const;
	virtual void          InsertCodePoint(uint32 nCodePoint, ByteArray& aBytes) const;
	virtual String GetName() const;

	static ExtractStatus ExtractChar(const ByteArray& aBytes, int32& nStartAt, uint32& nUcs4, ByteOrder eByteOrder);
private:
	ByteOrder m_eByteOrder;
};

class TextEncodingUtf32 : public TextEncoding
{
public:
	TextEncodingUtf32(ByteOrder eByteOrder);

	virtual ExtractStatus ExtractCodePoint(const ByteArray& aBytes, int32& nStart, uint32& nCodePoint) const;
	virtual void          InsertCodePoint(uint32 nCodePoint, ByteArray& aBytes) const;
	virtual String GetName() const;

private:
	ByteOrder m_eByteOrder;
};

} // namespace chustd

#endif // ndef CHUSTD_TEXTENCODING_H
