///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "Master.h"
#include "Tga.h"
#include "File.h"
#include "Math.h"

//////////////////////////////////////////////////////////////////////
using namespace chustd;
//////////////////////////////////////////////////////////////////////

Tga::Tga()
{
	Initialize();
}

Tga::~Tga()
{

}

void Tga::Initialize()
{
	m_nWidth = 0;
	m_nHeight = 0;
	m_nLastError = 0;

	m_eAttributesType = atUndef;
	m_epf = pfUnknown;
	m_pal.m_nCount = 0;
}

bool Tga::IsTga(IFile& file)
{
	// Unfortunately the TGA format does not include a signature in the header
	
	// At least read some constant values in the header
	uint8 aHeader[18];
	if( file.Read(aHeader, 18) != 18)
	{
		return false;
	}

	TgaFooter tf;
	if( ReadFooter(file, tf) )
	{
		// YES ! For sure it is a TGA file
		return true;
	}

	// Try what we can with the header :-/
	if( (aHeader[1] & 0xFE) != 0)
	{
		// ColorMapType can be 0 or 1 only
		return false;
	}

	if( (aHeader[2] & 0xC0) != 0)
	{
		// Image type from 0 to 33 (assume max = 63)
		return false;
	}

	uint8 nColorMapDepth = aHeader[7];
	if( !(nColorMapDepth == 0 || nColorMapDepth == 15 || nColorMapDepth == 16
		|| nColorMapDepth == 24 || nColorMapDepth == 32))
	{
		// Invalid palette color format
		return false;	
	}
	
	if( !(0 < aHeader[16] && aHeader[16] <= 64))
	{
		// Depth out of range
		return false;
	}

	// One can say for sure that TGA is not the state of the art in graphics format
	return true;
}

bool Tga::LoadFromFile(IFile& file)
{
	FreeBuffer();
	Initialize();
	
	// The TGA format is little endian
	TmpEndianMode tem(file, boLittleEndian);

	// I would have done a struct if the header was not such a mess of misaligned attributes
	uint8 aHeader[18];
	int32 nRead = file.Read(aHeader, 18);
	if( nRead != 18)
	{
		m_nLastError = uncompleteFile;
		return false;
	}

	uint8 nImageIdLength = aHeader[0];
	uint8 nColorMapType  = aHeader[1];
	uint8 nImageType     = aHeader[2];
	int16 nColorMapOrigin = MAKE16( aHeader[4], aHeader[3] );
	int16 nColorMapLength = MAKE16( aHeader[6], aHeader[5] );
	uint8 nColorMapDepth  = aHeader[7];
	int16 nXOrigin = MAKE16( aHeader[9], aHeader[8] );
	int16 nYOrigin = MAKE16( aHeader[11], aHeader[10] );
	int16 nWidth   = MAKE16( aHeader[13], aHeader[12] );
	int16 nHeight  = MAKE16( aHeader[15], aHeader[14] );
	uint8 nBitsPerPixel = aHeader[16];
	uint8 nImageDescriptor = aHeader[17];

	if( nColorMapType == 0)
	{
		// No color-map data is included with this image
	}
	else if( nColorMapType == 1)
	{
		// A color-map is included
	}
	else
	{
		// Error
		m_nLastError = errBadColorMapType;
		return false;
	}

	
	// Here comes the ImageId
	if( nImageIdLength > 0)
	{
		uint8 aImageId[255];
		nRead = file.Read(aImageId, nImageIdLength);
		if( nRead != nImageIdLength)
		{
			m_nLastError = uncompleteFile;
			return false;
		}
	
	}

	// Here comes the colormap (also called palette)
	if( nColorMapType == 1 && nColorMapLength > 0)
	{
		if( !ReadPalette(file, nColorMapOrigin, nColorMapLength, nColorMapDepth) )
		{
			return false;
		}
	}

	m_nWidth = nWidth;
	m_nHeight = nHeight;

	TgaImageOrigin eImageOrigin = TgaImageOrigin (nImageDescriptor >> 4);
	uint8 nAlphaBitCount = nImageDescriptor & 0x0f;

	// Retrieve additional information that can be used to know what to do with the attribute fields
	ReadNewTgaFormatUsefulInfo(file);

	bool bDecodeOk = false;

	TgaImageType eImageType = TgaImageType(nImageType);
	if( eImageType == titNoImage )
	{
		// So... ?	
		m_nLastError = errUnsupportedFormat;
		return false;
	}
	else if( eImageType == titUncompColorMapped )
	{
		// Uncompressed, color-mapped images
		if( nBitsPerPixel == 8 && nColorMapType == 1)
		{
			m_epf = pf8bppIndexed;
			bDecodeOk = DecodeRawPixels(file, 1);
		}
		else
		{
			m_nLastError = errUnsupportedFormat;
			return false;
		}
	}
	else if( eImageType == titUncompRgb )
	{
		// Uncompressed, RGB images
		if( nBitsPerPixel == 24 && nAlphaBitCount == 0)
		{
			m_epf = pf24bppBgr;
			bDecodeOk = DecodeRawPixels(file, 3);
		}
		else if( nBitsPerPixel == 32)
		{
			m_epf = pf32bppBgra;
			bDecodeOk = DecodeRawPixels(file, 4);
		}
		else if( nBitsPerPixel == 16)
		{
			// Note : the most significant bit stores an attribute, not an alpha channel, so we ignore it

			m_epf = pf16bppRgb555;
			bDecodeOk = DecodeRawPixels(file, 2);
			if( bDecodeOk)
			{
				// Arrange values from GGGBBBBB_ARRRRRGG to ARRRRRGG_GGGBBBBB in memory
				const int32 nPixelCount = m_nWidth * m_nHeight;
				IFile::Swap16( (uint16*) m_aPixelBuffer.GetPtr(), nPixelCount);
			}
		}
		else
		{
			m_nLastError = errUnsupportedFormat;
			return false;
		}
	}
	else if( eImageType == titUncompBaw )
	{
		// Uncompressed, black and white images
		if( nBitsPerPixel == 8)
		{
			m_epf = pf8bppGrayScale;
			bDecodeOk = DecodeRawPixels(file, 1);
		}
		else if( nBitsPerPixel == 1)
		{
			m_epf = pf1bppGrayScale;
			bDecodeOk = Decode1bppGrayScale(file);
		}
		else
		{
			m_nLastError = errUnsupportedFormat;
			return false;
		}
	}
	else if( eImageType == titRunlengthColorMapped )
	{
		// Runlength encoded color-mapped images
		if( nBitsPerPixel == 8 && nColorMapType == 1)
		{
			m_epf = pf8bppIndexed;
			bDecodeOk = DecodeRlePixels(file, 1);
		}
		else
		{
			m_nLastError = errUnsupportedFormat;
			return false;
		}
	}
	else if( eImageType == titRunlengthRgb )
	{
		// Runlength encoded RGB images
		if( nBitsPerPixel == 24 && nAlphaBitCount == 0)
		{
			m_epf = pf24bppBgr;
			bDecodeOk = DecodeRlePixels(file, 3);
		}
		else if( nBitsPerPixel == 32)
		{
			m_epf = pf32bppBgra;
			bDecodeOk = DecodeRlePixels(file, 4);
		}
		else if( nBitsPerPixel == 16)
		{
			// Note : the most significant bit stores an attribute, not an alpha channel, so we ignore it

			m_epf = pf16bppRgb555;
			bDecodeOk = DecodeRlePixels(file, 2);
			if( bDecodeOk)
			{
				// Arrange values from GGGBBBBB_ARRRRRGG to ARRRRRGG_GGGBBBBB in memory
				const int32 nPixelCount = m_nWidth * m_nHeight;
				IFile::Swap16( (uint16*) m_aPixelBuffer.GetPtr(), nPixelCount);
			}
		}
		else
		{
			m_nLastError = errUnsupportedFormat;
			return false;
		}
	}
	else if( eImageType == titCompBaw )
	{
		// Compressed, black and white images
		if( nBitsPerPixel == 8)
		{
			m_epf = pf8bppGrayScale;
			bDecodeOk = DecodeRlePixels(file, 1);
		}
		else
		{
			m_nLastError = errUnsupportedFormat;
			return false;
		}
	}
	else if( eImageType == titCompColorMapped )
	{
		// Compressed color-mapped data, using Huffman, Delta, and runlength encoding
		m_nLastError = errUnsupportedFormat;
		return false;
	}
	else if( eImageType == titCompColorMapped4Pass )
	{
		// Compressed color-mapped data, using Huffman, Delta, and
		// runlength encoding.  4-pass quadtree-type process
		m_nLastError = errUnsupportedFormat;
		return false;
	}

	if( !bDecodeOk)
	{
		FreeBuffer();
		return false;
	}

	
	/////////////////////////////////////////////////////////////////////////////
	// Perform additional treatment for 32 bits images

	if( nBitsPerPixel == 32 )
	{
		bool bDiscardAlpha = false;

		if( nAlphaBitCount == 8 )
		{
			// Ok, there are 8 bits to store an "attribute" with each color, but is it really
			// an alpha channel ?
			if( m_eAttributesType == atUndef)
			{
				// No additional information contained in the file, assume the attributes are alphas
			}
			else if( m_eAttributesType == atUsefulAlpha)
			{
				// Additional information available, this is a real alpha channel
			}
			else
			{
				// Discard alpha channel
				bDiscardAlpha = true;
			}
		}
		else
		{
			bDiscardAlpha = true;
		}

		if( bDiscardAlpha)
		{
			SetAlphaFullOpaque();
		}
	}
	
	/////////////////////////////////////////////////////////////////////////////
	// Change orientation

	if( eImageOrigin == tioBottomLeft)
	{
		FlipVertical();
	}
	else if( eImageOrigin == tioBottomRight)
	{
	
	}
	else if( eImageOrigin == tioTopLeft)
	{
		// Natural orientation, nothing to do
	}
	else if( eImageOrigin == tioTopRight)
	{
	
	}
	return true;
}

PixelFormat Tga::GetPixelFormat() const
{
	return m_epf;
}

const Palette* Tga::GetPalette() const
{
	return &m_pal;
}

String Tga::GetLastErrorString() const
{
	switch(m_nLastError)
	{
	case errBadColorMapType:
		return L"Bad color map type (0 or 1 expected)";
	case errUnsupportedFormat:
		return L"Unsupported format";
	case errUnsupportedColorMapDepth:
		return L"Unsupported color map depth (only 16, 24 and 32 are supported)";
	case errColorMapTooLarge:
		return L"Color map too large (more than 256 colors)";
	case errBadCompressedData:
		return L"Bad compressed data";
	}
	return ImageType::GetLastErrorString();
}

bool Tga::DecodeRawPixels(IFile& file, int32 nBytesPerPixel)
{
	const int32 nBufSize = m_nWidth * m_nHeight * nBytesPerPixel;
	if( !m_aPixelBuffer.SetSize(nBufSize))
	{
		m_nLastError = notEnoughMemory;
		return false;
	}
	uint8* pPixels = m_aPixelBuffer.GetPtr();

	int32 nRead = file.Read(pPixels, nBufSize);
	if( nRead != nBufSize)
	{
		m_nLastError = uncompleteFile;
		return false;
	}

	return true;
}

bool Tga::DecodeRlePixels(IFile& file, int32 nBytesPerPixel)
{
	const int32 nPixelCount = m_nWidth * m_nHeight;
	const int32 nOutBufSize = nPixelCount * nBytesPerPixel;
	if( !m_aPixelBuffer.SetSize(nOutBufSize))
	{
		m_nLastError = notEnoughMemory;
		return false;
	}
	uint8* pPixels = m_aPixelBuffer.GetPtr();

	int32 nDstOffset = 0;
	///////////////////////////////////////
	while( nDstOffset < nOutBufSize )
	{
		uint8 aInBuf[16];

		const int32 nMinRead = 1 + nBytesPerPixel;
		int32 nRead = file.Read(aInBuf, nMinRead);
		if( nRead != nMinRead)
		{
			m_nLastError = uncompleteFile;
			return false;
		}
		
		uint8 nRleHeader = aInBuf[0];

		uint8 nRunPixelCount = (nRleHeader & 0x7f) + 1;
		int32 nRunByteCount = nRunPixelCount * nBytesPerPixel;
		if( nDstOffset + nRunByteCount > nOutBufSize)
		{
			// Troubles here, too much data
			m_nLastError = errBadCompressedData;
			return false;
		}
		
		if( (nRleHeader & 0x80) == 0x80)
		{
			///////////////////////////////
			//     Run-length packet     //
			///////////////////////////////

			// Next is the color

			for(int i = 0; i < nRunPixelCount; ++i)
			{
				for(int iByte = 0; iByte < nBytesPerPixel; ++iByte)
				{
					pPixels[nDstOffset + iByte] = aInBuf[1 + iByte];
				}
				nDstOffset += nBytesPerPixel;
			}
		}
		else
		{
			///////////////////////////////
			//          Raw packet       //
			///////////////////////////////
		
			// Put first pixel
			for(int iByte = 0; iByte < nBytesPerPixel; ++iByte)
			{
				pPixels[nDstOffset + iByte] = aInBuf[1 + iByte];
			}
			nDstOffset += nBytesPerPixel;

			// Read next directly into dst buffer
			if( nRunPixelCount > 1)
			{
				// Read more

				// The "- nBytesPerPixel" si here because we already read the first pixel whose size is nBytesPerPixel
				int32 nWantedByteCount = nRunByteCount - nBytesPerPixel;
				nRead = file.Read(pPixels + nDstOffset, nWantedByteCount);
				if( nRead != nWantedByteCount)
				{
					m_nLastError = uncompleteFile;
					return false;
				}
				nDstOffset += nWantedByteCount;
			}
		}
	}

	return true;
}

bool Tga::ReadPalette(IFile& file, int16 nColorMapOrigin, int16 nColorMapLength, uint8 nColorMapDepth)
{
	if( !(nColorMapDepth == 15 || nColorMapDepth == 16 || nColorMapDepth == 24 || nColorMapDepth == 32) )
	{
		m_nLastError = errUnsupportedColorMapDepth;
		return false;
	}

	// Jump over unused colors
	file.SetPosition(nColorMapOrigin, IFile::posCurrent);
	int32 nColorCount = nColorMapLength;
	if( nColorCount > 256)
	{
		// No support for more than 256 colors in palette
		m_nLastError = errColorMapTooLarge;
		return false;
	}
	
	const int32 nBytesPerColor = nColorMapDepth / 8;
	const int32 nWantedBytes = (nColorMapLength - nColorMapOrigin) * nBytesPerColor;

	uint8 aCols[256 * 4];
	int32 nReadCol = file.Read(aCols, nWantedBytes);
	if( nReadCol != nWantedBytes)
	{
		m_nLastError = uncompleteFile;
		return false;
	}

	int32 iSrcColByte = 0;

	m_pal.m_nCount = nColorCount;
	for(int i = 0; i < nColorCount; ++i)
	{
		if( nColorMapDepth == 32)
		{
			uint8 b = aCols[iSrcColByte + 0];
			uint8 g = aCols[iSrcColByte + 1];
			uint8 r = aCols[iSrcColByte + 2];
			uint8 a = aCols[iSrcColByte + 3];

			m_pal.m_aColors[i] = Color32(r, g, b, a);
		}
		else if( nColorMapDepth == 24)
		{
			uint8 b = aCols[iSrcColByte + 0];
			uint8 g = aCols[iSrcColByte + 1];
			uint8 r = aCols[iSrcColByte + 2];
			
			m_pal.m_aColors[i] = Color32(r, g, b, 255);
		}
		else if( nColorMapDepth == 16 || nColorMapDepth == 15)
		{
			uint8 nLow = aCols[iSrcColByte];
			uint8 nHi = aCols[iSrcColByte + 1];
			uint16 nColor = (nHi << 8) | nLow;
			m_pal.m_aColors[i] = Color32::From16bppRgb555(nColor);
		}
		
		iSrcColByte += nBytesPerColor;
	}

	return true;
}

bool Tga::Decode1bppGrayScale(IFile& file)
{
	int32 nBytesPerRow = Math::DivCeil(m_nWidth, 8);
	int32 nBufSize = nBytesPerRow * m_nHeight;

	if( !m_aPixelBuffer.SetSize(nBufSize))
	{
		m_nLastError = notEnoughMemory;
		return false;
	}
	uint8* pPixels = m_aPixelBuffer.GetPtr();

	int32 nRead = file.Read(pPixels, nBufSize);
	if( nRead != nBufSize)
	{
		m_nLastError = uncompleteFile;
		return false;
	}

	return true;
}

// Returns false if no footer or cannot be read
bool Tga::ReadFooter(IFile& file, TgaFooter& tf)
{
	if( !file.SetPosition(-26, IFile::posEnd))
	{
		// Cannot seek
		return false;
	}

	if( file.Read(&tf, 26) == 26)
	{
		if( file.ShouldSwapBytes())
		{
			tf.nExtensionAreaOffset = IFile::Swap32(tf.nExtensionAreaOffset);
			tf.nDeveloperDirectoryOffset = IFile::Swap32(tf.nDeveloperDirectoryOffset);
		}

		if( tf.nDot == '.' && tf.nZero == 0x00 && Memory::IsEqual(tf.aSignature, "TRUEVISION-XFILE", 16))
		{
			// Footer read ok
			return true;
		}
	}
	return false;
}

// This can help to determine if in a 32 bits images the alpha channel actually stores alpha values
void Tga::ReadNewTgaFormatUsefulInfo(IFile& file)
{
	int64 nOldPos = file.GetPosition();
	if( nOldPos < 0)
	{
		// Cannot seek
		return;
	}

	TgaFooter tf;
	if( ReadFooter(file, tf) )
	{
		if( tf.nExtensionAreaOffset > 0)
		{
			if( file.SetPosition(tf.nExtensionAreaOffset + 494))
			{
				uint8 nAttributesTypeField24;
				if( file.Read8(nAttributesTypeField24))
				{
					// Damn, that was not simple to get
					m_eAttributesType = AttributesType(nAttributesTypeField24);
				}
			}
		}
	}

	// Back to first position
	file.SetPosition(nOldPos);
}
