///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "Master.h"
#include "TimeStamp.h"

namespace chustd {

void SplitChuTimeStamp(int64 nStamp, uint16& nYearOut, uint8& nMonth, uint8& nDay, 
	                       uint8& nHour, uint8& nMinute, uint8& nSecond, uint16& nMillisecond)
{
	int64 nNext = nStamp;

	nMillisecond = uint16(nNext % 1000);
	nNext /= 1000;

	nSecond = uint8(nNext % 60);
	nNext /= 60;

	nMinute = uint8(nNext % 60);
	nNext /= 60;

	nHour = uint8(nNext % 24);
	nNext /= 24;

	int32 nDays = int32(nNext);

	/////////////////////////////////////////////
	int32 nYear = 1601;

	// Leap year ?
	// Multiple of 4 : yes
	// Multiple of 100 : no
	// Multiple of 400 : yes

	/////////////////////////////////////////////////////////////////////////////////////////////
	// The gregorian calendar repeats itself every 400 years
	// The calendar is divided into an infinite number of parts, each part is a quadcentury
	//
	// ---- QuadCenturies ----
	// Count : 146097  146097  146097  146097 ...
	// Index :    0       1       2       3
	// Year  :  1601    2001    2401    2801

	int32 nDaysPer400Years = 400 * 365 + 100 - 4 + 1; // 146097
	
	const int32 nPartIndex0 = nDays / nDaysPer400Years;
	const int32 nDayIndex0 = nDays - nPartIndex0 * nDaysPer400Years;

	nYear += nPartIndex0 * 400;

	/////////////////////////////////////////////////////////////////////////////////////////////
	// The quadcentury is divided into 4 parts, each part is a century
	//
	// ---- Centuries ----
	// Count : 36524  36524  36524  36525 <-- always
	// Index :   0      1      2      3

	int32 nPartIndex1 = nDayIndex0 / 36524;
	if( nPartIndex1 > 3)
	{
		// Reajust
		nPartIndex1 = 3;
	}
	const int32 nDayIndex1 = nDayIndex0 - nPartIndex1 * 36524;

	nYear += nPartIndex1 * 100;

	/////////////////////////////////////////////////////////////////////////////////////////////
	// The century is divided into 25 parts, each part is a quadyear
	//
	// ---- QuadYears ----
	// Count : 1461  1461  1461... 1460 <-- regular last year of century (1700, 1800, 1900)
	// Count : 1461  1461  1461... 1461 <-- last year of quadcentury (2000, 2400, 2800)
	// Index :   0     1     2      24

	int32 nPartIndex2 = nDayIndex1 / 1461;

	const int32 nDayIndex2 = nDayIndex1 - nPartIndex2 * 1461;

	nYear += nPartIndex2 * 4;
	
	/////////////////////////////////////////////////////////////////////////////////////////////
	// The quadyear is divided into 4 parts, each part is a year
	// 
	// ---- Years ----
	// Count : 365 365 365 366 <-- regular last quadyear (1704, 1708, 1712)
	// Count : 365 365 365 366 <-- last year of quadcentury (2000, 2400, 2800)
	// Count : 365 365 365 365 <-- last year of century and not last year of quadcentury (1700, 1800, 1900)
	// Index :  0   1   2   3

	int32 nPartIndex3 = nDayIndex2 / 365;
	if( nPartIndex3 > 3)
	{
		// Readjust
		nPartIndex3 = 3;
	}
	const int32 nDayIndex3 = nDayIndex2 - nPartIndex3 * 365;

	nYear += nPartIndex3;
	///////////////////////////////////////////////////////////

	/////////////////////////////////////////////////////////////////////////////////////////////
	// Now that we got the year, let's compute the month

	static const uint8 aDaysPerMonth[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
	bool bLeapYear = ((nYear % 4) == 0) && !((nYear % 100) == 0) || ((nYear % 400) == 0);

	int32 nSumMonth = 0;
	int32 iMonth = 0;

	for(; iMonth < 12; ++iMonth)
	{
		int32 nDayCount = aDaysPerMonth[iMonth];
		if( bLeapYear && iMonth == 1)
		{
			nDayCount++;
		}

		if( nSumMonth + nDayCount > nDayIndex3)
		{
			break;
		}

		nSumMonth += nDayCount;
	}

	nYearOut = uint16(nYear);
	nMonth = uint8(1 + iMonth);
	nDay = uint8(1 + (nDayIndex3 - nSumMonth));
}

int64 MakeChuTimeStamp(uint16 nYear, uint8 nMonth, uint8 nDay, 
	                       uint8 nHour, uint8 nMinute, uint8 nSecond, uint16 nMillisecond)
{
	// The base date is 1601-01-01 00:00:00.000

	int32 nYearRange = nYear - 1601;
	if( nYearRange < 0)
	{
		// TODO: thinks wether we stop and return an error or just use 0
		return 0;
	}

	// Leap year ?
	// Multiple of 4 : yes
	// Multiple of 100 : no
	// Multiple of 400 : yes

	int32 nLeapCount = nYearRange / 4;
	nLeapCount -= nYearRange / 100;
	nLeapCount += nYearRange / 400;

	int64 nStamp = nYearRange;
	nStamp *= 365; // 365 days
	nStamp += nLeapCount;
		
	//////////////////////////////////////////////////////
	static const uint8 aDaysPerMonth[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
	bool bLeapYear = ((nYear % 4) == 0) && !((nYear % 100) == 0) || ((nYear % 400) == 0);

	int32 nMonthRange = nMonth;
	if( ! (1 <= nMonthRange && nMonthRange <= 12))
	{
		nMonthRange = 1;
	}
	nMonthRange -= 1;

	for(int32 iMonth = 0; iMonth < nMonthRange; ++iMonth)
	{
		uint8 nDayCount = aDaysPerMonth[iMonth];
		if( iMonth == 1 && bLeapYear)
		{
			nDayCount += 1;
		}
		nStamp += nDayCount;
	}
	
	/////////////////////////////////////////////////
	int32 nDayRange = nDay;
	if( ! (1 <= nDayRange && nDayRange <= 31))
	{
		nDayRange = 1;
	}
	nDayRange -= 1;

	nStamp += nDayRange;

	/////////////////////////////////////////////////
	nStamp *= 24;
	nStamp += nHour;

	nStamp *= 60;
	nStamp += nMinute;

	nStamp *= 60;
	nStamp += nSecond;

	nStamp *= 1000;
	nStamp += nMillisecond;

	return nStamp;
}

} // namespace chustd