///////////////////////////////////////////////////////////////////////////////
// This file is part of the chuwin32 library
// Copyright (C) 2005/2008 ChuTeam
// For conditions of distribution and use, see copyright notice in chuwin32.h
///////////////////////////////////////////////////////////////////////////////

#ifndef CHUWIN32_TRACECTL_H
#define CHUWIN32_TRACECTL_H

#if !defined(CHUSTD_VERSION)
#error "TraceCtl : you need the chustd library to build this class"
#endif

#include "Window.h" 
#include "DibBitmap.h"

// Class Version : 1.4
// 1.4
// - better thread safetey (corrected a deadlock with Vista)
// - better anti-aliased font rendering
// - return of the dib for one line of text in order to avoid flickering
// 1.3
// - thread safe by avoiding SendMessage in drawing functions
// - no more use of dib
// 1.2
// - chustd instead of dkstd
// 1.1 
// - support for unicode
// - use of the system icon font

namespace chuwin32 {

	class TraceCtl : public Window
{
public:
	bool Create(int x, int y, int w, int h, HWND hParentWnd, int nID);

	void AddText(const chustd::String& strText, COLORREF cr = 0);
	void AddLine(const chustd::String& strText, COLORREF cr = 0);
	void AddTextAtLine(int32 nLine, const chustd::String& strText, COLORREF cr = 0);
	void AddLink(const chustd::String& strText, const chustd::String& strUrl);
	void ClearLineAt(int32 nLine);
	void ClearLine();
	void Clear();

	void Resize(int cx, int cy);
	void ResetHrzPos();

	void SetFont(const HFONT& hFont);
	const HFONT& GetFont() const;

	TraceCtl();
	virtual ~TraceCtl();

protected:

	class TextPiece
	{
	public:
		chustd::String m_str;
		chustd::String m_strUrl; // If non-empty, the text piece is a link
		COLORREF m_cr;
	};

	class Line
	{
	public:
		chustd::Array<TextPiece> m_aPieces;

		TextPiece* AddOrReusePiece();
	};

protected:
	virtual int DrawTextPiece(HDC hDC, int x, int y, const chustd::String& str, COLORREF cr);
	
	enum DragResult
	{
		dragNone,
		dragMove,
		dragCopy
	};
	virtual int OnLinkDragBegin(const chustd::String& strUrl);
	virtual void OnLinkDoubleClick(const chustd::String& strUrl);

	int GetLineHeight() const { return m_nLineHeight; }
	int ComputeLineWidth(const Line& line);
	uint32 GetNewScrollPos32(int nSBType, HWND hScrollBar = 0);

	TextPiece* GetLinkFromPoint(int x, int y, RECT* pLocalRect = null);

	///////////////////////////////////////////////////////////////	
	void UpdateScrollSizeVert(bool bUpdatePosition = true);
	void InitScrollSizeVert();

	void UpdateScrollSizeHorz();
	void InitScrollSizeHorz();

	void Draw(HDC hDC);
	void Refresh();

	///////////////////////////////////////////////////////
	virtual LRESULT WndProc(UINT nMsg, WPARAM wParam, LPARAM lParam);

	int OnEraseBkgnd(WPARAM wParam, LPARAM lParam);
	int OnPaint(WPARAM wParam, LPARAM lParam);
	int OnVScroll(WPARAM wParam, LPARAM lParam);
	int OnHScroll(WPARAM wParam, LPARAM lParam);
	int OnSize(WPARAM wParam, LPARAM lParam);
	int OnRButtonDown(WPARAM wParam, LPARAM lParam);
	int OnLButtonDown(WPARAM wParam, LPARAM lParam);
	int OnLButtonUp(WPARAM wParam, LPARAM lParam);
	int OnMouseMove(WPARAM wParam, LPARAM lParam);
	int OnLButtonDblClk(WPARAM wParam, LPARAM lParam);
	int OnKeyDown(WPARAM wParam, LPARAM lParam);
	int OnMouseWheel(WPARAM wParam, LPARAM lParam);
	int OnAddText(WPARAM wParam, LPARAM lParam);
	int OnDestroy(WPARAM wParam, LPARAM lParam);
	///////////////////////////////////////////////////////

private:
	HFONT m_hFont;	   // La police courante
	HFONT m_hIconFont; // La police pour les icnes du bureau

	chustd::Array<Line> m_aLines;

	int m_nLineHeight;
	int m_nLargestLineWidth; // Largeur maximale en pixels
	int m_nMarginLeft;
	int m_nAvgCharWidth;

	bool m_bDragStart;
	TextPiece* m_pDragText;
	POINT m_ptDrag;

	chustd::CriticalSection m_cs;

	static uint32 ms_nWmAddText;

	enum Operation { opAddText, opAddTextAtLine, opAddLink, opClearLineAt, opClearAll };

	struct AddTextInfo
	{
		Operation operation;
		chustd::String strText;
		chustd::String strUrl;
		COLORREF cr;
		int nLine;
	};

	DibBitmap m_dib;

private:
	void DoAddText(const chustd::String& strText, COLORREF cr);
	void DoAddTextAtLine(int32 nLine, const chustd::String& strText, COLORREF cr);
	void DoAddLink(const chustd::String& strText, const chustd::String& strUrl);
	void DoClearLineAt(int32 nLine);
	void DoClear();
};

} // namespace chuwin32

#endif // ndef CHUWIN32_TRACECTL_H
